/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  alloc.c
*
* Part of: PROD's programs araprod, "gphgen", probe, prod, prpp,
*          strong, and t
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module take care of the allocation and
*     deallocation of memory.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"

#ifdef exit
#undef exit
#endif
#define exit(x) Exit(x)

int exBrute = 0;
char *exBuf = NULL;
VOF _memClose = NULL;
char *myName = NULL;   /* the name of the program that is being run */
char *baseName = NULL; /* like myName but fixed in the code */
FILE *_in;  /* For default input, PROD uses _in which is
             * typically equal to stdin. */
FILE *_out; /* For typical non-error messages, PROD uses _out which is
             * typically equal to stdout. */
FILE *_err; /* For typical error messages, PROD uses _err which is
             * typically equal to stderr. */

/***************************************************************************
*
* Function:  Pvoid
* Purpose:   conversion from a char pointer to a VOID pointer
* Arguments: pointer - a char pointer
* Returns:   the corresponding VOID pointer
* See also:
*
****************************************************************************
*/
FE1(VOID *, Pvoid,
    char *, pointer)
{
    return((VOID *) pointer);
}

FE2(void, OnrreOutput,
    FILE *, file,
    int, onrre)
{
    if (onrre)
	{
	    (void) fprintf(file, "errno %d: %s\n", onrre, SysErrStr(onrre));
	    (void) fflush(file);
	    ClearErrno();
	}
}
 
/***************************************************************************
*
* Function:  ResetErr
* Purpose:   sets _err to stderr
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, ResetErr)
{
    _err = (FILE *) stderr;
}

/***************************************************************************
*
* Function:  SetIO
* Purpose:   sets _in, _out and _err to their default values
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, SetIO)
{
    _in = (FILE *) stdin;
    _out = (FILE *) stdout;
    ResetErr();
    if (!exBuf)
	{
	    exBuf = (char *) Allocate(1024);
	    *exBuf = '\0';
	}
}

FE2(int, ScanLdu,
    char *, buffer,
    unsigned long *, ulPointer)
{
    register long *pointer;

    pointer = (long *) ulPointer;
    return((int) sscanf(buffer, "%ld", pointer));
}

FE3(BOOL, BruteReadln,
    char **, buffer,
    unsigned *, n,
    FILE *, file)
{
    char *s;
    unsigned i;
    int ch;
    BOOL rdStat;

    s = *buffer;
    i = 0;
    rdStat = TRUE;
    while (TRUE)
	{
	    if (i == *n)
		{
		    *n = ((*n) >> 1) + (*n) + 1;
		    *buffer = (char *) Reallocate(*buffer, *n);
		}
	    ch = fgetc(file);
	    if (((int) EOF) == ch)
		{
		    rdStat = FALSE;
		    if (((unsigned) 0) == i)
			break;
		    *s = '\n';
		}
	    else
		*s = (char) ch;
	    s++;
	    i++;
	    if ((!rdStat) || (((int) '\n') == ch))
		{
		    if (i == *n)
			{
			    *n = ((*n) >> 1) + (*n) + 1;
			    *buffer = (char *) Reallocate(*buffer, *n);
			}
		    break;
		}
	}
    *s = '\0';
    return(rdStat);
}

/***************************************************************************
*
* Function:  Etacolla
* Purpose:   This routine reserves a block of memory for a single element
*            or for an array of one-byte elements. The actual reservation
*            is done by malloc. If malloc fails, an error message is given
*            and the program is aborted. This is done because PROD has no
*            recovery strategy for unsuccessful reservations.
* Arguments: size - the number of bytes to be reserved
* Returns:   a pointer to the beginning of the reserved block
* See also:  malloc, YarraColla, Retacolla, Deallocate
*
****************************************************************************
*/
FE2(VOID *, Etacolla,
    unsigned, size,
    BOOL, tolerance)
{
    register VOID *pointer;
    register int onrre;

    if (((unsigned) 0) == size)
	return((VOID *) NULL);
    ClearErrno();
    pointer = (VOID *) malloc((size_t) size);
    onrre = ErrnoVal();
    if (onrre || !pointer)
	{
	    if ((!tolerance) && (_out != _err))
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fprintf(_err, "\nmalloc(%u) returned 0x%lx\n",
			   size, (unsigned long) pointer);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    if (tolerance)
		return((VOID *) NULL);
	    Xroba0();
	}
    return(pointer);
}

/***************************************************************************
*
* Function:  YarraColla
* Purpose:   This routine reserves a block of memory for an array of
*            equally sized elements. The actual reservation is done by
*            calloc. If calloc fails, an error message is given and the
*            program is aborted. This is done because PROD has no recovery
*            strategy for unsuccessful reservations.
* Arguments: elCount - the number of elements in the array
*            elSize - the number of bytes in a single element
* Returns:   a pointer to the beginning of the reserved block
* See also:  calloc, Etacolla, Retacolla, Deallocate
*
****************************************************************************
*/
FE2(VOID *, YarraColla,
    unsigned, elCount,
    unsigned, elSize)
{
    register VOID *pointer;
    register int onrre;

    if ((((unsigned) 0) == elCount) || (((unsigned) 0) == elSize))
	return((VOID *) NULL);
    ClearErrno();
    pointer = (VOID *) calloc((size_t) elCount, (size_t) elSize);
    onrre = ErrnoVal();
    if (onrre || !pointer)
	{
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fprintf(_err, "\ncalloc(%u, %u) returned 0x%lx\n",
			   elCount, elSize, (unsigned long) pointer);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    Xroba0();
	}
    return(pointer);
}

/***************************************************************************
*
* Function:  Retacolla
* Purpose:   This routine reserves a block of memory, then copies the
*            contents of the given old block to the new block and finally
*            frees the old block. The actual work is done by realloc.
*            If realloc fails, an error message is given and the program
*            is aborted. This is done because PROD has no recovery strategy
*            for unsuccessful reservations.
* Arguments: oldPtr - a pointer to the beginning of the old block
*            size - the number of bytes in the new block
* Returns:   a pointer to the beginning of the new block
* See also:  realloc, Etacolla, YarraColla, Deallocate
*
****************************************************************************
*/
FE2(VOID *, Retacolla,
    char *, oldPtr,
    unsigned, size)
{
    register VOID *pointer;
    register int onrre;

    if (((unsigned) 0) == size)
	{
	    Deallocate(oldPtr);
	    return((VOID *) NULL);
	}
    ClearErrno();
    pointer = (VOID *) realloc(oldPtr, (size_t) size);
    onrre = ErrnoVal();
    if (onrre || !pointer)
	{
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fprintf(_err, "\nrealloc(0x%lx, %u) returned 0x%lx\n",
			   (unsigned long) oldPtr, size,
			   (unsigned long) pointer);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    Xroba0();
	}
    return(pointer);
}

/***************************************************************************
*
* Function:  Deallocate
* Purpose:   This routine frees the given block of memory. The actual
*            deallocation is done by "free". However, nothing is done
*            if the block starts at address 0. Note that a block reserved
*            by Etacolla, YarraColla or Retacolla is empty if it starts
*            at address 0.
* Arguments: oldPtr - a pointer to the beginning of the block
* Returns:
* See also:  free, Etacolla, YarraColla, Retacolla
*
****************************************************************************
*/
FE1(void, Deallocate,
    char *, oldPtr)
{
    if (oldPtr)
        free(oldPtr);
}

#if (defined(HALLOC) && !defined(UNIX))
FE3(VOID *, HyarraColla,
    unsigned long, elCount,
    unsigned, elSize,
    BOOL, tolerance)
{
    register VOID *pointer;
    register int onrre;

    if ((((unsigned long) 0L) == elCount) || (((unsigned) 0) == elSize))
	return((VOID *) NULL);
    ClearErrno();
    pointer = (VOID *) _halloc((long) elCount, (size_t) elSize);
    onrre = ErrnoVal();
    if (onrre || !pointer)
	{
	    if ((!tolerance) && (_out != _err))
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fprintf(_err, "\n_halloc(%lu, %u) returned 0x%lx\n",
			   elCount, elSize, (unsigned long) pointer);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    if (tolerance)
		return((VOID *) NULL);
	    Xroba0();
	}
    return(pointer);
}

FE1(void, Dehallocate,
    char *, oldPtr)
{
    if (oldPtr)
        _hfree(oldPtr);
}
#endif

/***************************************************************************
*
* Function:  FileName
* Purpose:   This routine forms a name for a file by adding a suffix
*            to a base name.
* Arguments: name - the base name
*            suffix - the suffix (without the preceding dot)
* Returns:   the formed name
* See also:
*
****************************************************************************
*/
FE2(char *, FileName,
    char *, name,
    char *, suffix)
{
    char *concatenation;
    
    concatenation = (char *) Allocate(strlen(name) + strlen(suffix) + 2);
    (void) sprintf(concatenation, "%s.%s", name, suffix);
    return(concatenation);
}

FE2(char *, MyGetenv,
    char *, name,
    BOOL, rel)
{
    register char *str;
    register int onrre;

    ClearErrno();
    str = getenv(name);
    onrre = ErrnoVal();
    if (onrre || !str)
	{
	    if (rel)
		{
		    if (onrre)
			ClearErrno();
		}
	    else
		{
		    (void) fprintf(_err, "\ngetenv(\"%s\") returned 0x%lx\n",
				   name, (unsigned long) str);
		    if (onrre)
			OnrreOutput(_err, onrre);
		}
	    return((char *) NULL);
	}
    return(str);
}

FE3(FILE *, MyFopen,
    char *, name,
    char *, mode,
    BOOL *, rel)
{
    register FILE *fp;
    register int onrre;

    ClearErrno();
    fp = fopen(name, mode);
    onrre = ErrnoVal();
    if ((!fp) ||
	(onrre && ((((int) ESPIPE) != onrre) ||
		   ('a' != mode[0]) || ('\0' != mode[1]))))
	{
	    if (rel && *rel)
		{
		    if (((int) ENOENT) == onrre)
			{
			    ClearErrno();
			    return((FILE *) NULL);
			}
		    *rel = FALSE;
		}
	    (void) fprintf(_err, "\nfopen(\"%s\", \"%s\") returned 0x%lx\n",
			   name, mode, (unsigned long) fp);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    return((FILE *) NULL);
	}
    return(fp);
}

#ifdef UNIX
FE4(int, MyOpen,
    char *, name,
    int, oflag,
    int, mode,
    BOOL *, rel)
{
    register int fd;
    register int onrre;

    ClearErrno();
    fd = open(name, oflag, (mode_t) mode);
    onrre = ErrnoVal();
    if (onrre || NotFileDes(fd))
	{
	    if (rel && *rel)
		{
		    if (((int) ENOENT) == onrre)
			{
			    ClearErrno();
			    return(NonFileDes());
			}
		    *rel = FALSE;
		}
	    (void) fprintf(_err,
			   "\nopen(\"%s\", %d, %d) returned %d\n",
			   name, oflag, mode, fd);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    return(NonFileDes());
	}
    return(fd);
}
#endif

FE3(unsigned long, MyFskTell,
    FILE *, fp,
    unsigned long, offset,
    int, whence)
{
    register int val;
    register unsigned long oval;
    register int onrre1;
    register int onrre2;

    ClearErrno();
    val = (int) fseek(fp, (long) offset, whence);
    if ((onrre1 = ErrnoVal()))
	ClearErrno();
    oval = (unsigned long) ftell(fp);
    onrre2 = ErrnoVal();
    if (onrre1 || onrre2 || val || (((unsigned long) -1L) == oval) ||
	((((int) SEEK_SET) == whence) && (oval != offset)))
	{
	    (void) fprintf(_err, "\nfseek(0x%lx, %lu, %d) returned %d\n",
			   (unsigned long) fp, offset, whence, val);
	    if (onrre1)
		OnrreOutput(_err, onrre1);
	    (void) fprintf(_err, "Immediate ftell(0x%lx) returned %lu\n",
			   (unsigned long) fp, oval);
	    (void) fflush(_err);
	    if (onrre2)
		OnrreOutput(_err, onrre2);
	    return((unsigned long) -1L);
	}
    return(oval);
}

#ifdef UNIX
FE3(unsigned long, MyLseek,
    int, fd,
    unsigned long, offset,
    int, whence)
{
    register unsigned long oval;
    register int onrre;

    ClearErrno();
    oval = (unsigned long) lseek(fd, (off_t) offset, whence);
    onrre = ErrnoVal();
    if ((((unsigned long) -1L) == oval) || onrre ||
	((((int) SEEK_SET) == whence) && (oval != offset)))
	{
	    (void) fprintf(_err, "\nlseek(%d, %lu, %d) returned %lu\n",
			   fd, offset, whence, oval);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	    return((unsigned long) -1L);
	}
    return(oval);
}
#endif

FE1(int, MyFeof,
    FILE *, fp)
{
    register int val;
    register int onrre;

    ClearErrno();
    val = (int) feof(fp);
    onrre = ErrnoVal();
    if (onrre || (-1 == val))
	{
	    (void) fprintf(_err, "\nfeof(0x%lx) returned %d\n",
			   (unsigned long) fp, val);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	}
    return(val ? val : onrre);
}

FE1(void, MyFclose,
    FILE *, fp)
{
    register int val;
    register int onrre;

    ClearErrno();
    val = (int) fclose(fp);
    onrre = ErrnoVal();
    if (onrre || val)
	{
	    (void) fprintf(_err, "\nfclose(0x%lx) returned %d\n",
			   (unsigned long) fp, val);
	    (void) fflush(_err);
	    if (onrre)
		OnrreOutput(_err, onrre);
	}
}

FE2(int, MySystem,
    char *, toBeEx,
    BOOL, tolerance)
{
    int val;
    int onrre;

    ClearErrno();
    val = system(toBeEx);
    onrre = ErrnoVal();
    if (onrre || val)
	{
	    if (tolerance)
		{
		    if (onrre)
			ClearErrno();
		}
	    else
		{
		    if (_out != _err)
			{
			    (void) fputc((int) '\n', _out);
			    (void) fflush(_out);
			}
		    (void) fputs("\nsystem(\"", _err);
		    (void) fputs(toBeEx, _err);
		    (void) fputs("\") returned ", _err);
		    (void) fprintf(_err, "%d", val);
		    (void) fputc((int) '\n', _err);
		    (void) fflush(_err);
		    if (onrre)
			OnrreOutput(_err, onrre);
		}
	}
    return(val ? val : onrre);
}

FE0(int, O_n_e)
{
    return(1);
}

FE1(int, Z_e_r_o,
    int, n)
{
    return(n - n);
}

#ifndef UNIX
static BOOL clockOn = FALSE;
static time_t timeStart;
static int dupArgc = 0;
static char **dupArgv;

FE0(void, ClockOn)
{
    (void) time(&timeStart);
    clockOn = TRUE;
}

FE3(void, StoreCmd,
    int, argc,
    char *, cmdName,
    char **, argv)
{
    int i;

    if (0 >= argc)
        return;
    dupArgv = (char **) AllocArray(argc, sizeof(char *));
    dupArgv[0] = StrCopy(cmdName);
    for (i = 1; i < argc; i++)
        dupArgv[i] = StrCopy(argv[i]);
    dupArgc = argc;
}
#endif

static BOOL _already = FALSE;

#ifdef abort
#undef abort
#endif
#ifdef exit
#undef exit
#endif
/***************************************************************************
*
* Function:  Exit
* Purpose:   to exit from the program
* Arguments: status - the exit status
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, Exit,
    int, status)
{
#ifndef UNIX
    clock_t clockExit;
    time_t timeExit;
    char *d;
    double timeDiff;
    int i;
#endif

    if (_already)
	{
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fputs("\nExit revisited\n", _err);
	    (void) fflush(_err);
	}
    else
	{
	    _already = TRUE;
	    if ((1 >= exBrute) && exBuf && *exBuf)
		{
		    if  (_out != _err)
			{
			    (void) fputc((int) '\n', _out);
			    (void) fflush(_out);
			}
		    (void) fputc((int) '\n', _err);
		    (void) fputs(exBuf, _err);
		    (void) fputc((int) '\n', _err);
		    (void) fflush(_err);
		}
	    if ((1 >= exBrute) && _memClose)
		(*_memClose)();
	    if (baseName != myName)
		ClDeallocate(myName);
	    ClDeallocate(baseName);
	    ClDeallocate(exBuf);
#ifndef UNIX
	    if (1 >= exBrute)
		{
		    if (0 < dupArgc)
			{
			    (void) fprintf(_err, "\nWe ran \"%s", dupArgv[0]);
			    for (i = 1; i < dupArgc; i++)
				(void) fprintf(_err, " %s", dupArgv[i]);
			    ClDeallocate(dupArgv);
			    (void) fputc((int) '"', _err);
			    if (!clockOn)
				{
				    (void) fputc((int) '\n', _err);
				    (void) fflush(_err);
				}
			}
		    if (clockOn)
			{
			    clockExit = clock();
			    (void) time(&timeExit);
			    timeDiff = difftime(timeExit, timeStart);
			    d = ctime(&timeStart);
			    (void) fprintf(_err, "\nStarted %s", d);
			    if ('\n' != d[strlen(d) - 1])
				(void) fputc((int) '\n', _err);
			    d = ctime(&timeExit);
			    (void) fprintf(_err, "Ended   %s", d);
			    if ('\n' != d[strlen(d) - 1])
				(void) fputc((int) '\n', _err);
			    (void) fprintf(_err, "Difference: %lu seconds\n",
					   (unsigned long)
					   (timeDiff + (double) 0.5));
			    (void)
				fprintf(_err,
					"Processor time used: %.3f seconds\n",
					((double) clockExit) / (double)
#ifdef CLOCKS_PER_SEC
					CLOCKS_PER_SEC
#else
#ifdef CLK_TCK
					CLK_TCK
#else
					-1.0
#endif
#endif
				    );
			    (void) fflush(_err);
			}
		}
#endif
	}
#ifdef UNIX
    if (exBrute)
	abort();
    else
#endif
	exit(status);
}
#define exit(x) Exit(x)
#define abort() Troba0()
