/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  araprod2.c
*
* Part of: PROD's program araprod
*
* Authors: Lasse Anderson, Robert Jurgens and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See araprod.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"
#include "genesig.h"
#include "cmdnm.h"
#include "araprod.h"

static FILE *auxFile = NULL;
static char *auxName = NULL;
static LIST *ltsList = NULL;
static LIST *hideList = NULL;

/***************************************************************************
*
* Function:  Instances
* Purpose:   This routine constructs low and high level transitions
*            corresponding to a visible action.
* Arguments: lacCursor - the cursor of ActGUsed(arpAction) as expected
*                        by the caller
* Returns:
* See also:  TauInstances, TstrTauInstances, Accepted
*
****************************************************************************
*/
FE1(void, Instances,
    VOID *, lacCursor)
{
    arpLaction = (LACTION *) LstThis(ActGUsed(arpAction));
    if (arpLaction == (LACTION *) LstFirst(ActGUsed(arpAction)))
	{
	    if (ActHidden(arpAction))
		{
		    HiTrNmOutput(IName(), arpHiTr);
		    NetOutput(netFile, "/* hidden \"%s\" */\n",
			      ActGName(arpAction));
		    if (arpHiTr != 0)
			{
			    NetOutput(netFile,
				      "/* previous invisible instances with the same flow are not repeated */\n");
			}
		}
	    else
		HiTrNmOutput(ActGName(arpAction), (unsigned) 0);
	    if (!arpLaction)
		{
		    NetOutput(netFile, "    gate (0 != 0);\n#endtr\n");
		    return;
		}
	    NetOutput(netFile, "    in  { ");
	    if ((!OnTheFly()) ||
		(LacGSystem((LACTION *) LstFirst(ActGUsed(arpAction)))
		 != tstrSys))
		{
		    NetOutput(netFile, "p: ");
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			{
			    if (OnTheFly() &&
				(LacGSystem(arpLaction) == tstrSys))
				break;
			    if (arpLaction != (LACTION *)
				LstFirst(ActGUsed(arpAction)))
				{
				    NetOutput(netFile, " + ");
				}
			    NetOutput(netFile, "<.");
			    NetOutput(netFile, DOMFORMAT,
				      LacGSystem(arpLaction));
			    NetOutput(netFile, ", y%lu.>",
				      (unsigned long) LacGSystem(arpLaction));
			}
		    NetOutput(netFile, "; ");
		}
	    if (OnTheFly() &&
		(LacGSystem((LACTION *) LstLast(ActGUsed(arpAction)))
		 == tstrSys))
		{
		    NetOutput(netFile, "q: <.y%lu.>; ",
			      (unsigned long) tstrSys);
		}
	    NetOutput(netFile, "}\n    out { ");
	    if ((!OnTheFly()) ||
		(LacGSystem((LACTION *) LstFirst(ActGUsed(arpAction)))
		 != tstrSys))
		{
		    NetOutput(netFile, "p: ");
		    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
			{
			    if (OnTheFly() &&
				(LacGSystem(arpLaction) == tstrSys))
				break;
			    if (arpLaction != (LACTION *)
				LstFirst(ActGUsed(arpAction)))
				{
				    NetOutput(netFile, " + ");
				}
			    NetOutput(netFile, "<.");
			    NetOutput(netFile, DOMFORMAT,
				      LacGSystem(arpLaction));
			    NetOutput(netFile, ", z%lu.>",
				      (unsigned long) LacGSystem(arpLaction));
			}
		    NetOutput(netFile, "; ");
		}
	    if (OnTheFly() &&
		(LacGSystem((LACTION *) LstLast(ActGUsed(arpAction)))
		 == tstrSys))
		{
		    NetOutput(netFile, "q: <.z%lu.>; ",
			      (unsigned long) tstrSys);
		}
	    NetOutput(netFile, "}\n");
	    LSTFOR (ActGUsed(arpAction), arpLaction, LACTION)
		{
		    if (!LstNext(ActGUsed(arpAction)))
			break;
		    NetOutput(netFile, "    gate ");
		    LSTFOR (LacGLisTransitions(arpLaction), arpTrans, TRANS)
			{
			    if (arpTrans != (TRANS *)
				LstFirst(LacGLisTransitions(arpLaction)))
				{
				    NetOutput(netFile, " || ");
				}
			    NetOutput(netFile, "(y%lu == ",
				      (unsigned long) LacGSystem(arpLaction));
			    NetOutput(netFile, DOMFORMAT,
				      TransGSource(arpTrans));
			    NetOutput(netFile, ")");
			}
		    NetOutput(netFile, ";\n");
		}
	    NetOutput(netFile, "    comp {");
	    if (ActHidden(arpAction))
		{
		    NetOutput(netFile, " Invisible();");
		}
	    NetOutput(netFile, "\n");
	    arpLaction = (LACTION *) LstBegin(ActGUsed(arpAction));
	    arpIndex = 0;
	    arpCounter = 0L;
	    goEnd = FALSE;
	}
    (void) LstInsLast(sysotList, (char *)
		      SysotCons(LacGSystem(arpLaction),
				(TRANS *) NULL, (DOMAIN) MINDOMAIN));
    if (LstNext(ActGUsed(arpAction)))
	{
	    if (LstFirst(LacGLisTransitions(arpLaction)) !=
		LstLast(LacGLisTransitions(arpLaction)))
		(void) LstInsFirst(goList, (char *)
				   LngCons((unsigned long) -1L));
	    LSTFOR (LacGLisTransitions(arpLaction), arpTrans, TRANS)
		{
		    if (LstNext(LacGLisTransitions(arpLaction)))
			{
			    NetOutput(netFile, "        if (y%lu != ",
				      (unsigned long) LacGSystem(arpLaction));
			    NetOutput(netFile, DOMFORMAT,
				      TransGSource(arpTrans));
			    NetOutput(netFile, ") goto g%lu;\n", arpCounter);
			    *((unsigned long *) LstFirst(goList)) =
				arpCounter++;
			}
		    SysotSTrans((SYSOT *) LstLast(sysotList), arpTrans);
		    arpVars[arpIndex] = TransGSource(arpTrans);
		    (void) LstStep(ActGUsed(arpAction));
		    arpIndex++;
		    Instances(LstGet(ActGUsed(arpAction)));
		    arpIndex--;
		    LstSet(ActGUsed(arpAction), lacCursor);
		    arpLaction = (LACTION *) LstThis(ActGUsed(arpAction));
		    if (LstNext(LacGLisTransitions(arpLaction)))
			{
			    NetOutput(netFile,
				      "        goto end;\n    g%lu: ;\n",
				      *((unsigned long *) LstFirst(goList)));
			    goEnd = TRUE;
			}
		}
	    if (LstFirst(LacGLisTransitions(arpLaction)) !=
		LstLast(LacGLisTransitions(arpLaction)))
		(void) LstDelFirst(goList, TRUE);
	}
    else
	{
	    LSTFOR (LacGLisTransitions(arpLaction), arpTrans, TRANS)
		{
		    NetOutput(netFile, "        if (y%lu == ",
			      (unsigned long) LacGSystem(arpLaction));
		    NetOutput(netFile, DOMFORMAT, TransGSource(arpTrans));
		    NetOutput(netFile, ") {");
		    SysotSTrans((SYSOT *) LstLast(sysotList), arpTrans);
		    arpVars[arpIndex] = TransGSource(arpTrans);
		    (void) LstBegin(sysotList);
		    arpIndex++;
		    Accepted(LstGet(sysotList));
		    arpIndex--;
		    if (LstNext(LacGLisTransitions(arpLaction)))
			{
			    NetOutput(netFile, " goto end; }\n");
			    goEnd = TRUE;
			}
		    else
			NetOutput(netFile, " }\n");
		}
	}
    (void) LstDelLast(sysotList, TRUE);
    if (arpLaction == (LACTION *) LstFirst(ActGUsed(arpAction)))
	{
	    if (goEnd)
		{ NetOutput(netFile, "    end: ; }\n#endtr\n"); }
	    else
		{ NetOutput(netFile, "    }\n#endtr\n"); }
	}
}

/***************************************************************************
*
* Function:  ReadLTS
* Purpose:   This routine reads an LTS from a file.
* Arguments: name - the name of the LTS file
*            number - the number of the LTS
*            file - the LTS file
*            tstr - the LTS is a tester if and only if tstr is TRUE
* Returns:
* See also:  ReadMonitor, ReadHidden
*
****************************************************************************
*/
FE4(void, ReadLTS,
    char *, name,
    DOMAIN, number,
    FILE *, file,
    BOOL, tstr)
{
    unsigned szBuffer;
    char *buffer;
    char *s;
    char *t;
    DOMAIN root;
    DOMAIN source;
    DOMAIN target;
    unsigned actNumber;
    BOOL rdRoot;
    BOOL rdSymbols;
    BOOL rdArcs;
    BOOL bruteStat;
    unsigned long lineNumber;
    unsigned long a = 0L;

    rdRoot = FALSE;
    rdSymbols = FALSE;
    rdArcs = FALSE;
    lineNumber = 0L;
    szBuffer = (unsigned) BUFSIZ;
    buffer = (char *) Allocate(szBuffer);
    bruteStat = TRUE;
 PLAIN: ;
    if (!bruteStat)
	goto END_OF_FILE;
    bruteStat = BruteReadln(&buffer, &szBuffer, file);
    if ((!bruteStat) && !*buffer)
	goto END_OF_FILE;
    for (s = buffer; *s != '\0'; s++)
	{
	    if (isalpha((UCHAR) *s) && isupper((UCHAR) *s))
		*s = (char) tolower((UCHAR) *s);
	}
    lineNumber++;
    s = buffer;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((s[0] == 'r') && (s[1] == 'o') && (s[2] == 'o') && (s[3] == 't'))
	{
	    if (!rdRoot)
		{
		    rdRoot = TRUE;
		    s += 4;
		    goto ROOT;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nROOT multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(23);
	}
    if ((s[0] == 's') && (s[1] == 'y') && (s[2] == 'm') && (s[3] == 'b') &&
	(s[4] == 'o') && (s[5] == 'l') && (s[6] == 's'))
	{
	    if (!rdSymbols)
		{
		    rdSymbols = TRUE;
		    s += 7;
		    goto SYMBOLS;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nSYMBOLS multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(23);
	}
    if ((s[0] == 'a') && (s[1] == 'r') && (s[2] == 'c') && (s[3] == 's'))
	{
	    if (!rdArcs)
		{
		    rdArcs = TRUE;
		    s += 4;
		    goto ARCS;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nARCS multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(23);
	}
    goto PLAIN;
 ROOT: ;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((*s == '\0') || (*s == '\n') || (*s == '_') || (*s == '$')
	|| isalnum((UCHAR) *s))
	goto BAD_ROOT;
    t = s + 1;
    while (*t != *s)
	{
	    if ((*t == '\n') || (*t == '\0'))
		goto BAD_ROOT;
	    t++;
	}
    *t = '\0';
    s++;
    if (ScanLdu(s, &a) != 1)
	goto BAD_ROOT;
    root = (DOMAIN) a;
    if (tstr)
	{
	    if (root < tstrMin)
		tstrMin = root;
	    if (root > tstrMax)
		tstrMax = root;
	}
    else
	{
	    if (root < arpMin)
		arpMin = root;
	    if (root > arpMax)
		arpMax = root;
	}
    (void) LstInsLast(RootList(), (char *) DomCons(root));
    goto PLAIN;
 SYMBOLS: ;
    if (!bruteStat)
	goto BAD_SYMBOLS;
    bruteStat = BruteReadln(&buffer, &szBuffer, file);
    if ((!bruteStat) && !*buffer)
	goto BAD_SYMBOLS;
    for (s = buffer; *s != '\0'; s++)
	{
	    if (isalpha((UCHAR) *s) && isupper((UCHAR) *s))
		*s = (char) tolower((UCHAR) *s);
	}
    lineNumber++;
    s = buffer;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') && (s[3] == '_') &&
	(s[4] == 's') && (s[5] == 'y') && (s[6] == 'm') && (s[7] == 'b') &&
	(s[8] == 'o') && (s[9] == 'l') && (s[10] == 's'))
	goto PLAIN;
    if (ScanLdu(s, &a) != 1)
	goto SYMBOLS;
    actNumber = (unsigned) a;
    while ((*s == '-') || (*s == '+') || (*s == ' ') || (*s == '\t'))
	s++;
    while (isdigit((UCHAR) *s))
	s++;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((*s != '_') && (*s != '$') && !isalpha((UCHAR) *s))
	goto BAD_SYMBOLS;
    t = s;
    while ((*t == '_') || (*t == '$') || isalnum((UCHAR) *t))
	t++;
    *t = '\0';
    if (strcmp(s, IName()) == 0)
	{
	    if (actNumber != ICode())
		{
		    ClDeallocate(buffer);
		    (void)
			fprintf(_err,
				"\n%s, line %lu: %s must have the code %u\n",
				name, lineNumber, IName(), ICode());
		    (void) fflush(_err);
		    Exit(26);
		}
	}
    else if (actNumber == ICode())
	{
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\n%s, line %lu: only %s can have the code %u\n",
			   name, lineNumber, IName(), ICode());
	    (void) fflush(_err);
	    Exit(26);
	}
    arpAction = (ACTION *) BalTrElement(actNumTree, (char *) &actNumber);
    if (arpAction)
	{
	    (void)
		fprintf
		(_err,
		 "\n%s, line %lu: %u cannot be the code of both\n%s and %s in a single LTS\n",
		 name, lineNumber, actNumber,
		 ActGName(arpAction), s);
	    (void) fflush(_err);
	    ClDeallocate(buffer);
	    Exit(26);
	}
    arpAction = (ACTION *) BalTrElement(actTree, (char *) s);
    if (arpAction)
	{
	    arpLaction = (LACTION *) LstLast(ActGUsed(arpAction));
	    if (LacGSystem(arpLaction) == number)
		{
		    if (ActGNumber(arpAction) == actNumber)
			goto SYMBOLS;
		    (void)
			fprintf
			(_err,
			 "\n%s, line %lu: %s cannot have both the code %u\nand the code %u in a single LTS\n",
			 name, lineNumber, s,
			 ActGNumber(arpAction), actNumber);
		    (void) fflush(_err);
		    ClDeallocate(buffer);
		    Exit(26);
		}
	    ActSNumber(arpAction, actNumber);
	}
    else
	{
	    arpAction = ActCons(actNumber, StrCopy(s),
				(!OnTheFly()) && (actNumber == ICode()));
	    (void) BalTrInsert(actTree, (char *) arpAction, (char *) s, TRUE);
	    if (actNumber == ICode())
		SetIAction(arpAction);
	}
    (void) LstInsLast(ActGUsed(arpAction), (char *) LacCons(number));
    (void) BalTrInsert(actNumTree, (char *) arpAction,
		       (char *) &actNumber, TRUE);
    goto SYMBOLS;
 ARCS: ;
    if (!bruteStat)
	goto BAD_ARCS;
    bruteStat = BruteReadln(&buffer, &szBuffer, file);
    if ((!bruteStat) && !*buffer)
	goto BAD_ARCS;
    for (s = buffer; *s != '\0'; s++)
	{
	    if (isalpha((UCHAR) *s) && isupper((UCHAR) *s))
		*s = (char) tolower((UCHAR) *s);
	}
    lineNumber++;
    s = buffer;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') && (s[3] == '_') &&
	(s[4] == 'a') && (s[5] == 'r') && (s[6] == 'c') && (s[7] == 's'))
	goto PLAIN;
    if (ScanLdu(s, &a) != 1)
	goto ARCS;
    source = (DOMAIN) a;
    while ((*s == '-') || (*s == '+') || (*s == ' ') || (*s == '\t'))
	s++;
    while (isdigit((UCHAR) *s))
	s++;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if (ScanLdu(s, &a) != 1)
	goto BAD_ARCS;
    target = (DOMAIN) a;
    while ((*s == '-') || (*s == '+') || (*s == ' ') || (*s == '\t'))
	s++;
    while (isdigit((UCHAR) *s))
	s++;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if (ScanLdu(s, &a) != 1)
	goto BAD_ARCS;
    actNumber = (unsigned) a;
    arpAction = (ACTION *) BalTrElement(actNumTree, (char *) &actNumber);
    if (!arpAction)
	{
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\n%s, line %lu: no action has the code %u\n",
			   name, lineNumber, actNumber);
	    (void) fflush(_err);
	    Exit(27);
	}
    arpLaction = (LACTION *) LstLast(ActGUsed(arpAction));
    arpTrans = (TRANS *) BalTrElement(LacGBalTransitions(arpLaction),
				      (char *) &source);
    if (!arpTrans)
	{
	    arpTrans = (TRANS *) TransCons(source);
	    (void) BalTrInsert(LacGBalTransitions(arpLaction),
			       (char *) arpTrans, (char *) &source, TRUE);
	    if (tstr)
		{
		    if (source < tstrMin)
			tstrMin = source;
		    if (source > tstrMax)
			tstrMax = source;
		}
	    else
		{
		    if (source < arpMin)
			arpMin = source;
		    if (source > arpMax)
			arpMax = source;
		}
	}
    arpTarget = DomCons(target);
    if (BalTrInsert(TransGBalTargets(arpTrans),
		    (char *) arpTarget, (char *) arpTarget, TRUE))
	ClDeallocate(arpTarget);
    else
	{
	    if (tstr)
		{
		    if (target < tstrMin)
			tstrMin = target;
		    if (target > tstrMax)
			tstrMax = target;
		}
	    else
		{
		    if (target < arpMin)
			arpMin = target;
		    if (target > arpMax)
			arpMax = target;
		}
	}
    goto ARCS;
 BAD_ROOT: ;
    ClDeallocate(buffer);
    (void) fprintf(_err, "\nBad ROOT in %s, line %lu\n", name, lineNumber);
    (void) fflush(_err);
    Exit(24);
 BAD_SYMBOLS: ;
    ClDeallocate(buffer);
    (void) fprintf(_err, "\nBad SYMBOLS in %s, line %lu\n",
		   name, lineNumber);
    (void) fflush(_err);
    Exit(24);
 BAD_ARCS: ;
    ClDeallocate(buffer);
    (void) fprintf(_err, "\nBad ARCS in %s, line %lu\n", name, lineNumber);
    (void) fflush(_err);
    Exit(24);
 END_OF_FILE: ;
    ClDeallocate(buffer);
    if (rdRoot)
	return;
    (void) fprintf(_err, "\nNo ROOT in %s\n", name);
    (void) fflush(_err);
    Exit(25);
}

/***************************************************************************
*
* Function:  ReadMonitor
* Purpose:   This routine reads information about monitor states
*            of a tester from a file.
* Arguments: name - the name of the file
*            file - the file
* Returns:
* See also:  ReadLTS, ReadHidden
*
****************************************************************************
*/
FE2(void, ReadMonitor,
    char *, name,
    FILE *, file)
{
    unsigned szBuffer;
    char *buffer;
    char *s;
    BOOL bruteStat;
    DOMAIN state;
    DOMAIN high;
    unsigned rdRejects;
    unsigned rdDeadlocks;
    unsigned rdLivelocks;
    unsigned rdInfinites;
    unsigned long lineNumber;
    unsigned long a = 0L;

    rdRejects = 0;
    rdDeadlocks = 0;
    rdLivelocks = 0;
    rdInfinites = 0;
    lineNumber = 0L;
    state = high = (DOMAIN) MINDOMAIN;
    szBuffer = (unsigned) BUFSIZ;
    buffer = (char *) Allocate(szBuffer);
    bruteStat = TRUE;
 PLAIN: ;
    if (!bruteStat)
	{
	    ClDeallocate(buffer);
	    return;
	}
    bruteStat = BruteReadln(&buffer, &szBuffer, file);
    if ((!bruteStat) && !*buffer)
	{
	    ClDeallocate(buffer);
	    return;
	}
    for (s = buffer; *s != '\0'; s++)
	{
	    if (isalpha((UCHAR) *s) && isupper((UCHAR) *s))
		*s = (char) tolower((UCHAR) *s);
	}
    lineNumber++;
    s = buffer;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((*s == '\0') || (*s == '\n') || ((s[0] == '-') && (s[1] == '-')))
	goto PLAIN;
    if ((s[0] == 'r') && (s[1] == 'e') && (s[2] == 'j') && (s[3] == 'e') &&
	(s[4] == 'c') && (s[5] == 't') && (s[6] == 's'))
	{
	    rdRejects += 2;
	    if (rdRejects == 2)
		{
		    s += 7;
		    goto MONITOR;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nREJECTS multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(29);
	}
    if ((s[0] == 'd') && (s[1] == 'e') && (s[2] == 'a') && (s[3] == 'd') &&
	(s[4] == 'l') && (s[5] == 'o') && (s[6] == 'c') && (s[7] == 'k') &&
	(s[8] == 's'))
	{
	    rdDeadlocks += 2;
	    if (rdDeadlocks == 2)
		{
		    s += 9;
		    goto MONITOR;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nDEADLOCKS multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(29);
	}
    if ((s[0] == 'l') && (s[1] == 'i') && (s[2] == 'v') && (s[3] == 'e') &&
	(s[4] == 'l') && (s[5] == 'o') && (s[6] == 'c') && (s[7] == 'k') &&
	(s[8] == 's'))
	{
	    rdLivelocks += 2;
	    if (rdLivelocks == 2)
		{
		    s += 9;
		    goto MONITOR;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nLIVELOCKS multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(29);
	}
    if ((s[0] == 'i') && (s[1] == 'n') && (s[2] == 'f') && (s[3] == 'i') &&
	(s[4] == 'n') && (s[5] == 'i') && (s[6] == 't') && (s[7] == 'e') &&
	(s[8] == 's'))
	{
	    rdInfinites += 2;
	    if (rdInfinites == 2)
		{
		    s += 9;
		    goto MONITOR;
		}
	    ClDeallocate(buffer);
	    (void) fprintf(_err,
			   "\nINFINITES multiply defined in %s, line %lu\n",
			   name, lineNumber);
	    (void) fflush(_err);
	    Exit(29);
	}
    goto BAD_MONITOR;
 MONITOR: ;
    if (!bruteStat)
	goto BAD_EOF;
    bruteStat = BruteReadln(&buffer, &szBuffer, file);
    if ((!bruteStat) && !*buffer)
	goto BAD_EOF;
    for (s = buffer; *s != '\0'; s++)
	{
	    if (isalpha((UCHAR) *s) && isupper((UCHAR) *s))
		*s = (char) tolower((UCHAR) *s);
	}
    lineNumber++;
    s = buffer;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((*s == '\0') || (*s == '\n') || ((s[0] == '-') && (s[1] == '-')))
	goto MONITOR;
    if (rdRejects == 2)
	{
	    if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') &&
		(s[3] == '_') && (s[4] == 'r') && (s[5] == 'e') &&
		(s[6] == 'j') && (s[7] == 'e') && (s[8] == 'c') &&
		(s[9] == 't') && (s[10] == 's'))
		{
		    rdRejects = 1;
		    if (Reject())
			{
			    NetOutput(netFile, ")");
			}
		    goto PLAIN;
		}
	}
    else if (rdDeadlocks == 2)
	{
	    if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') &&
		(s[3] == '_') && (s[4] == 'd') && (s[5] == 'e') &&
		(s[6] == 'a') && (s[7] == 'd') && (s[8] == 'l') &&
		(s[9] == 'o') && (s[10] == 'c') && (s[11] == 'k') &&
		(s[12] == 's'))
		{
		    rdDeadlocks = 1;
		    if (StabFail())
			{
			    NetOutput(netFile, ")");
			}
		    goto PLAIN;
		}
	}
    else if (rdLivelocks == 2)
	{
	    if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') &&
		(s[3] == '_') && (s[4] == 'l') && (s[5] == 'i') &&
		(s[6] == 'v') && (s[7] == 'e') && (s[8] == 'l') &&
		(s[9] == 'o') && (s[10] == 'c') && (s[11] == 'k') &&
		(s[12] == 's'))
		{
		    rdLivelocks = 1;
		    if (Divergence())
			{
			    NetOutput(netFile, ")");
			}
		    goto PLAIN;
		}
	}
    else if ((s[0] == 'e') && (s[1] == 'n') && (s[2] == 'd') &&
	     (s[3] == '_') && (s[4] == 'i') && (s[5] == 'n') &&
	     (s[6] == 'f') && (s[7] == 'i') && (s[8] == 'n') &&
	     (s[9] == 'i') && (s[10] == 't') && (s[11] == 'e') &&
	     (s[12] == 's'))
	{
	    rdInfinites = 1;
	    if (Infinite())
		{
		    NetOutput(netFile, ")");
		}
	    goto PLAIN;
	}
    if (ScanLdu(s, &a) != 1)
	goto BAD_MONITOR;
    state = (DOMAIN) a;
    if ((state < tstrMin) || (state > tstrMax))
	goto BAD_STATE;
    if (rdRejects == 2)
	{
	    if (Reject())
		{ NetOutput(netFile, "+"); }
	    else
		{
		    SetReject();
		    NetOutput(netFile, " \\\nreject(");
		}
	}
    else if (rdDeadlocks == 2)
	{
	    if (StabFail())
		{ NetOutput(netFile, "+"); }
	    else
		{
		    SetStabFail();
		    NetOutput(netFile, " \\\ndeadlock(");
		}
	}
    else if (rdLivelocks == 2)
	{
	    if (Divergence())
		{ NetOutput(netFile, "+"); }
	    else
		{
		    SetDivergence();
		    NetOutput(netFile, " \\\nlivelock(");
		}
	}
    else if (Infinite())
	{ NetOutput(netFile, "+"); }
    else
	{
	    SetInfinite();
	    NetOutput(netFile, " \\\ninfinite(");
	}
    while ((*s == '-') || (*s == '+') || (*s == ' ') || (*s == '\t'))
	s++;
    while (isdigit((UCHAR) *s))
	s++;
    while ((*s == ' ') || (*s == '\t'))
	s++;
    if ((*s == '\0') || (*s == '\n') || ((s[0] == '-') && (s[1] == '-')))
	high = state;
    else
	{
	    if ((s[0] != '.') && (s[1] != '.'))
		goto BAD_MONITOR;
	    s += 2;
	    while ((*s == ' ') || (*s == '\t'))
		s++;
	    if (ScanLdu(s, &a) != 1)
		goto BAD_MONITOR;
	    high = (DOMAIN) a;
	    if ((high < tstrMin) || (high > tstrMax))
		{
		    state = high;
		    goto BAD_STATE;
		}
	    if (state > high)
		goto BAD_RANGE;
	}
    NetOutput(netFile, "<.");
    NetOutput(netFile, DOMFORMAT, state);
    if (state != high)
	{
	    NetOutput(netFile, "..");
	    NetOutput(netFile, DOMFORMAT, high);
	}
    NetOutput(netFile, ".>");
    while (TRUE)
	{
	    TupSField(arpSingle, (unsigned) 0, state);
	    (void) TupEncode(MonPlInd(), arpSingle,
			     EngABuffer(), EngABufSize());
	    arpPlc = GetPlc(EngBuffer());
	    if (rdRejects == 2)
		PlcSReject(arpPlc);
	    else if (rdDeadlocks == 2)
		PlcSDeadlock(arpPlc);
	    else if (rdLivelocks == 2)
		PlcSLivelock(arpPlc);
	    else
		PlcSInfinite(arpPlc);
	    if (state == high)
		break;
	    state++;
	}
    goto MONITOR;
 BAD_MONITOR: ;
    ClDeallocate(buffer);
    (void) fprintf(_err,
		   "\nUnexpected input in %s at line %lu\n",
		   name, lineNumber);
    (void) fflush(_err);
    Exit(30);
 BAD_EOF: ;
    ClDeallocate(buffer);
    (void) fprintf(_err,
		   "\nUnexpected end of file in %s at line %lu\n",
		   name, lineNumber);
    (void) fflush(_err);
    Exit(30);
 BAD_STATE: ;
    ClDeallocate(buffer);
    (void) fprintf(_err, "\n%s, line %lu: ", name, lineNumber);
    (void) fprintf(_err, DOMFORMAT, state);
    (void) fputs(" is not in range ", _err);
    (void) fprintf(_err, DOMFORMAT, tstrMin);
    (void) fputs("..", _err);
    (void) fprintf(_err, DOMFORMAT, tstrMax);
    (void) fputs("\n", _err);
    (void) fflush(_err);
    Exit(31);
 BAD_RANGE: ;
    ClDeallocate(buffer);
    (void) fprintf(_err, "\n%s, line %lu: ", name, lineNumber);
    (void) fprintf(_err, DOMFORMAT, state);
    (void) fputs(" > ", _err);
    (void) fprintf(_err, DOMFORMAT, high);
    (void) fputs("\n", _err);
    (void) fflush(_err);
    Exit(31);
}

/***************************************************************************
*
* Function:  ReadHidden
* Purpose:   This routine reads from a file the names of the actions to be
*            made invisible. This routine is called only if there is no
*            tester. 
* Arguments: name - the name of the file
*            file - the file
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, ReadHidden,
    char *, name,
    FILE *, file)
{
    unsigned szBuffer;
    char *buffer;
    int ch;
    unsigned long lineNumber;
    char *s = NULL;

    lineNumber = 1L;
    szBuffer = (unsigned) BUFSIZ;
    buffer = (char *) Allocate(szBuffer);
 PLAIN: ;
    if ((ch = fgetc(file)) == EOF)
	{
	    ClDeallocate(buffer);
	    return;
	}
    if (ch == (int) '-')
	goto MINUS;
    if ((ch == (int) '_') || (ch == (int) '$') || isalpha((UCHAR) ch))
	{
	    if (isalpha((UCHAR) ch) && isupper((UCHAR) ch))
		ch = tolower((UCHAR) ch);
	    s = buffer;
	    goto ACT;
	}
    if (ch == (int) '\n')
	lineNumber++;
    goto PLAIN;
 MINUS: ;
    if ((ch = fgetc(file)) == EOF)
	{
	    ClDeallocate(buffer);
	    return;
	}
    if (ch == (int) '-')
	goto COMMENT;
    if (ch == (int) '\n')
	lineNumber++;
    goto PLAIN;
 COMMENT: ;
    if ((ch = fgetc(file)) == EOF)
	{
	    ClDeallocate(buffer);
	    return;
	}
    if (ch != (int) '\n')
	goto COMMENT;
    lineNumber++;
    goto PLAIN;
 ACT: ;
    if (s == buffer + szBuffer)
	{
	    szBuffer = (szBuffer >> 1) + szBuffer + 1;
	    buffer = (char *) Reallocate(buffer, szBuffer);
	}
    *s++ = (char) ch;
    if ((ch = fgetc(file)) == EOF)
	{
	    if (s == buffer + szBuffer)
		{
		    szBuffer = (szBuffer >> 1) + szBuffer + 1;
		    buffer = (char *) Reallocate(buffer, szBuffer);
		}
	    *s = '\0';
	    arpAction = (ACTION *) BalTrElement(actTree, (char *) buffer);
	    if (arpAction)
		ActHide(arpAction);
	    else
		{
		    (void) fprintf(_err,
				   "\n%s, line %lu: %s\nis no action\n",
				   name, lineNumber, buffer);
		    (void) fflush(_err);
		}
	    ClDeallocate(buffer);
	    return;
	}
    if ((ch == (int) '_') || (ch == (int) '$') || isalnum((UCHAR) ch))
	{
	    if (isalpha((UCHAR) ch) && isupper((UCHAR) ch))
		ch = tolower((UCHAR) ch);
	    goto ACT;
	}
    if (s == buffer + szBuffer)
	{
	    szBuffer = (szBuffer >> 1) + szBuffer + 1;
	    buffer = (char *) Reallocate(buffer, szBuffer);
	}
    *s = '\0';
    arpAction = (ACTION *) BalTrElement(actTree, (char *) buffer);
    if (arpAction)
	ActHide(arpAction);
    else
	{
	    (void) fprintf(_err,
			   "\n%s, line %lu: %s\nis no action\n",
			   name, lineNumber, buffer);
	    (void) fflush(_err);
	}
    if (ch == (int) '-')
	goto MINUS;
    if (ch == (int) '\n')
	lineNumber++;
    goto PLAIN;
}

/***************************************************************************
*
* Function:  TrnFold
* Purpose:   This routine puts the correspondence between trn and an
*            action into a correspondence table.
* Arguments: trn - a transition in a low level net
* Returns:
* See also:  TreeTrnFold, LTSOutput (graph.c)
*
****************************************************************************
*/
FE1(BOOL, TrnFold,
    TRN *, trn)
{
    arpPointer = TrnGKey(trn);
    GetValue(unsigned, arpHiTr, arpPointer);
    PtrBlPut(ArpFoldings(), (char *) (actTbl[arpHiTr]), arpCounter);
    arpCounter++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  TreeTrnFold
* Purpose:   This routine constructs a table that maps each transition of
*            a low level net to the corresponding action.
* Arguments:
* Returns:
* See also:  TrnFold, LTSOutput (graph.c)
*
****************************************************************************
*/
FE0(void, TreeTrnFold)
{
    _arpFoldings = PtrBlCreate(BalTrGElCount(TrnBalTree(0)), &actTools);
    arpCounter = 0L;
    (void) BalTrApply(TrnBalTree(0), TrnFold);
}

/***************************************************************************
*
* Function:  Misusage
* Purpose:   This routine handles most of the cases where the command line
*            of the program is incorrect.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, Misusage)
{
    MsgExit("[-.] [-B] [-binteger] [-C] [-c]\n[-D] [-d] [-E] [-e] [-f] [-ginteger] [-hname]... [-Mname] [-minteger]\n-oname [-rinteger] [-S] [-s] [-Tname] [[-t]name]... [-x] [-y] [-Z] [-z]",
	    (int) 1);
}

FE0(void, AraMemClose)
{
    GphClose(theGraph);
    ClFclose(netFile);
    ClFclose(auxFile);
    PtClose();
    BaseLTSClose();
    DynamClose();
    PthDiscard(thePath);
    PthDiscard(the2Path);
    PthDiscard(the3Path);
    PthDiscard(nowPath);
    StaDiscard(theState);
    StaDiscard(theFolState);
    StaDiscard(theEvalState);
    BalTrDestroy(theNodeTree, FALSE);
    BalTrDestroy(actNumTree, FALSE);
    BalTrDestroy(actTree, lacB);
    BalTrDestroy(flowTree, TRUE);
    LstDestroy(hideList, FALSE);
    LstDestroy(ltsList, FALSE);
    LstDestroy(hideList, FALSE);
    LstDestroy(sysotList, TRUE);
    LstDestroy(goList, TRUE);
    LstDestroy(_rootList, TRUE);
    LstDestroy(_actList, TRUE);
    PtrBlDestroy(_arpFoldings, FALSE);
    ClDeallocate(actTbl);
    ClDeallocate(auxName);
    ClDeallocate(_iName);
    EngDBuffer();
    TupDiscard(arpSingle);
    TupDiscard(arpDouble);
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments: argc - the number of command line arguments given to the
*                   program
*            argv - an array containing the arguments
* Returns:   0 if everything succeeds, and something else otherwise
*            (PROD has no systematic convention for non-zero exits.)
* See also:
*
****************************************************************************
*/
MAIN2()
{
    int argc = 0;
    char **argv = NULL;
    BINARYFILE bFile = NonBinaryFile();
    int i;
    unsigned long gCount = 0L, a = 0L, seedVal = 0L, gphGCount, createFlag;
    unsigned bits = 0, u, turn;
    DOMAIN systNumber, *root;
    char *argument, *bs, *gs, *ms, *rs, *name, *outName, *tstrName, *monName;
    BOOL bdds, cont;
    unsigned long mapLen = 0L;

#ifndef UNIX
    ClockOn();
#endif
    SetIO();
    ClearErrno();
    baseName = StrCopy("araprod");
#ifdef ARGV0
    myName = CmdNm(_argv_[0]);
#else
    myName = baseName;
#endif
#ifndef UNIX
    StoreCmd(_argc_, myName, (char **) _argv_);
#endif
    ExpArgs(_argc_, (char **) _argv_, &argc, &argv);
    IgnoreSignals();
    continueGeneration = TRUE;
    stopFlag = FALSE;
    SetSigHandlers();
    _memClose = (VOF) AraMemClose;
    DynamInit();
    outName = NULL;
    tstrName = NULL;
    monName = NULL;
    bs = NULL;
    gs = NULL;
    ms = NULL;
    rs = NULL;
    bdds = FALSE;
    cont = FALSE;
    theDots = FALSE;
    AraSetProd();
    ResetAra();
    ResetPep();
    ClElimination();
    ResetDfs();
    ResetFreeExpansion();
    ClPtStub();
    ClPtSymm();
    ClPtVerbose();
    ClPtSleep();
    ResetPlcIMkSymm();
    ResetPlRnd();
    ClPtLazy();
    ResetTrRnd();
    ClPtDelAlg();
    ClPtTryAgain();
    ClPtCFFD();
    SetPtDefVis();
    ClOnTheFly();
    ClBlocked();
#ifdef SMODELS
    ClPtSmodels();
    ClPtLongPlay();
#endif
    ltsList = LstCreate(&strTools);
    hideList = LstCreate(&strTools);
    for (i = 1; i < argc; i++)
	{
	    argument = argv[i];
	    if (*argument == '-')
		{
		    argument++;
		    if ((*argument == '\0') &&
			(i + 1 == argc))
			Misusage();
		    if (*argument == '\0')
			{
			    argument = argv[i + 1];
			    i++;
			}
		    while (*argument != '\0')
			{
			    if (*argument == '.')
				theDots = TRUE;
			    else if (*argument == 'B')
				bdds = TRUE;
			    else if (*argument == 'b')
				{
				    argument++;
				    if ((bs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    bs = argument;
				    if (ScanLdu(bs, &a) != 1)
					Misusage();
				    else
					bits = (unsigned) a;
				    break;
				}
			    else if (*argument == 'C')
				{
				    SetPtCFFD();
				    SetPtStub();
				    SetDfs();
				}
			    else if (*argument == 'c')
				cont = TRUE;
			    else if (*argument == 'D')
				SetDfs();
			    else if (*argument == 'd')
				{
				    SetPtStub();
				    SetPtDelAlg();
				}
			    else if (*argument == 'E')
				{
				    SetPtSymm();
				    SetPlcIMkSymm();
				}
			    else if (*argument == 'e')
				SetPtSymm();
			    else if (*argument == 'f')
				{
				    SetDfs();
				    SetFreeExpansion();
				}
			    else if (*argument == 'g')
				{
				    argument++;
				    if ((gs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    gs = argument;
				    if (ScanLdu(gs, &gCount) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'h')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsLast(hideList,
						      (char *) argument);
				    break;
				}
			    else if (*argument == 'M')
				{
				    argument++;
				    if ((monName != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    monName = argument;
				    break;
				}
			    else if (*argument == 'm')
				{
				    argument++;
				    if ((ms != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    ms = argument;
				    if (ScanLdu(ms, &mapLen) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'o')
				{
				    argument++;
				    if ((outName != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    outName = argument;
				    break;
				}
			    else if (*argument == 'r')
				{
				    argument++;
				    if ((rs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    rs = argument;
				    if (ScanLdu(rs, &seedVal) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'S')
				SetPtSleep();
			    else if (*argument == 's')
				SetPtStub();
			    else if (*argument == 'T')
				{
				    argument++;
				    if ((tstrName != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    tstrName = argument;
				    break;
				}
			    else if (*argument == 't')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsLast(ltsList,
						      (char *) argument);
				    break;
				}
			    else if (*argument == 'x')
				SetPlRnd();
			    else if (*argument == 'y')
				SetPtLazy();
			    else if (*argument == 'Z')
				SetPtTryAgain();
			    else if (*argument == 'z')
				{
				    SetPtLazy();
				    SetTrRnd();
				}
			    else
				Misusage();
			    argument++;
			}
		}
	    else
		(void) LstInsLast(ltsList, (char *) argument);
	}
    if (outName == NULL)
	Misusage();
    if (tstrName != NULL)
	{
	    if (monName == NULL)
		{
		    (void) fputs("\n-T is illegal without -M\n", _err);
		    Exit(1);
		}
	    SetOnTheFly();
	    SetDfs();
	    if (!LstEmpty(hideList))
		{
		    (void)
			fputs
			("\nActions are not hidden in on-the-fly verification\n",
			 _err);
		    (void) fflush(_err);
		}
	    LstDestroy(hideList, FALSE);
	    (void) LstInsLast(ltsList, (char *) tstrName);
	}
    else if (monName != NULL)
	{
	    (void) fputs("\n-M is illegal without -T\n", _err);
	    (void) fflush(_err);
	    Exit(1);
	}
    i = (int) strlen(outName);
    if ((i > 4) && (strcmp(outName + i - 4, ".lts") == 0))
	outName[i - 4] = '\0';
    if (gs == NULL)
	gCount = (unsigned long) -1L;
    else if (gCount == 0L)
	{
	    gCount = 1L;
	    (void) fprintf(_err, "\n-g %s transformed to -g %lu\n",
			   gs, gCount);
	    (void) fflush(_err);
	}
    if (cont) 
	{
	    theGraph = GphOpen(outName, GPH_GENERATE, FALSE);
	    if (theGraph == NULL)
		{
		    Exit(4);
		}
	    if (bdds)
		{
		    (void)
			fputs
			("\nToo late for BDD-based graph generation\n",
			 _err);
		    (void) fflush(_err);
		    bdds = FALSE;
		}
	    if (!GphUnfold(theGraph))
		{
		    (void) fputs("\nToo late for unfolding\n", _err);
		    (void) fflush(_err);
		    Exit(11);
		}
	    if (ms != NULL)
		GphMapToMem(theGraph, mapLen);
	    u = GphBits(theGraph);
	    if ((bs != NULL) && (bits != u))
		{
		    (void) fprintf(_err,
				   "\n-b %s ignored (old hash table used)\n",
				   bs);
		    (void) fflush(_err);
		}
	    if (GphSymm(theGraph))
		{
		    if (GphIMkSymm(theGraph))
			{
			    if (PtSymm() && !PlcIMkSymm())
				{
				    (void)
					fputs
					("\nStill requiring symmetricity of the initial marking\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetPlcIMkSymm();
			}
		    else if (PlcIMkSymm())
			{
			    (void)
				fputs
				("\nToo late to require symmetricity of the initial marking\n",
				 _err);
			    (void) fflush(_err);
			    ResetPlcIMkSymm();
			}
		    SetPtSymm();
		}
	    else if (PtSymm())
		{
		    (void) fputs("\nToo late for symmetry method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtSymm();
		}
	    if (GphStub(theGraph))
		{
		    if (GphDelAlg(theGraph))
			{
			    if (PtStub() && !PtDelAlg())
				{
				    (void)
					fputs
					("\nContinuing with deletion algorithm\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetPtDelAlg();
			}
		    else if (PtDelAlg())
			{
			    (void)
				fputs
				("\nContinuing with incremental algorithm\n",
				 _err);
			    (void) fflush(_err);
			    ClPtDelAlg();
			}
		    if (GphCFFD(theGraph))
			{
			    if (PtLazy() && TrRnd() && !PtDelAlg())
				{
				    (void)
					fputs
					("\n-z is ignored because of CFFD-preserving incremental algorithm\n",
					 _err);
				    (void) fflush(_err);
				    ClPtLazy();
				    ResetTrRnd();
				}
			    SetPtCFFD();
			    SetDfs();
			}
		    else if (PtCFFD())
			{
			    (void)
				fputs
				("\nToo late for CFFD-preservation\n",
				 _err);
			    (void) fflush(_err);
			    ClPtCFFD();
			}
		    SetPtStub();
		}
	    else if (PtStub())
		{
		    (void) fputs("\nToo late for stubborn set method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtStub();
		    ClPtDelAlg();
		    ClPtCFFD();
		}
	    if (GphSleep(theGraph))
		SetPtSleep();
	    else if (PtSleep())
		{
		    (void) fputs("\nToo late for sleep set method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtSleep();
		}
	    if (GphDfs(theGraph))
		{
		    if (GphFreeExpansion(theGraph))
			{
			    if (Dfs() && !FreeExpansion())
				{
				    (void)
					fputs
					("\nContinuing with free expansion\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetFreeExpansion();
			}
		    else if (FreeExpansion())
			{
			    (void) fputs("\nToo late for free expansion\n",
					 _err);
			    (void) fflush(_err);
			    ResetFreeExpansion();
			}
		    SetDfs();
		}
	    else if (Dfs())
		{
		    (void) fputs("\nToo late for depth first search\n",
				 _err);
		    (void) fflush(_err);
		    ResetDfs();
		    ResetFreeExpansion();
		}
	    if (!GphDefVis(theGraph))
		{
		    (void)
			fputs
			("\nTransition instances are invisible by default\n",
			 _err);
		    (void) fflush(_err);
		    Exit(11);
		}
	    EngSBufSize(GphBufSize(theGraph));
	    EngSBuffer((DATATYPE *) AllocArray(EngBufSize(),
					       sizeof(DATATYPE)));
	    if (rs != NULL)
		GphSetSeed(theGraph, (UNS32) seedVal);
	}
    else
	{
	    if (bs == NULL)
		bits = 10;
	    else if ((bits == 0) || (bits > 30))
		{
		    if (bits == 0)
			bits = 1;
		    else
			bits = 30;
		    (void) fprintf(_err, "\n-b %s transformed to -b %u\n",
				   bs, bits);
		    (void) fflush(_err);
		}
	    EngSBufSize((unsigned) 1024);
	    EngSBuffer((DATATYPE *) AllocArray(EngBufSize(),
					       sizeof(DATATYPE)));
	    if ((OnTheFly() || PtCFFD() || PtSleep()) && PtSymm())
		{
		    (void)
			fputs
			("\n-E and -e are ignored because of on-the-fly verification or sleep set\nmethod or CFFD-preserving stubborn set method\n",
			 _err);
		    (void) fflush(_err);
		    ClPtSymm();
		}
	    if (PtSleep() && PtCFFD())
		{
		    (void)
			fputs
			("\nCFFD-preserving stubborn set method prevents from\nusing sleep set method\n",
			 _err);
		    (void) fflush(_err);
		    ClPtSleep();
		}
	    if (PtCFFD() && PtLazy() && TrRnd() && !PtDelAlg())
		{
		    (void)
			fputs
			("\n-z is ignored because of CFFD-preserving incremental algorithm\n",
			 _err);
		    (void) fflush(_err);
		    ClPtLazy();
		    ResetTrRnd();
		}
	    if (PtSleep())
		{
		    SetDfs();
		    SetFreeExpansion();
		}
	}
    systNumber = (DOMAIN) 0L;
    if (LstEmpty(ltsList))
	_rootList = NULL;
    else
	{
	    transB = TRUE;
	    lacB = TRUE;
	    _iCode = 0;
	    _iName = StrCopy("i");
	    SetIAction((ACTION *) NULL);
	    actTree = BalTrCreate(&actBalTools);
	    _rootList = LstCreate(&domTools);
	    arpMin = (DOMAIN) MAXDOMAIN;
	    arpMax = (DOMAIN) MINDOMAIN;
	    name = (char *) LstBegin(ltsList);
	    while (TRUE)
		{
		    if (!(auxFile = MyFopen(name, "r", (BOOL *) NULL)))
			Exit(2);
		    actNumTree = BalTrCreate(&actNumBalTools);
		    ReadLTS(name, systNumber, auxFile,
			    OnTheFly() && !LstNext(ltsList));
		    ClFclose(auxFile);
		    BalTrDestroy(actNumTree, FALSE);
		    name = (char *) LstStep(ltsList);
		    if (!name)
			break;
		    systNumber++;
		}
	    if (OnTheFly())
		{
		    tstrSys = systNumber;
		    systNumber--;
		}
	}
    LstDestroy(ltsList, FALSE);
    if (!OnTheFly())
	{
	    LSTFOR (hideList, name, char)
		{
		    if (!(auxFile = MyFopen(name, "r", (BOOL *) NULL)))
			Exit(2);
		    ReadHidden(name, auxFile);
		    ClFclose(auxFile);
		}
	    LstDestroy(hideList, FALSE);
	}
    if (cont)
	netFile = (FILE *) NULL;
    else
	{
	    auxName = FileName(outName, "net");
	    if (!(netFile = MyFopen(auxName, "w", (BOOL *) NULL)))
		Exit(2);
	    ClDeallocate(auxName);
	}
    if (RootList())
	{
	    arpSingle = TupCons((unsigned) 1);
	    arpDouble = TupCons((unsigned) 2);
	    transB = FALSE;
	    lacB = FALSE;
	    _actList = LstCreate(&actTools);
	    (void) BalTrApply(actTree, ActTrLst);
	    BalTrDestroy(actTree, FALSE);
	    arpHiTr = 0;
	    if (IAction())
		{
		    if ((!OnTheFly()) ||
			(LacGSystem((LACTION *) LstFirst(ActGUsed(IAction())))
			 != tstrSys))
			arpHiTr = 1;
		    if (OnTheFly() &&
			(LacGSystem((LACTION *) LstLast(ActGUsed(IAction())))
			 == tstrSys))
			arpHiTr++;
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		if (arpAction != IAction())
		    arpHiTr++;
	    BaseLTSInit((unsigned) ((OnTheFly() && (tstrSys != (DOMAIN) 0L))
				    ? 2 : 1),
			arpHiTr);
	    if ((!OnTheFly()) || (tstrSys != (DOMAIN) 0L))
		MagIProfile((unsigned) 0, (unsigned) 2);
	    if (OnTheFly())
		MagIProfile((unsigned) ((tstrSys != (DOMAIN) 0L) ? 1 : 0),
			    (unsigned) 1);
	    PtBaseInit();
	    if ((!OnTheFly()) || (tstrSys != (DOMAIN) 0L))
		{
		    NetOutput(netFile, "#place p lo(<.");
		    NetOutput(netFile, DOMFORMAT,
#if MINDOMAIN
			      (systNumber >= (DOMAIN) 0L)
			      ? ((DOMAIN) 0L) : ((DOMAIN) MINDOMAIN)
#else
			      (DOMAIN) 0L
#endif
			      );
		    NetOutput(netFile, ", ");
		    NetOutput(netFile, DOMFORMAT, arpMin);
		    NetOutput(netFile, ".>) hi(<.");
		    NetOutput(netFile, DOMFORMAT,
#if MINDOMAIN
			      (systNumber >= (DOMAIN) 0L)
			      ? systNumber : ((DOMAIN) MAXDOMAIN)
#else
			      systNumber
#endif
			      );
		    NetOutput(netFile, ", ");
		    NetOutput(netFile, DOMFORMAT, arpMax);
		    NetOutput(netFile, ".>) mk(");
		    systNumber = (DOMAIN) 0L;
		    LSTFOR (RootList(), root, DOMAIN)
			{
			    if (OnTheFly() && !LstNext(RootList()))
				break;
			    TupSField(arpDouble, (unsigned) 0, systNumber);
			    TupSField(arpDouble, (unsigned) 1, *root);
			    (void) TupEncode((unsigned) 0, arpDouble,
					     EngABuffer(), EngABufSize());
			    PlcSCount(GetPlc(EngBuffer()), (unsigned long) 1L);
			    if (root != (DOMAIN *) LstFirst(RootList()))
				{
				    NetOutput(netFile, " + ");
				}
			    NetOutput(netFile, "<.");
			    NetOutput(netFile, DOMFORMAT, systNumber);
			    NetOutput(netFile, ", ");
			    NetOutput(netFile, DOMFORMAT, *root);
			    NetOutput(netFile, ".>");
			    systNumber++;
			}
		    NetOutput(netFile, ")\n");
		}
	    if (OnTheFly())
		{
		    root = (DOMAIN *) LstLast(RootList());
		    _monPlInd = (tstrSys != (DOMAIN) 0L) ? 1 : 0;
		    TupSField(arpSingle, (unsigned) 0, *root);
		    (void) TupEncode(MonPlInd(), arpSingle,
				     EngABuffer(), EngABufSize());
		    PlcSCount(GetPlc(EngBuffer()), (unsigned long) 1L);
		    NetOutput(netFile, "#place q lo(<.");
		    NetOutput(netFile, DOMFORMAT, tstrMin);
		    NetOutput(netFile, ".>) hi(<.");
		    NetOutput(netFile, DOMFORMAT, tstrMax);
		    NetOutput(netFile, ".>) mk(<.");
		    NetOutput(netFile, DOMFORMAT, *root);
		    NetOutput(netFile, ".>)\n#tester q");
		    if (monName != NULL)
			{
			    if (!(auxFile =
				  MyFopen(monName, "r", (BOOL *) NULL)))
				Exit(2);
			    ReadMonitor(monName, auxFile);
			    ClFclose(auxFile);
			}
		    NetOutput(netFile, "\n");
		}
	    actTbl = (ACTION **) AllocArray(IngGNofTransitions(),
					    sizeof(ACTION *));
	    arpHiTr = 0;
	    u = 0;
	    if (IAction())
		{
		    if ((!OnTheFly()) ||
			(LacGSystem((LACTION *) LstFirst(ActGUsed(IAction())))
			 != tstrSys))
			{
			    IngIProfile((unsigned) 0, (unsigned) 3);
			    actTbl[0] = IAction();
			    u = 3;
			    arpHiTr = 1;
			}
		    if (OnTheFly() &&
			(LacGSystem((LACTION *) LstLast(ActGUsed(IAction())))
			 == tstrSys))
			{
			    IngIProfile(arpHiTr, (unsigned) 3);
			    actTbl[arpHiTr] = IAction();
			    if (u < IngGProfile(arpHiTr))
				u = IngGProfile(arpHiTr);
			    arpHiTr++;
			}
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || !ActHidden(arpAction))
			continue;
		    IngIProfile(arpHiTr, 2 * LstSize(ActGUsed(arpAction)));
		    actTbl[arpHiTr] = IAction();
		    if (u < IngGProfile(arpHiTr))
			u = IngGProfile(arpHiTr);
		    arpHiTr++;
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || ActHidden(arpAction))
			continue;
		    IngIProfile(arpHiTr, 2 * LstSize(ActGUsed(arpAction)));
		    actTbl[arpHiTr] = arpAction;
		    if (u < IngGProfile(arpHiTr))
			u = IngGProfile(arpHiTr);
		    arpHiTr++;
		}
	    if (u != 0)
		arpVars = (DOMAIN *) AllocArray(u, sizeof(DOMAIN));
	    else
		arpVars = NULL;
	    flowTree = BalTrCreate(&flowBalTools);
	    sysotList = LstCreate(&sysotTools);
	    goList = LstCreate(&lngTools);
	    arpHiTr = 0;
	    if (IAction())
		{
		    if ((!OnTheFly()) ||
			(LacGSystem((LACTION *) LstFirst(ActGUsed(IAction())))
			 != tstrSys))
			{
			    TauInstances();
			    arpHiTr = 1;
			}
		    if (OnTheFly() &&
			(LacGSystem((LACTION *) LstLast(ActGUsed(IAction())))
			 == tstrSys))
			{
			    TstrTauInstances();
			    arpHiTr++;
			}
		}
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || !ActHidden(arpAction))
			continue;
		    ActSNumber(arpAction, ICode());
		    (void) LstBegin(ActGUsed(arpAction));
		    Instances(LstGet(ActGUsed(arpAction)));
		    arpHiTr++;
		}
	    BalTrDestroy(flowTree, TRUE);
	    u = 0;
	    LSTFOR (ActList(), arpAction, ACTION)
		{
		    if ((arpAction == IAction()) || ActHidden(arpAction))
			continue;
		    if (u == ICode())
			u++;
		    ActSNumber(arpAction, u);
		    u++;
		    (void) LstBegin(ActGUsed(arpAction));
		    Instances(LstGet(ActGUsed(arpAction)));
		    arpHiTr++;
		}
	    TupDiscard(arpSingle);
	    TupDiscard(arpDouble);
	    if (arpVars)
		ClDeallocate(arpVars);
	    LstDestroy(sysotList, TRUE);
	    LstDestroy(goList, TRUE);
	}
    else
	{
	    BaseLTSInit((unsigned) 0, (unsigned) 0);
	    PtBaseInit();
	}
    if (cont) 
	{
	    if (RootList())
		{
		    LstDestroy(_rootList, TRUE);
		    LSTFOR (ActList(), arpAction, ACTION)
			LstDestroy(ActGUsed(arpAction), TRUE);
		}
	    PlcFill();
	    TreeTrnFold();
	    TrnFill();
	    if (!NTrVisible())
		ClInfinite();
	    if (PtSymm())
		CompSymm(TRUE);
	    thePath = PthCreate((unsigned long) -1L);
	}
    else 
	{
	    ClFclose(netFile);
	    auxName = FileName(outName, "def");
	    if (!(auxFile = MyFopen(auxName, "w", (BOOL *) NULL)))
		Exit(2);
	    ClDeallocate(auxName);
	    ClFclose(auxFile);
	    auxName = FileName(outName, "dat");
	    if (NotBinaryFile(bFile = CreateWrite(auxName)))
		Exit(2);
	    ClDeallocate(auxName);
	    DatOutput(bFile);
	    OwnClose(&bFile);
	    if (RootList())
		{
		    LSTFOR (ActList(), arpAction, ACTION)
			LstDestroy(ActGUsed(arpAction), TRUE);
		}
	    auxName = FileName(outName, "fld");
	    if (NotBinaryFile(bFile = CreateWrite(auxName)))
		Exit(5);
	    ClDeallocate(auxName);
	    PtOutput(bFile);
	    OwnClose(&bFile);
	    PlcFill();
	    TreeTrnFold();
	    TrnFill();
	    if (!NTrVisible())
		ClInfinite();
	    if (PtSleep() && (Divergence() || Infinite()))
		{
		    (void)
			fputs
			("\n-f replaces -S because of loop-searching\n",
			 _err);
		    (void) fflush(_err);
		    ClPtSleep();
		}
	    if (PtSymm())
		CompSymm(TRUE);
	    if (rs == NULL)
		seedVal = 1L;
	    createFlag = 0L;
	    createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtSymm())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PlcIMkSymm())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtStub())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtDelAlg())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtCFFD())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtDefVis())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtSleep())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (bdds)
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (Dfs())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (FreeExpansion())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (Infinite() != 0L)
		createFlag |= 1L;
	    createFlag <<= 1;
	    theGraph = GphCreate(outName, bits, EngBufSize(), (UNS32) seedVal,
				 createFlag);
	    if (theGraph == NULL)
		{
		    Exit(3);
		}
	    if (ms != NULL)
		GphMapToMem(theGraph, mapLen);
	    u = PlcEncode(TblPlc(), EngABuffer(), EngABufSize());
	    thePath = PthCreate(GphPutNode(theGraph, EngBuffer(), u,
					   (unsigned long *) NULL));
	    if (PtSleep())
		GphPush(theGraph, PthEnd(thePath), (BALTREE *) NULL);
	    else if (OnTheFly() && PtFlyPlc())
		{
		    nowPath = thePath;
		    if (PlcReject(PtFlyPlc()))
			PtC1((unsigned long) -1L, PthEnd(thePath));
		    else if (PlcDeadlock(PtFlyPlc()))
			PtC2((unsigned long) -1L, PthEnd(thePath));
		}
	}
    if (PtStub())
	PtSRandom(GphSeed(theGraph));
    a = 0L;
    theInner = 0;
    theOuter = 0;
    if (Divergence() || PtCFFD())
	theNodeTree = BalTrCreate(&lngBalTools);
    turn = 0;
    if (Infinite())
	{
	    the2Path = PthCreate(PthEnd(thePath));
	    the3Path = PthCreate((unsigned long) -1L);
	    if (!Divergence())
		turn = 2;
	}
    do
	{    
	    if ((a != 0L) && stopFlag)
		{
		    if (PtStub())
			GphSetSeed(theGraph, PtTerm());
		    GphFlush(theGraph);
		    Suspend();
		    continueGeneration = TRUE;
		    stopFlag = FALSE;
		    SetSigHandlers();
		}
	    gphGCount = GphGenCount(theGraph);
	    if ((!Infinite()) ||
		((turn != 2) && ((turn == 0) ||
				 (PthEnd(the2Path) == (unsigned long) -1L))))
		{
		    PthSEnd(thePath, GphGet(theGraph, EngABuffer(),
					    EngABufSize(),
					    PtASlpSet(), PtA2SlpSet()));
		    if (PthEnd(thePath) == (unsigned long) -1L)
			turn = 2;
		    else
			{
			    if (OnTheFly())
				{
				    PlcDecode(EngBuffer());
				    PtOnTheFly();
				}
			    else
				{
				    PlcDecode(EngBuffer());
				    if (PtCFFD())
					PtProCFFD();
				    else
					PtFollowers();
				}
			}
		}
	    else if (PthEnd(the2Path) != (unsigned long) -1L)
		PtInfinite();
	    if (gphGCount < GphGenCount(theGraph))
		a += GphGenCount(theGraph) - gphGCount;
	} while ((a < gCount) && continueGeneration &&
		 ((turn != 2) || (Infinite() &&
				  (PthEnd(the2Path) != (unsigned long) -1L))));
    IgnSigHandlers();
    if ((!theDots) || (theOuter != 0) || (a <= 9L))
	{
	    (void) fputc((int) '\n', _out);
	    (void) fflush(_out);
	}
    if (PtStub())
	GphSetSeed(theGraph, PtTerm());
    LTSOutput(theGraph);
    GphClose(theGraph);
    Exit(0);
    return(0);
}
