/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  checkctl.h
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: checkctl.h,v 3.1 1997/04/17 15:43:54 kepa Exp $
*
* Description:
*     See checkctl.c.
*
* Version history:
*     $Log: checkctl.h,v $
 * Revision 3.1  1997/04/17  15:43:54  kepa
 * *** empty log message ***
 *
***************************************************************************/

typedef int CTL_FORMULATYPE;

#define EVALUATED_FORMULA ((CTL_FORMULATYPE)0)
#define ATOMIC_FORMULA ((CTL_FORMULATYPE)1)
#define AND_FORMULA ((CTL_FORMULATYPE)2)
#define OR_FORMULA ((CTL_FORMULATYPE)3)
#define NOT_FORMULA ((CTL_FORMULATYPE)4)
#define EX_FORMULA ((CTL_FORMULATYPE)5)
#define AX_FORMULA ((CTL_FORMULATYPE)6)
#define EU_FORMULA ((CTL_FORMULATYPE)7)
#define AU_FORMULA ((CTL_FORMULATYPE)8)

typedef struct _ctl_formula
{
    CTL_FORMULATYPE formulatype;   /* Type of the formula to be checked. */
    struct _ctl_formula *left_subformula;
    struct _ctl_formula *right_subformula;

    BITARRAY *result;
    
    BOOL top_level;                /* Is this the top_level formula,
				      for which a counterexample or a
				      witness must be evaluated. */
    
    BOOL negated;                  /* Is the formula inside an odd number
				      of negations. */
    
    EXPRESSION *p;                 /* The atomic proposition expression
				      to be evaluated. */
    
} CTL_FORMULA;

#define ctl_formula_type(f) \
((f)->formulatype)

#define left_subformula_result(f) \
((f)->left_subformula->result)

#define right_subformula_result(f) \
((f)->right_subformula->result)

#define formula_result(f) \
((f)->result)

#define set_result(f) \
(f)->result = (f)->right_subformula->result; \
(f)->right_subformula->result = NULL

#define mark_evaluated(f) \
(f)->formulatype = EVALUATED_FORMULA;

#define proposition(f) \
((f)->p)

#define is_toplevel(f) \
((f)->top_level)

#define is_negated(f) \
((f)->negated)

extern TOOLS CTLFormulaTools;

DE3(BOOL, ModelCheckCTL,
    FORMULA *, formula,
    GRAPH *, graph,
    NODE, state)
DE1(BOOL, RemoveNegations,
    LIST *, formulas)
DE3(CTL_FORMULA *, InitVisitFormulas,
    FORMULA *, formula,
    LIST *, formulas,
    unsigned long, num_states)
DE5(BOOL, CheckCTL,
    LIST *, formulas,
    NODE, s,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    STACK *, path)
DE6(BOOL, CheckCTLFormula,
    CTL_FORMULA *, f,
    NODE, s,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
DE5(BOOL, LabelAnd,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
DE5(BOOL, LabelOr,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
DE4(BOOL, LabelNot,
    NODE, s,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
DE1(void, CTLFormulaDeallocate,
    CTL_FORMULA *, f)
#ifdef DEBUG
DE2(void, CTLFormulaDump,
    CTL_FORMULA *, f,
    unsigned, depth)
#endif
