/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  comline.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko, Tomi Janhunen, Kimmo Valkealahti,
*          and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "formula.h"
#include "mbufs.h"
#include "macptree.h"
#include "comline.h"
#include "probe.h"
#include "redirect.h"

unsigned long lineLength;
static char outpBuffer[50];
static char dupBuffer[50];
static unsigned outpBLen = 0;
static unsigned long cursor = 0L;

#define MB_SIZE (16*BUFSIZ)
#define MELIMIT 10000

unsigned symCount = 0;
char **plaNames = NULL, **traNames = NULL, ***traVarNames = NULL;
char **symNames = NULL, **priNames = NULL;
BALTREE *resTree = NULL, *plaTree = NULL, *traTree = NULL;
BALTREE **traVarTrees = NULL, *symTree = NULL, *priTree = NULL;
BOOL transform = FALSE;
BOOL evalStuff = FALSE;
MBUF *ioBuf = NULL;
MBUF *evalBuf = NULL;
static MBUF *swBuf = NULL;
static MBUF *qwBuf0 = NULL;
static MBUF *qwBuf1 = NULL;

/***************************************************************************
*
* Function:  PrtCaret
* Purpose:
* Arguments: count -
*            textLen -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PrtCaret,
    unsigned, count,
    unsigned, textLen)
{
    if (textLen == 0)
	{
	    textLen++;
	    count++;
	}
    *(ioBuf->_written) = '\0';
    (void) fputs(ioBuf->_buf, _err);
    while (count-- != textLen)
	(void) fputc((int) ' ', _err);
    while (textLen-- != 0)
	(void) fputc((int) '^', _err);
    (void) fputc((int) '\n', _err);
    (void) fflush(_err);
}

/***************************************************************************
*
* Function:  CLineInit
* Purpose:
* Arguments: datFile -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, CLineInit,
    BINARYFILE, datFile)
{
    unsigned long pos = 0L;
    unsigned i;
    unsigned j;
    unsigned len = 0;
    char *s;

    ReadLng(&pos, 1, datFile);
    BaseInit(datFile, FALSE);
    ClOnTheFly();
    SeekAbs(datFile, pos);
    if (MagGNofPlaces())
	plaNames = (char **) AllocArray(MagGNofPlaces(), sizeof(char *));
    else
	plaNames = (char **) NULL;
    plaTree = BalTrCreate(&refBalTools);
    if (IngGNofTransitions())
	{
	    traNames = (char **)
		AllocArray(IngGNofTransitions(), sizeof(char *));
	    traVarNames = (char ***)
		AllocArray(IngGNofTransitions(), sizeof(char **));
	    traVarTrees = (BALTREE **)
		AllocArray(IngGNofTransitions(), sizeof(BALTREE *));
	}
    else
	{
	    traNames = (char **) NULL;
	    traVarNames = (char ***) NULL;
	    traVarTrees = (BALTREE **) NULL;
	}
    traTree = BalTrCreate(&refBalTools);
    if (IngGNofClasses() > 1)
	priNames = (char **) AllocArray(IngGNofClasses() - 1, sizeof(char *));
    else
	priNames = (char **) NULL;
    priTree = BalTrCreate(&refBalTools);
    for (i = 0; i < IngGNofTransitions(); i++)
	{
	    ReadUns(&len, 1, datFile);
	    traNames[i] = (char *) Allocate(len + 1);
	    ReadChr(traNames[i], len, datFile);
	    (void) BalTrInsert(traTree, (char *)
			       RefCons(traNames[i], (unsigned long) i),
			       (char *) (traNames[i]), TRUE);
	    if (IngGProfile(i))
		traVarNames[i] = (char **)
		    AllocArray(IngGProfile(i), sizeof(char *));
	    else
		traVarNames[i] = (char **) NULL;
	    traVarTrees[i] = BalTrCreate(&refBalTools);
	    for (j = 0; j < IngGProfile(i); j++)
		{
		    ReadUns(&len, 1, datFile);
		    traVarNames[i][j] = (char *) Allocate(len + 1);
		    ReadChr(traVarNames[i][j], len, datFile);
		    (void) BalTrInsert(traVarTrees[i], (char *)
				       RefCons(traVarNames[i][j],
					       (unsigned long) j),
				       (char *) (traVarNames[i][j]), TRUE);
		}
	}
    ReadUns(&symCount, 1, datFile);
    if (symCount)
	symNames = (char **) AllocArray(symCount, sizeof(char *));
    else
	symNames = (char **) NULL;
    symTree = BalTrCreate(&refBalTools);
    for (i = 0; i < symCount; i++)
	{
	    ReadUns(&len, 1, datFile);
	    symNames[i] = (char *) Allocate(len + 1);
	    ReadChr(symNames[i], len, datFile);
	    (void) BalTrInsert(symTree, (char *)
			       RefCons(symNames[i], (unsigned long) i),
			       (char *) (symNames[i]), TRUE);
	}
    for (i = 0; i < IngGNofClasses() - 1; i++)
	{
	    ReadUns(&len, 1, datFile);
	    priNames[i] = (char *) Allocate(len + 1);
	    ReadChr(priNames[i], len, datFile);
	    (void) BalTrInsert(priTree, (char *)
			       RefCons(priNames[i], (unsigned long) i),
			       (char *) (priNames[i]), TRUE);
	}
    for (i = 0; i < MagGNofPlaces(); i++)
	{
	    ReadUns(&len, 1, datFile);
	    plaNames[i] = (char *) Allocate(len + 1);
	    ReadChr(plaNames[i], len, datFile);
	    (void) BalTrInsert(plaTree, (char *)
			       RefCons(plaNames[i], (unsigned long) i),
			       (char *) (plaNames[i]), TRUE);
	}
    probeMacros = BalTrCreate(&macpBalTools);
    probeHelps = NULL;
    qwBuf0 = MbufCreate(MB_SIZE);
    qwBuf1 = MbufCreate(MB_SIZE);
    ioBuf = qwBuf0;
    evalBuf = qwBuf1;
    resTree = BalTrCreate(&strBalTools);
    s = StrCopy("af"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("ag"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("all"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("allcomps"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("and"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("arrow"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("au"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("ax"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("bpath"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("bspan"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("build"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("calc"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("card"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("check"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("clear"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("clearall"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("define"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("defs"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("dpath"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("dspan"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("ef"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("eg"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("empty"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("eu"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("ex"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("faithful"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("false"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("field"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("fire"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("goto"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("help"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("load"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("look"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("mute"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("next"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("node"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("not"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("or"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("preclass"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("pred"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("preds"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("prev"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("query"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("quit"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("sets"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("sgpred"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("sgsucc"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("showends"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("shrink"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("statistics"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("step"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("succ"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("succs"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("superverbose"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("termcomps"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("true"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("undef"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("undefall"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("verbose"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
    s = StrCopy("volatile"); (void) BalTrInsert(resTree, (char *) s, (char *) s, TRUE);
}

static char *command = NULL;
static COMLINE *comLine = NULL;
static int redirStat = 0;
static REDIRECT redirect;

FE0(void, CLineClose)
{
    unsigned i, j;

    if (redirStat)
	{
	    if (2 == redirStat)
		(void) CloseRedirect(&redirect);
	    (void) CleanupRedirect(&redirect);
	    redirStat = 0;
	}
    if (plaNames)
	{
	    for (i = 0; i < MagGNofPlaces(); i++)
		ClDeallocate(plaNames[i]);
	    ClDeallocate(plaNames);
	}
    BalTrDestroy(plaTree, TRUE);
    if (traNames)
	{
	    for (i = 0; i < IngGNofTransitions(); i++)
		ClDeallocate(traNames[i]);
	    ClDeallocate(traNames);
	}
    if (traVarNames)
	{
	    for (i = 0; i < IngGNofTransitions(); i++)
		{
		    if (traVarNames[i])
			{
			    for (j = 0; j < IngGProfile(i); j++)
				ClDeallocate(traVarNames[i][j]);
			    ClDeallocate(traVarNames[i]);
			}
		}
	    ClDeallocate(traVarNames);

	}
    if (traVarTrees)
	{
	    for (i = 0; i < IngGNofTransitions(); i++)
		BalTrDestroy(traVarTrees[i], TRUE);
	    ClDeallocate(traVarTrees);
	}
    BalTrDestroy(traTree, TRUE);
    if (priNames)
	{
	    for (i = 0; i < IngGNofClasses() - 1; i++)
		ClDeallocate(priNames[i]);
	    ClDeallocate(priNames);
	}
    BalTrDestroy(priTree, TRUE);
    if (symNames)
	{
	    for (i = 0; i < symCount; i++)
		ClDeallocate(symNames[i]);
	    ClDeallocate(symNames);
	}
    BalTrDestroy(symTree, TRUE);
    BalTrDestroy(probeMacros, TRUE);
    BalTrDestroy(probeHelps, TRUE);
    MbufDiscard(qwBuf0);
    MbufDiscard(qwBuf1);
    BalTrDestroy(resTree, TRUE);
    ClDeallocate(command);
    CmlDiscard(comLine);
    BaseClose();
}

/***************************************************************************
*
* Function:  CLine
* Purpose:
* Arguments: inFile -
* Returns:
* See also:
*
****************************************************************************
*/
#ifdef USE_READLINE
FE2(void, CLine,
    FILE *, inFile,
    char *, prompt)
#else
FE1(void, CLine,
    FILE *, inFile)
#endif
{
    char c = '\0', *buf, *b2;
    unsigned bufSize;

    MbufInit(ioBuf);
#ifdef USE_READLINE
    if ((inFile != _in) || !isatty(fileno(inFile)))
	MbufReadln(ioBuf, inFile, (char *) NULL, FALSE);
    else
	MbufReadln(ioBuf, inFile, prompt, TRUE);
#else
    MbufReadln(ioBuf, inFile);
#endif
    if (!continueEvaluation || (errorFlag != NO_ERROR))
	return;
    bufSize = MbufSize(ioBuf);
    if (bufSize == 0)
	return;
    evalStuff = FALSE;
    MacpParse();
    if (!evalStuff || (errorFlag != NO_ERROR) || !continueEvaluation)
	return;
    command = strncpy((char *) Allocate(bufSize), ioBuf->_buf,
		      (size_t) (bufSize - 1));
    command[bufSize - 1] = '\0';
    mEvalTimes = 0;
    mEvalLimit = MELIMIT;
    ioBuf->_read = ioBuf->_buf;
    do
	{
	    MbufInit(evalBuf);
	    transform = FALSE;
	    MacroEval();
	    swBuf = ioBuf;
	    ioBuf = evalBuf;
	    evalBuf = swBuf;
	} while (transform && continueEvaluation && (errorFlag == NO_ERROR));
    if ((errorFlag != NO_ERROR) || !continueEvaluation)
	{
	    ClDeallocate(command);
	    return;
	}
    b2 = ioBuf->_buf;
    do
	{
	    for (buf = ioBuf->_read;
		 (buf < ioBuf->_written) && (*buf != '#');
		 buf++)
		;
	    if (buf < ioBuf->_written)
		{
		    *buf = '\n';
		    c = *(buf + 1);
		    *(buf + 1) = '\0';
		}
	    else
		*buf = '\0';
	    errorFlag = NO_ERROR;
	    if (continueEvaluation)
		{
		    FogSGraph(theGraph);
		    FogSWorld(PthEnd(thePath));
		    FogSTransition((unsigned) -1);
		    FogSInstance((DOMAIN *) NULL);
		    FogSPriority(0);
		    FogSFaithful(TRUE);
		    FogSTuple((TUPLE *) NULL);
		    redirStat = 1;
		    comLine = CmlParse(&redirect);
		    if (comLine != NULL)
			{
			    if ((errorFlag == NO_ERROR) &&
				continueEvaluation)
				{
				    redirStat = 2;
				    (void) InitRedirect(&redirect);
				    Execute(comLine, command);
				    (void) CloseRedirect(&redirect);
				    redirStat = 1;
				}
			    CmlDiscard(comLine);
			}
		    (void) CleanupRedirect(&redirect);
		    redirStat = 0;
		}
	    if (buf < ioBuf->_written)
		{
		    *(buf + 1) = c;
		    ioBuf->_read = buf + 1;
		    ioBuf->_buf = buf + 1;
		}
	} while ((buf < ioBuf->_written) && thePath &&
		 continueEvaluation);
    ClDeallocate(command);
    ioBuf->_buf = b2;
}

/***************************************************************************
*
* Function:  LngToPth
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, LngToPth,
    unsigned long *, lng)
{
    return((char *) PthCreate(*lng));
}

static BALTREE *sopTree, *sopTree2;

/***************************************************************************
*
* Function:  SopCut
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, SopCut,
    PATH *, path)
{
    if (BalTrElement(sopTree2, (char *) path))
	(void) BalTrInsert(sopTree, (char *) PthCopy(path),
			   (char *) path, TRUE);
    return(!continueEvaluation);
}

/***************************************************************************
*
* Function:  SopMinus
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, SopMinus,
    PATH *, path)
{
    (void) BalTrDelete(sopTree, (char *) NULL, (char *) path, TRUE);
    return(!continueEvaluation);
}

/***************************************************************************
*
* Function:  SopUnion
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, SopUnion,
    PATH *, path)
{
    PATH *aux;

    aux = PthCopy(path);
    if (BalTrInsert(sopTree, (char *) aux, (char *) aux, TRUE))
	PthDiscard(aux);
    return(!continueEvaluation);
}

/***************************************************************************
*
* Function:  SopEval
* Purpose:
* Arguments: setOper -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BALTREE *, SopEval,
    SETOPER *, setOper)
{
    PTHSET *pthSet;
    BALTREE *tree1, *tree2;
    unsigned long number;

    switch (setOper->operator)
	{
	case SE_SET:
	    number = XprEval(setOper->us.expression);
	    if ((errorFlag != NO_ERROR) || !continueEvaluation)
		return(NULL);
	    if ((pthSet = (PTHSET *) BalTrElement(clbleSets,
						  (char *) (&number))) ||
		(pthSet = (PTHSET *) BalTrElement(specSets,
						  (char *) (&number))))
		return(PtsGPaths(pthSet));
	    IsntSet(_err, number);
	    errorFlag = SEM_ERROR;
	    return(NULL);
	case SE_COMP:
	    number = XprEval(setOper->us.expression);
	    if ((errorFlag != NO_ERROR) || !continueEvaluation)
		return(NULL);
	    if (((unsigned long) 0L) == compCount)
		{
		    StrongNotComp(_err);
		    errorFlag = SEM_ERROR;
		    return(NULL);
		}
	    if (number >= compCount)
		{
		    IsntSgc(_err, number);
		    errorFlag = SEM_ERROR;
		    return(NULL);
		}
	    if (sgsInMemory)
		return(BalTrImage((BALTREE *) PtrBlGet(comps, number),
				  &pthBalTools, (CFUNC) LngToPth));
	    tree1 = GphSgGet(theGraph, number);
	    tree2 = BalTrImage(tree1, &pthBalTools, (CFUNC) LngToPth);
	    BalTrDestroy(tree1, TRUE);
	    return(tree2);
	case SE_CUT:
	    if (!(tree1 = SopEval(setOper->us.os.sub1s)))
		return(NULL);
	    if (!(tree2 = SopEval(setOper->us.os.sub2s)))
		{
		    if (setOper->us.os.sub1s->operator != SE_SET)
			BalTrDestroy(tree1, TRUE);
		    return(NULL);
		}
	    sopTree = BalTrCreate(&pthBalTools);
	    sopTree2 = tree2;
	    (void) BalTrApply(tree1, SopCut);
	    if (setOper->us.os.sub1s->operator != SE_SET)
		BalTrDestroy(tree1, TRUE);
	    if (setOper->us.os.sub2s->operator != SE_SET)
		BalTrDestroy(tree2, TRUE);
	    if (!continueEvaluation)
		{
		    BalTrDestroy(sopTree, TRUE);
		    return(NULL);
		}
	    return(sopTree);
	case SE_MINUS:
	case SE_UNION:
	    if (!(tree1 = SopEval(setOper->us.os.sub1s)))
		return(NULL);
	    if (!(tree2 = SopEval(setOper->us.os.sub2s)))
		{
		    if (setOper->us.os.sub1s->operator != SE_SET)
			BalTrDestroy(tree1, TRUE);
		    return(NULL);
		}
	    if (setOper->us.os.sub1s->operator == SE_SET)
		sopTree = BalTrCopy(tree1, TRUE);
	    else
		sopTree = tree1;
	    if (setOper->operator == SE_MINUS)
		(void) BalTrApply(tree2, SopMinus);
	    else
		(void) BalTrApply(tree2, SopUnion);
	    if (setOper->us.os.sub2s->operator != SE_SET)
		BalTrDestroy(tree2, TRUE);
	    if (!continueEvaluation)
		{
		    BalTrDestroy(sopTree, TRUE);
		    return(NULL);
		}
	    return(sopTree);
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  SopIn
* Purpose:
* Arguments: setOper -
*            world -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(BOOL, SopIn,
    SETOPER *, setOper,
    unsigned long, world)
{
    BALTREE *paths;
    PATH *path, *aux;
    unsigned long number;

    if (setOper->operator == SE_COMP)
	{
	    number = XprEval(setOper->us.expression);
	    if ((errorFlag != NO_ERROR) || !continueEvaluation)
		return(FALSE);
	    if (((unsigned long) 0L) == compCount)
		{
		    StrongNotComp(_err);
		    errorFlag = SEM_ERROR;
		    return(FALSE);
		}
	    if (number >= compCount)
		{
		    IsntSgc(_err, number);
		    errorFlag = SEM_ERROR;
		    return(FALSE);
		}
	    return (GphCompNumber(theGraph, world) == number);
	}
    if (!(paths = SopEval(setOper)))
	return(FALSE);
    path = PthCreate(world);
    aux = (PATH *) BalTrGrEqPrefix(paths, (char *) path, TRUE);
    PthDiscard(path);
    if (setOper->operator != SE_SET)
	BalTrDestroy(paths, TRUE);
    return(aux != NULL);
}

/***************************************************************************
*
* Function:  SopCons
* Purpose:
* Arguments: operator -
*            sop1 -
*            sop2 -
* Returns:
* See also:
*
****************************************************************************
*/
FE3(SETOPER *, SopCons,
    unsigned, operator,
    SETOPER *, sop1,
    SETOPER *, sop2)
{
    SETOPER *setOper;

    setOper = SopAlloc();
    setOper->operator = operator;
    switch (operator)
	{
	case SE_SET:
	case SE_COMP:
	    setOper->us.expression = (EXPRESSION *) sop1;
	    break;
	case SE_UNION:
	case SE_CUT:
	case SE_MINUS:
	    setOper->us.os.sub1s = sop1;
	    setOper->us.os.sub2s = sop2;
	    break;
	}
    return(setOper);
}

/***************************************************************************
*
* Function:  SopAux
* Purpose:
* Arguments: operator -
*            sop1 -
*            sop2 -
* Returns:
* See also:
*
****************************************************************************
*/
FE3(SETOPER *, SopAux,
    unsigned, operator,
    SETOPER *, sop1,
    SETOPER *, sop2)
{
    if (errorFlag == NO_ERROR)
	return(SopCons(operator, sop1, sop2));
    switch (operator)
	{
	case SE_SET:
	case SE_COMP:
	    if (sop1)
		XprCastDiscard(sop1);
	    break;
	case SE_UNION:
	case SE_CUT:
	case SE_MINUS:
	    if (sop1)
		SopDiscard(sop1);
	    if (sop2)
		SopDiscard(sop2);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  SopCopy
* Purpose:
* Arguments: setOper -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(SETOPER *, SopCopy,
    SETOPER *, setOper)
{
    SETOPER *copy;

    copy = SopAlloc();
    copy->operator = setOper->operator;
    switch (setOper->operator)
	{
	case SE_SET:
	case SE_COMP:
	    copy->us.expression = XprCopy(setOper->us.expression);
	    break;
	case SE_UNION:
	case SE_CUT:
	case SE_MINUS:
	    copy->us.os.sub1s = SopCopy(setOper->us.os.sub1s);
	    copy->us.os.sub2s = SopCopy(setOper->us.os.sub2s);
	    break;
	}
    return(copy);
}

/***************************************************************************
*
* Function:  SopDracsid
* Purpose:
* Arguments: setOper -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, SopDracsid,
    SETOPER *, setOper)
{
    if (setOper)
	{
	    switch (setOper->operator)
		{
		case SE_SET:
		case SE_COMP:
		    XprDiscard(setOper->us.expression);
		    break;
		case SE_UNION:
		case SE_CUT:
		case SE_MINUS:
		    SopDiscard(setOper->us.os.sub1s);
		    SopDiscard(setOper->us.os.sub2s);
		    break;
		}
	    Deallocate((char *) setOper);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  SopDump
* Purpose:
* Arguments: setOper -
*            depth -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, SopDump,
    SETOPER *, setOper,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "SETOPER: operator %u\n", setOper->operator);
    (void) fflush(_err);
    switch (setOper->operator)
	{
	case SE_SET:
	case SE_COMP:
	    XprDump(setOper->us.expression, depth + 1);
	    break;
	case SE_UNION:
	case SE_CUT:
	case SE_MINUS:
	    SopDump(setOper->us.os.sub1s, depth + 1);
	    SopDump(setOper->us.os.sub2s, depth + 1);
	    break;
	}
    DumpIndent(depth);
    (void) fputs("End of SETOPER\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  CmlCons
* Purpose:
* Arguments: operator -
*            options -
*            ep1 -
*            ep2 -
* Returns:
* See also:
*
****************************************************************************
*/
FE4(COMLINE *, CmlCons,
    unsigned, operator,
    unsigned, options,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    COMLINE *comline;

    comline = CmlAlloc();
    comline->operator = operator;
    comline->options = options;
    switch(operator)
	{
	case LI_CALC :
	    if (options & TG_MARKING)
		comline->ul.markexpr = (MARKEXPR *) ep1;
	    else
		comline->ul.expression = ep1;
	    break;
	case LI_GOTO :
	case LI_LOOK :
	case LI_SUCC :
	case LI_PRED :
	case LI_NEXT :
	case LI_CLEAR :
	case LI_SHRINK :
	case LI_SGSUCC :
	case LI_SGPRED :
	    comline->ul.expression = ep1;
	    if (operator == LI_SUCC)
		comline->expr2 = ep2;
	    break;
	case LI_CHECK :
	case LI_QUERY :
	    comline->ul.formula = (FORMULA *) ep1;
	    break;
	case LI_BUILD :
	case LI_SHOWENDS :
	    comline->ul.setOper = (SETOPER *) ep1;
	    break;
	}
    return(comline);
}

/***************************************************************************
*
* Function:  CmlAux
* Purpose:
* Arguments: operator -
*            options -
*            ep1 -
*            ep2 -
* Returns:
* See also:
*
****************************************************************************
*/
FE4(COMLINE *, CmlAux,
    unsigned, operator,
    unsigned, options,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    if (errorFlag == NO_ERROR)
	return(CmlCons(operator, options, ep1, ep2));
    switch(operator)
	{
	case LI_CALC :
	    if (ep1)
		{
		    if (options & TG_MARKING)
			MexCastDiscard(ep1);
		    else
			XprDiscard(ep1);
		}
	    break;
	case LI_GOTO :
	case LI_LOOK :
	case LI_SUCC :
	case LI_PRED :
	case LI_NEXT :
	case LI_CLEAR :
	case LI_SHRINK :
	case LI_SGSUCC :
	case LI_SGPRED :
	    if (ep1)
		XprDiscard(ep1);
	    if (ep2)
		{
		    if (operator == LI_SUCC)
			XprDiscard(ep2);
		}
	    break;
	case LI_CHECK :
	case LI_QUERY :
	    if (ep1)
		ForCastDiscard(ep1);
	    break;
	case LI_BUILD :
	case LI_SHOWENDS :
	    if (ep1)
		SopCastDiscard(ep1);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  CmlDracsid
* Purpose:
* Arguments: comline -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, CmlDracsid,
    COMLINE *, comline)
{
    if (comline)
	{
	    switch (comline->operator)
		{
		case LI_CALC :
		    if (comline->options & TG_MARKING)
			{
			    if (comline->ul.markexpr != NULL)
				MexDiscard(comline->ul.markexpr);
			}
		    else
			{
			    if (comline->ul.expression != NULL)
				XprDiscard(comline->ul.expression);
			}
		    break;
		case LI_GOTO :
		case LI_LOOK :
		case LI_SUCC :
		case LI_PRED :
		case LI_NEXT :
		case LI_CLEAR :
		case LI_SHRINK :
		case LI_SGSUCC :
		case LI_SGPRED :
		    if (comline->ul.expression != NULL)
			XprDiscard(comline->ul.expression);
		    if (comline->operator == LI_SUCC)
			{
			    if (comline->expr2 != NULL)
				XprDiscard(comline->expr2);
			}
		    break;
		case LI_CHECK :
		case LI_QUERY :
		    if (comline->ul.formula != NULL)
			{
			    if (forIsOpen)
				{
				    forIsOpen = FALSE;
				    ForClose(comLine->ul.formula, theGraph);
				}
			    ForDiscard(comline->ul.formula);
			}
		    break;
		case LI_BUILD :
		case LI_SHOWENDS :
		    if (comline->ul.setOper != NULL)
			SopDiscard(comline->ul.setOper);
		    break;
		}
	    Deallocate((char *) comline);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  CmlDump
* Purpose:
* Arguments: comline -
*            depth -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, CmlDump,
    COMLINE *, comline,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "COMLINE: operator %u\n", comline->operator);
    (void) fflush(_err);
    DumpIndent(depth);
    (void) fprintf(_err, "COMLINE: options %u\n", comline->options);
    (void) fflush(_err);
    switch (comline->operator)
	{
	case LI_CALC :
	    if (comline->options & TG_MARKING)
		{
		    if (comline->ul.markexpr != NULL)
			MexDump(comline->ul.markexpr, depth + 1);
		}
	    else
		{
		    if (comline->ul.expression != NULL)
			XprDump(comline->ul.expression, depth + 1);
		}
	    break;
	case LI_GOTO :
	case LI_LOOK :
	case LI_SUCC :
	case LI_PRED :
	case LI_NEXT :
	case LI_CLEAR :
	case LI_SHRINK :
	case LI_SGSUCC :
	case LI_SGPRED :
	    if (comline->ul.expression != NULL)
		XprDump(comline->ul.expression, depth + 1);
	    if (comline->operator == LI_SUCC)
		{
		    if (comline->expr2 != NULL)
			XprDump(comline->expr2, depth + 1);
		}
	    break;
	case LI_QUERY :
	    if (comline->ul.formula != NULL)
		ForDump(comline->ul.formula, depth + 1);
	    break;
	case LI_BUILD :
	case LI_SHOWENDS :
	    if (comline->ul.setOper != NULL)
		SopDump(comline->ul.setOper, depth + 1);
	    break;
	}
    DumpIndent(depth);
    (void) fputs("End of COMLINE\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  PthOutput
* Purpose:
* Arguments: path -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PthOutput,
    PATH *, path,
    FILE *, f)
{
    STEP *step;

    if (path == NULL)
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) outpBLen <= lineLength)
			(void) fprintf(f, "%s\n", outpBuffer);
		    else
			(void) fprintf(f, "\n%s\n", outpBuffer);
		}
	    outpBLen = 0;
	    cursor = 0L;
	}
    else
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) (outpBLen + 6) < lineLength)
			{
			    (void) fprintf(f, "%s,     ", outpBuffer);
			    cursor += (unsigned long) (outpBLen + 6);
			}
		    else if (cursor + (unsigned long) (outpBLen + 1)
			     <= lineLength)
			{
			    (void) fprintf(f, "%s,\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s,     ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 6);
			}
		}
	    LSTNEST(path->steps, step, STEP)
		;
	    LSTUNNEST (path->steps, step, STEP)
		{
		    (void) sprintf(outpBuffer, "%4lu",
				   GphOrdinal(theGraph, step->world));
		    outpBLen = (unsigned) strlen(outpBuffer);
		    if (cursor + (unsigned long) (outpBLen + 1) < lineLength)
			{
			    (void) fprintf(f, "%s ", outpBuffer);
			    cursor += (unsigned long) (outpBLen + 1);
			}
		    else if (cursor + (unsigned long) outpBLen <= lineLength)
			{
			    (void) fprintf(f, "%s\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 1);
			}
		    (void) sprintf(dupBuffer, "[%lu>",
				   GphAwOrdinal(theGraph, step->awAddr));
		    (void) sprintf(outpBuffer, "%4s", dupBuffer);
		    outpBLen = (unsigned) strlen(outpBuffer);
		    if (cursor + (unsigned long) (outpBLen + 1) < lineLength)
			{
			    (void) fprintf(f, "%s ", outpBuffer);
			    cursor += (unsigned long) (outpBLen + 1);
			}
		    else if (cursor + (unsigned long) outpBLen <= lineLength)
			{
			    (void) fprintf(f, "%s\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 1);
			}
		}
	    (void) sprintf(outpBuffer, "%4lu",
			   GphOrdinal(theGraph, path->end));
	    outpBLen = (unsigned) strlen(outpBuffer);
	}
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  NdOutput
* Purpose:
* Arguments: nd -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, NdOutput,
    unsigned long, nd,
    FILE *, f)
{
    if (nd == (unsigned long) -1L)
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) outpBLen <= lineLength)
			(void) fprintf(f, "%s\n", outpBuffer);
		    else
			(void) fprintf(f, "\n%s\n", outpBuffer);
		}
	    outpBLen = 0;
	    cursor = 0L;
	}
    else
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) (outpBLen + 6) < lineLength)
			{
			    (void) fprintf(f, "%s,     ", outpBuffer);
			    cursor += (unsigned long) outpBLen + 6;
			}
		    else if (cursor + (unsigned long) (outpBLen + 1)
			     <= lineLength)
			{
			    (void) fprintf(f, "%s,\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s,     ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 6);
			}
		}
	    (void) sprintf(outpBuffer, "%4lu", nd);
	    outpBLen = (unsigned) strlen(outpBuffer);
	}
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  AwSgOutput
* Purpose:
* Arguments: ordfa -
*            compfa -
*            aw -
*            ordson -
*            compson -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE6(void, AwSgOutput,
    unsigned long, ordfa,
    unsigned long, compfa,
    unsigned long, aw,
    unsigned long, ordson,
    unsigned long, compson,
    FILE *, f)
{
    register char *p;

    if (ordfa == (unsigned long) -1L)
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) outpBLen <= lineLength)
			(void) fprintf(f, "%s\n", outpBuffer);
		    else
			(void) fprintf(f, "\n%s\n", outpBuffer);
		}
	    outpBLen = 0;
	    cursor = 0L;
	}
    else
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) (outpBLen + 3) < lineLength)
			{
			    (void) fprintf(f, "%s,  ", outpBuffer);
			    cursor += (unsigned long) (outpBLen + 3);
			}
		    else if (cursor + (unsigned long) (outpBLen + 1)
			     <= lineLength)
			{
			    (void) fprintf(f, "%s,\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s,  ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 3);
			}
		}
	    p = dupBuffer;
	    (void) sprintf(p, "%lu(", ordfa);
	    while (*p)
		p++;
	    BfSgcDenote(p, compfa);
	    while (*p)
		p++;
	    (void) sprintf(p, ") [%lu> %lu(", aw, ordson);
	    while (*p)
		p++;
	    BfSgcDenote(p, compson);
	    while (*p)
		p++;
	    *p++ = ')';
	    *p = '\0';
	    (void) sprintf(outpBuffer, "%30s", dupBuffer);
	    outpBLen = (unsigned) strlen(outpBuffer);
	}
}

/***************************************************************************
*
* Function:  NamSgOutput
* Purpose:
* Arguments: sg -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, NamSgOutput,
    unsigned long, sg,
    FILE *, f)
{
    if (sg == (unsigned long) -1L)
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) outpBLen <= lineLength)
			(void) fprintf(f, "%s\n", outpBuffer);
		    else
			(void) fprintf(f, "\n%s\n", outpBuffer);
		}
	    outpBLen = 0;
	    cursor = 0L;
	}
    else
	{
	    if (outpBLen != 0)
		{
		    if (cursor + (unsigned long) (outpBLen + 4) < lineLength)
			{
			    (void) fprintf(f, "%s,   ", outpBuffer);
			    cursor += (unsigned long) (outpBLen + 4);
			}
		    else if (cursor + (unsigned long) (outpBLen + 1)
			     <= lineLength)
			{
			    (void) fprintf(f, "%s,\n", outpBuffer);
			    cursor = 0L;
			}
		    else
			{
			    (void) fprintf(f, "\n%s,   ", outpBuffer);
			    cursor = (unsigned long) (outpBLen + 4);
			}
		}
	    BfSgcDenote(dupBuffer, sg);
	    (void) sprintf(outpBuffer, "%6s", dupBuffer);
	    outpBLen = (unsigned) strlen(outpBuffer);
	}
}

/***************************************************************************
*
* Function:  DomOutput
* Purpose:
* Arguments: n -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, DomOutput,
    DOMAIN, n,
    FILE *, f)
{
    if (symCount > (unsigned) (((DOMAIN) MAXDOMAIN) - n))
	(void) fputs(symNames[(unsigned) (((DOMAIN) MAXDOMAIN) - n)], f);
    else
	(void) fprintf(f, DOMFORMAT, n);
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  PriOutput
* Purpose:
* Arguments: n -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PriOutput,
    unsigned, n,
    FILE *, f)
{
    if (n != 0)
	(void) fputs(priNames[n - 1], f);
    else
	(void) fputc((int) '0', f);
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  MarOutput
* Purpose:
* Arguments: marking -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, MarOutput,
    MARKING *, marking,
    FILE *, f)
{
    MULTUP *multup;
    unsigned i;
    DOMAIN n;
    char *s;

    if (MarEmpty(marking))
	{
	    if (cursor + 5L <= lineLength)
		(void) fputs("empty", f);
	    else
		(void) fputs("\nempty", f);
	}
    MARFOR (marking, multup)
	{
	    if (multup != MarFirst(marking))
		{
		    if (cursor + 3L < lineLength)
			{
			    (void) fputs(" + ", f);
			    cursor += 3L;
			}
		    else if (cursor + 2L <= lineLength)
			{
			    (void) fputs(" +\n", f);
			    cursor = 0L;
			}
		    else
			{
			    (void) fputs("\n+ ", f);
			    cursor = 2L;
			}
		}
	    if (!MultupGArity(multup))
		{
		    if (MultupGCount(multup) != 1L)
			{
			    (void) sprintf(outpBuffer, "%lu",
					   MultupGCount(multup));
			    outpBLen = (unsigned) strlen(outpBuffer);
			    if (cursor + (unsigned long) (outpBLen + 4)
				<= lineLength)
				{
				    (void) fprintf(f, "%s<..>", outpBuffer);
				    cursor += (unsigned long) (outpBLen + 4);
				}
			    else
				{
				    (void) fprintf(f, "\n%s<..>", outpBuffer);
				    cursor = (unsigned long) (outpBLen + 4);
				}
			}
		    else
			{
			    if (cursor + 4L <= lineLength)
				{
				    (void) fputs("<..>", f);
				    cursor += 4L;
				}
			    else
				{
				    (void) fputs("\n<..>", f);
				    cursor = 4L;
				}
			}
		}
	    else
		{
		    n = MultupGField(multup, 0);
		    if (symCount > (unsigned) (((DOMAIN) MAXDOMAIN) - n))
			s = symNames[(unsigned) (((DOMAIN) MAXDOMAIN) - n)];
		    else
			{
			    (void) sprintf(dupBuffer, "%lu",
					   (unsigned long) n);
			    s = dupBuffer;
			}
		    outpBLen = (unsigned) strlen(s);
		    if (MultupGArity(multup) == 1)
			{
			    if (MultupGCount(multup) != 1L)
				{
				    (void) sprintf(outpBuffer, "%lu",
					   MultupGCount(multup));
				    outpBLen += (unsigned) strlen(outpBuffer);
				    if (cursor + (unsigned long) (outpBLen + 4)
					<= lineLength)
					{
					    (void) fprintf(f, "%s<.%s.>",
							   outpBuffer, s);
					    cursor += (unsigned long)
						(outpBLen + 4);
					}
				    else
					{
					    (void) fprintf(f, "\n%s<.%s.>",
							   outpBuffer, s);
					    cursor = (unsigned long)
						(outpBLen + 4);
					}
				}
			    else
				{
				    if (cursor + (unsigned long) (outpBLen + 4)
					<= lineLength)
					{
					    (void) fprintf(f, "<.%s.>", s);
					    cursor += (unsigned long)
						(outpBLen + 4);
					}
				    else
					{
					    (void) fprintf(f, "\n<.%s.>", s);
					    cursor = (unsigned long)
						(outpBLen + 4);
					}
				}
			}
		    else
			{
			    if (MultupGCount(multup) != 1L)
				{
				    (void) sprintf(outpBuffer, "%lu",
						   MultupGCount(multup));
				    outpBLen += (unsigned) strlen(outpBuffer);
				    if (cursor + (unsigned long) (outpBLen + 3)
					<= lineLength)
					{
					    (void) fprintf(f, "%s<.%s,",
							   outpBuffer, s);
					    cursor += (unsigned long)
						(outpBLen + 3);
					}
				    else
					{
					    (void) fprintf(f, "\n%s<.%s,",
							   outpBuffer, s);
					    cursor = (unsigned long)
						(outpBLen + 3);
					}
				}
			    else
				{
				    if (cursor + (unsigned long) (outpBLen + 3)
					<= lineLength)
					{
					    (void) fprintf(f, "<.%s,", s);
					    cursor += (unsigned long)
						(outpBLen + 3);
					}
				    else
					{
					    (void) fprintf(f, "\n<.%s,", s);
					    cursor = (unsigned long)
						(outpBLen + 3);
					}
				}
			    for (i = 1; i < MultupGArity(multup) - 1; i++)
				{
				    n = MultupGField(multup, i);
				    if (symCount > (unsigned)
					(((DOMAIN) MAXDOMAIN) - n))
					s = symNames[(unsigned)
						     (((DOMAIN) MAXDOMAIN)
						      - n)];
				    else
					{
					    (void) sprintf(dupBuffer,
							   "%lu",
							   (unsigned long) n);
					    s = dupBuffer;
					}
				    outpBLen = (unsigned) strlen(s);
				    if (cursor + (unsigned long) (outpBLen + 1)
					<= lineLength)
					{
					    (void) fprintf(f, "%s,", s);
					    cursor += (unsigned long)
						(outpBLen + 1);
					}
				    else
					{
					    (void) fprintf(f, "\n%s,", s);
				    cursor = (unsigned long)
					(outpBLen + 1);
					}
				}
			    n = MultupGField(multup, MultupGArity(multup) - 1);
			    if (symCount > (unsigned)
				(((DOMAIN) MAXDOMAIN) - n))
				s = symNames[(unsigned)
					     (((DOMAIN) MAXDOMAIN) - n)];
			    else
				{
				    (void) sprintf(dupBuffer, "%lu",
						   (unsigned long) n);
				    s = dupBuffer;
				}
			    outpBLen = (unsigned) strlen(s);
			    if (cursor + (unsigned long) (outpBLen + 2)
				<= lineLength)
				{
				    (void) fprintf(f, "%s.>", s);
				    cursor += (unsigned long) (outpBLen + 2);
				}
			    else
				{
				    (void) fprintf(f, "\n%s.>", s);
				    cursor = (unsigned long) (outpBLen + 2);
				}
			}
		}
	}
    (void) fputc((int) '\n', f);
    outpBLen = 0;
    cursor = 0L;
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  WldVerbose
* Purpose:
* Arguments: world -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, WldVerbose,
    unsigned long, world,
    FILE *, f)
{
    unsigned i;
    STATE *state;

    state = StaAlloc();
    BStaDecode(GphNode(theGraph, world), state);
    NodeBelongs(f, theGraph, world);
    for (i = 0; i < MagGNofPlaces(); i++)
	{
	    if (!MarEmpty(StaGMarking(state, i)))
		{
		    cursor = (unsigned long) strlen(plaNames[i]);
		    if (cursor + 4L < lineLength)
			{
			    (void) fprintf(f, "  %s: ", plaNames[i]);
			    cursor += 4L;
			}
		    else
			{
			    (void) fprintf(f, "  %s:\n", plaNames[i]);
			    cursor = 0L;
			}
		    MarOutput(StaGMarking(state, i), f);
		}
	}
    StaDiscard(state);
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  ArwVerbose
* Purpose:
* Arguments: awNr -
*            awAddr -
* Returns:
* See also:
*
****************************************************************************
*/
FE3(void, ArwVerbose,
    unsigned long, awNr,
    unsigned long, awAddr,
    FILE *, f)
{
    unsigned transition = 0;
    unsigned priority = 0;
    unsigned i;
    BOOL faithful = FALSE;
    DOMAIN *instance;
    DATATYPE *pointer;

    (void) fprintf(f, "Arrow %lu: ", awNr);
    pointer = GphArrow(theGraph, awAddr);
    do
	{
	    instance =
		BInsDecode(&transition, &priority, &faithful, &pointer);
	    (void) fprintf(f, "transition %s, precedence class ",
			   traNames[transition]);
	    PriOutput(priority, f);
	    (void) fputc((int) '\n', f);
	    for (i = 0; i < IngGProfile(transition); i++)
		{
		    (void) fprintf(f, "  %s = ",
				   traVarNames[transition][i]);
		    DomOutput(instance[i], f);
		    (void) fputc((int) '\n', f);
		}
	    Deallocate((char *) instance);
	    if (symmMethod)
		{
		    if (faithful)
			(void) fputs("produces the successor marking\n", f);
		    else
			(void)
			    fputs
			    ("produces a marking equivalent but not equal to\nthe successor marking\n",
			     f);
		}
	} while (pointer);
    (void) fflush(f);
}

/***************************************************************************
*
* Function:  PthVerbose
* Purpose:
* Arguments: path -
*            f -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PthVerbose,
    PATH *, path,
    FILE *, f)
{
    STEP *step;

    (void) fputs("PATH\n", f);
    (void) fflush(f);
    LSTNEST(path->steps, step, STEP)
	;
    LSTUNNEST (path->steps, step, STEP)
	{
	    WldVerbose(step->world, f);
	    ArwVerbose(GphAwOrdinal(theGraph, step->awAddr), step->awAddr, f);
	}
    WldVerbose(path->end, f);
    CutVerbose(f);
}

/***************************************************************************
*
* Function:  BStaDecode
* Purpose:
* Arguments: buffer -
*            state -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, BStaDecode,
    DATATYPE *, buffer,
    STATE *, state)
{
    if (!unfold)
	StaDecode(buffer, state);
    else
	{
	    TUPLE *tuple;
	    unsigned place;
	    register DATATYPE *pointer;
	    register unsigned long a;

	    for (place = 0; place < MagGNofPlaces(); place++)
		StaIMarking(state, place, MarCreate());
	    pointer = buffer;
	    while (*pointer)
		{
		    GetValue(unsigned long, a, pointer);
		    if (a >= PtrBlGCount(placeFoldings))
			Troba3("Illegal data for BStaDecode: %lu >= %lu",
			       a, PtrBlGCount(placeFoldings));
		    tuple = TupAlloc();
		    TupDecode(&place, (DATATYPE *)
			      PtrBlGet(placeFoldings, a), tuple);
		    if (CntAt(pointer))
			{
			    CntPass(pointer);
			    GetValue(unsigned long, a, pointer);
			    MarInsLast(StaGMarking(state, place),
				       MultupCons(a + 2L, tuple));
			}
		    else
			MarInsLast(StaGMarking(state, place),
				   MultupCons((unsigned long) 1L, tuple));
		}
	}
}

/***************************************************************************
*
* Function:  BInsDecode
* Purpose:
* Arguments: trIndex -
*            priority -
*            faithful -
*            argPointer -
* Returns:
* See also:
*
****************************************************************************
*/
FE4(DOMAIN *, BInsDecode,
    unsigned *, trIndex,
    unsigned *, priority,
    BOOL *, faithful,
    DATATYPE **, argPointer)
{
#ifdef WEIRD
    DATATYPE *pointer;
#else
    register DATATYPE *pointer;
#endif
    register unsigned long a;

#ifndef WEIRD
    pointer = *argPointer;
    *argPointer = NULL;
#endif
    *faithful = TRUE;
    if (unfold)
	{
#ifdef WEIRD
	    pointer = *argPointer;
#endif
	    GetValue(unsigned long, a, pointer);
	    if (*pointer)
		*argPointer = pointer;
#ifdef WEIRD
	    else
		*argPointer = NULL;
#endif
	    if (symmMethod)
		{
		    if (a & 1L)
			*faithful = FALSE;
		    a >>= 1;
		}
	    if (a >= PtrBlGCount(transFoldings))
		Troba3("Illegal data for BInsDecode: %lu >= %lu",
		       a, PtrBlGCount(transFoldings));
	    pointer = (DATATYPE *) PtrBlGet(transFoldings, a);
#ifdef WEIRD
	    return(InsDecode(trIndex, priority, &pointer));
#endif
	}
    return(InsDecode(trIndex, priority,
#ifdef WEIRD
		     argPointer
#else
		     pointer
#endif
		     ));
}
