/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  expr.h
*
* Part of: PROD's programs probe and prpp
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See expr.c.
*
* Version history:
*     $Log$
***************************************************************************/

#define NO_ERROR 0
#define CML_ERROR 1
#define FOR_ERROR 2
#define XPR_ERROR 3
#define MEX_ERROR 4
#define SOP_ERROR 5
#define MAC_ERROR 6
#define SYM_ERROR 7
#define SEM_ERROR 8
#define BUF_ERROR 9

extern int errorFlag;
extern BOOL continueEvaluation;
extern TUPLE *fogTuple;
#define FogGTuple() fogTuple
#define FogSTuple(tuple) (fogTuple = (tuple))

typedef struct _xrange 
{
    struct _expression *lower, *upper;
} XRANGE;

#define TRANGE LIST

typedef struct _markexpr 
{
    unsigned operator;
    union 
	{
	    MARKING *marking;
	    unsigned place;
	    TRANGE *trange;
	    struct 
		{
		    struct _expression *esub;
		    struct _markexpr *msub;
		} em;
	    struct 
		{
		    struct _markexpr *sub1m, *sub2m;
		} mm;
	} um;
} MARKEXPR;

typedef struct _expression 
{
    unsigned operator;
    union 
	{
	    DOMAIN value;
	    struct 
		{
		    MARKEXPR *msub1, *msub2;
		} me;
	    struct 
		{
		    struct _expression *sub1e, *sub2e, *sub3e;
		} oe;
	} ue;
} EXPRESSION;

typedef MARKING *(*MARF)();

typedef struct 
{
    unsigned kind;
    MARF eval;
    UNSF reduce;
} MEXOPENTRY;

typedef struct 
{
    unsigned kind;
    DOMF eval;
    UNSF reduce;
} XPROPENTRY;

#define MO_CONSTANT  0
#define MO_PLACE     1
#define MO_TRANGE    2
#define MO_TIMES     3
#define MO_CARTES    4
#define MO_SUM       5
#define MO_DIFF      6
#define MO_SELECT    7
#define MO_INTERS    8
#define SZ_MO 9
extern MEXOPENTRY mexOpTable[SZ_MO];

#define MK_CONSTANT  0
#define MK_PLACE     1
#define MK_TRANGE    2
#define MK_EXPRMARK  3
#define MK_MARKMARK  4

#define EO_DOM       0
#define EO_FIELD     1
#define EO_CARD      2
#define EO_NEG       3
#define EO_NOT       4
#define EO_BNOT      5
#define EO_TIMES     6
#define EO_DIV       7
#define EO_REM       8
#define EO_MINUS     9
#define EO_PLUS     10
#define EO_LSHIFT   11
#define EO_RSHIFT   12
#define EO_LESS     13
#define EO_LEQUAL   14
#define EO_EQUAL    15
#define EO_NEQUAL   16
#define EO_BAND     17
#define EO_BXOR     18
#define EO_BOR      19
#define EO_AND      20
#define EO_OR       21
#define EO_IF       22
#define EO_MLESS    23
#define EO_MLEQUAL  24
#define EO_MEQUAL   25
#define EO_MNEQUAL  26
#define EO_TRAVAR   27
#define EO_PRECLASS 28
#define EO_FAITHFUL 29
#define SZ_EO 30
extern XPROPENTRY xprOpTable[SZ_EO];

#define EK_CONSTANT 0
#define EK_CARD     1
#define EK_MCOMP    2
#define EK_NULLARY  3
#define EK_UNARY    4
#define EK_BINARY   5
#define EK_TRINARY  6

#define RE_CONSTANT ((unsigned) 0)
#define RE_GLOBAL   ((unsigned) 1)
#define RE_LOCAL    ((unsigned) 2)
    
#define XrgAlloc() ((XRANGE *) Allocate(sizeof(XRANGE)))
extern TOOLS xrgTools;
#define XrgGLower(xrange) ((xrange)->lower)
#define XrgGUpper(xrange) ((xrange)->upper)

#define TrgCreate() LstCreate(&xrgTools)
#define TrgInsert(trange, xrange) ((void) LstInsLast(trange,(char *)(xrange)))
#define TrgCopy(trange) LstCopy(trange, TRUE)
#define TrgDiscard(trange) LstDestroy(trange, TRUE)
#define TrgCastDiscard(trange) \
(LstDstroy((LIST *) trange, TRUE), (trange) = NULL)
#define TrgArity(trange) LstSize(trange)
#define TRGFOR(trange, xrange) LSTFOR(trange, xrange, XRANGE)
#define TRGNUM(trange, xrange, num) LSTNUM(trange, xrange, XRANGE, num)

#define MexAlloc() ((MARKEXPR *) Allocate(sizeof(MARKEXPR)))
#define MexKind(markexpr) (mexOpTable[(markexpr)->operator].kind)
#define MexEval(markexpr) \
    ((*(mexOpTable[(markexpr)->operator].eval))(markexpr))
#define SetMexEval(operator,_eval) (mexOpTable[operator].eval = (_eval))
#define MexReduce(markexpr) \
    ((*(mexOpTable[(markexpr)->operator].reduce))(markexpr))
#define SetMexReduce(operator,_reduce) \
    (mexOpTable[operator].reduce = (_reduce))

#define XprAlloc() ((EXPRESSION *) Allocate(sizeof(EXPRESSION)))
#define XprKind(expression) (xprOpTable[(expression)->operator].kind)
#define XprEval(expression) \
    ((*(xprOpTable[(expression)->operator].eval))(expression))
#define SetXprEval(operator,_eval) (xprOpTable[operator].eval = (_eval))
#define XprReduce(expression) \
    ((*(xprOpTable[(expression)->operator].reduce))(expression))
#define SetXprReduce(operator,_reduce) \
    (xprOpTable[operator].reduce = (_reduce))

DE4(MARKEXPR *, MexCons,
    unsigned, operator,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2,
    unsigned, place)
DE4(MARKEXPR *, MexAux,
    unsigned, operator,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2,
    unsigned, place)
DE1(MARKEXPR *, MexCopy,
    MARKEXPR *, markexpr)
DE2(void, MexForce,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
DE1(void, MexDracsid,
    MARKEXPR *, markexpr)
#define MexDiscard(markexpr) \
(MexDracsid(markexpr), (markexpr) = (MARKEXPR *) NULL)
#define MexCastDiscard(markexpr) \
(MexDracsid((MARKEXPR *) (markexpr)), (markexpr) = NULL)
#ifdef DEBUG
DE2(void, MexDump,
    MARKEXPR *, markexpr,
    unsigned, depth)
#endif
DE0(unsigned, MexRPlace)
DE5(EXPRESSION *, XprCons,
    unsigned, operator,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2,
    EXPRESSION *, ep3,
    DOMAIN, value)
DE5(EXPRESSION *, XprAux,
    unsigned, operator,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2,
    EXPRESSION *, ep3,
    DOMAIN, value)
DE1(EXPRESSION *, XprCopy,
    EXPRESSION *, expression)
DE2(void, XprForce,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
DE1(void, XprDracsid,
    EXPRESSION *, expression)
#define XprDiscard(expression) \
(XprDracsid(expression), (expression) = (EXPRESSION *) NULL)
#define XprCastDiscard(expression) \
(XprDracsid((EXPRESSION *) (expression)), (expression) = NULL)
#ifdef DEBUG
DE2(void, XprDump,
    EXPRESSION *, expression,
    unsigned, depth)
#endif
DE2(XRANGE *, XrgCons,
    EXPRESSION *, lower,
    EXPRESSION *, upper)
DE2(XRANGE *, XrgAux,
    EXPRESSION *, lower,
    EXPRESSION *, upper)
DE1(XRANGE *, XrgCopy,
    XRANGE *, xrange)
DE1(void, XrgDracsid,
    XRANGE *, xrange)
#define XrgDiscard(xrange) (XrgDracsid(xrange), (xrange) = (XRANGE *) NULL)
#ifdef DEBUG
DE2(void, XrgDump,
    XRANGE *, xrange,
    unsigned, depth)
DE2(void, TrgDump,
    TRANGE *, trange,
    unsigned, depth)
#endif
DE1(MARKING *, TrgMarking,
    TRANGE *, trange)
