/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  formula.h
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko, Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See formula.c.
*
* Version history:
*     $Log$
***************************************************************************/

#ifndef FORMULA_H
#define FORMULA_H

extern GRAPH *fogGraph;
extern unsigned long fogWorld;
extern unsigned fogTransition;
extern DOMAIN *fogInstance;
extern unsigned fogPriority;
extern BOOL fogFaithful;
#define FogGGraph() fogGraph
#define FogSGraph(graph) (fogGraph = (graph))
#define FogGWorld() fogWorld
#define FogSWorld(world) (fogWorld = (world))
#define FogGTransition() fogTransition
#define FogATransition() (&fogTransition)
#define FogSTransition(transition) (fogTransition = (transition))
#define FogGInstance(i) fogInstance[i]
#define FogSInstance(instance) (fogInstance = (instance))
#define FogGPriority() fogPriority
#define FogAPriority() (&fogPriority)
#define FogSPriority(priority) (fogPriority = (priority))
#define FogGFaithful() fogFaithful
#define FogAFaithful() (&fogFaithful)
#define FogSFaithful(faithful) (fogFaithful = (faithful))

typedef struct _fclass 
{
    unsigned operator;
    union 
	{
	    struct 
		{
		    unsigned transition;
		    EXPRESSION *expression;
		} tc;
	    struct {
		struct _fclass *sub1c, *sub2c;
	    } sc;
	} uc;
} FCLASS;

typedef struct _formula 
{
    unsigned operator;
    union
	{
	    BOOL visited;
	    struct
		{
		    int count;
		    BALTREE *paths;
		} sf;
	    struct
		{
		    unsigned long state, wdCount;
		} nf;
	    struct 
		{
		    unsigned long awCount, awAddr, succ;
		} af;
	    struct 
		{
		    LIST *front;
		    int phase;
		    BITSET *used;
		} bf;
	    struct 
		{
		    PATH *path;
		    int phase;
		    BITSET *used;
		    BITSET *inPath;
		} df;
	    struct
		{
		    unsigned long present, head, tail;
		    PATH *path, *next;
		} pf;
	} mf;
    union 
	{
	    struct _setOper *setOper;
	    EXPRESSION *expression;
	    struct 
		{
		    FCLASS *fclass;
		    struct _formula *sub;
		} ff;
	    struct 
		{
		    struct _formula *sub1f, *sub2f, *sub3f;
		} of;
	} uf;
} FORMULA;

typedef struct 
{
    unsigned _kind;
    VOF _init;
    VOF _clear;
    PTHF _eval;
    VOF _close;
} FOROPENTRY;

#define FC_ATOM      0
#define FC_NOT       1
#define FC_OR        2
#define FC_AND       3

#define FP_NOLOOP    0
#define FP_ISLOOP    1
#define FP_LOOPDONE  2
#define FP_FORWARD   3
#define FP_BACKTRACK 4

#define FO_FALSE      0
#define FO_TRUE       1
#define FO_NOT        2
#define FO_NODE       3
#define FO_STEP       4
#define FO_FIRE       5
#define FO_BPATH      6
#define FO_DPATH      7
#define FO_BSPAN      8
#define FO_DSPAN      9
#define FO_AT        10
#define FO_INSET     11
#define FO_EXPR      12
#define FO_AND       13
#define FO_OR        14
#define FO_CTL_FALSE 15
#define FO_CTL_TRUE  16
#define FO_CTL_NOT   17
#define FO_CTL_EF    18
#define FO_CTL_EG    19
#define FO_CTL_EU    20
#define FO_CTL_EX    21
#define FO_CTL_AF    22
#define FO_CTL_AG    23
#define FO_CTL_AU    24
#define FO_CTL_AX    25
#define FO_CTL_EXPR  26
#define FO_CTL_AND   27
#define FO_CTL_OR    28
#define SZ_FO 29
extern FOROPENTRY forOpTable[SZ_FO];

#define FK_SET        0
#define FK_FIRE       1
#define FK_EXPRESSION 2
#define FK_NULLARY    3
#define FK_UNARY      4
#define FK_BINARY     5
#define FK_TRINARY    6

#define FclAlloc() ((FCLASS *) Allocate(sizeof(FCLASS)))
#define ForAlloc() ((FORMULA *) Allocate(sizeof(FORMULA)))
#define ForKind(formula) (forOpTable[(formula)->operator]._kind)
#define ForInit(formula, graph) \
    ((*(forOpTable[(formula)->operator]._init))(formula, graph))
#define ForClear(formula, graph, world) \
    ((*(forOpTable[(formula)->operator]._clear))(formula, graph, world))
#define ForEval(formula, graph, world) \
    (((formula)->operator == FO_TRUE) ? ForETrue(formula, world) \
     : (*(forOpTable[(formula)->operator]._eval))(formula, graph, world))
#define ForClose(formula, graph) \
    ((*(forOpTable[(formula)->operator]._close))(formula, graph))

#define SetExMex Set2ExMex

DE4(FCLASS *, FclCons,
    unsigned, operator,
    FCLASS *, fcl1,
    FCLASS *, fcl2,
    unsigned, transition)
DE4(FCLASS *, FclAux,
    unsigned, operator,
    FCLASS *, fcl1,
    FCLASS *, fcl2,
    unsigned, transition)
DE1(FCLASS *, FclCopy,
    FCLASS *, fclass)
DE1(void, FclDracsid,
    FCLASS *, fclass)
#define FclDiscard(fclass) (FclDracsid(fclass), (fclass) = (FCLASS *) NULL)
#define FclCastDiscard(fclass) \
(FclDracsid((FCLASS *) (fclass)), (fclass) = NULL)
#ifdef DEBUG
DE2(void, FclDump,
    FCLASS *, fclass,
    unsigned, depth)
#endif
DE1(BOOL, FclInClass,
    FCLASS *, fclass)
DE4(FORMULA *, ForCons,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2,
    FORMULA *, fp3)
DE4(FORMULA *, ForAux,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2,
    FORMULA *, fp3)
DE1(FORMULA *, ForCopy,
    FORMULA *, formula)
DE1(void, ForCardisd,
    FORMULA *, formula)
#define ForDiscard(formula) \
(ForCardisd(formula), (formula) = (FORMULA *) NULL)
#define ForCastDiscard(formula) \
(ForCardisd((FORMULA *) (formula)), (formula) = NULL)
#ifdef DEBUG
DE2(void, ForDump,
    FORMULA *, formula,
    unsigned, depth)
#endif
DE2(PATH *, ForETrue,
    FORMULA *, formula,
    unsigned long, world)
DE0(void, SetExMex)

#endif
