/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  lexer.c
*
* Part of: PROD's program prpp
*
* Authors: Heikki Suonsivu and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module performs most of the lexical analysis of the
*     net description.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "util.h"
#include "prpp.h"
#include "dynamic.h"
#include "lexer.h"

#define TEXT_STATE int

#define INQUOTE 1
#define INNAME 2
#define INNUMBER 3
#define INSPACE 4
#define INCOMMENT 5
#define INNONE 6

PP_TOKEN_TYPE closing_bracket_pair[4] =
{
    PPCLOSING_BRACKET, PPCLOSING_SQUARE_BRACKET, PPCLOSING_PARENTHESIS,
    PPTUPLE_RIGHT
};

/***************************************************************************
*
* Function:  finish_token
* Purpose:
* Arguments: buffer -
*            state -
* Returns:
* See also:
*
****************************************************************************
*/
FS2(PP_TOKEN *, finish_token,
    char *, buffer,
    TEXT_STATE, state)
{
    register PP_TOKEN *token;
    register int c;

    token = PpTokCons(buffer, PPNOTHING);
    Deallocate((char *) buffer);
    switch (state)
	{
	case INQUOTE:
	    token->type = PPQUOTED_EXPRESSION;
	    break;

	case INNAME:
	    token->type = PPNAME;
	    break;

	case INNUMBER:
	    token->type = PPNUMBER;
	    break;

	case INSPACE:
	    token->type = PPSPACE;
	    break;

	case INCOMMENT:
	    token->type = PPCOMMENT;
	    break;

	case INNONE:
	    switch (*(token->token))
		{
		case (int) '#':
		    token->type = PPDV;
		    break;

		case (int) '\n':
		    token->type = PPNEWLINE;
		    break;

		case (int) '(':
		    token->type = PPOPENING_PARENTHESIS;
		    break;

		case (int) '[':
		    token->type = PPOPENING_SQUARE_BRACKET;
		    break;

		case (int) '{':
		    token->type = PPOPENING_BRACKET;
		    break;

		case (int) ')':
		    token->type = PPCLOSING_PARENTHESIS;
		    break;

		case (int) ']':
		    token->type = PPCLOSING_SQUARE_BRACKET;
		    break;

		case (int) '}':
		    token->type = PPCLOSING_BRACKET;
		    break;

		case (int) ',':
		    token->type = PPCOMMA;
		    break;

		case (int) '+':
		    token->type = PPPLUS;
		    break;

		case (int) '-':
		    token->type = PPMINUS;
		    break;

		case (int) '*':
		    token->type = PPMULTIPLY;
		    break;

		case (int) '/':
		    token->type = PPDIVIDE;
		    break;

		case (int) '%':
		    token->type = PPREMAINDER;
		    break;

		case (int) '!':
		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '=')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("!=");
				    token->type = PPNOT_EQUAL;
				    break;
				}
			    c_ppio_unget(c);
			}

		    token->type = PPNOT;
		    break;

		case (int) '^':
		    token->type = PPBINARY_XOR;
		    break;

		case (int) '~':
		    token->type = PPBINARY_NOT;
		    break;

		case (int) '?':
		    token->type = PPCONDITIONAL;
		    break;

		case (int) ':':
		    token->type = PPCOLON;
		    break;

		case (int) ';':
		    token->type = PPSEMICOLON;
		    break;

		case (int) '<':

		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '=')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("<=");
				    token->type = PPLESS_THAN_OR_EQUAL;
				    break;
				}
			    else if (c == (int) '<')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("<<");
				    token->type = PPSHIFT_LEFT;
				    break;
				}
			    else if (c == (int) '.')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("<.");
				    token->type = PPTUPLE_LEFT;
				    break;
				}
			    c_ppio_unget(c);
			}
		    token->type = PPLESS_THAN;
		    break;

		case (int) '>':

		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '=')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy(">=");
				    token->type = PPMORE_THAN_OR_EQUAL;
				    break;
				}
			    else if (c == (int) '>')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy(">>");
				    token->type = PPSHIFT_RIGHT;
				    break;
				}
			    c_ppio_unget(c);
			}
		    token->type = PPMORE_THAN;
		    break;

		case (int) '.':

		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '>')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy(".>");
				    token->type = PPTUPLE_RIGHT;
				    break;
				}
			    else if (c == (int) '.')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("..");
				    token->type = PPTUPLE_RANGE;
				    break;
				}
			    c_ppio_unget(c);
			}
		    token->type = PPSPECIAL;
		    break;
		case (int) '=':
		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '=')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("==");
				    token->type = PPEQUAL;
				    break;
				}
			    c_ppio_unget(c);
			}

		    token->type = PPSPECIAL;
		    break;

		case (int) '&':
		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '&')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("&&");
				    token->type = PPAND;
				    break;
				}
			    c_ppio_unget(c);
			}

		    token->type = PPBINARY_AND;
		    break;

		case (int) '|':
		    c = ppio_getc();
		    if (c != EOF)
			{
			    if (c == (int) '|')
				{
				    Deallocate((char *) token->token);
				    token->token = StrCopy("||");
				    token->type = PPOR;
				    break;
				}

			    c_ppio_unget(c);
			}

		    token->type = PPBINARY_OR;
		    break;

		default:
		    token->type = PPSPECIAL;
		}
	    break;

	default:
	    (void) sprintf(errStr, "Unknown state %d in get_token", state);
	    panic(errStr);

	}
    return token;
}

/***************************************************************************
*
* Function:  get_token
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(PP_TOKEN *, get_token)
{
  register int c, quotechar;
  TEXT_STATE state = INNONE;
  BUFFER_DEFINITION(register);

  INITIALIZE_BUFFER(8);

  quotechar = EOF;
  for (c = ppio_getc(); c != EOF; c = ppio_getc())
    {
      switch (state)
	{
	 case INQUOTE:
	  if (c == quotechar)
	    {

	      ADD_BUFFER(c);
	      ADD_BUFFER('\0');
	      return finish_token(buffer, state);
	    }
	  else if (c == (int) '\\')
	    {

	      c = ppio_getc();
	      if (c == EOF)
		{
		  panic("Premature end inside a string constant.");

		}
	      ADD_BUFFER('\\');
	    }

	  ADD_BUFFER(c);
	  break;

	 case INNAME:

	  if ((!isalnum((UCHAR) c)) && (c !=  '_') && (c != '$'))
	      {
		  c_ppio_unget(c);
		  ADD_BUFFER('\0');
		  return finish_token(buffer, state);
	      }
	  ADD_BUFFER(c);
	  break;

	 case INNUMBER:

	  if (!isdigit((UCHAR) c))
	      {
		  c_ppio_unget(c);
		  ADD_BUFFER('\0');
		  return finish_token(buffer, state);
	      }
	  ADD_BUFFER(c);
	  break;

	 case INSPACE:

	  if (c != ' ' && c != '\t')
	      {
		  c_ppio_unget(c);
		  ADD_BUFFER('\0');
		  return finish_token(buffer, state);
	      }
	  ADD_BUFFER(c);
	  break;

	 case INCOMMENT:
	  {
            register unsigned long line = 0L;
	    register char *filename = NULL;


	    may_have_lost_line = TRUE;
	    if (curFile)
	      {
		filename = curFile->name;
		line = curFile->line;
	      }
	    while (c != EOF)
	      {
		ADD_BUFFER(c);
		if (c == (int) '*')
		  {
		    c = ppio_getc();

		    if (c == (int) '/')
		      {
			ADD_BUFFER(c);
			break;
		      }
		  }
		else
		  c = ppio_getc();
	      }
	    if (c == EOF)
	      {
		if (filename)
		    {
			(void) sprintf(errStr,
                                       "Unterminated comment starts in %s at line %lu",
				       filename, line);
			error(errStr);
		    }
		else
		    error("Unterminated comment, I have lost start information?");
	      }
	    ADD_BUFFER('\0');
	    if (preserve_comments)
	      return finish_token(buffer, state);
	    else if (ansi)
		{
		    Deallocate((char *) buffer);
		    return finish_token(StrCopy(" "), INSPACE);
		}
	    p = buffer;
	    state = INNONE;
	    break;
	  }

	 case INNONE:

	  if (isdigit((UCHAR) c))
	    {
	      ADD_BUFFER(c);
	      state = INNUMBER;
	      continue;
	    }

	  if (isalpha((UCHAR) c) || (c == (int) '_') || (c == (int) '$'))
	    {
	      ADD_BUFFER(c);
	      state = INNAME;
	      continue;
	    }

	  if (c == (int) ' ' || c == (int) '\t')
	    {
	      ADD_BUFFER(c);
	      state = INSPACE;
	      continue;
	    }

	  switch (c)
	    {
	     case (int) '\'':
	     case (int) '"':
	      ADD_BUFFER(c);
	      state = INQUOTE;
	      quotechar = c;
	      continue;

	     case (int) '\\':

	      if (parsing_control)
		{
		  c = ppio_getc();


		  if (c == EOF)
		    {
		      c = (int) '\\';
		      break;
		    }

		  if (c == (int) '\n')
		    {

		      p = buffer;
		      continue;
		    }
		  c_ppio_unget(c);
		  c = (int) '\\';
		}
	      break;

	     case (int) '/':

	      c = ppio_getc();
	      if (c == (int) '*')
		  {
		      ADD_BUFFER('/');
		      ADD_BUFFER('*');
		      state = INCOMMENT;
		      continue;
		  }
	      else
		  {
		      c_ppio_unget(c);
		      c = (int) '/';
		  }
	      break;
	    }


	  ADD_BUFFER(c);
	  ADD_BUFFER('\0');
	  return finish_token(buffer, state);

	 default:
	  (void) sprintf(errStr, "Unknown state %d in get_token", state);
	  panic(errStr);

	}
    }

  if (p != buffer)
    {
      ADD_BUFFER('\0');
      return finish_token(buffer, state);
    }
  Deallocate((char *) buffer);
  return NULL;
}

/***************************************************************************
*
* Function:  PpTokCons
* Purpose:
* Arguments: s -
*            type -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(PP_TOKEN *, PpTokCons,
    char *, s,
    PP_TOKEN_TYPE, type)
{
    PP_TOKEN *token;

    token = (PP_TOKEN *) Allocate(sizeof(PP_TOKEN));
    token->token = StrCopy(s);
    token->type = type;
    (void) LstInsFirst(ppTokList, (char *) token);
    return token;
}

/***************************************************************************
*
* Function:  PpTokDracsid
* Purpose:
* Arguments: token -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, PpTokDracsid,
    PP_TOKEN *, token)
{
    PP_TOKEN *ppToken;
    VOID *ppCursor;

    if (((PP_TOKEN *) NULL) == token)
	return;
    if (((LIST *) NULL) != ppTokList)
	{
	    for (ppCursor = (VOID *) NULL,
		 ppToken = (PP_TOKEN *) LstBegin(ppTokList);
		 (((PP_TOKEN *) NULL) != ppToken) && (token != ppToken);
		 ppCursor = LstGet(ppTokList),
		 ppToken = (PP_TOKEN *) LstStep(ppTokList))
		;
	    if (((PP_TOKEN *) NULL) != ppToken)
		{
		    LstSet(ppTokList, ppCursor);
		    (void) LstDelNext(ppTokList, FALSE);
		}
	}
    Deallocate(token->token);
    Deallocate((char *) token);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  PpTokDump
* Purpose:
* Arguments: token -
*            depth -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PpTokDump,
    PP_TOKEN *, token,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fputs("PP_TOKEN: ", _err);
    if (((PP_TOKEN *) NULL) != token)
	{
	    (void) fputs("token ", _err);
	    if (((char *) NULL) != token->token)
		{
		    (void) fputc((int) '"', _err);
		    (void) fputs(token->token, _err);
		    (void) fputc((int) '"', _err);
		}
	    else
		(void) fputs("NULL", _err);
	    (void) fputs(", type ", _err);
	    (void) fprintf(_err, "%lu", (unsigned long) (token->type));
	}
    else
	(void) fputs("NULL", _err);
    (void) fputc((int) '\n', _err);
    (void) fflush(_err);
}
#endif

/* A collection of routines for the PP_TOKEN type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS ppTokTools =
{ (CFUNC) NULL, PpTokDracsid
#ifdef DEBUG
, PpTokDump
#endif
};

/***************************************************************************
*
* Function:  unget_token
* Purpose:
* Arguments: token -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, unget_token,
    PP_TOKEN *, token)
{
    if (token)
	s_ppio_unget(token->token);
}
