/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  m.c
*
* Part of: PROD's program "gphgen"
*
* Authors: Lasse Anderson, Robert Jurgens, Marko Rauhamaa,
*          and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This is the main module for the reachability graph generator
*     program of a net.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "inst.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"
#include "genesig.h"
#include "cmdnm.h"

static char *fileName = NULL;
static BOOL unfold = FALSE;

/***************************************************************************
*
* Function:  Misusage
* Purpose:   This routine handles most of the cases where the command line
*            of the program is incorrect.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, Misusage)
{
    MsgExit(
#ifdef ARGV0
	    "[-.] [-A] [-B] [-binteger] [-C] [-c]\n[-D] [-d] [-E] [-e] [-f] [-ginteger] [-i] [-L] [-minteger] [-P] [-q]\n[-rinteger] [-S] [-s] [-u] [-V] [-v] [-X] [-x] [-Y] [-y] [-Z] [-z]"
#else
	    "[-.] [-A] [-B] [-binteger] [-C] [-c]\n[-D] [-d] [-E] [-e] [-f] [-ginteger] [-i] [-L] [-minteger] [-P] [-q]\n[-rinteger] [-S] [-s] [-t]name [-u] [-V] [-v] [-X] [-x] [-Y] [-y] [-Z] [-z]"
#endif
	    , (int) 1);
}

FE0(void, GenMemClose)
{
    if (!unfold)
	_ClFollowers();
    GphClose(theGraph);
    GphClose(the2Graph);
    DelAraNames(TRUE);
    DelPlaNames();
    PthDiscard(thePath);
    PthDiscard(the2Path);
    PthDiscard(the3Path);
    PthDiscard(nowPath);
    StaDiscard(theState);
    StaDiscard(theFolState);
    StaDiscard(theEvalState);
    BalTrDestroy(theNodeTree, FALSE);
    IniStClose();
    PtClose();
    BaseClose();
    DynamClose();
    ClDeallocate(fileName);
    ClDeallocate(araArcName);
    EngDBuffer();
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments: argc - the number of command line arguments given to the
*                   program
*            argv - an array containing the arguments
* Returns:   0 if everything succeeds, and something else otherwise
*            (PROD has no systematic convention for non-zero exits.)
* See also:
*
****************************************************************************
*/
MAIN2()
{
    int argc = 0;
    char **argv = NULL;
    BINARYFILE bFile = NonBinaryFile();
    int i;
    unsigned long gCount, seedVal = 0L, a = 0L, gphGCount, createFlag;
    unsigned bits = 0, u, turn;
    char *argument, *bs, *gs, *ms, *rs, *str;
    BOOL bdds, cont, blk;
    PLC *plc;
    unsigned long mapLen = 0L;

#ifndef UNIX
    ClockOn();
#endif
    SetIO();
    ClearErrno();
    _buchiOK = _BuchiOK;
    baseName = StrCopy("gphgen");
    _memClose = (VOF) GenMemClose;
#ifdef ARGV0
#ifdef UNIX
    fileName = StrCopy(_argv_[0]);
#else
    u = 0;
    while (_argv_[0][u])
	u++;
    while (TRUE)
	{
	    if ((u < 2) ||
		(_argv_[0][u - 1] == DIRCHAR) || (_argv_[0][u - 1] == ':'))
		{
		    while (_argv_[0][u])
			u++;
		    u++;
		    break;
		}
	    if (_argv_[0][u - 1] == '.')
		break;
	    u--;
	}
    fileName = (char *) Allocate(u);
    u--;
    fileName[u] = '\0';
    while (u--)
	if (isalpha((UCHAR) (_argv_[0][u])) &&
	    isupper((UCHAR) (_argv_[0][u])))
	    fileName[u] = tolower((UCHAR) (_argv_[0][u]));
	else
	    fileName[u] = _argv_[0][u];
#endif
    myName = CmdNm(_argv_[0]);
#else
    myName = baseName;
    fileName = NULL;
#endif
#ifndef UNIX
    StoreCmd(_argc_, myName, (char **) _argv_);
#endif
    ExpArgs(_argc_, (char **) _argv_, &argc, &argv);
    IgnoreSignals();
    continueGeneration = TRUE;
    stopFlag = FALSE;
    SetSigHandlers();
    DynamInit();
    bs = NULL;
    gs = NULL;
    gCount = (unsigned long) -1L;
    ms = NULL;
    rs = NULL;
    bdds = FALSE;
    cont = FALSE;
    theDots = FALSE;
    AraResetProd();
    ResetAra();
    ResetPep();
    ClElimination();
    ResetDfs();
    ResetFreeExpansion();
    ClPtStub();
    ClPtSymm();
    ClPtVerbose();
    ClPtSleep();
    ResetPlcIMkSymm();
    ResetPlRnd();
    ClPtLazy();
    ResetTrRnd();
    ClPtDelAlg();
    ClPtTryAgain();
    ClPtCFFD();
    SetPtDefVis();
    ClOnTheFly();
    ClBlocked();
#ifdef SMODELS
    ClPtSmodels();
    ClPtLongPlay();
#endif
    blk = FALSE;
    for (i = 1; i < argc; i++)
	{
	    argument = argv[i];
	    if (*argument == '-')
		{
		    argument++;
		    if ((*argument == '\0') &&
			(i + 1 == argc))
			Misusage();
		    if (*argument == '\0')
			{
			    argument = argv[i + 1];
			    i++;
			}
		    while (*argument != '\0')
			{
			    if (*argument == '.')
			        theDots = TRUE;
#ifdef WEIRD
			    else if (*argument == '@')
				_c_o_m_p_a_c_t_ = TRUE;
#endif
			    else if (*argument == 'A')
				SetAra();
			    else if (*argument == 'B')
			        bdds = TRUE;
			    else if (*argument == 'b')
				{
				    argument++;
				    if ((bs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    bs = argument;
				    if (ScanLdu(bs, &a) != 1)
					Misusage();
				    else
					bits = (unsigned) a;
				    break;
				}
			    else if (*argument == 'C')
				{
				    SetPtCFFD();
				    SetPtStub();
				    SetDfs();
				}
			    else if (*argument == 'c')
				cont = TRUE;
			    else if (*argument == 'D')
				SetDfs();
			    else if (*argument == 'd')
				{
				    SetPtStub();
				    SetPtDelAlg();
				}
			    else if (*argument == 'E')
				{
				    SetPtSymm();
				    SetPlcIMkSymm();
				}
			    else if (*argument == 'e')
				SetPtSymm();
			    else if (*argument == 'f')
				{
				    SetDfs();
				    SetFreeExpansion();
				}
			    else if (*argument == 'g')
				{
				    argument++;
				    if ((gs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    gs = argument;
				    if (ScanLdu(gs, &gCount) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'i')
				ClPtDefVis();
			    else if (*argument == 'L')
				{
#ifdef SMODELS
				    SetPtStub();
				    SetPtDelAlg();
				    SetPtSmodels();
#else
				    (void)
					fprintf
					(_err,
					 "\n%s refuses to work with -L\nbecause the Smodels library has not been included appropriately.\n",
					 myName);
				    (void) fflush(_err);
				    Exit(1);
#endif
				}
			    else if (*argument == 'm')
				{
				    argument++;
				    if ((ms != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    ms = argument;
				    if (ScanLdu(ms, &mapLen) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'P')
				SetPep();
			    else if (*argument == 'q')
				SetElimination();
			    else if (*argument == 'r')
				{
				    argument++;
				    if ((rs != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    rs = argument;
				    if (ScanLdu(rs, &seedVal) != 1)
					Misusage();
				    break;
				}
			    else if (*argument == 'S')
				SetPtSleep();
			    else if (*argument == 's')
				SetPtStub();
#ifndef ARGV0
			    else if (*argument == 't')
				{
				    argument++;
				    if ((fileName != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    fileName = argument;
				    break;
				}
#endif
			    else if (*argument == 'u')
				unfold = TRUE;
			    else if (*argument == 'V')
				SetPtSuperVerbose();
			    else if (*argument == 'v')
				SetPtVerbose();
			    else if (*argument == 'X')
				{
#ifdef SMODELS
				    SetPtStub();
				    SetPtDelAlg();
				    SetPtSmodels();
				    SetPtLongPlay();
#else
				    (void)
					fprintf
					(_err,
					 "\n%s refuses to work with -X\nbecause the Smodels library has not been included appropriately.\n",
					 myName);
				    (void) fflush(_err);
				    Exit(1);
#endif
				}
			    else if (*argument == 'x')
				SetPlRnd();
                            else if (*argument == 'Y')
                                {
                                    SetBlocked();
                                    blk = TRUE;
                                }
			    else if (*argument == 'y')
				SetPtLazy();
			    else if (*argument == 'Z')
				SetPtTryAgain();
			    else if (*argument == 'z')
				{
				    SetPtLazy();
				    SetTrRnd();
				}
			    else
				Misusage();
			    argument++;
			}
		}
#ifndef ARGV0
	    else if (fileName == NULL)
		fileName = argument;
#endif
	    else
		Misusage();
	}
#ifndef ARGV0
    if (fileName == NULL)
	Misusage();
    i = strlen(fileName);
    if ((i > 4) && (strcmp(fileName + i - 4, ".gph") == 0))
	fileName[i - 4] = '\0';
#endif
    if ((gs != NULL) && (gCount == 0L))
	{
	    gCount = 1L;
	    (void) fprintf(_err, "\n-g %s transformed to -g %lu\n",
			   gs, gCount);
	    (void) fflush(_err);
	}
    str = FileName(fileName, "dat");
    bFile = OpenRead(str, (BOOL *) NULL);
    ClDeallocate(str);
    if (NotBinaryFile(bFile))
	Exit(2);
    SeekAbs(bFile, (unsigned long) sizeof(unsigned long));
    BaseInit(bFile, TRUE);
    if (IngGNofClasses() != 1)
	{
#ifdef SMODELS
	    if (PtLongPlay())
		{
		    (void)
			fputs
			("\n-L replaces -X because of precedence classes\n",
			 _err);
		    (void) fflush(_err);
		    ClPtLongPlay();
		}
#endif
	    if (PtStub() && !PtDelAlg())
		{
		    (void)
			fputs
			("\n-d is forced because of precedence classes\n",
			 _err);
		    (void) fflush(_err);
		    SetPtDelAlg();
		}
	    if (PtSymm())
		{
		    (void)
			fputs
			("\nPrecedence classes prevent from\nusing symmetries\n",
			 _err);
		    (void) fflush(_err);
		    ClPtSymm();
		}
	    if (Pep())
		{
		    (void)
			fputs
			("\n-P is ignored because of precedence classes\n",
			 _err);
		    (void) fflush(_err);
		    ResetPep();
		}
	}
    if (OnTheFly())
	{
	    if (Pep())
		{
		    (void)
			fputs
			("\n-P is ignored because of on-the-fly verification\n",
			 _err);
		    (void) fflush(_err);
		    ResetPep();
		}
	    SetDfs();
	}
    if (PtStub() || PtSymm() || PtSleep())
	unfold = TRUE;
    if (cont) 
	{
	    theGraph = GphOpen(fileName, GPH_GENERATE, FALSE);
	    if (theGraph == NULL)
		Exit(4);
	    if (bdds)
		{
		    (void)
			fputs
			("\nToo late for BDD-based graph generation\n",
			 _err);
		    (void) fflush(_err);
		    bdds = FALSE;
		}
	    if (ms != NULL)
		GphMapToMem(theGraph, mapLen);
	    u = GphBits(theGraph);
	    if ((bs != NULL) && (bits != u))
		{
		    (void)
			fprintf
			(_err,
			 "\n-b %s is ignored (old hash table used)\n",
			 bs);
		    (void) fflush(_err);
		}
	    if (GphSymm(theGraph))
		{
		    if (GphIMkSymm(theGraph))
			{
			    if (PtSymm() && !PlcIMkSymm())
				{
				    (void)
					fputs
					("\nStill requiring symmetricity of the initial marking\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetPlcIMkSymm();
			}
		    else if (PlcIMkSymm())
			{
			    (void)
				fputs
				("\nToo late to require symmetricity of the initial marking\n",
				 _err);
			    (void) fflush(_err);
			    ResetPlcIMkSymm();
			}
		    SetPtSymm();
		}
	    else if (PtSymm())
		{
		    (void) fputs("\nToo late for symmetry method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtSymm();
		}
	    if (GphStub(theGraph))
		{
		    if (GphDelAlg(theGraph))
			{
			    if (PtStub() && !PtDelAlg())
				{
				    (void)
					fputs
					("\nContinuing with deletion algorithm\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetPtDelAlg();
			}
		    else if (PtDelAlg())
			{
			    (void)
				fputs
				("\nContinuing with incremental algorithm\n",
				 _err);
			    (void) fflush(_err);
			    ClPtDelAlg();
			}
		    if (GphCFFD(theGraph))
			{
			    if (PtLazy() && TrRnd() && !PtDelAlg())
				{
				    (void)
					fputs
					("\n-z is ignored because of CFFD-preserving incremental algorithm\n",
					 _err);
				    (void) fflush(_err);
				    ClPtLazy();
				    ResetTrRnd();
				}
			    SetPtCFFD();
			    SetDfs();
			}
		    else if (PtCFFD())
			{
			    (void)
				fputs
				("\nToo late for CFFD-preservation\n",
				 _err);
			    (void) fflush(_err);
			    ClPtCFFD();
			}
		    SetPtStub();
		}
	    else if (PtStub())
		{
		    (void) fputs("\nToo late for stubborn set method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtStub();
		    ClPtDelAlg();
#ifdef SMODELS
		    ClPtSmodels();
		    ClPtLongPlay();
#endif
		    ClPtCFFD();
		}
	    if (GphSleep(theGraph))
		SetPtSleep();
	    else if (PtSleep())
		{
		    (void) fputs("\nToo late for sleep set method\n",
				 _err);
		    (void) fflush(_err);
		    ClPtSleep();
		}
	    if (GphUnfold(theGraph))
		unfold = TRUE;
	    else if (unfold)
		{
		    (void) fputs("\nToo late for unfolding\n", _err);
		    (void) fflush(_err);
		    unfold = FALSE;
		}
	    if (GphDfs(theGraph))
		{
		    if (GphFreeExpansion(theGraph))
			{
			    if (Dfs() && !FreeExpansion())
				{
				    (void)
					fputs
					("\nContinuing with free expansion\n",
					 _err);
				    (void) fflush(_err);
				}
			    SetFreeExpansion();
			}
		    else if (FreeExpansion())
			{
			    (void) fputs("\nToo late for free expansion\n",
					 _err);
			    (void) fflush(_err);
			    ResetFreeExpansion();
			}
		    SetDfs();
		}
	    else if (Dfs())
		{
		    (void) fputs("\nToo late for depth first search\n",
				 _err);
		    (void) fflush(_err);
		    ResetDfs();
		    ResetFreeExpansion();
		}
	    if (!GphDefVis(theGraph))
		ClPtDefVis();
	    else if (!PtDefVis())
		{
		    (void)
			fputs
			("\nToo late to make transition instances invisible by default\n",
			 _err);
		    (void) fflush(_err);
		    SetPtDefVis();
		}
	    if (Ara())
		{
		    (void) fputs("\n-A is ignored because of -c\n",
				 _err);
		    (void) fflush(_err);
		    ResetAra();
		}
	    if (Pep())
		{
		    (void) fputs("\n-P is ignored because of -c\n",
				 _err);
		    (void) fflush(_err);
		    ResetPep();
		}
	    EngSBufSize(GphBufSize(theGraph));
	    EngSBuffer((DATATYPE *) AllocArray(EngBufSize(),
					       sizeof(DATATYPE)));
	    if (rs != NULL)
		GphSetSeed(theGraph, (UNS32) seedVal);
	    if (unfold)
		{
		    PtInit(bFile);
		    if (PtSymm() && PtVerbose())
			RdPlaNames(bFile);
		    OwnClose(&bFile);
		    if (Buchi() && PtStub())
			{
			    the2Graph = GphOpen(fileName, GPH_GENERATE, TRUE);
			    if (the2Graph == NULL)
				Exit(4);
			    if (ms != NULL)
				GphMapToMem(the2Graph, mapLen);
			    if (rs != NULL)
				GphSetSeed(the2Graph, (UNS32) seedVal);
			}
		    _CompInst();
		    PlcFill();
		    _CompPrec();
		    TrnFill();
		    if ((!NTrVisible()) && !Buchi())
			ClInfinite();
		    if (PtSymm())
			CompSymm(TRUE);
		}
	    else
		{
		    if (OnTheFly())
			{
			    theState = InitialState(bFile);
			    for (u = 0; u < MagGNofPlaces(); u++)
				MarDiscard(StaGMarking(theState, u));
			}
		    else
			theState = StaAlloc();
		    OwnClose(&bFile);
		}
	    thePath = PthCreate((unsigned long) -1L);
	}
    else 
	{
	    if (bs == NULL)
		bits = 10;
	    else if ((bits == 0) || (bits > 30))
		{
		    if (bits == 0)
			bits = 1;
		    else
			bits = 30;
		    (void) fprintf(_err, "\n-b %s transformed to -b %u\n",
				   bs, bits);
		    (void) fflush(_err);
		}
	    EngSBufSize((unsigned) 1024);
	    EngSBuffer((DATATYPE *) AllocArray(EngBufSize(),
					       sizeof(DATATYPE)));
#ifdef WEIRD
            if (_c_o_m_p_a_c_t_ && Ara())
		{
		    (void) fputs("\n-A is ignored because of -@\n",
				 _err);
		    (void) fflush(_err);
		    ResetAra();
		}
#endif
	    if (unfold)
		{
		    PtInit(bFile);
#ifndef DEBUG
		    if (Ara() || Pep())
#endif
			AraNames(bFile);
#ifndef DEBUG
		    if (Pep() || (PtSymm() && PtVerbose()))
#endif
			RdPlaNames(bFile);
		    OwnClose(&bFile);
		    if (Buchi() && PtStub())
			{
			    SetPtCFFD();
			    SetDfs();
			}
		    if ((OnTheFly() || PtCFFD() || PtSleep()) && PtSymm())
			{
			    (void)
				fputs
				("\n-E and -e are ignored because of on-the-fly verification or sleep set\nmethod or CFFD-preserving stubborn set method\n",
				 _err);
			    (void) fflush(_err);
			    ClPtSymm();
			}
		    if (PtSleep() && PtCFFD())
			{
			    (void)
				fputs
				("\nCFFD-preserving stubborn set method prevents from\nusing sleep set method\n",
				 _err);
			    (void) fflush(_err);
			    ClPtSleep();
			}
		    if (PtCFFD() && PtLazy() && TrRnd() && !PtDelAlg())
			{
			    (void)
				fputs
				("\n-z is ignored because of CFFD-preserving incremental algorithm\n",
				 _err);
			    (void) fflush(_err);
			    ClPtLazy();
			    ResetTrRnd();
			}
		    if (PtSleep())
			{
			    SetDfs();
			    SetFreeExpansion();
			}
		    _CompInst();
		    str = FileName(fileName, "fld");
		    bFile = CreateWrite(str);
		    ClDeallocate(str);
		    if (NotBinaryFile(bFile))
			Exit(5);
		    PtOutput(bFile);
		    OwnClose(&bFile);
		    PlcFill();
		    _CompPrec();
		    if (Pep())
			PtPepOutput(fileName);
		    if (Ara())
			{
			    LTSBegin(fileName, FALSE);
			    PtAraOutput();
			}
#ifndef DEBUG
		    if (Ara() || Pep())
			DelAraNames((!Pep()) && ((!PtSymm()) || !PtVerbose()));
		    if (Pep() && ((!PtSymm()) || !PtVerbose()))
			DelPlaNames();
#endif
		    TrnFill();
		    if ((!NTrVisible()) && !Buchi())
			ClInfinite();
		    if (PtSleep() && (Divergence() || Infinite()))
			{
			    (void)
				fputs
				("\n-f replaces -S because of loop-searching\n",
				 _err);
			    (void) fflush(_err);
			    ClPtSleep();
			}
		    if (PtSymm())
			CompSymm(TRUE);
		}
	    else
		{
		    if (Pep())
			{
			    (void)
				fputs
				("\n-P is ignored because the net is not unfolded\n",
				 _err);
			    (void) fflush(_err);
			    ResetPep();
			}
		    theState = InitialState(bFile);
		    if (Ara())
			AraNames(bFile);
		    OwnClose(&bFile);
		    if (Ara())
			LTSBegin(fileName, TRUE);
		}
	    if (rs == NULL)
		seedVal = 1L;
	    createFlag = 0L;
	    if (unfold)
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtSymm())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PlcIMkSymm())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtStub())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtDelAlg())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtCFFD())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtDefVis())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (PtSleep())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (bdds)
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (Dfs())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (FreeExpansion())
		createFlag |= 1L;
	    createFlag <<= 1;
	    if (Infinite() != 0L)
		createFlag |= 1L;
	    createFlag <<= 1;
	    theGraph = GphCreate(fileName, bits, EngBufSize(), (UNS32) seedVal,
				 createFlag);
	    if (theGraph == NULL)
		Exit(3);
	    if (Buchi() && PtStub())
		{
		    createFlag = 0L;
		    if (unfold)
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtSymm())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PlcIMkSymm())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtStub())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtDelAlg())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtCFFD())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtDefVis())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (PtSleep())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (bdds)
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (Dfs())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (FreeExpansion())
			createFlag |= 1L;
		    createFlag <<= 1;
		    if (Infinite() != 0L)
			createFlag |= 1L;
		    createFlag <<= 1;
		    createFlag |= 1L;
		    the2Graph = GphCreate(fileName, bits, EngBufSize(),
					  (UNS32) seedVal, createFlag);
		    if (the2Graph == NULL)
			Exit(3);
		}
	    if (ms != NULL)
		{
		    GphMapToMem(theGraph, mapLen);
		    if (Buchi() && PtStub())
			GphMapToMem(the2Graph, mapLen);
		}
	    if (unfold)
		{
		    u = PlcEncode(TblPlc(), EngABuffer(), EngABufSize());
		    thePath = PthCreate(GphPutNode(theGraph, EngBuffer(), u,
						   (unsigned long *) NULL));
		    if (Buchi() && PtStub())
			{
			    plc = PtFlyPlc();
			    PlcSCount(plc, (unsigned long) 0L);
			    u = PlcEncode(TblPlc(), EngABuffer(),
					  EngABufSize());
			    (void) GphPutNode(the2Graph, EngBuffer(), u,
					      (unsigned long *) NULL);
			    PlcSCount(plc, (unsigned long) 1L);
			    PtSFlyPlc(plc);
			}
		    if (PtSleep())
			GphPush(theGraph, PthEnd(thePath), (BALTREE *) NULL);
		    else if (OnTheFly() && PtFlyPlc())
			{
			    nowPath = thePath;
			    if (PlcReject(PtFlyPlc()))
				PtC1((unsigned long) -1L, PthEnd(thePath));
			    else if (PlcDeadlock(PtFlyPlc()))
				PtC2((unsigned long) -1L, PthEnd(thePath));
			}
		}
	    else
		{
		    u = StaEncode(theState, EngABuffer(), EngABufSize());
		    thePath = PthCreate(GphPutNode(theGraph, EngBuffer(), u,
						   (unsigned long *) NULL));
		    if (OnTheFly())
			{
			    theNdAddr = PthEnd(thePath);
			    theFolState = theState;
			    theAwAddr = (unsigned long) -1L;
			    if (Reject())
				{
				    nowPath = thePath;
				    IllTrace();
				}
			}
		    for (u = 0; u < MagGNofPlaces(); u++)
			MarDiscard(StaGMarking(theState, u));
		}
	}
    if (Elimination() &&
	((!unfold) || (NTrVisible() == PtrBlGCount(TblTrn())) ||
	 Ara() || bdds || PlRnd() || PtLazy() ||
	 (((BOOL) Buchi()) ? (!PtStub()) : (BOOL) Infinite())))
	{
	    (void)
		fprintf
		(_err, "\n-q is ignored because %s\n",
		 (!unfold) ? "the net is not unfolded"
		 : ((NTrVisible() == PtrBlGCount(TblTrn()))
		    ? "all transition instances are visible"
		    : (Ara() ? "of -A"
		       : (bdds ? "of -B"
			  : (PlRnd() ? "of -x"
			     : (PtLazy() ? "of -z"
				: (Buchi() ?
				   "a formula is verified without the stubborn set method"
				   :
				   "of infinite path monitor states")))))));
	    (void) fflush(_err);
	    ClElimination();
	}
    if (PtStub())
	PtSRandom(GphSeed(theGraph));
    if (blk)
        {
            if (!Blocked())
                {
                    (void) fputs("\n-Y is ignored\n", _err);
                    (void) fflush(_err);
                }
            else if (PtTryAgain())
                {
                    (void) fputs("\n-Z is ignored because of -Y\n", _err);
                    (void) fflush(_err);
                    ClPtTryAgain();
                }
        }
    a = 0L;
    theInner = 0;
    theOuter = 0;
    if ((PtDelAlg() && OnTheFly()) || Divergence() || PtCFFD())
	theNodeTree = BalTrCreate(&lngBalTools);
    if (Ara() && !unfold)
	SetAraTree();
    turn = 0;
    if (Infinite())
	{
	    the2Path = PthCreate(PthEnd(thePath));
	    the3Path = PthCreate((unsigned long) -1L);
	    if (!Divergence())
		turn = 2;
	    if (!Buchi())
		{
		    (void)
			fputs
			("\nWARNING: algorithms for user-defined infinite path monitor states\nare unreliable!\n",
			 _err);
		    (void) fflush(_err);
		}
	}
    do
	{    
	    if ((a != 0L) && stopFlag)
		{
		    if (PtStub())
			{
			    GphSetSeed(theGraph, PtTerm());
			    if (Buchi())
				{
				    GphSetSeed(the2Graph, PtTerm());
				    GphFlush(the2Graph);
				}
			}
		    GphFlush(theGraph);
		    Suspend();
		    continueGeneration = TRUE;
		    stopFlag = FALSE;
		    SetSigHandlers();
		}
	    gphGCount = GphGenCount(theGraph);
	    if ((!Infinite()) ||
		((turn != 2) && ((turn == 0) ||
				 (PthEnd(the2Path) == (unsigned long) -1L))))
		{
		    PthSEnd(thePath, GphGet(theGraph, EngABuffer(),
					    EngABufSize(),
					    PtASlpSet(), PtA2SlpSet()));
		    if (PthEnd(thePath) == (unsigned long) -1L)
			turn = 2;
		    else
			{
			    if (OnTheFly())
				{
				    if (unfold)
					{
					    PlcDecode(EngBuffer());
					    PtOnTheFly();
					}
				    else
					{
					    StaDecode(EngBuffer(), theState);
					    DoOnTheFly();
					}
				}
			    else if (unfold)
				{
				    PlcDecode(EngBuffer());
				    if (PtCFFD())
					PtProCFFD();
				    else
					PtFollowers();
				}
			    else
				{
				    theAwCount = 0L;
				    theAwFirst = (unsigned long) -1L;
				    StaDecode(EngBuffer(), theState);
				    nowPath = thePath;
				    _Followers();
				    if (theAwCount != 0L)
					GphUpdate(theGraph, PthEnd(thePath),
						  theAwCount, theAwFirst,
						  (BALTREE *) NULL);
				    else
					GphAddTermNode(theGraph,
						       PthEnd(thePath));
				    GphIncGenCount(theGraph);
				    if (theDots)
					DotOutput();
				    for (u = 0; u < MagGNofPlaces(); u++)
					MarDiscard(StaGMarking(theState, u));
				}
			}
		}
	    else if (PthEnd(the2Path) != (unsigned long) -1L)
		{
		    if (unfold)
			{
			    if (Buchi())
				PtBuchi();
			    else
				PtInfinite();
			}
		    else
			{
			    if (Buchi())
				DoBuchi();
			    else
				DoInfinite();
			}
		}
	    if (gphGCount < GphGenCount(theGraph))
		a += GphGenCount(theGraph) - gphGCount;
	} while ((a < gCount) && continueGeneration &&
		 ((turn != 2) || (Infinite() &&
				  (PthEnd(the2Path) != (unsigned long) -1L))));
    if (Ara())
	LTSEnd();
    IgnSigHandlers();
    if ((!theDots) || (theOuter != 0) || (a <= 9L))
	{
	    (void) fputc((int) '\n', _out);
	    (void) fflush(_out);
	}
    if (PtStub())
	{
	    GphSetSeed(theGraph, PtTerm());
	    if (Buchi())
		GphSetSeed(the2Graph, PtTerm());
	}
    GphClose(theGraph);
    if (Buchi() && PtStub())
	GphClose(the2Graph);
    Exit(0);
    return(0);
}
