/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  mbufs.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko, Tomi Janhunen and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "expr.h"
#include "formula.h"
#include "mbufs.h"
#include "macptree.h"
#include "comline.h"
#include "probe.h"
#include "error.h"

#define MB_EOF       ((int) EOF)
#define MB_ASTERISK  ((int) '*')
#define MB_BACKSLASH ((int) '\\')
#define MB_EOL       ((int) '\n')
#define MB_SLASH     ((int) '/')
#define MB_SPACE     ((int) ' ')

#define WAIT_NONE   0
#define WAIT_AST   1
#define WAIT_SLASH 2

FE1(MBUF *, MbufCreate,
    unsigned, size)
{
    MBUF *mbuf;

    mbuf = (MBUF *) Allocate(sizeof(MBUF));
    mbuf->_buf = (char *) Allocate(size);
    mbuf->_dupBuf = mbuf->_buf;
    mbuf->_written = mbuf->_buf;
    mbuf->_read = mbuf->_buf;
    mbuf->_max = mbuf->_buf + size - 1;
    return(mbuf);
}

FE1(void, MbufDracsid,
    MBUF *, mbuf)
{
    if (mbuf)
	{
	    mbuf->_buf = (char *) NULL;
	    mbuf->_written = (char *) NULL;
	    mbuf->_read = (char *) NULL;
	    mbuf->_max = (char *) NULL;
	    ClDeallocate(mbuf->_dupBuf);
	    Deallocate((char *) mbuf);
	}
}

FE2(void, MbufPut,
    MBUF *, mbuf,
    int, ch)
{
    if ((ch == (int) '\r') || (ch == /* Ctrl-Z */ (int) '\032'))
	return;
    if (mbuf->_written < mbuf->_max)
	*(mbuf->_written++) = (char) ch;
    else 
	{
	    if (errorFlag == NO_ERROR)
		{
		    (void) fputs("Line buffer overflow\n", _err);
		    (void) fflush(_err);
		    errorFlag = BUF_ERROR;
		}
	}
}

FE2(void, MbufStr,
    MBUF *, mbuf,
    char *, str)
{
    while (*str)
	MbufPut(mbuf, *(str++));
}

#ifdef USE_READLINE
UCHAR *rdlnBffr = NULL;
UCHAR *rdlnPntr = NULL;
BOOL rdlnEof = FALSE;
BOOL rdlnHist = FALSE;

FE0(void, MbufRlInit)
{
    register int onrre;

    if (!rdlnHist)
	{
	    rdlnHist = TRUE;
	    ClearErrno();
	    using_history();
	    onrre = ErrnoVal();
	    if (onrre)
		{
		    (void) fputs("\nusing_history():\n", _err);
		    OnrreOutput(_err, onrre);
		}
	}
}

FE0(void, MbufRlClose)
{
    register int onrre;

    ClDeallocate(rdlnBffr);
    if (rdlnHist)
	{
	    ClearErrno();
	    clear_history();
	    onrre = ErrnoVal();
	    if (onrre)
		{
		    (void) fputs("\nclear_history():\n", _err);
		    OnrreOutput(_err, onrre);
		}
	}
}

FE1(int, MbufFgetc,
    char *, prompt)
{
    register int onrre;

    if (!rdlnBffr)
	{
	    ClearErrno();
	    rdlnPntr = rdlnBffr = (UCHAR *) readline(prompt);
	    onrre = ErrnoVal();
	    if (onrre)
		{
		    if ((((int) ENOENT) == onrre) ||
			(((int) EINTR) == onrre))
			ClearErrno();
		    else
			{
			    (void) fputs("\nreadline(", _err);
			    if (prompt)
				{
				    (void) fputc((int) '"', _err);
				    (void) fputs(prompt, _err);
				    (void) fputc((int) '"', _err);
				}
			    else
				(void) fputs("NULL", _err);
			    (void) fputs(") returned ", _err);
			    if (rdlnBffr)
				{
				    (void) fputc((int) '"', _err);
				    (void) fputs((char *) rdlnBffr, _err);
				    (void) fputc((int) '"', _err);
				}
			    else
				(void) fputs("NULL", _err);
			    (void) fputc((int) '\n', _err);
			    (void) fflush(_err);
			    OnrreOutput(_err, onrre);
			}
		}
	    if (!rdlnBffr)
		{
		    rdlnEof = TRUE;
		    return(MB_EOF);
		}
	    if (*rdlnBffr)
		{
		    add_history((char *) rdlnBffr);
		    onrre = ErrnoVal();
		    if (onrre)
			{
			    (void) fputs("\nadd_history(\"", _err);
			    (void) fputs((char *) rdlnBffr, _err);
			    (void) fputs("\"):\n", _err);
			    OnrreOutput(_err, onrre);
			}
		}
	}
    if (*rdlnPntr)
	return((int) (*rdlnPntr++));
    ClDeallocate(rdlnBffr);
    return(MB_EOL);
}

FE4(void, MbufReadln,
    MBUF *, mbuf,
    FILE *, file,
    char *, prompt,
    BOOL, use_readline)
#else
FE2(void, MbufReadln,
    MBUF *, mbuf,
    FILE *, file)
#endif
{
    int ch, cmtDepth, cmtScan;
    BOOL skipEol, endOfLine;

    cmtDepth = 0;
    cmtScan = WAIT_NONE;
    skipEol = FALSE;
    endOfLine = FALSE;
    do 
	{
#ifdef USE_READLINE
	    if (use_readline)
		{
		    ch = MbufFgetc(prompt);
		    prompt = (char *) NULL;
		}
	    else
#endif
		ch = fgetc(file);
	    if (echoInput)
		{
		    if (ch == MB_EOF)
			(void) fputc(MB_EOL, _out);
		    else
			(void) fputc(ch, _out);
		}
	    if (skipEol)
		{
		    if (ch == MB_EOL)
			{
			    if (MbufSize(mbuf) != 0)
				MbufPut(mbuf, MB_SPACE);
			}
		    else
			{
			    MbufPut(mbuf, MB_BACKSLASH);
			    skipEol = FALSE;
			}
		}
	    if (skipEol)
		skipEol = FALSE;
	    else
		{
		    switch(ch)
			{
			case MB_EOF:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    break;
				case WAIT_AST:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_SLASH);
				    cmtScan = WAIT_NONE;
				    break;
				case WAIT_SLASH:
				    cmtScan = WAIT_NONE;
				    break;
				}
			    if (MbufSize(mbuf) != 0)
				MbufPut(mbuf, MB_EOL);
			    endOfLine = TRUE;
			    if (cmtDepth != 0)
				{
				    (void)
					fputs
					("Incomplete comment at the end of input\n",
					 _err);
				    (void) fflush(_err);
				}
			    break;
			case MB_EOL:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    break;
				case WAIT_AST:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_SLASH);
				    cmtScan = WAIT_NONE;
				    break;
				case WAIT_SLASH:
				    cmtScan = WAIT_NONE;
				    break;
				}
			    if (cmtDepth == 0)
				{
				    if (MbufSize(mbuf) != 0)
					MbufPut(mbuf, MB_EOL);
				    endOfLine = TRUE;
				}
			    break;
			case MB_BACKSLASH:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    break;
				case WAIT_AST:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_SLASH);
				    cmtScan = WAIT_NONE;
				    break;
				case WAIT_SLASH:
				    cmtScan = WAIT_NONE;
				    break;
				}
			    if (cmtDepth == 0)
				skipEol = TRUE;
			    break;
			case MB_SLASH:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    cmtScan = WAIT_AST;
				    break;
				case WAIT_AST:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_SLASH);
				    break;
				case WAIT_SLASH:
				    cmtDepth--;
				    if ((cmtDepth == 0) &&
					(MbufSize(mbuf) != 0))
					MbufPut(mbuf, MB_SPACE);
				    cmtScan = WAIT_NONE;
				    break;
				}
			    break;
			case MB_ASTERISK:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_ASTERISK);
				    else
					cmtScan = WAIT_SLASH;
				    break;
				case WAIT_AST:
				    cmtDepth++;
				    cmtScan = WAIT_NONE;
				    break;
				case WAIT_SLASH:
				    break;
				}
			    break;
			default:
			    switch(cmtScan)
				{
				case WAIT_NONE:
				    break;
				case WAIT_AST:
				    if (cmtDepth == 0)
					MbufPut(mbuf, MB_SLASH);
				    cmtScan = WAIT_NONE;
				    break;
				case WAIT_SLASH:
				    cmtScan = WAIT_NONE;
				    break;
				}
			    if (cmtDepth == 0)
				MbufPut(mbuf, ch);
			    break;
			}
		}
	} while (!endOfLine);
    if (echoInput)
	(void) fflush(_out);
}
