/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ppio.c
*
* Part of: PROD's program prpp
*
* Authors: Johannes Helander, Heikki Suonsivu, and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains the actual reading of input from the
*     net description file.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ppio.h"
#include "util.h"
#include "prpp.h"
#include "token.h"
#include "output.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"

LIST *include_path = NULL;
PPIOFILE *curFile;

typedef int buftype;		

static buftype *buffer = NULL;
static int bufsize;
static int bufindex;		
static LIST *file_stack = NULL;
static BALTREE *current_names = NULL;

/***************************************************************************
*
* Function:  ppio_init
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, ppio_init)
{
    bufsize = 1024 * sizeof(buftype);
    buffer = (buftype *) Allocate((unsigned) bufsize); 
    bufindex = -1;		
    file_stack = LstCreate((TOOLS *) NULL);
    curFile = NULL;
    include_path = LstCreate(&strTools);
    current_names = BalTrCreate(&strBalTools);
}

/***************************************************************************
*
* Function:  ppio_close
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, ppio_close)
{
    PPIOFILE *ppiofile;

    ClDeallocate(buffer);
    if (((LIST *) NULL) != file_stack)
	{
	    LSTFOR (file_stack, ppiofile, PPIOFILE)
		{
		    ClFclose(ppiofile->fp);
		    Deallocate((char *) (ppiofile->name));
		    Deallocate((char *) ppiofile);
		}
	    LstDestroy(file_stack, FALSE);
	}
    LstDestroy(include_path, TRUE);
    BalTrDestroy(current_names, FALSE);
}

/***************************************************************************
*
* Function:  ppio_push_file
* Purpose:
* Arguments: file_name - 
*            angle_brackets - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, ppio_push_file,
    char **, file_name,
    BOOL, angle_brackets)
{
    unsigned len;
    char *directory;
    char *name = NULL;
    FILE *fp = NULL;
    PPIOFILE *ppiofile;
    BOOL rel;

    if (BalTrElement(current_names, (char *) *file_name))
	{
	    (void) sprintf(errStr, "Cyclic inclusion: %s", *file_name);
	    error(errStr);
	    LSTFOR (file_stack, ppiofile, PPIOFILE)
		{
		    if (ppiofile != (PPIOFILE *) LstFirst(file_stack))
			(void) fputs("which ", _err);
		    (void) fprintf(_err, "is included in %s\n",
				   ppiofile->name);
		    if (!strcmp(ppiofile->name, *file_name))
			break;
		}
	    (void) fflush(_err);
	    ClDeallocate(*file_name);
	    Exit(1);
	}
    directory = (char *) LstBegin(include_path);
    rel = (BOOL) (((char *) NULL) != directory);
    if (angle_brackets && !rel)
	{
	    (void)
		sprintf(errStr,
			"Cannot open file %s for reading\nEmpty include_path",
			*file_name);
	    ClDeallocate(*file_name);
	    panic(errStr);
	}
    if (!angle_brackets)
	{
	    name = StrCopy(*file_name);
	    fp = MyFopen(name, "r", &rel);
	    ClDeallocate(name);
	    if ((!fp) && !rel)
		{
		    ClDeallocate(*file_name);
		    *errStr = '\0';
		    panic(errStr);
		}
	}
    while (directory && !fp)
	{
	    len = (unsigned) strlen(directory);
	    if ((len != 0) && (directory[len - 1] == DIRCHAR))
		{
		    name = (char *) Allocate(len + (unsigned)
					     (strlen(*file_name) + 1));
		    (void) sprintf(name, "%s%s", directory, *file_name);
		}
	    else
		{
		    name = (char *) Allocate(len + (unsigned)
					     (strlen(*file_name) + 2));
		    (void) sprintf(name, "%s%c%s", 
				   directory, DIRCHAR, *file_name);
		}
	    directory = (char *) LstStep(include_path);
	    rel = (BOOL) (((char *) NULL) != directory);
	    fp = MyFopen(name, "r", &rel);
	    ClDeallocate(name);
	    if ((!fp) && !rel)
		{
		    ClDeallocate(*file_name);
		    *errStr = '\0';
		    panic(errStr);
		}
	}
    curFile = (PPIOFILE *) Allocate(sizeof(PPIOFILE));
    curFile->fp = fp;
    curFile->name = StrCopy(*file_name);
    curFile->line = 0L;
    curFile->newlines_pushed = 0L;
    ClDeallocate(*file_name);
    (void) LstInsFirst(file_stack, (char *) curFile);
    (void) BalTrInsert(current_names, (char *) (curFile->name),
		       (char *) (curFile->name), TRUE);
}

/***************************************************************************
*
* Function:  ppio_pop_file
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, ppio_pop_file)
{
    if (!curFile)
	panic("ppio_pop_file(): Internal error: No file open!!");
    ClFclose(curFile->fp);
    (void) LstDelFirst(file_stack, FALSE);
    (void) BalTrDelete(current_names, (char *) NULL,
		       (char *) (curFile->name), TRUE);
    Deallocate((char *) curFile);
    curFile = (PPIOFILE *) LstFirst(file_stack);
}

static BOOL newlineProblem = FALSE;

/***************************************************************************
*
* Function:  ppio_getc
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(int, ppio_getc)
{
    int c = 0;
  
    if (!curFile)
	panic("ppio_getc(): Internal error: No file open!!");
    if (bufindex < 0)		
	{
	    for (;;)
		{
		    if ((c = fgetc(curFile->fp)) == (int) EOF)
			{
			    if (curFile !=
				(PPIOFILE *) LstLast(file_stack))
				{
				    if (newlineProblem)
					{
					    error("Unexpected end of file");
					    warning("Upper level input continues the incomplete line\n(spaceless concatenation)");
					    newlineProblem = FALSE;
					}
				    ppio_pop_file();
				    insert_line_tag();
				}
			    else
				{
				    newlineProblem = FALSE;
				    break;
				}
			}
		    else if ((c != (int) '\r') &&
			     (c != /* Ctrl-Z */ (int) '\032'))
			break;
		}
	    if (c != (int) EOF) 
		{
		    if (c == (int) '\n')
			{
			    newlineProblem = FALSE;
			    curFile->line++;
			    curFile->newlines_pushed = 0L;
			}
		    else
			newlineProblem = TRUE;
		}
	}
    else
	{
	    c = buffer[bufindex];
	    bufindex--;		
	    if ((c != (int) EOF) && (c == (int) '\n'))
		if (curFile->newlines_pushed)
		    curFile->newlines_pushed--;
	}
    return(c);
}

/***************************************************************************
*
* Function:  c_ppio_unget
* Purpose:
* Arguments: c - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, c_ppio_unget,
    int, c)
{
    bufindex++;
    if (((int) (bufindex * sizeof(buftype))) >= bufsize)
	{
	    bufsize += 4096 * sizeof(buftype); 
	    buffer = (buftype *) Reallocate((char *) buffer,
					    (unsigned)bufsize);
	}
    buffer[bufindex] = c;		
    if (c == '\n')
	curFile->newlines_pushed++;
}

/***************************************************************************
*
* Function:  s_ppio_unget
* Purpose:
* Arguments: s - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, s_ppio_unget,
    char *, s)
{
    char *c;

    for(c = s; *c; c++)		
	;
    for(c--; c >= s; c--)		
	c_ppio_unget((int) *c);
}

/***************************************************************************
*
* Function:  d_ppio_unget
* Purpose:
* Arguments: value - 
*            strict - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, d_ppio_unget,
    DOMAIN, value,
    BOOL, strict)
{
    UDOMAIN uval;

    uval = (UDOMAIN) value;
    (void) sprintf(numStr,
		   (strict && (uval > (UDOMAIN) ~uval))
		   ? DOMXFORMAT : UDOMFORMAT, uval);
    s_ppio_unget(numStr);
}

/***************************************************************************
*
* Function:  u_ppio_unget
* Purpose:
* Arguments: value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, u_ppio_unget,
    unsigned, value)
{
    (void) sprintf(numStr, "%u", value);
    s_ppio_unget(numStr);
}
