/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  pplst.c
*
* Part of: PROD's program prpp
*
* Authors: Johannes Helander and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module transforms net information into forms understood
*     by the later programs.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "output.h"
#include "place.h"
#include "prpp.h"
#include "lexer.h"
#include "macro.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "net.h"
#include "prec.h"
#include "prpexpr.h"
#include "pplst.h"

LIST *PlaceLst = NULL, *PrecLst = NULL, *LshLst = NULL, *RshLst = NULL;
LIST *TransLst = NULL, *EnumLst = NULL, *PriLst = NULL, *BlkLngLst = NULL;
BALTREE *enumTree = NULL, *priTree = NULL;
BINARYFILE datFile = NonBinaryFile();
char *monName = NULL;
PLACE *monPlace = NULL;
BOOL verify = FALSE;
char *extBuchi = NULL;

/***************************************************************************
*
* Function:  pplst_init
* Purpose:   This routine initializes the global variables in which the
*            net information is kept.
* Arguments:
* Returns:
* See also:  pplst_close    
*
****************************************************************************
*/
FE0(void, pplst_init)
{
    DynamInit();
    PlaceLst = LstCreate(&plaTools);
    PrecLst = LstCreate(&preTools);
    LshLst = LstCreate(&shfTools);
    RshLst = LstCreate(&shfTools);
    TransLst = LstCreate(&traTools);
    EnumLst = LstCreate(&strTools);
    PriLst = LstCreate(&strTools);
    BlkLngLst = LstCreate(&lngTools);
    enumTree = BalTrCreate(&refBalTools);
    priTree = BalTrCreate(&refBalTools);
    monitors = BalTrCreate(&monBalTools);
}

/***************************************************************************
*
* Function:  pplst_close
* Purpose:   This routine destroys structures initialized by pplst_init.
* Arguments:
* Returns:
* See also:  pplst_init
*
****************************************************************************
*/
FE0(void, pplst_close)
{
    PLACE *pl;

    if (((char *) NULL) != monName)
        {
	    if ((((PLACE *) NULL) == monPlace) ||
		(monName != PlaGName(monPlace)))
	        {
		    LSTFOR (PlaceLst, pl, PLACE)
		        {
			    if (PlaGName(pl) == monName)
			        break;
			}
		    if (((PLACE *) NULL) == pl)
		        Deallocate(monName);
		}
	    monName = (char *) NULL;
	}
    if (((PLACE *) NULL) != monPlace)
        {
	     LSTFOR (PlaceLst, pl, PLACE)
	         {
		     if (pl == monPlace)
		         break;
		 }
	     if (((PLACE *) NULL) == pl)
	         PlaDiscard(monPlace);
	     monPlace = (PLACE *) NULL;
	}
    LstDestroy(PlaceLst, TRUE);
    LstDestroy(PrecLst, TRUE);
    LstDestroy(LshLst, TRUE);
    LstDestroy(RshLst, TRUE);
    LstDestroy(TransLst, TRUE);
    LstDestroy(EnumLst, TRUE);
    LstDestroy(PriLst, TRUE);
    LstDestroy(BlkLngLst, TRUE);
    BalTrDestroy(enumTree, TRUE);
    BalTrDestroy(priTree, TRUE);
    BalTrDestroy(monitors, TRUE);
    DynamClose();
}

/***************************************************************************
*
* Function:  pplst_gen
* Purpose:   This routine transforms the net information into a form
*            understood by the C compiler.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, pplst_gen)
{
    NET *net;
    TRANSITION *tran;
    ARC *arc, *arc2;
    SYMMULTUP *symmultup = NULL, *syt2 = NULL;
    VARIABLE *var;
    unsigned i, j;
    VOID *cursor;
    char *str;
    LIST *tmp_list;
#ifdef TRAIN
    PLACE *place, *pla2, *iplace, *cplace;
    char *str2;
    unsigned negCount, phase;
    MACRO *macrop;
    unsigned long a = 0L;
    unsigned long ival;
    MULTUP *imultup, *cmultup;
#endif

    if (monPlace)
        for (cursor = NULL,
	     tran = (TRANSITION *) LstBegin(TransLst); tran;
	     cursor = LstGet(TransLst),
	     tran = (TRANSITION *) LstStep(TransLst))
	    {
		LSTFOR (TraGArcs(tran), arc, ARC)
		    {
			if (ArcGPlace(arc) != monPlace)
			    continue;
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    if (SytGArity(symmultup) != 1)
				break;
			if (symmultup)
			    break;
		    }
		if (arc)
		    {
			(void) sprintf(errStr,
				       "Transition %s: every tuple in the tester \"%s\" must be unary",
				       TraGName(tran),
				       PlaGName(monPlace));
			error(errStr);
			LstSet(TransLst, cursor);
			(void) LstDelNext(TransLst, TRUE);
			continue;
		    }
		LSTFOR (TraGInputs(tran), arc, ARC)
		    if (ArcGPlace(arc) == monPlace)
			{
			    symmultup = (SYMMULTUP *)
				LstFirst(ArcGSymsum(arc));
			    break;
			}
		LSTFOR (TraGOutputs(tran), arc2, ARC)
		    if (ArcGPlace(arc2) == monPlace)
			{
			    syt2 = (SYMMULTUP *)
				LstFirst(ArcGSymsum(arc2));
			    break;
			}
		if (arc || arc2)
		    {
			if ((!arc) || (!arc2) ||
			    (LstSize(ArcGSymsum(arc)) !=
			     LstSize(ArcGSymsum(arc2))))
			    {
				(void) sprintf(errStr,
					       "Transition %s: attempt to change",
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "the multiplicity of the tester \"%s\"",
					       PlaGName(monPlace));
				error(errStr);
				LstSet(TransLst, cursor);
				(void) LstDelNext(TransLst, TRUE);
				continue;
			    }
			if ((LstSize(ArcGSymsum(arc)) != 1)
			    || ((!SytGCffSym(symmultup)) &&
				(SytGMulti(symmultup)
				 != 1L)))
			    {
				(void) sprintf(errStr,
					       "Transition %s: strange input",
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "from the tester \"%s\"",
					       PlaGName(monPlace));
				error(errStr);
				LstSet(TransLst, cursor);
				(void) LstDelNext(TransLst, TRUE);
				continue;
			    }
			if ((!SytGCffSym(symmultup)) 
			    && (!SytGCffSym(syt2))
			    && (SytGMulti(symmultup) !=
				SytGMulti(syt2)))
			    {
				(void) sprintf(errStr,
					       "Transition %s: attempt to change",
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "the multiplicity of the tester \"%s\"",
					       PlaGName(monPlace));
				error(errStr);
				LstSet(TransLst, cursor);
				(void) LstDelNext(TransLst, TRUE);
				continue;
			    }
			if ((SytGCffSym(symmultup) 
			     && !SytGCffSym(syt2)) ||
			    ((!SytGCffSym(symmultup)) 
			     && SytGCffSym(syt2)))
			    {
				(void) sprintf(errStr,
					       "Transition %s: potential attempt to change",
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "the multiplicity of the tester \"%s\"",
					       PlaGName(monPlace));
				error(errStr);
				LstSet(TransLst, cursor);
				(void) LstDelNext(TransLst, TRUE);
				continue;
			    }
		    }
	    }
#ifdef TRAIN
    if (train)
	{
	    LSTFOR (PlaceLst, place, PLACE)
		{
		    if (place == monPlace)
			continue;
		    for (negCount = 0, str = PlaGName(place); *str; str++)
			{
			    if ((*str == '_') && (*(str + 1) == 'n'))
				{
				    if (negCount)
					{
					    if (negCount == 1)
						{
						    (void) sprintf(errStr,
								   "Place %s: more than one _n",
								   PlaGName(place));
						    error(errStr);
						}
					    *str = '$';
					}
				    negCount++;
				}
			}
		    cursor = LstGet(PlaceLst);
		    iplace = NULL;
		    cplace = NULL;
		    LSTFOR (PlaceLst, pla2, PLACE)
			{
			    if (pla2 == place)
				break;
			    if ((pla2 == monPlace) ||
				PlaGComplement(pla2))
				continue;
			    str = PlaGName(place);
			    str2 = PlaGName(pla2);
			    for (;;)
				{
				    if (negCount)
					{
					    if ((*str == '_') &&
						(*(str + 1) == 'n'))
						str += 2;
					}
				    else
					{
					    if ((*str2 == '_') &&
						(*(str2 + 1) == 'n'))
						str2 += 2;
					}
				    if ((!*str) || (!*str2) || (*str != *str2))
					break;
				    str++;
				    str2++;
				}
			    if ((!*str) && (!*str2))
				{
				    if (negCount)
					{
					    iplace = pla2;
					    cplace = place;
					}
				    else
					{
					    iplace = place;
					    cplace = pla2;
					}
				    break;
				}
			}
		    LstSet(PlaceLst, cursor);
		    if (!iplace)
			continue;
		    PlaSComplement(iplace, cplace);
		    PlaSComplement(cplace, iplace);
		    ival = ~((unsigned long) 0L);
		    str = (char *)
			Allocate((unsigned) strlen(PlaGName(iplace)) + 2);
		    *str = 'k';
		    *(str + 1) = '\0';
		    (void) strcat(str, PlaGName(iplace));
		    for (phase = 0; phase < 2; phase++)
			{
			    if (1 == phase)
				*str = 'i';
			    macrop = ismacro(str);
			    if (!macrop)
				{
				    (void) sprintf(errStr,
						   "Macro %s: undefined",
						   str);
				    error(errStr);
				}
			    else
				{
				    str2 = simple_expansion(macrop);
				    if (!str2)
					{
					    (void) sprintf(errStr,
							   "Macro %s: unacceptable definition",
							   str);
					    error(errStr);
					}
				    else
					{
					    if ((sscanf(str2, "%lu", &a)
						 != 1) ||
						(a > (unsigned long) ~a) ||
						((1 == phase) ?
						 (PlaGCapacity(iplace) < a) :
						 (0L == a)))
						{
						    (void) sprintf(errStr,
								   "Macro %s: unacceptable definition",
								   str);
						    error(errStr);
						}
					    else if (1 == phase)
						ival = a;
					    else
						PlaSCapacity(iplace, a);
					    Deallocate(str2);
					}
				}
			}
		    Deallocate(str);
		    PlaSCapacity(cplace, PlaGCapacity(iplace));
		    imultup = MarBegin(PlaGMarking(iplace));
		    cmultup = MarBegin(PlaGMarking(cplace));
		    if (imultup && (MarStep(PlaGMarking(iplace)) ||
				    MultupGArity(imultup)))
			{
			    (void) sprintf(errStr,
					   "Place %s: unacceptable mk",
					   PlaGName(iplace));
			    error(errStr);
			}
		    else if ((ival != ~((unsigned long) 0L)) &&
			     ((!imultup) ? (0L != ival) :
			      (MultupGCount(imultup) != ival)))
			{
			    (void) sprintf(errStr,
					   "Place %s initially contradicts i%s",
					   PlaGName(iplace),
					   PlaGName(iplace));
			    error(errStr);
			}
		    if (cmultup && (MarStep(PlaGMarking(cplace)) ||
				    MultupGArity(cmultup)))
			{
			    (void) sprintf(errStr,
					   "Place %s: unacceptable mk",
					   PlaGName(cplace));
			    error(errStr);
			}
		    else if ((ival != ~((unsigned long) 0L)) &&
			     ((!cmultup) ? (ival != PlaGCapacity(iplace)) :
			      (MultupGCount(cmultup) + ival
			       != PlaGCapacity(iplace))))
			{
			    (void) sprintf(errStr,
					   "Place %s initially contradicts k%s - i%s",
					   PlaGName(cplace),
					   PlaGName(iplace),
					   PlaGName(iplace));
			    error(errStr);
			}
		}
	    LSTFOR (PlaceLst, place, PLACE)
		{
		    if (place == monPlace)
			continue;
		    if (!PlaGComplement(place))
			{
			    (void) sprintf(errStr,
					   "Place %s: no specified complement place",
					   PlaGName(place));
			    error(errStr);
			}
		}
	    for (cursor = NULL,
		 tran = (TRANSITION *) LstBegin(TransLst); tran;
		 cursor = LstGet(TransLst),
		 tran = (TRANSITION *) LstStep(TransLst))
		{
		    LSTFOR (TraGArcs(tran), arc, ARC)
			{
			    if (!PlaGComplement(ArcGPlace(arc)))
				continue;
			    symmultup = (SYMMULTUP *)
				LstBegin(ArcGSymsum(arc));
			    if ((!symmultup) ||	LstStep(ArcGSymsum(arc)) ||
				SytGArity(symmultup) ||
				SytGCffSym(symmultup))
				break;
			}
		    if (arc)
			{
			    (void) sprintf(errStr,
					   "Transition %s: place %s misused",
					   TraGName(tran),
					   PlaGName(ArcGPlace(arc)));
			    error(errStr);
			    LstSet(TransLst, cursor);
			    (void) LstDelNext(TransLst, TRUE);
			    continue;
			}
		    LSTFOR (TraGInputs(tran), arc, ARC)
			{
			    if (!PlaGComplement(ArcGPlace(arc)))
				continue;
			    symmultup = (SYMMULTUP *)
				LstBegin(ArcGSymsum(arc));
			    PlaIDown(ArcGPlace(arc), SytGMulti(symmultup));
			}
		    LSTFOR (TraGOutputs(tran), arc, ARC)
			{
			    if (!PlaGComplement(ArcGPlace(arc)))
				continue;
			    symmultup = (SYMMULTUP *)
				LstBegin(ArcGSymsum(arc));
			    PlaIUp(ArcGPlace(arc), SytGMulti(symmultup));
			}
		    LSTFOR (TraGArcs(tran), arc, ARC)
			{
			    if (!PlaGComplement(ArcGPlace(arc)))
				continue;
			    if (PlaGDown(ArcGPlace(arc)) +
				PlaGDown(PlaGComplement(ArcGPlace(arc))) !=
				PlaGUp(ArcGPlace(arc)) +
				PlaGUp(PlaGComplement(ArcGPlace(arc))))
				{
				    (void) sprintf(errStr,
						   "Transition %s: attempt to change the joint multiplicity of\nplaces %s and %s",
						   TraGName(tran),
						   PlaGName(ArcGPlace(arc)),
						   PlaGName(PlaGComplement(ArcGPlace(arc))));
				    break;
				}
			    if (PlaGDown(ArcGPlace(arc)) +
				PlaGDown(PlaGComplement(ArcGPlace(arc))) >
				PlaGCapacity(ArcGPlace(arc)))
				{
				    (void) sprintf(errStr,
						   "Transition %s: joint multiplicity of\n%s and %s was to be %lu",
						   TraGName(tran),
						   PlaGName(ArcGPlace(arc)),
						   PlaGName(PlaGComplement(ArcGPlace(arc))),
						   PlaGCapacity(ArcGPlace(arc)));
				    break;
				}
			}
		    LSTFOR (TraGInputs(tran), arc2, ARC)
			PlaCDown(ArcGPlace(arc2));
		    LSTFOR (TraGOutputs(tran), arc2, ARC)
			PlaCUp(ArcGPlace(arc2));
		    if (arc)
			{
			    error(errStr);
			    LstSet(TransLst, cursor);
			    (void) LstDelNext(TransLst, TRUE);
			    continue;
			}
		}
	}
#endif
    shf_expansion();
    LstDestroy(LshLst, TRUE);
    LshLst = (LIST *) NULL;
    LstDestroy(RshLst, TRUE);
    RshLst = (LIST *) NULL;
    pre_loops();
    tmp_list = verify ? BuTra() : (LIST *) NULL;
    net = NetCons(PlaceLst, TransLst, PrecLst, LstSize(PriLst) + 1,
		  monPlace, tmp_list);
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    LSTNUM (TraGVariables(tran), var, VARIABLE, i)
		{
		    (void) sprintf(errStr, "%s%s%s_v[%u]",
				   VarGName(var) ? "/*" : "",
				   VarGName(var) ? VarGName(var) : "",
				   VarGName(var) ? "*/" : "", i);
		    VarSLabel(var, StrCopy(errStr));
		}
	    if (TraGComputation(tran))
		for (cursor = NULL,
		     str = (char *) LstBegin(TraGComputation(tran)); str;
		     cursor = LstGet(TraGComputation(tran)),
		     str = (char *) LstStep(TraGComputation(tran)))
		    if ((var = TrVarLookup(str, tran)))
			{
			    LstSet(TraGComputation(tran), cursor);
			    (void) LstDelNext(TraGComputation(tran), TRUE);
			    (void) LstInsNext(TraGComputation(tran), (char *)
					      StrCopy(VarGLabel(var)));
			}
	    LSTNUM (TraGInputs(tran), arc, ARC, i)
		{
		    ArcSNumber(arc, i);
		    LSTNUM (ArcGSymsum(arc), symmultup, SYMMULTUP, j)
			{
			    SytSNumber(symmultup, j);
			    if (LstFirst(ArcGSymsum(arc)) ==
				LstLast(ArcGSymsum(arc)))
				(void) sprintf(errStr,
					       "/*mtuple from %s*/",
					       PlaGName(ArcGPlace(arc)));
			    else
				(void) sprintf(errStr,
					       "/*%u%s mtuple from %s*/",
					       j + 1, ((j % 10) == 0) ? "st"
					       : (((j % 10) == 1)
						  ? "nd" : "th"),
					       PlaGName(ArcGPlace(arc)));
			    SytSLabel(symmultup, StrCopy(errStr));
			}
		}
	    LSTNUM (TraGOutputs(tran), arc, ARC, i)
		{
		    ArcSNumber(arc, i);
		    LSTNUM (ArcGSymsum(arc), symmultup, SYMMULTUP, j)
			{
			    SytSNumber(symmultup, j);
			    if (LstFirst(ArcGSymsum(arc)) ==
				LstLast(ArcGSymsum(arc)))
				(void) sprintf(errStr,
					       "/*mtuple to %s*/",
					       PlaGName(ArcGPlace(arc)));
			    else
				(void) sprintf(errStr,
					       "/*%u%s mtuple to %s*/",
					       j + 1, ((j % 10) == 0) ? "st"
					       : (((j % 10) == 1)
						  ? "nd" : "th"),
					       PlaGName(ArcGPlace(arc)));
			    SytSLabel(symmultup, StrCopy(errStr));
			}
		}
	}
    Generate(net);
    NetDiscard(net);
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    LSTFOR (TraGVariables(tran), var, VARIABLE)
		{
		    Deallocate((char *) VarGLabel(var));
		    VarSLabel(var, (char *) NULL);
		}
	    if (TraGComputation(tran))
		TraDComputation(tran);
	    LSTFOR (TraGArcs(tran), arc, ARC)
		LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
		    {
			Deallocate((char *) SytGLabel(symmultup));
			SytSLabel(symmultup, (char *) NULL);
		    }
	}
}

static unsigned monLen;
static unsigned long monPos;
static DATATYPE *monBuffer;

/***************************************************************************
*
* Function:  MonOut
* Purpose:   This routine writes to a binary file, datFile, information
*            about a monitor state of the tester.
* Arguments: mon - the monitor state
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, MonOut,
    MON *, mon)
{
    register unsigned i, len;
    register UDOMAIN value;
    register DATATYPE *pointer;
    char *format;

    if (PlaGLow(monPlace) && LstFirst(PlaGLow(monPlace)) &&
	(TupGArity((TUPLE *) LstFirst(PlaGLow(monPlace))) == 1) &&
	(MonGState(mon) < TupGField((TUPLE *) LstFirst(PlaGLow(monPlace)), 0)))
	{
	    (void) sprintf(errStr, "tester %%s: %s is below low limit %s",
			   DOMFORMAT, DOMFORMAT);
	    format = StrCopy(errStr);
	    (void) sprintf(errStr, format, PlaGName(monPlace), MonGState(mon),
			   TupGField((TUPLE *)
				     LstFirst(PlaGLow(monPlace)), 0));
	    Deallocate((char *) format);
	    error(errStr);
	    return(FALSE);
	}
    if (PlaGHigh(monPlace) && LstFirst(PlaGHigh(monPlace)) &&
	(TupGArity((TUPLE *) LstFirst(PlaGHigh(monPlace))) == 1) &&
	(MonGState(mon) > TupGField((TUPLE *)
				    LstFirst(PlaGHigh(monPlace)), 0)))
	{
	    (void) sprintf(errStr, "tester %%s: %s is above high limit %s",
			   DOMFORMAT, DOMFORMAT);
	    format = StrCopy(errStr);
	    (void) sprintf(errStr, format, PlaGName(monPlace), MonGState(mon),
			   TupGField((TUPLE *)
				     LstFirst(PlaGHigh(monPlace)), 0));
	    Deallocate((char *) format);
	    error(errStr);
	    return(FALSE);
	}
    len = 0;
    value = MonGState(mon);
    value = Encode((DOMAIN) value);
    IncLen(value, len);
    i = MonGFlag(mon);
    IncLen(i, len);
    monLen = len;
    WriteUns(&monLen, 1, datFile);
    monPos += (unsigned long) sizeof(unsigned);
    monBuffer = (DATATYPE *) AllocArray(monLen, sizeof(DATATYPE));
    pointer = monBuffer;
    value = MonGState(mon);
    value = Encode((DOMAIN) value);
    PutValue(value, pointer);
    i = MonGFlag(mon);
    PutValue(i, pointer);
    WriteData(monBuffer, monLen, datFile);
    monPos += (unsigned long) (monLen * sizeof(DATATYPE));
    Deallocate((char *) monBuffer);
    return(FALSE);
}

/***************************************************************************
*
* Function:  pplst_dat
* Purpose:   This routine writes to a binary file, datFile, information
*            about the net.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, pplst_dat)
{
    PLACE *place;
    TRANSITION *tran;
    VARIABLE *var;
    char *str;
    unsigned p, sLen;
    unsigned long pos;
    MULTUP *multup;
    DATATYPE *buffer;
    register DATATYPE *pointer;
    register unsigned i, len;
    register unsigned long count;
    register UDOMAIN value;

    pos = 0L;
    WriteLng(&pos, 1, datFile);
    pos += (unsigned long) sizeof(unsigned long);
    if (monPlace)
	p = LstWhich(PlaceLst, (char *) monPlace) - 1;
    else
	p = (unsigned) -1;
    WriteUns(&p, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    p = LstSize(PriLst) + 1;
    WriteUns(&p, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    len = 0;
    i = LstSize(PlaceLst);
    IncLen(i, len);
    LSTFOR (PlaceLst, place, PLACE)
	{
	    i = PlaGArity(place);
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buffer;
    i = LstSize(PlaceLst);
    PutValue(i, pointer);
    LSTFOR (PlaceLst, place, PLACE)
	{
	    i = PlaGArity(place);
	    PutValue(i, pointer);
	}
    WriteData(buffer, sLen, datFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    Deallocate((char *) buffer);
    len = 0;
    i = LstSize(TransLst);
    IncLen(i, len);
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    i = LstSize(TraGVariables(tran));
	    if (verify)
		{
		    if (TraGKind(tran) == TK_INITIAL)
			i--;
		    else
			i -= 2;
		}
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buffer;
    i = LstSize(TransLst);
    PutValue(i, pointer);
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    i = LstSize(TraGVariables(tran));
	    if (verify)
		{
		    if (TraGKind(tran) == TK_INITIAL)
			i--;
		    else
			i -= 2;
		}
	    PutValue(i, pointer);
	}
    WriteData(buffer, sLen, datFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    Deallocate((char *) buffer);
    len = 0;
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    i = TraGFdWidth(tran);
	    IncLen(i, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buffer;
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    i = TraGFdWidth(tran);
	    PutValue(i, pointer);
	}
    WriteData(buffer, sLen, datFile);
    pos += (unsigned long) (sLen * sizeof(DATATYPE));
    Deallocate((char *) buffer);
    LSTFOR (TransLst, tran, TRANSITION)
        {
            if (TraGRange(tran) && !LstEmpty(TraGRange(tran)))
                LSTFOR (TraGRange(tran), buffer, DATATYPE)
                    {
                        sLen = DataLen(buffer);
                        WriteUns(&sLen, 1, datFile);
                        pos += (unsigned long) sizeof(unsigned);
                        WriteData(buffer, sLen, datFile);
                        pos += (unsigned long) (sLen * sizeof(DATATYPE));
                    }
            sLen = 0;
            WriteUns(&sLen, 1, datFile);
            pos += (unsigned long) sizeof(unsigned);
        }
    LSTNUM (PlaceLst, place, PLACE, p)
	{
	    if (MarEmpty(PlaGMarking(place)))
		continue;
	    MARFOR (PlaGMarking(place), multup)
		{
		    len = 0;
		    i = p;
		    IncLen(i, len);
		    if (MultupGArity(multup) != PlaGArity(place))
			{
			    len++;
			    i = MultupGArity(multup);
			    IncLen(i, len);
			}
		    if (MultupGArity(multup))
			{
			    for (i = 0; i < MultupGArity(multup); i++)
				{
				    value = MultupGField(multup, i);
				    value = Encode((DOMAIN) value);
				    IncLen(value, len);
				}
			    if (MultupGCount(multup) != 1L)
				{
				    len++;
				    count = MultupGCount(multup) - 2L;
				    IncLen(count, len);
				}
			}
		    else
			{
			    count = MultupGCount(multup) - 1L;
			    IncLen(count, len);
			}
		    sLen = len;
		    WriteUns(&sLen, 1, datFile);
		    pos += (unsigned long) sizeof(unsigned);
		    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
		    pointer = buffer;
		    i = p;
		    PutValue(i, pointer);
		    if (MultupGArity(multup) != PlaGArity(place))
			{
			    AriCode(pointer);
			    i = MultupGArity(multup);
			    PutValue(i, pointer);
			}
		    if (MultupGArity(multup))
			{
			    for (i = 0; i < MultupGArity(multup); i++)
				{
				    value = MultupGField(multup, i);
				    value = Encode((DOMAIN) value);
				    PutValue(value, pointer);
				}
			    if (MultupGCount(multup) != 1L)
				{
				    CntCode(pointer);
				    count = MultupGCount(multup) - 2L;
				    PutValue(count, pointer);
				}
			}
		    else
			{
			    count = MultupGCount(multup) - 1L;
			    PutValue(count, pointer);
			}
		    WriteData(buffer, sLen, datFile);
		    pos += (unsigned long) (sLen * sizeof(DATATYPE));
		    Deallocate((char *) buffer);
		}
	}
    sLen = (unsigned) -1;
    WriteUns(&sLen, 1, datFile);
    pos += (unsigned long) sizeof(unsigned);
    if (monPlace)
	{
	    monPos = pos;
	    (void) BalTrApply(monitors, MonOut);
	    pos = monPos;
	    sLen = (unsigned) -1;
	    WriteUns(&sLen, 1, datFile);
	    pos += (unsigned long) sizeof(unsigned);
	    if (verify)
		OutBuchi(&pos);
	    else
		{
		    sLen = 0;
		    WriteUns(&sLen, 1, datFile);
		    pos += (unsigned long) sizeof(unsigned);
		}
	}
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    sLen = (unsigned) strlen(TraGName(tran));
	    WriteUns(&sLen, 1, datFile);
	    WriteChr(TraGName(tran), sLen, datFile);
	    LSTFOR (TraGVariables(tran), var, VARIABLE)
		{
		    if (!VarGName(var))
			continue;
		    sLen = (unsigned) strlen(VarGName(var));
		    WriteUns(&sLen, 1, datFile);
		    WriteChr(VarGName(var), sLen, datFile);
		}
	}
    p = LstSize(EnumLst);
    WriteUns(&p, 1, datFile);
    LSTFOR (EnumLst, str, char)
	{
	    sLen = (unsigned) strlen(str);
	    WriteUns(&sLen, 1, datFile);
	    WriteChr(str, sLen, datFile);
	}
    LSTFOR (PriLst, str, char)
	{
	    sLen = (unsigned) strlen(str);
	    WriteUns(&sLen, 1, datFile);
	    WriteChr(str, sLen, datFile);
	}
    LSTFOR (PlaceLst, place, PLACE)
	{
	    sLen = (unsigned) strlen(PlaGName(place));
	    WriteUns(&sLen, 1, datFile);
	    WriteChr(PlaGName(place), sLen, datFile);
	}
    SeekAbs(datFile, (unsigned long) 0L);
    WriteLng(&pos, 1, datFile);
}

/***************************************************************************
*
* Function:  TrVarLookup
* Purpose:
* Arguments: name - 
*            tran - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(VARIABLE *, TrVarLookup,
    char *, name,
    TRANSITION *, tran)
{
    VARIABLE *var;

    LSTFOR(TraGVariables(tran), var, VARIABLE)
	if(VarGName(var) && !strcmp(VarGName(var), name))
	    return(var);
    return(NULL);
}
