/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  probe.c
*
* Part of: PROD's program probe
*
* Authors: Lasse Anderson, Keijo Heljanko, Tomi Janhunen, Robert Jurgens,
*          Ismo Kangas, Marko Rauhamaa, Kimmo Valkealahti,
*          and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The "probe" program answers to questions about the reachability graph
*     of a net.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "formula.h"
#include "mbufs.h"
#include "macptree.h"
#include "comline.h"
#include "probe.h"
#include "probesig.h"
#include "cmdnm.h"
#include "bitarr.h"
#include "stackpro.h"
#include "temporal.h"
#include "prepctl.h"
#include "checkctl.h"

BOOL echoInput;
char *prefDefault = NULL;
char *sourceName = NULL;
char *probHelp = NULL;
BOOL unfold;
BOOL symmMethod;
PTRBL *placeFoldings = NULL;
PTRBL *transFoldings = NULL;
BOOL sgsInMemory;
BOOL forIsOpen = FALSE;
BALTREE *specSets = NULL, *clbleSets = NULL, *freeSets = NULL;
BALTREE *trivMace = NULL;
static BOOL dots;
static COMLINE *shComLine = NULL;
static PATH *shPath = NULL;
static unsigned long shCounter;
PTRBL *comps = NULL;
static LIST *largvList = NULL, *nameList = NULL, *fileList = NULL;
static BALTREE *termComps = NULL;
static BALTREE *tmPaths = NULL;
static unsigned long untscc;
unsigned long compCount = 0L;

/***************************************************************************
*
* Function:  PtsCons
* Purpose:
* Arguments: number -
*            description -
*            paths -
* Returns:
* See also:
*
****************************************************************************
*/
FE3(PTHSET *, PtsCons,
    unsigned long, number,
    char *, description,
    BALTREE *, paths)
{
    PTHSET *pthSet;

    pthSet = PtsAlloc();
    pthSet->number = number;
    pthSet->description = description;
    pthSet->paths = paths;
    return(pthSet);
}

/***************************************************************************
*
* Function:  PtsCopy
* Purpose:
* Arguments: pthSet -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(PTHSET *, PtsCopy,
    PTHSET *, pthSet)
{
    PTHSET *copy;

    copy = PtsAlloc();
    copy->number = pthSet->number;
    copy->description = StrCopy(pthSet->description);
    copy->paths = BalTrCopy(pthSet->paths, TRUE);
    return(copy);
}

/***************************************************************************
*
* Function:  PtsDracsid
* Purpose:
* Arguments: pthSet -
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, PtsDracsid,
    PTHSET *, pthSet)
{
    if (pthSet)
	{
	    ClDeallocate(pthSet->description);
	    BalTrDestroy(pthSet->paths, TRUE);
	    ClDeallocate(pthSet);
	}
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  PtsDump
* Purpose:
* Arguments: pthSet -
*            depth -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PtsDump,
    PTHSET *, pthSet,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "PTHSET number %lu\n", pthSet->number);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "description %s\n", pthSet->description);
    DumpIndent(depth + 1);
    (void) fputs("paths\n", _err);
    BalTrDump(pthSet->paths, depth + 2);
    DumpIndent(depth);
    (void) fputs("End of PTHSET\n", _err);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  PtsKeyCmp
* Purpose:
* Arguments: pthSet -
*            key -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(int, PtsKeyCmp,
    PTHSET *, pthSet,
    unsigned long *, key)
{
    return((pthSet->number == *key) ? 0 : ((pthSet->number < *key) ? -1 : 1));
}

/***************************************************************************
*
* Function:  PtsPrefix
* Purpose:
* Arguments: pthSet -
*            prefix -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(BOOL, PtsPrefix,
    PTHSET *, pthSet,
    unsigned long *, prefix)
{
    return(pthSet->number == *prefix);
}

BALTOOLS ptsBalTools =
{ (CFUNC) PtsCopy, PtsDracsid
#ifdef DEBUG
, PtsDump
#endif
, PtsKeyCmp, PtsPrefix };

static unsigned long newSetNumber = 0L;

/***************************************************************************
*
* Function:  AddSet
* Purpose:
* Arguments: paths -
*            command -
*            sets -
* Returns:
* See also:
*
****************************************************************************
*/
FS3(unsigned long, AddSet,
    BALTREE *, paths,
    char *, command,
    BALTREE *, sets)
{
    unsigned long number, *p;

    if ((p = (unsigned long *) BalTrFirst(freeSets)))
	{
	    number = *p;
	    (void) BalTrDelete(freeSets, (char *) NULL, (char *) p, TRUE);
	}
    else
	number = newSetNumber++;
    (void) BalTrInsert(sets, (char *) PtsCons(number, command, paths),
		       (char *) (&number), TRUE);
    return(number);
}

/***************************************************************************
*
* Function:  ClearSet
* Purpose:
* Arguments: number -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, ClearSet,
    unsigned long, number)
{
    if (BalTrDelete(clbleSets, (char *) NULL, (char *) (&number), TRUE))
	{
	    if (number + 1L == newSetNumber)
		newSetNumber--;
	    else
		(void) BalTrInsert(freeSets, (char *) LngCons(number),
				   (char *) (&number), TRUE);
	}
    else
	{
	    if (BalTrElement(specSets, (char *) (&number)))
		{
		    (void) fputs("Special set cannot be cleared\n", _err);
		    (void) fflush(_err);
		}
	    else
		IsntSet(_err, number);
	    errorFlag = SEM_ERROR;
	}
}

/***************************************************************************
*
* Function:  ClearAll
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, ClearAll)
{
    if (BalTrGElCount(clbleSets) != 0L)
	{
	    PTHSET *pthSet;

	    BalTrDestroy(clbleSets, TRUE);
	    BalTrDestroy(freeSets, TRUE);
	    clbleSets = BalTrCreate(&ptsBalTools);
	    freeSets = BalTrCreate(&lngBalTools);
	    if ((pthSet = (PTHSET *) BalTrLast(specSets)))
		{
		    unsigned long i;

		    i = PtsGNumber(pthSet);
		    if ((newSetNumber = i + 1L)	!= BalTrGElCount(specSets))
			{
			    while (i--)
				{
				    if (!BalTrElement(specSets,
						      (char *) (&i)))
					(void) BalTrInsert(freeSets, (char *)
							   LngCons(i), (char *)
							   (&i), TRUE);
				}
			}
		}
	    else
		newSetNumber = 0L;
	}
}

static BALTREE *shrTree = NULL;

/***************************************************************************
*
* Function:  ShrPath
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShrPath,
    PATH *, path)
{
    if (!(shPath && (path->end == shPath->end)))
	{
	    shPath = PthCreate(path->end);
	    (void) BalTrInsert(shrTree, (char *) shPath,
			       (char *) shPath, TRUE);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  ShrinkSet
* Purpose:
* Arguments: number -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, ShrinkSet,
    unsigned long, number)
{
    PTHSET *pthSet;

    if ((pthSet = (PTHSET *) BalTrElement(clbleSets, (char *) (&number))))
	{
	    shrTree = BalTrCreate(&pthBalTools);
	    shPath = (PATH *) NULL;
	    (void) BalTrApply(PtsGPaths(pthSet), ShrPath);
	    BalTrDestroy(PtsGPaths(pthSet), TRUE);
	    PtsSPaths(pthSet, shrTree);
	}
    else
	{
	    if (BalTrElement(specSets, (char *) (&number)))
		{
		    (void) fputs("Special set cannot be shrinked\n", _err);
		    (void) fflush(_err);
		}
	    else
		IsntSet(_err, number);
	    errorFlag = SEM_ERROR;
	}
}

/***************************************************************************
*
* Function:  ShowNamSg
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowNamSg,
    unsigned long *, lng)
{
    NamSgOutput(*lng, _out);
    return(!continueEvaluation);
}

static unsigned long shSgNumber;

/***************************************************************************
*
* Function:  ShSgPred
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShSgPred,
    unsigned long *, lng)
{
    unsigned long sgpr, ordaw, ordchi, awAddr = 0L, pred = 0L;

    ordchi = GphOrdinal(theGraph, *lng);
    GphPrLast(theGraph, *lng, &awAddr, &pred);
    while (continueEvaluation && (awAddr != (unsigned long) -1L))
	{
	    if ((sgpr =
		 GphCompNumber(theGraph, pred)) != shSgNumber)
		{
		    shCounter++;
		    ordaw = GphAwOrdinal(theGraph, awAddr);
		    if (shComLine->options & TG_SUPERVERBOSE)
			{
			    WldVerbose(pred, _out);
			    ArwVerbose(ordaw, awAddr, _out);
			    WldVerbose(*lng, _out);
			    CutVerbose(_out);
			}
		    else
			AwSgOutput(GphOrdinal(theGraph, pred),
				   sgpr, ordaw, ordchi,
				   shSgNumber, _out);
		}
	    GphPrPrec(theGraph, &awAddr, &pred);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  ShSgSucc
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShSgSucc,
    unsigned long *, lng)
{
    unsigned long sgfl, i, ordfa, awCount, awAddr = 0L, succ = 0L;

    awCount = GphAwCount(theGraph, *lng);
    if ((awCount != 0L) && (awCount != (unsigned long) -1L))
	{
	    ordfa = GphOrdinal(theGraph, *lng);
	    GphFlFirst(theGraph, *lng, &awAddr, &succ);
	    if (FreeExpansion())
		{
		    for (i = awCount; (i != 0L) && continueEvaluation;)
			{
			    i--;
			    if ((sgfl =
				 GphCompNumber(theGraph, succ)) != shSgNumber)
				{
				    shCounter++;
				    if (shComLine->options & TG_SUPERVERBOSE)
					{
					    WldVerbose(*lng, _out);
					    ArwVerbose(i, awAddr, _out);
					    WldVerbose(succ, _out);
					    CutVerbose(_out);
					}
				    else
					AwSgOutput(ordfa, shSgNumber, i,
						   GphOrdinal(theGraph, succ),
						   sgfl, _out);
				}
			    if (i != 0L)
				GphFlNext(theGraph, &awAddr, &succ);
			}
		}
	    else
		{
		    for (i = 0L; (i < awCount) && continueEvaluation; i++)
			{
			    if ((sgfl =
				 GphCompNumber(theGraph, succ)) != shSgNumber)
				{
				    shCounter++;
				    if (shComLine->options & TG_SUPERVERBOSE)
					{
					    WldVerbose(*lng, _out);
					    ArwVerbose(i, awAddr, _out);
					    WldVerbose(succ, _out);
					    CutVerbose(_out);
					}
				    else
					AwSgOutput(ordfa, shSgNumber, i,
						   GphOrdinal(theGraph, succ),
						   sgfl, _out);
				}
			    if (i + 1L < awCount)
				GphFlNext(theGraph, &awAddr, &succ);
			}
		}
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  SgPrFl
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, SgPrFl,
    unsigned long *, lng)
{
    return(*lng == shSgNumber);
}

/***************************************************************************
*
* Function:  SgPred
* Purpose:
* Arguments: number -
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, SgPred,
    BALTREE *, component,
    unsigned long, number)
{
    BALTREE *followers, *sg;
    unsigned long prCount, ordfa;

    prCount = 0L;
    shSgNumber = number;
    for (ordfa = 0L; (ordfa < compCount) && continueEvaluation; ordfa++)
	{
	    followers = GphSgSucc(theGraph, ordfa);
	    if (BalTrApply(followers, SgPrFl))
		{
		    if (prCount++ == 0L)
			{
			    (void) fputs("The predecessor components of ",
					 _out);
			    SgcDenote(_out, number);
			    (void) fputs(":\n", _out);
			}
		    NamSgOutput(ordfa, _out);
		}
	    BalTrDestroy(followers, TRUE);
	}
    NamSgOutput((unsigned long) -1L, _out);
    if (continueEvaluation)
	{
	    SgcDenote(_out, number);
	    (void) fprintf(_out, " has %lu predecessor components\n",
			   prCount);
	    (void) fflush(_out);
	}
    if ((prCount != 0L) && !(shComLine->options & TG_MUTE) &&
	(shComLine->options & TG_VERBOSE) && continueEvaluation)
	{
	    (void) fputs("Arrows going to ", _out);
	    SgcDenote(_out, number);
	    (void) fputs(" from other components:\n", _out);
	    (void) fflush(_out);
	    if (shComLine->options & TG_SUPERVERBOSE)
		CutVerbose(_out);
	    if (component)
		sg = component;
	    else if (sgsInMemory)
		sg = (BALTREE *) PtrBlGet(comps, number);
	    else
		sg = GphSgGet(theGraph, number);
	    shCounter = 0L;
	    (void) BalTrApply(sg, ShSgPred);
	    if (!(component || sgsInMemory))
		BalTrDestroy(sg, TRUE);
	    if (!(shComLine->options & TG_SUPERVERBOSE))
		AwSgOutput((unsigned long) -1L, (unsigned long) -1L,
			   (unsigned long) -1L, (unsigned long) -1L,
			   (unsigned long) -1L, _out);
	    if (continueEvaluation)
		{
		    (void) fprintf(_out, "%lu arrows go to ",
				   shCounter);
		    SgcDenote(_out, number);
		    (void) fputs(" from other components\n", _out);
		    (void) fflush(_out);
		}
	    shCounter = number;
	}
    if (continueEvaluation && !(shComLine->options & TG_MUTE))
	CutVerbose(_out);
}

/***************************************************************************
*
* Function:  SgSucc
* Purpose:
* Arguments: component -
*            number -
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, SgSucc,
    BALTREE *, component,
    unsigned long, number)
{
    BALTREE *followers, *sg;
    unsigned long flCount;

    if (shComLine->options & TG_MUTE)
	flCount = GphSgFlCount(theGraph, number);
    else
	{
	    followers = GphSgSucc(theGraph, number);
	    if ((flCount = BalTrGElCount(followers)) != 0L)
		{
		    (void) fputs("The successor components of ",
				 _out);
		    SgcDenote(_out, number);
		    (void) fputs(":\n", _out);
		    (void) BalTrApply(followers, ShowNamSg);
		    NamSgOutput((unsigned long) -1L, _out);
		}
	    BalTrDestroy(followers, TRUE);
	}
    if (continueEvaluation)
	{
	    SgcDenote(_out, number);
	    (void) fprintf(_out, " has %lu successor components\n",
			   flCount);
	    (void) fflush(_out);
	}
    if ((flCount != 0L) && (!(shComLine->options & TG_MUTE)) &&
	(shComLine->options & TG_VERBOSE) && continueEvaluation)
	{
	    (void) fputs("Arrows going from ", _out);
	    SgcDenote(_out, number);
	    (void) fputs(" to other components:\n", _out);
	    (void) fflush(_out);
	    if (shComLine->options & TG_SUPERVERBOSE)
		CutVerbose(_out);
	    if (component)
		sg = component;
	    else if (sgsInMemory)
		sg = (BALTREE *) PtrBlGet(comps, number);
	    else
		sg = GphSgGet(theGraph, number);
	    shSgNumber = number;
	    shCounter = 0L;
	    (void) BalTrApply(sg, ShSgSucc);
	    if (!(component || sgsInMemory))
		BalTrDestroy(sg, TRUE);
	    if (!(shComLine->options & TG_SUPERVERBOSE))
		AwSgOutput((unsigned long) -1L, (unsigned long) -1L,
			   (unsigned long) -1L, (unsigned long) -1L,
			   (unsigned long) -1L, _out);
	    if (continueEvaluation)
		{
		    (void) fprintf(_out, "%lu arrows go from ",
				   shCounter);
		    SgcDenote(_out, number);
		    (void) fputs(" to other components\n", _out);
		    (void) fflush(_out);
		}
	    shCounter = number;
	}
    if (continueEvaluation && !(shComLine->options & TG_MUTE))
	CutVerbose(_out);
}

/***************************************************************************
*
* Function:  NdPred
* Purpose:
* Arguments: world -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, NdPred,
    unsigned long, world)
{
    unsigned long awAddr = 0L, pred = 0L, found;

    GphPrLast(theGraph, world, &awAddr, &pred);
    for (found = 0L;
	 continueEvaluation && (awAddr != (unsigned long) -1L);
	 found++)
	{
	    if (!(shComLine->options & TG_MUTE))
		{
		    if (shComLine->options & TG_VERBOSE)
			WldVerbose(pred, _out);
		    else
			{
			    (void) fprintf(_out, "from node %lu via\n",
					   GphOrdinal(theGraph, pred));
			    (void) fflush(_out);
			}
		    ArwVerbose(GphAwOrdinal(theGraph, awAddr), awAddr, _out);
		    CutVerbose(_out);
		}
	    GphPrPrec(theGraph, &awAddr, &pred);
	}
    if (continueEvaluation)
	{
	    (void) fprintf(_out, "Node %lu has %lu predecessor arrows\n",
			   GphOrdinal(theGraph, world), found);
	    (void) fflush(_out);
	    if (!(shComLine->options & TG_MUTE))
		CutVerbose(_out);
	}
}

/***************************************************************************
*
* Function:  NdSucc
* Purpose:
* Arguments: world -
*            transexpr -
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, NdSucc,
    unsigned long, world,
    EXPRESSION *, transexpr)
{
    unsigned long trans, i, awCount, awAddr = 0L, succ = 0L;

    awCount = GphAwCount(theGraph, world);
    if (awCount == (unsigned long) -1L)
	DoesntKnow(_out, theGraph, world);
    else if (transexpr == NULL)
	{
	    if (!(shComLine->options & TG_MUTE) && (awCount != 0L))
		{
		    GphFlFirst(theGraph, world, &awAddr, &succ);
		    if (FreeExpansion())
			{
			    for (i = awCount; (i != 0L) && continueEvaluation;)
				{
				    i--;
				    ArwVerbose(i, awAddr, _out);
				    if (shComLine->options & TG_VERBOSE)
					WldVerbose(succ, _out);
				    else
					{
					    (void)
						fprintf
						(_out, "to node %lu\n",
						 GphOrdinal(theGraph,
							    succ));
					    (void) fflush(_out);
					}
				    CutVerbose(_out);
				    if (i != 0L)
					GphFlNext(theGraph, &awAddr, &succ);
				}
			}
		    else
			{
			    for (i = 0L; (i < awCount) && continueEvaluation;
				 i++)
				{
				    ArwVerbose(i, awAddr, _out);
				    if (shComLine->options & TG_VERBOSE)
					WldVerbose(succ, _out);
				    else
					{
					    (void)
						fprintf
						(_out,
						 "to node %lu\n",
						 GphOrdinal(theGraph,
							    succ));
					    (void) fflush(_out);
					}
				    CutVerbose(_out);
				    if (i + 1L < awCount)
					GphFlNext(theGraph, &awAddr, &succ);
				}
			}
		}
	    if (continueEvaluation)
		{
		    (void) fprintf(_out,
				   "Node %lu has %lu successor arrows\n",
				  GphOrdinal(theGraph, world), awCount);
		    (void) fflush(_out);
		}
	}
    else
	{
	    trans = (unsigned long) XprEval(transexpr);
	    if ((trans < awCount) && (errorFlag == NO_ERROR) &&
		continueEvaluation)
		{
		    if (!(shComLine->options & TG_MUTE))
			{
			    GphFlFirst(theGraph, world, &awAddr, &succ);
			    if (FreeExpansion())
				{
				    for (i = awCount - 1L;
					 (i != trans) && continueEvaluation;
					 i--)
					{
					    if (i != 0L)
						GphFlNext(theGraph,
							  &awAddr, &succ);
					}
				}
			    else
				{
				    for (i = 0L;
					 (i != trans) && continueEvaluation;
					 i++)
					{
					    if (i + 1L < awCount)
						GphFlNext(theGraph,
							  &awAddr, &succ);
					}
				}
			    if (continueEvaluation)
				{
				    ArwVerbose(i, awAddr, _out);
				    if (shComLine->options & TG_VERBOSE)
					WldVerbose(succ, _out);
				    else
					{
					    (void)
						fprintf
						(_out,
						 "to node %lu\n",
						 GphOrdinal(theGraph, succ));
					    (void) fflush(_out);
					}
				}
			}
		}
	    else
		{
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    IsntArrow(_err, theGraph, world, trans);
			    errorFlag = SEM_ERROR;
			}
		}
	}
    if (continueEvaluation && (errorFlag == NO_ERROR) &&
	!(shComLine->options & TG_MUTE))
	CutVerbose(_out);
}

/***************************************************************************
*
* Function:  ShowNode
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowNode,
    unsigned long *, lng)
{
    if (!(shComLine->options & TG_MUTE))
	{
	    if (shComLine->options & TG_SUPERVERBOSE)
		{
		    WldVerbose(*lng, _out);
		    CutVerbose(_out);
		}
	    else
		NdOutput(GphOrdinal(theGraph, *lng), _out);
	}
    return((!continueEvaluation) || (shComLine->options & TG_MUTE));
}

/***************************************************************************
*
* Function:  ShowEnd
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowEnd,
    PATH *, path)
{
    if (!(shPath && (path->end == shPath->end)))
	{
	    shPath = path;
	    shCounter++;
	    if (shComLine->options & TG_MUTE)
		NodeBelongs(_out, theGraph, path->end);
	    else
		{
		    WldVerbose(path->end, _out);
		    CutVerbose(_out);
		}
	    if (shComLine->options & TG_PREDS)
		NdPred(path->end);
	    if (shComLine->options & TG_SUCCS)
		NdSucc(path->end, (EXPRESSION *) NULL);
	}
    return(!continueEvaluation);
}

FS2(void, AuxShowSg,
    BALTREE *, sg,
    unsigned long, number)
{
    (void) fputs("Component ", _out);
    SgcDenote(_out, number);
    (void) fputc((int) '\n', _out);
    (void) fflush(_out);
    if (!(shComLine->options & TG_MUTE))
	{
	    CutVerbose(_out);
	    (void) BalTrApply(sg, ShowNode);
	    if (!(shComLine->options & TG_SUPERVERBOSE))
		NdOutput((unsigned long) -1L, _out);
	}
    if (continueEvaluation)
	{
	    SgcDenote(_out, number);
	    (void) fprintf(_out, " has %lu nodes\n", BalTrGElCount(sg));
	    (void) fflush(_out);
	}
    if (continueEvaluation && !(shComLine->options & TG_MUTE))
	CutVerbose(_out);
    if (continueEvaluation && (shComLine->options & TG_PREDS))
	SgPred(sg, number);
    if (continueEvaluation && (shComLine->options & TG_SUCCS) &&
	(shComLine->operator != LI_TERMCOMPS))
	SgSucc(sg, number);
    if (!sgsInMemory)
	BalTrDestroy(sg, TRUE);
}

/***************************************************************************
*
* Function:  ShowSg
* Purpose:
* Arguments: sg -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowSg,
    BALTREE *, sg)
{
    AuxShowSg(sg, shCounter);
    shCounter++;
    return(!continueEvaluation);
}

/***************************************************************************
*
* Function:  ShowLSg
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowLSg,
    unsigned long *, lng)
{
    BALTREE *sg;

    if (sgsInMemory)
	sg = (BALTREE *) PtrBlGet(comps, *lng);
    else
	sg = GphSgGet(theGraph, *lng);
    AuxShowSg(sg, *lng);
    return(!continueEvaluation);
}

/***************************************************************************
*
* Function:  ShowPath
* Purpose:
* Arguments: path -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowPath,
    PATH *, path)
{
    if (!(shComLine->options & TG_MUTE))
	{
	    if (shComLine->options & TG_VERBOSE)
		PthVerbose(path, _out);
	    else PthOutput(path, _out);
	}
    return((!continueEvaluation) || (shComLine->options & TG_MUTE));
}

/***************************************************************************
*
* Function:  ShowSet
* Purpose:
* Arguments: pthSet -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ShowSet,
    PTHSET *, pthSet)
{
    (void) fputs("  ", _out);
    SetDenote(_out, PtsGNumber(pthSet));
    (void) fputs(": ", _out);
    (void) fputs(PtsGDescription(pthSet), _out);
    (void) fputc((int) '\n', _out);
    return(FALSE);
}

/***************************************************************************
*
* Function:  Execute
* Purpose:
* Arguments: comLine -
*            command -
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, Execute,
    COMLINE *, comLine,
    char *, command)
{
    static unsigned long state, world, ndCount, awCount, awAddr;
    static unsigned long succ, trans, i, count;
    static DOMAIN value;
    static char *str = NULL;
    static BALTREE *paths = NULL;
    static PATH *path = NULL;
    static MARKING *marking = NULL;
    static int inner, outer;

    shComLine = comLine;
    switch (comLine->operator)
	{
	case LI_ALLCOMPS:
	    if (((unsigned long) 0L) == compCount)
		{
		    if (errorFlag == NO_ERROR)
			{
			    StrongNotComp(_err);
			    errorFlag = SEM_ERROR;
			}
		}
	    else
		{
		    if (sgsInMemory)
			{
			    shCounter = 0L;
			    (void) PtrBlApply(comps, ShowSg);
			}
		    else
			{
			    for (i = 0L;
				 (i < compCount) && continueEvaluation; i++)
				(void) ShowLSg(&i);
			}
		    if (continueEvaluation)
			{
			    (void)
				fprintf
				(_out,
				 "%lu strongly connected components\n",
				 compCount);
			    (void) fflush(_out);
			}
		    if (continueEvaluation && (errorFlag == NO_ERROR) &&
			!(comLine->options & TG_MUTE))
			CutVerbose(_out);
		}
	    break;
	case LI_BUILD:
	    if (comLine->ul.setOper != NULL)
		{
		    if ((paths = SopEval(comLine->ul.setOper)))
			{
			    if (!(comLine->options & TG_MUTE))
				(void) BalTrApply(paths, ShowPath);
			    if (!(comLine->options & TG_MUTE) &&
				!(comLine->options & TG_VERBOSE))
				PthOutput((PATH *) NULL, _out);
			    (void) fprintf(_out, "%lu paths\n",
					   BalTrGElCount(paths));
			    (void) fflush(_out);
			    if (comLine->options & TG_VOLATILE)
				{
				    if (comLine->ul.setOper->operator !=
					SE_SET)
					BalTrDestroy(paths, TRUE);
				}
			    else
				{
				    (void) fputs("Built set ", _out);
				    SetDenote
					(_out,
					 AddSet(((comLine->ul.setOper->operator
						  == SE_SET)
						 ? BalTrCopy(paths,
							     TRUE)
						 : paths),
						StrCopy(command),
						clbleSets));
				    (void) fputc((int) '\n', _out);
				    (void) fflush(_out);
				}
			}
		}
	    if ((errorFlag == NO_ERROR) && !(comLine->options & TG_MUTE))
		CutVerbose(_out);
	    break;
	case LI_CALC:
	    if (comLine->options & TG_MARKING)
		{
		    marking = MexEval(comLine->ul.markexpr);
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			MarOutput(marking, _out);
		    if (marking != NULL)
			MarDiscard(marking);
		}
	    else
		{
		    value = XprEval(comLine->ul.expression);
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    (void) fprintf(_out, DOMFORMAT, value);
			    (void) fputc((int) '\n', _out);
			    (void) fflush(_out);
			}
		}
	    break;
	case LI_CHECK:
	    if (comLine->ul.formula == NULL)
		break;
	    (void) fputs("Check CTL formula.\n", _err);
	    (void) fflush(_err);
	    if (PrepareCTL(theGraph) == FALSE)
		{
		    (void)
			fputs
			("Preparing for CTL model checking failed.\n",
			 _err);
		    (void) fflush(_err);
		}
	    else
		{
		    /* Preparing went OK, now let's do the model checking. */
		    (void) ModelCheckCTL(comLine->ul.formula,
					 theGraph, (NODE)
					 GphOrdinal(theGraph,
						    PthEnd(thePath)));
		}
	    break;
	case LI_CLEAR:
	    i = (unsigned long) XprEval(comLine->ul.expression);
	    if ((errorFlag == NO_ERROR) && continueEvaluation)
		ClearSet(i);
	    break;
	case LI_CLEARALL:
	    ClearAll();
	    break;
	case LI_GOTO:
	    state = (unsigned long) XprEval(comLine->ul.expression);
	    if ((state < GphNdCount(theGraph)) && (errorFlag == NO_ERROR) &&
		continueEvaluation)
		{
		    PthDiscard(thePath);
		    thePath = PthCreate(GphId(theGraph, state));
		}
	    else
		{
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    IsntNode(_err, state);
			    errorFlag = SEM_ERROR;
			}
		}
	    break;
	case LI_LOOK:
	    state = (unsigned long) XprEval(comLine->ul.expression);
	    if ((state < GphNdCount(theGraph)) && (errorFlag == NO_ERROR) &&
		continueEvaluation)
		{
		    world = GphId(theGraph, state);
		    WldVerbose(world, _out);
		}
	    else
		{
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    IsntNode(_err, state);
			    errorFlag = SEM_ERROR;
			}
		}
	    if ((errorFlag == NO_ERROR) && continueEvaluation)
		CutVerbose(_out);
	    break;
	case LI_NEXT:
	    trans = (unsigned long) XprEval(comLine->ul.expression);
	    if ((errorFlag != NO_ERROR) || !continueEvaluation)
		break;
	    awCount = GphAwCount(theGraph, PthEnd(thePath));
	    if (awCount == (unsigned long) -1L)
		DoesntKnow(_out, theGraph, PthEnd(thePath));
	    else if (trans < awCount)
		{
		    GphFlFirst(theGraph, PthEnd(thePath), &awAddr, &succ);
		    if (FreeExpansion())
			{
			    for (i = awCount - 1L;
				 (i != trans) && continueEvaluation; i--)
				{
				    if (i != 0L)
					GphFlNext(theGraph, &awAddr, &succ);
				}
			}
		    else
			{
			    for (i = 0L; (i != trans) && continueEvaluation;
				 i++)
				{
				    if (i + 1L < awCount)
					GphFlNext(theGraph, &awAddr, &succ);
				}
			}
		    if (continueEvaluation)
			PthInsLast(thePath, awAddr, succ);
		}
	    else
		{
		    if (errorFlag == NO_ERROR)
			{
			    IsntArrow(_err, theGraph,
				      PthEnd(thePath), trans);
			    errorFlag = SEM_ERROR;
			}
		}
	    break;
	case LI_PRED:
	    state = (unsigned long) XprEval(comLine->ul.expression);
	    if ((state < GphNdCount(theGraph)) && (errorFlag == NO_ERROR) &&
		continueEvaluation)
		NdPred(GphId(theGraph, state));
	    else
		{
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    IsntNode(_err, state);
			    errorFlag = SEM_ERROR;
			}
		}
	    break;
	case LI_PREV:
	    if (PthLast(thePath))
		PthDelLast(thePath);
	    else
		{
		    if (errorFlag == NO_ERROR)
			{
			    (void) fputs("No predecessor\n", _err);
			    (void) fflush(_err);
			    errorFlag = SEM_ERROR;
			}
		}
	    break;
	case LI_QUERY:
	    if (comLine->ul.formula == NULL)
		break;
	    paths = (comLine->options & TG_VOLATILE)
		? ((BALTREE *) NULL) : BalTrCreate(&pthBalTools);
	    ForInit(comLine->ul.formula, theGraph);
	    forIsOpen = TRUE;
	    i = 0L;
	    count = 0L;
            inner = 0;
            outer = 0;
	    while (TRUE)
		{
		    if (!(path = ForEval(comLine->ul.formula, theGraph,
					 PthEnd(thePath))))
			break;
		    if ((errorFlag != NO_ERROR) || !continueEvaluation)
			{
			    PthDiscard(path);
			    break;
			}
		    if (comLine->options & TG_MUTE)
                        {
		            if (dots)
		                {
		                    if (inner == 9)
		                        {
		                            inner = 0;
		                            (void) fputc((int) '.', _out);
		                            if (outer == 24)
		                                {
		                                    outer = 0;
		                                    (void)
							fputc((int) '\n',
							      _out);
		                                }
		                            else
		                                outer++;
		                            (void) fflush(_out);
		                        }
		                    else
		                        inner++;
		                }
                        }
		    else if (comLine->options & TG_VERBOSE)
			PthVerbose(path, _out);
		    else
			PthOutput(path, _out);
		    if (comLine->options & TG_VOLATILE)
			PthDiscard(path);
		    else if (BalTrInsert(paths, (char *) path,
					 (char *) path, TRUE))
			{
			    PthDiscard(path);
			    count++;
			}
		    i++;
		}
            if ((comLine->options & TG_MUTE) && dots && (outer != 0))
                {
                    (void) fputc((int) '\n', _out);
                    (void) fflush(_out);
                }
	    if (!((comLine->options & TG_MUTE) ||
		  (comLine->options & TG_VERBOSE)))
		PthOutput((PATH *) NULL, _out);
	    forIsOpen = FALSE;
	    ForClose(comLine->ul.formula, theGraph);
	    if (errorFlag == NO_ERROR)
		{
		    if (count != 0L)
			{
			    (void) fprintf(_out,
					   "%lu paths (%lu duplicates)\n",
					  i, count);
			    (void) fflush(_out);
			}
		    else
			{
			    (void) fprintf(_out, "%lu paths\n", i);
			    (void) fflush(_out);
			}
		    if (!(comLine->options & TG_VOLATILE))
			{
			    if (continueEvaluation)
				str = StrCopy(command);
			    else
				{
				    str = (char *) Allocate((unsigned)
						   (strlen(command) + 16));
				    (void) sprintf(str, "%s (BROKEN QUERY)",
						   command);
				}
			    (void) fputs("Built set ", _out);
			    SetDenote(_out, AddSet(paths, str, clbleSets));
			    (void) fputc((int) '\n', _out);
			    (void) fflush(_out);
			}
		}
	    else if (!(comLine->options & TG_VOLATILE))
		BalTrDestroy(paths, TRUE);
	    if ((errorFlag == NO_ERROR) &&
		!(comLine->options & TG_MUTE))
		CutVerbose(_out);
	    break;
	case LI_SETS:
	    if (((unsigned long) 0L) == compCount)
		StrongNotComp(_out);
	    else if (((unsigned long) 1L) == compCount)
		{
		    (void) fputs("One strongly connected component: ",
				 _out);
		    SgcDenote(_out, (unsigned long) 0L);
		    (void) fputc((int) '\n', _out);
		}
	    else
		{
		    (void) fputs("Strongly connected components: ",
				 _out);
		    SgcDenote(_out, (unsigned long) 0L);
		    (void) fputs("..", _out);
		    SgcDenote(_out, compCount - 1L);
		    (void) fputc((int) '\n', _out);
		}
	    if (BalTrGElCount(specSets) != 0L)
		{
		    (void) fputs("Special sets:\n", _out);
		    (void) BalTrApply(specSets, ShowSet);
		}
	    if (BalTrGElCount(clbleSets) != 0L)
		{
		    (void) fputs("Clearable sets:\n", _out);
		    (void) BalTrApply(clbleSets, ShowSet);
		}
	    (void) fflush(_out);
	    break;
	case LI_SGPRED:
	case LI_SGSUCC:
	    i = (unsigned long) XprEval(comLine->ul.expression);
	    if ((errorFlag == NO_ERROR) && continueEvaluation)
		{
		    if (((unsigned long) 0L) == compCount)
			{
			    StrongNotComp(_err);
			    errorFlag = SEM_ERROR;
			}
		    else if (i >= compCount)
			{
			    IsntSgc(_err, i);
			    errorFlag = SEM_ERROR;
			}
		    else if (comLine->operator == LI_SGPRED)
			SgPred((BALTREE *) NULL, i);
		    else
			SgSucc((BALTREE *) NULL, i);
		}
	    break;
	case LI_SHOWENDS:
	    if (comLine->ul.setOper == NULL)
		{
		    ndCount = GphNdCount(theGraph);
		    for (i = 0L; (i < ndCount) && continueEvaluation; i++)
			{
			    world = GphId(theGraph, i);
			    if (comLine->options & TG_MUTE)
				NodeBelongs(_out, theGraph, world);
			    else
				{
				    WldVerbose(world, _out);
				    CutVerbose(_out);
				}
			    if (comLine->options & TG_PREDS)
				NdPred(world);
			    if (comLine->options & TG_SUCCS)
				NdSucc(world, (EXPRESSION *) NULL);
			}
		    if (continueEvaluation)
			{
			    (void) fprintf(_out, "%lu nodes (all)\n", i);
			    (void) fflush(_out);
			}
		}
	    else
		{
		    if ((paths = SopEval(comLine->ul.setOper)))
			{
			    shPath = (PATH *) NULL;
			    shCounter = 0L;
			    (void) BalTrApply(paths, ShowEnd);
			    if (continueEvaluation)
				{
				    (void) fprintf(_out,
						   "%lu path end nodes\n",
						  shCounter);
				    (void) fflush(_out);
				}
			    if (comLine->ul.setOper->operator != SE_SET)
				BalTrDestroy(paths, TRUE);
			}
		}
	    if (continueEvaluation && (errorFlag == NO_ERROR) &&
		!(comLine->options & TG_MUTE))
		CutVerbose(_out);
	    break;
	case LI_SHRINK:
	    i = (unsigned long) XprEval(comLine->ul.expression);
	    if ((errorFlag == NO_ERROR) && continueEvaluation)
		ShrinkSet(i);
	    break;
	case LI_STATISTICS:
	    GphStatistics(theGraph, _out);
	    break;
	case LI_SUCC:
	    state = (unsigned long) XprEval(comLine->ul.expression);
	    if ((state < GphNdCount(theGraph)) && (errorFlag == NO_ERROR) &&
		continueEvaluation)
		NdSucc(GphId(theGraph, state), comLine->expr2);
	    else
		{
		    if ((errorFlag == NO_ERROR) && continueEvaluation)
			{
			    IsntNode(_err, state);
			    errorFlag = SEM_ERROR;
			}
		}
	    break;
	case LI_TERMCOMPS:
	    if (((unsigned long) 0L) == compCount)
		{
		    if (errorFlag == NO_ERROR)
			{
			    StrongNotComp(_err);
			    errorFlag = SEM_ERROR;
			}
		}
	    else
		{
		    (void) BalTrApply(termComps, ShowLSg);
		    if (continueEvaluation)
			{
			    (void)
				fprintf
				(_out,
				 "%lu nontrivial terminal strongly connected components\n",
				 GphTmSgCount(theGraph));
			    if (sgsInMemory)
				{
				    SetDenote(_out, untscc);
				    (void) fputs(" is their union\n",
						 _out);
				}
			    (void) fflush(_out);
			}
		    if (continueEvaluation && (errorFlag == NO_ERROR) &&
			!(comLine->options & TG_MUTE))
			CutVerbose(_out);
		}
	    break;
	case LI_QUIT:
	    PthDiscard(thePath);
	    break;
	default:
	    (void) fputs("Sorry, not implemented yet!\n", _out);
	    (void) fflush(_out);
	}
}

/***************************************************************************
*
* Function:  GetSg
* Purpose:
* Arguments: balTree -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, GetSg,
    BALTREE **, balTree)
{
    *balTree = GphSgGet(theGraph, shCounter++);
    return(FALSE);
}

/***************************************************************************
*
* Function:  AddToTmPaths
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, AddToTmPaths,
    unsigned long *, lng)
{
    PATH *path;

    path = PthCreate(*lng);
    (void) BalTrInsert(tmPaths, (char *) path, (char *) path, TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  BuildTmPaths
* Purpose:
* Arguments: lng -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, BuildTmPaths,
    unsigned long *, lng)
{
    (void) BalTrApply((BALTREE *) PtrBlGet(comps, *lng), AddToTmPaths);
    return(FALSE);
}

/***************************************************************************
*
* Function:  GetReserved
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, GetReserved,
    char *, name)
{
    BALTREE *tree, *paths;
    BINARYFILE bFile = NonBinaryFile();
    char *str;
    PATH *pth;
    unsigned long lng2[2];

    lng2[0] = 0L;
    lng2[1] = 0L;
    tree = GphTNGet(theGraph);
    paths = BalTrImage(tree, &pthBalTools, (CFUNC) LngToPth);
    BalTrDestroy(tree, TRUE);
    (void) AddSet(paths, StrCopy("** terminal nodes **"), specSets);
    if (GphErr(theGraph))
	{
	    str = FileName(name, "err");
	    if (NotBinaryFile(bFile = OpenRead(str, (BOOL *) NULL)))
		ClDeallocate(str);
	    else
		{
		    ClDeallocate(str);
		    pth = PthCreate((unsigned long) -1L);
		    while (TRUE)
			{
			    ReadLng(lng2, 2, bFile);
			    if (lng2[1] == (unsigned long) -1L)
				break;
			    PthInsFirst(pth, lng2[0], lng2[1]);
			}
		    PthSEnd(pth, lng2[0]);
		    OwnClose(&bFile);
		    paths = BalTrCreate(&pthBalTools);
		    (void) BalTrInsert(paths, (char *) pth,
				       (char *) pth, TRUE);
		    (void) AddSet(paths, StrCopy("** illegal path **"),
				  specSets);
		}
	}
    if ((compCount = GphSgCount(theGraph)) != (unsigned long) 0L)
	{
	    termComps = GphTmSgGet(theGraph);
	    if (sgsInMemory)
		{
		    comps = PtrBlCreate(compCount, &balTrTools);
		    shCounter = 0L;
		    (void) PtrBlAddrApply(comps, GetSg);
		    tmPaths = BalTrCreate(&pthBalTools);
		    (void) BalTrApply(termComps, BuildTmPaths);
		    untscc =
			AddSet
			(tmPaths,
			 StrCopy
			 ("** union of nontrivial terminal strongly connected components **"),
			 specSets);
		}
	}
}

static BINARYFILE fldFile = NonBinaryFile();

/***************************************************************************
*
* Function:  ReadFld
* Purpose:
* Arguments: s -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, ReadFld,
    DATATYPE **, s)
{
    unsigned len = 0;

    ReadUns(&len, 1, fldFile);
    *s = (DATATYPE *) AllocArray(len + 1, sizeof(DATATYPE));
    ReadData(*s, len, fldFile);
    return(FALSE);
}

/***************************************************************************
*
* Function:  FoldingInput
* Purpose:
* Arguments: name -
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, FoldingInput,
    char *, name)
{
    char *str;
    unsigned long a = 0L;

    str = FileName(name, "fld");
    if (NotBinaryFile(fldFile = OpenRead(str, (BOOL *) NULL)))
        {
	    ClDeallocate(str);
	    Exit(4);
	}
    ClDeallocate(str);
    ReadLng(&a, 1, fldFile);
    placeFoldings = PtrBlCreate(a, &dataTools);
    (void) PtrBlAddrApply(placeFoldings, ReadFld);
    ReadLng(&a, 1, fldFile);
    transFoldings = PtrBlCreate(a, &dataTools);
    (void) PtrBlAddrApply(transFoldings, ReadFld);
    OwnClose(&fldFile);
}

/***************************************************************************
*
* Function:  Misusage
* Purpose:   This routine handles most of the cases where the command line
*            of the program is incorrect.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, Misusage)
{
    MsgExit("[-.] [-d] [-e] [-G] [-hname] [-i]\n[-lname]... [-M] [-q] [-t]name [-winteger]",
	    (int) 1);
}

FE0(void, PrbMemClose)
{
    MacpClean();
    MaceClean();
    RulesClean();
    CLineClose();
    GphClose(theGraph);
    BalTrDestroy(termComps, TRUE);
    PtrBlDestroy(comps, TRUE);
    PtrBlDestroy(placeFoldings, TRUE);
    PtrBlDestroy(transFoldings, TRUE);
    BalTrDestroy(specSets, TRUE);
    BalTrDestroy(clbleSets, TRUE);
    BalTrDestroy(freeSets, TRUE);
    PthDiscard(thePath);
    PthDiscard(the2Path);
    PthDiscard(the3Path);
    PthDiscard(nowPath);
    StaDiscard(theState);
    StaDiscard(theFolState);
    StaDiscard(theEvalState);
    DynamClose();
    ClDeallocate(probHelp);
    ClDeallocate(sourceName);
#ifdef USE_READLINE
    MbufRlClose();
#endif
    LstDestroy(largvList, FALSE);
    LstDestroy(nameList, TRUE);
    LstDestroy(fileList, FALSE);
    BalTrDestroy(trivMace, TRUE);
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments: argc - the number of command line arguments given to the
*                   program
*            argv - an array containing the arguments
* Returns:   0 if everything succeeds, and something else otherwise
*            (PROD has no systematic convention for non-zero exits.)
* See also:
*
****************************************************************************
*/
MAIN2()
{
    int argc = 0;
    char **argv = NULL;
    char *argument, *ws, *str;
    BINARYFILE datFile = NonBinaryFile();
    FILE *source, *src;
    int i;
    BOOL echoLoadInput, echo_In, echoState;
    BOOL mapToMem;
#ifdef USE_READLINE
    char prompt[64];
#endif

#ifndef UNIX
    ClockOn();
#endif
    SetIO();
    _err = _out;
    ClearErrno();
    SetExMex();
    baseName = StrCopy("probe");
#ifdef ARGV0
    myName = CmdNm(_argv_[0]);
#else
    myName = baseName;
#endif
#ifndef UNIX
    StoreCmd(_argc_, myName, (char **) _argv_);
#endif
    ExpArgs(_argc_, (char **) _argv_, &argc, &argv);
    IgnoreSignals();
    _memClose = (VOF) PrbMemClose;
    DynamInit();
#ifdef USE_READLINE
    MbufRlInit();
#endif
    prefDefault = NULL;
    dots = FALSE;
    echo_In = FALSE;
    echoLoadInput = FALSE;
    echoState = TRUE;
    ws = NULL;
    sgsInMemory = FALSE;
    largvList = LstCreate(&strTools);
    mapToMem = FALSE;
    for (i = 1; i < argc; i++)
	{
	    argument = argv[i];
	    if (*argument == '-')
		{
		    argument++;
		    if ((*argument == '\0') &&
			(i + 1 == argc))
			Misusage();
		    if (*argument == '\0')
			{
			    argument = argv[i + 1];
			    i++;
			}
		    while (*argument != '\0')
			{
			    if (*argument == '.')
			        dots = TRUE;
			    else if (*argument == 'd')
				ResetErr();
			    else if (*argument == 'e')
				echo_In = TRUE;
			    else if (*argument == 'G')
				{
				    mapToMem = TRUE;
				    sgsInMemory = FALSE;
				}
			    else if (*argument == 'h')
				{
				    argument++;
				    if ((((char *) NULL) != probHelp) ||
					((*argument == '\0') &&
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    probHelp = StrCopy(argument);
				    break;
				}
			    else if (*argument == 'i')
				echoLoadInput = TRUE;
			    else if (*argument == 'l')
				{
				    argument++;
				    if ((*argument == '\0') &&
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsLast(largvList,
						      (char *) argument);
				    break;
				}
			    else if (*argument == 'M')
				{
				    if (!mapToMem)
					sgsInMemory = TRUE;
				}
			    else if (*argument == 'q')
				echoState = FALSE;
			    else if (*argument == 't')
				{
				    argument++;
				    if ((prefDefault != NULL) ||
					((*argument == '\0') &&
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    prefDefault = argument;
				    break;
				}
			    else if (*argument == 'w')
				{
				    argument++;
				    if ((ws != NULL) ||
					((*argument == '\0') &&
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    ws = argument;
				    if (ScanLdu(ws, &lineLength) != 1)
					Misusage();
				    break;
				}
			    else
				Misusage();
			    argument++;
			}
		}
	    else if (prefDefault == NULL)
		prefDefault = argument;
	    else
		Misusage();
	}
    if (prefDefault == NULL)
	Misusage();
    if (((char *) NULL) == probHelp)
	probHelp = StrCopy("probhelp");
    i = (int) strlen(prefDefault);
    if ((i > 4) && (strcmp(prefDefault + i - 4, ".gph") == 0))
	prefDefault[i - 4] = '\0';
    if (ws == NULL)
	lineLength = 79L;
    else if (lineLength < 39L)
	{
	    lineLength = 39L;
	    (void) fprintf(_err, "-w %s transformed to -w %lu\n",
			   ws, lineLength);
	    (void) fflush(_err);
	}
    echoInput = echoLoadInput;
    sourceName = FileName(prefDefault, "dat");
    if (NotBinaryFile(datFile = OpenRead(sourceName, (BOOL *) NULL)))
        {
	    Exit(2);
	}
    ClDeallocate(sourceName);
    CLineInit(datFile);
    OwnClose(&datFile);
    theGraph = GphOpen(prefDefault, GPH_PROBE, FALSE);
    if (theGraph == NULL)
        {
	    Exit(3);
	}
    if (mapToMem)
	GphMapToMem(theGraph, (unsigned long) 0L);
    unfold = GphUnfold(theGraph);
    if (unfold)
	FoldingInput(prefDefault);
    symmMethod = GphSymm(theGraph);
    if (GphFreeExpansion(theGraph))
	SetFreeExpansion();
    else
	ResetFreeExpansion();
    thePath = PthCreate(GphStart(theGraph));
    specSets = BalTrCreate(&ptsBalTools);
    clbleSets = BalTrCreate(&ptsBalTools);
    freeSets = BalTrCreate(&lngBalTools);
    GetReserved(prefDefault);
    nameList = LstCreate(&strTools);
    fileList = LstCreate(&strTools);
    LSTFOR (largvList, argument, char)
	{
	    sourceName = StrCopy(argument);
	    if (!(source = OpenText((char *) NULL, sourceName, FALSE, FALSE)))
		{
		    ClDeallocate(sourceName);
		    continue;
		}
	    (void) LstInsFirst(nameList, (char *) sourceName);
	    sourceName = (char *) NULL;
	    while (thePath)
		{
		    while (MyFeof(source))
			{
			    MyFclose(source);
			    (void) LstDelFirst(nameList, TRUE);
			    if (LstEmpty(nameList))
				break;
			    source = (FILE *) LstDelFirst(fileList, FALSE);
			}
		    if (LstEmpty(nameList))
			break;
		    if (echoState && echoLoadInput)
			{
			    (void) fprintf(_out, "%lu#",
				   GphOrdinal(theGraph, PthEnd(thePath)));
			    (void) fflush(_out);
			}
		    errorFlag = NO_ERROR;
		    continueEvaluation = TRUE;
		    sourceName = (char *) NULL;
		    SetSigHandlers();
#ifdef USE_READLINE
		    CLine(source, (char *) NULL);
		    if (rdlnEof)
			PthDiscard(thePath);
#else
		    CLine(source);
#endif
		    if (!sourceName)
			continue;
		    LSTFOR (nameList, str, char)
			if (strcmp(str, sourceName) == 0)
			    break;
		    if (str)
			{
			    ClDeallocate(sourceName);
			    (void) fprintf(_err, "Load loop: %s\n", str);
			    LSTFOR (nameList, sourceName, char)
				{
				    if (sourceName != (char *)
					LstFirst(nameList))
					(void) fputs("which ", _err);
				    (void) fprintf(_err, "is called by %s\n",
					    sourceName);
				    if (sourceName == str)
					break;
				}
			    sourceName = (char *) NULL;
			    (void) fflush(_err);
			}
		    else if (!(src = OpenText((char *) NULL,
					      sourceName, FALSE, FALSE)))
			ClDeallocate(sourceName);
		    else
			{
			    (void) LstInsFirst(nameList, (char *) sourceName);
			    sourceName = (char *) NULL;
			    (void) LstInsFirst(fileList, (char *) source);
			    source = src;
			}
		}
	    if (!thePath)
		{
		    IgnSigHandlers();
		    while (TRUE)
			{
			    MyFclose(source);
			    (void) LstDelFirst(nameList, TRUE);
			    if (LstEmpty(nameList))
				break;
			    source = (FILE *)
				LstDelFirst(fileList, FALSE);
			}
		    GphClose(theGraph);
		    break;
		}
	}
    if (!thePath)
        {
	    Exit(0);
	}
    LstDestroy(largvList, FALSE);
    source = _in;
    while (TRUE)
	{
	    if (!thePath)
		{
		    IgnSigHandlers();
		    while (TRUE)
			{
			    if (source != _in)
				MyFclose(source);
			    if (LstEmpty(nameList))
				break;
			    (void) LstDelFirst(nameList, TRUE);
			    source = (FILE *)
				LstDelFirst(fileList, FALSE);
			}
		    GphClose(theGraph);
		    break;
		}
	    while (MyFeof(source))
		{
		    if (source != _in)
			MyFclose(source);
		    if (LstEmpty(nameList))
			{
			    IgnSigHandlers();
			    GphClose(theGraph);
			    PthDiscard(thePath);
			    break;
			}
		    (void) LstDelFirst(nameList, TRUE);
		    source = (FILE *) LstDelFirst(fileList, FALSE);
		}
	    if (!thePath)
		break;
	    if (source == _in)
		echoInput = echo_In;
	    else
		echoInput = echoLoadInput;
	    if (echoState && (echoLoadInput || (source == _in)))
		{
#ifdef USE_READLINE
		    if ((source != _in) || !isatty(fileno(source)))
			{
#endif
			    (void) fprintf(_out, "%lu#",
					   GphOrdinal(theGraph,
						      PthEnd(thePath)));
			    (void) fflush(_out);
#ifdef USE_READLINE
			}
		    else
			(void) sprintf(prompt, "%lu#",
				       GphOrdinal(theGraph,
						  PthEnd(thePath)));
#endif
		}
	    errorFlag = NO_ERROR;
	    continueEvaluation = TRUE;
	    sourceName = (char *) NULL;
	    SetSigHandlers();
#ifdef USE_READLINE
	    CLine(source, prompt);
	    if (rdlnEof)
		PthDiscard(thePath);
#else
	    CLine(source);
#endif
	    if (!sourceName)
		continue;
	    LSTFOR (nameList, str, char)
		if (strcmp(str, sourceName) == 0)
		    break;
	    if (str)
		{
		    ClDeallocate(sourceName);
		    (void) fprintf(_err, "Load loop: %s\n", str);
		    LSTFOR (nameList, sourceName, char)
			{
			    if (sourceName != (char *) LstFirst(nameList))
				(void) fputs("which ", _err);
			    (void) fprintf(_err, "is called by %s\n",
					   sourceName);
			    if (sourceName == str)
				break;
			}
		    sourceName = (char *) NULL;
		    (void) fflush(_err);
		}
	    else if (!(src = OpenText((char *) NULL, sourceName,
				      FALSE, FALSE)))
		ClDeallocate(sourceName);
	    else
		{
		    (void) LstInsFirst(nameList, (char *) sourceName);
		    sourceName = (char *) NULL;
		    (void) LstInsFirst(fileList, (char *) source);
		    source = src;
		}
	}
    Exit(0);
    return(0);
}
