/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  probesig.c
*
* Part of: PROD's program probe
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module takes care of the signal handling of the probe program.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "probesig.h"

/***************************************************************************
*
* Function:  IgnoreSignals
* Purpose:   This routine forces certain signals to be ignored.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, IgnoreSignals)
{
#ifndef _WINDOWS
    (void) signal(SIGINT, SIG_IGN);
#endif
#ifdef UNIX
    (void) signal(SIGHUP, SIG_IGN);
    (void) signal(SIGTERM, SIG_IGN);
#endif
}

#if (defined(UNIX) || !defined(_WINDOWS))
/***************************************************************************
*
* Function:  KillEvaluation
* Purpose:   This routine handles the SIGINT signal.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, KillEvaluation,
    int, sig)
{
    if (((int) SIGINT) == sig)
	{
	    (void) signal(SIGINT, SIG_IGN);
	    continueEvaluation = FALSE;
	}
}
#endif

/***************************************************************************
*
* Function:  SetSigHandlers
* Purpose:   This routine sets the signal handlers.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, SetSigHandlers)
{
#ifndef _WINDOWS
    (void) signal(SIGINT, KillEvaluation);
#endif
}

/***************************************************************************
*
* Function:  IgnSigHandlers
* Purpose:   This routine disables the signal handlers. The corresponding
*            signals are then ignored.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, IgnSigHandlers)
{
#ifndef _WINDOWS
    (void) signal(SIGINT, SIG_IGN);
#endif
}
