/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prodmac.c
*
* Part of: PROD's program prod
*
* Authors: James Reilly and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "prodtok.h"

/*
 * prod's "Macro package" that handles building a macro
 *        table from both user-defined and prodfile macros,
 *	  expansion of macros, etc.  Macros are basically
 *	  a (Name, Value) tuple.   Targets were also implemented
 *	  using the handy macro package.
 */
/***************************************************************************
*
* Function:  tail_of_macro
* Purpose:
* Arguments: macro - 
* Returns:
* See also:
*
****************************************************************************
*/
FS1(Macro *, tail_of_macro,
    Macro *, macro)
{
    Macro * m = macro;
    if (m == NULL)
	return NULL;
    else {
	while (m -> next != NULL)
		m = m -> next;
    }
    return m;
}

/***************************************************************************
*
* Function:  new_macro
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(Macro *, new_macro)
{
    /* alloc and initialize a new Macro */

    Macro * m = (Macro *) Allocate (sizeof *m);
    m -> name  = NULL;
    m -> value = NULL;
    m -> has_value = HS_UNDEFINED;
    m->matched = FALSE;
    m -> next  = NULL;
    return m;
}

/***************************************************************************
*
* Function:  stringAlloc
* Purpose:
* Arguments: s2 - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, stringAlloc,
    char *, s2)
{
	/* $(X) -> X for name ? */
	/* check if name is present already ? */
	char * s1 = (char *) Allocate ((unsigned) (strlen (s2) + 1));
	(void) strcpy (s1, s2);
	return s1;
}

/*
 * argument value of macro 
 */
/***************************************************************************
*
* Function:  argv_macro
* Purpose:
* Arguments: macro - 
*            string - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(Macro *, argv_macro,
    Macro *, macro,
    char *, string)
{
    Macro * m = tail_of_macro (macro);
    char * s = NULL;

    if (m == NULL)  {
	m = new_macro();

	if ((s = strchr (string, '=')) != NULL) { 
	   *s++ = '\0';
	   m -> value = stringAlloc (s);
	}
	else
	   m -> value = stringAlloc ("1");
	m -> name = stringAlloc (string);
	m -> has_value = HS_USER;	/* user defined */
	return m;
   }
   else 
	return (m -> next = argv_macro (m -> next, string));
}

/*
 * argument value of target 
 */
/***************************************************************************
*
* Function:  argv_target
* Purpose:
* Arguments: macro - 
*            string - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(Macro *, argv_target,
    Macro *, macro,
    char *, string)
{
    Macro * m = tail_of_macro (macro);

    if (m == NULL)  {
	m = new_macro();
	m -> value = stringAlloc ("");
	m -> name = stringAlloc (string);	/* target prefix */
	m -> has_value = HS_USER;			/* user defined */
	return m;
   }
   else 
	return (m -> next = argv_target (m -> next, string));
}

/***************************************************************************
*
* Function:  macro2value
* Purpose:
* Arguments: search - 
*            name - 
* Returns:
* See also:
*
****************************************************************************
*/
FS2(char *, macro2value,
    Macro *, search,
    char *, name)
{
    while (search != NULL) {
	if (!strcmp (name, search -> name))
	    return search -> value;
        search = search -> next;
    }
    return NULL;	/* MACRO NOT FOUND */
}

/*
 * s is a string that has one or more macros.
 * expand the macros in s.
 */
char expansion[BUFFER_SIZE];
char ex_name[BUFFER_SIZE];

/***************************************************************************
*
* Function:  expandMacro
* Purpose:
* Arguments: lookup - 
*            str - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(char *, expandMacro,
    Macro *,lookup,
    char *, str)
{
    char *s = str, * aux;
    char *buf=expansion, *name=ex_name, *value=NULL;

    *buf = *name = '\0'; /* init */

    while (*s != '\0') {
        if (*s == '$' && *(s+1) == '(') {	/* macro found */
            *(name = ex_name) = '\0';	/* re-init */
            for (s=s+2; *s != ')'; s++) {
		if (*s == '\0') {	/* error */
		    (void) fprintf (_err, 
				    "Error: macro %s missing ')'\n", str);
                    return NULL;	
                }
                else {
                    *name++ = *s ;
                }
	    }
            *name = '\0' ;
            if ((value = macro2value (lookup, ex_name))) {
		aux = value;
		while (*aux != '\0')
		    {
			if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
			    {
				(void) fprintf(_err,
					       "\n%s: line buffer overflow\n",
					       myName);
				(void) fflush(_err);
				Exit(1);
			    }
			*buf++ = *aux++;
		    }
            }
	    s++;	/* move past ')' in original macro */
        }
	else if (*s == '$' && *(s+1) == '*')
	    {	/* internal macro found */
		aux = cuTaBaseName;
		while (*aux != '\0')
		    {
			if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
			    {
				(void) fprintf(_err,
					       "\n%s: line buffer overflow\n",
					       myName);
				(void) fflush(_err);
				Exit(1);
			    }
			*buf++ = *aux++;
		    }
		s++; s++;	/* advance past $* */
        }
	else if (*s == '$' && *(s+1) == '@')
	    {	/* internal macro found */
		aux = cuTaBaseName;
		while (*aux != '\0')
		    {
			if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
			    {
				(void) fprintf(_err,
					       "\n%s: line buffer overflow\n",
					       myName);
				(void) fflush(_err);
				Exit(1);
			    }
			*buf++ = *aux++;
		    }
		if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
		    {
			(void) fprintf(_err,
				       "\n%s: line buffer overflow\n", myName);
			(void) fflush(_err);
			Exit(1);
		    }
		*buf++ = '.';
		aux = cuTaSuffix;
		while (*aux != '\0')
		    {
			if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
			    {
				(void) fprintf(_err,
					       "\n%s: line buffer overflow\n",
					       myName);
				(void) fflush(_err);
				Exit(1);
			    }
			*buf++ = *aux++;
		    }
		s++; s++;	/* advance past $@ */
        }
	else {
	    if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
		{
		    (void) fprintf(_err,
				   "\n%s: line buffer overflow\n", myName);
		    (void) fflush(_err);
		    Exit(1);
		}
	    *buf++ = *s++ ;
        }
    }
    if (((unsigned) (buf - expansion)) >= BUFFER_SIZE)
	{
	    (void) fprintf(_err,
			   "\n%s: line buffer overflow\n", myName);
	    (void) fflush(_err);
	    Exit(1);
	}
    *buf = '\0' ;	/* terminate */
    return expansion;
}

static Macro * true_head;	/* true head for expandMacro () */
/***************************************************************************
*
* Function:  push_macro
* Purpose:
* Arguments: macro - 
*            name - 
*            value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE3(Macro *, push_macro,
    Macro *, macro,
    char *, name,
    char *, value)
{
    Macro * m = tail_of_macro (macro);

    if (m == NULL) {
	m = new_macro();
	m -> name = stringAlloc (name);
	m -> value = stringAlloc (expandMacro (true_head, value));
        if (*m -> value != '\0')
	    m -> has_value = HS_PROGRAM;
        m -> next = NULL;
	return m;
    }
    else 
	return (m -> next = push_macro (m -> next, name, value));
}

/***************************************************************************
*
* Function:  macro_find
* Purpose:
* Arguments: search - 
*            name - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(Macro *, macro_find,
    Macro *, search,
    char *, name)
{
    while (search != NULL) {
	if (!strcmp (name, search -> name))
	    return search ;
        search = search -> next;
    }
    return NULL;	/* MACRO NOT FOUND */
}

/***************************************************************************
*
* Function:  add_macro
* Purpose:
* Arguments: head - 
*            name - 
*            value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE3(Macro *, add_macro,
    Macro *, head,
    char *, name,
    char *, value)
{
    Macro * m=NULL;

    true_head = head;	/* true head for expanding in push_macro */

    if (macro2value (head, name) == NULL)
	return push_macro (head, name, value);
    else { /* it already exists */
        m = macro_find (head, name);
	/* replace existing macro, unless user defined it */
        if (m->value && m->has_value != HS_USER) {
	    Deallocate((char *) (m->value));  m->value = NULL;
	    m->value = stringAlloc (expandMacro (head, value));
        }
        return tail_of_macro(head);
    }
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  print_macro
* Purpose:
* Arguments: macro - 
*            strm - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, print_macro,
    Macro *, macro,
    FILE *, strm)
{
    Macro * m ;
    for (m = macro; m != NULL; m = m -> next) {
        (void) fprintf (strm, "\"%s\" := \"%s\"\n", m -> name, m -> value);
    }
}
#endif
