/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prodmain.c
*
* Part of: PROD's program prod
*
* Authors: Lasse Anderson, Robert Jurgens, James Reilly,
*          and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The "prod" program is a general purpose batch program though
*     its main purpose should be to help PROD users.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "prodtok.h"

/***************************************************************************
*
* Function:  Misusage
* Purpose:   This routine handles most of the cases where the command line
*            of the program is incorrect.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, Misusage)
{
    MsgExit("[-mname[=defstr]]... [-pname] [[-t]name]...", (int) 2);
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments: argc - the number of command line arguments given to the
*                   program
*            argv - an array containing the arguments
* Returns:   0 if everything succeeds, and something else otherwise
*            (PROD has no systematic convention for non-zero exits.)
* See also:
*
****************************************************************************
*/
MAIN2()
{
    int argc = 0;
    char **argv = NULL;
    int i;
    char * s = NULL, * prodfile = NULL, * argument;
    Macro  * mhead = NULL , * mtail = NULL;
    Macro * thead = NULL , * ttail = NULL; /* kludgey targets ? jfr */
    Token  token;			/* token passed back by regexp() */
    
    SetIO();
    ClearErrno();
    baseName = StrCopy("prod");
#ifdef ARGV0
    myName = CmdNm(_argv_[0]);
#else
    myName = baseName;
#endif
    ExpArgs(_argc_, (char **) _argv_, &argc, &argv);
    /* 
     * Handle command line arguments.
     *   build up macro table, target table.
     *   Command line is of form :
     *     prod [-m macro]* [-p prodfile] [[-t] target]*
     *
     *   Target names are of form :   name.suffix
     *   the suffix is matched against targets in prodfile
     *   to find correct batch execution.
     */
    
    for (i = 1; i < argc; i++)
	{
	    argument = argv[i];
	    if (*argument == '-')
		{
		    argument++;
		    if ((*argument == '\0') &&
			(i + 1 == argc))
			Misusage();
		    if (*argument == '\0')
			{
			    argument = argv[i + 1];
			    i++;
			}
		    while (*argument != '\0')
			{
			    if (*argument == 'm')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    mtail = argv_macro (mhead, argument);
				    if (mhead == NULL)
					mhead = mtail;
				    break;
				}
			    else if (*argument == 'p')
				{
				    argument++;
				    if ((prodfile != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    prodfile = argument;
				    break;
				}
			    else if (*argument == 't')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    ttail = argv_target (thead, argument);
				    if (thead == NULL) 
					thead = ttail;
				    break;
				}
			    else
				Misusage();
			    argument++;
			}
		}
	    else
		{
		    ttail = argv_target (thead, argument);
		    if (thead == NULL) 
			thead = ttail;
		}
	}
    /* Prodfile used */
    if (prodfile == NULL)
	prodfile = "prodfile";
    if (!(prodStream = OpenText((char *) NULL, prodfile, FALSE, TRUE)))
	Exit(3);
#ifdef DOSBATCH
    if (!(batchFile = MyFopen("tmp.bat", "w", (BOOL *) NULL)))
	Exit(4);
    (void) fputs("@echo on\n@if errorlevel 1 clerr.exe\n", batchFile);
#endif
    /* 
     * Parse input prodfile
     *	- continue building macro table.
     *  - When target rules are encountered, try and
     *    handle them for list of user's targets in target table.
     */
    while ((s = ReadLine()))
	{
	    token = * regexp (s);
	    switch (token.type)
		{
		case TK_IGNORE:
		    break;	/* next line */
		case TK_MACRO:
		    mtail = add_macro (mhead, token.u_tok.m.name,
				       token.u_tok.m.value);
		    if (mhead == NULL)
			mhead = mtail;
		    Deallocate((char *) (token.u_tok.m.name));
		    Deallocate((char *) (token.u_tok.m.value));
		    break;
		case TK_TARGET:
		    /* handle all targets that match this token */
		    /* if no target matches, skip to ';' in stream */
		    handle_target (thead, mhead, s);
		    break;
		case TK_CONDITIONAL:
		case TK_OTHER:
		default:
		    (void) fprintf (_err,
				    "\nError: %s - Unknown expression %s\n", 
				    prodfile, s);
		    (void) fflush(_err);
		    break;
		}
	}
#ifdef DOSBATCH
    (void) fputs(":end\n", batchFile);
    MyFclose(batchFile);
#endif
    i = 0;
    for (ttail = thead; ttail; ttail = ttail->next)
	if (!(ttail->matched))
	    {
		(void) fprintf(_err, "\n%s has no suffix from %s\n",
			       ttail->name, prodfile);
		(void) fflush(_err);
	    }
    Exit(0);
    return(0);
}
