/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prodrexp.c
*
* Part of: PROD's program prod
*
* Authors: James Reilly and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif



#include "prodtok.h"

Token grep_token;

/***************************************************************************
*
* Function:  eat_space
* Purpose:
* Arguments: s - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, eat_space,
    char *, s)
{
     while (*s != '\0') {
     	if (*s == ' ' || *s == '\t')
     		s++;
     	else 
     		break;
    }
    return s;
}

/***************************************************************************
*
* Function:  regexp
* Purpose:
* Arguments: s - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(Token *, regexp,
    char *, s)
{
    char *value, *name, * nf, * nl;

    s = eat_space (s);
    if (*s == '\0')
        grep_token.type = TK_IGNORE;
    else if (*s != '$')
	grep_token.type = TK_OTHER;
    else
	{
	    switch (*++s)
		{
		case '*' :	/* Handle target rules */
		    grep_token.type = TK_TARGET;
		    return(&grep_token); /* let caller handle_target() */
		case '(' :	/* Handle Macros */
		    nf = s;
		    nl = strchr (s, (int) ')');	/* name bounded by ( ) */
		    if (nl != NULL)
			{
			    grep_token.type = TK_MACRO;
			    name = (char *)
				Allocate ((unsigned) (nl - nf + 1));
			    s++;	/* 1st char of name */
			    for (nf = name; *s != '\0' && *s != ')';
				 s++, nf++)
				*nf = *s;
			    *nf = '\0';	/* null terminate */
			    grep_token.u_tok.m.next = NULL;
			    grep_token.u_tok.m.name = name;
			    /* eat white spaces after ')' */
			    s++;
			    s = eat_space (s);
			    if (*s != '=')
				{
				    (void) fprintf (_err,
						    "Error: '=' missing in macro (%s) assignment.\n",
						    name);
				    grep_token.u_tok.m.value = NULL;
				}
			    /* eat white space after '=' */
			    s++;
			    nf = eat_space (s);
			    value = (char *)
				Allocate ((unsigned) (strlen (nf) + 1));
			    (void) strcpy (value, nf);
			    grep_token.u_tok.m.value = value;
			}
		    else
			{
			    grep_token.type = TK_OTHER;
			    grep_token.u_tok.m.next = NULL;
			    grep_token.u_tok.m.name = NULL;
			    grep_token.u_tok.m.value = NULL;
			}
		    break;
		default:
		    (void) fprintf (_err,
				    "Error - Unknown prodfile directive : %s\n",
				    s);
		    grep_token.type = TK_OTHER;
		    break;
		}
	}
    return(&grep_token);
}
