/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prodtarg.c
*
* Part of: PROD's program prod
*
* Authors: James Reilly and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "prodtok.h"

#define SUBSTRCMP(sub, s2) (!strncmp(sub, s2, strlen(sub)))

/* globals for $* and $@ internal macros */
char * cuTaBaseName = NULL, * cuTaSuffix = NULL;
char ioBfr[BUFFER_SIZE];
char dupBuf[BUFFER_SIZE];
FILE *prodStream;
static int endOfStream = 0;
static unsigned long posOfStream = 0L;
#ifdef DOSBATCH
FILE *batchFile;
#endif

/***************************************************************************
*
* Function:  ReadLine
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(char *, ReadLine)
{
    unsigned i;
    int c, d, skip;
    char s[3];

#define ReadOne() (1L == (unsigned long) \
fread(s, (size_t) 1, (size_t) 1, prodStream)) \
? (posOfStream++, (int) *s) : ((int) EOF)

    if (endOfStream)
	return(NULL);
    c = ReadOne();
    if (c == EOF)
	{
	    endOfStream = 1;
	    return(NULL);
	}
    i = 0;
    skip = 0;
    while (TRUE)
	{
	    if (i >= BUFFER_SIZE)
		{
		    (void) fprintf(_err,
				   "\n%s: line buffer overflow\n", myName);
		    (void) fflush(_err);
		    Exit(1);
		}
	    if ((c == EOF) || (c == (int) '\n'))
		{
		    if (c == EOF)
			endOfStream = 1;
		    ioBfr[i] = '\0';
		    break;
		}
	    if (skip)
		{
		    if (c == (int) '\\')
			{
			    c = ReadOne();
			    if ((c != EOF) && (c == (int) '\n'))
				{
				    c = ReadOne();
				}
			}
		    else
			{
			    c = ReadOne();
			}
		}
	    else
		{
		    if (c == (int) '\\')
			{
			    c = ReadOne();
			    if ((c != EOF) && (c == (int) '\n'))
				{
				    ioBfr[i] = ' ';
				    i++;
				    c = ReadOne();
				}
			    else
				{
				    ioBfr[i] = '\\';
				    i++;
				}
			}
		    else if (c == (int) '#')
			{
			    c = ReadOne();
			    if ((c != EOF) && (c == (int) '#'))
				{
				    ioBfr[i] = '#';
				    i++;
				    c = ReadOne();
				}
			    else
				skip = 1;
			}
		    else
			{
			    d = ReadOne();
			    if ((c != (int) '\r') ||
				((d != EOF) && (d != (int) '\n')))
				{
				    ioBfr[i] = (char) c;
				    i++;
				}
			    c = d;
			}
		}
	}
    return(ioBfr);
}

/*
 * system_command :
 *   Execute a string that represents a host OS command,
 *   Exit if host executed command failed.
 */
/***************************************************************************
*
* Function:  system_command
* Purpose:
* Arguments: string - 
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, system_command,
    char *, string)
{
    int silent = 0;
    int ignore = FALSE;
#ifdef DOSBATCH
    char *t;
#else
    int val;
#endif
    
    if (string)
	{
	    /* bullet proofing */
	    if (*string != '\0')
		{
		    /* bullet proofing */
		    string = eat_space (string);
		    while (*string == '-' || *string == '@')
			{
			    if (*string == '-')
				{
				    ignore = TRUE;
				    string++;
				}
			    else if (*string == '@')
				{
				    silent = 1;
				    string++;
				}
			    else        break;
			}
#ifdef DOSBATCH
		    for (t = string; (*t != '\0') && (*t != '|'); t++)
			;
		    if (*t == '|')
			{
			    if (_out != _err)
				{
				    (void) fputc((int) '\n', _out);
				    (void) fflush(_out);
				}
			    (void) fputc((int) '\n', _err);
			    (void) fputs(myName, _err);
			    (void)
				fputs
				(": '|' is not allowed in a command\n",
				 _err);
			    (void) fflush(_err);
			    Exit(1);
			}
		    if (silent)
			(void) fputc((int) '@', batchFile);
		    (void) fputs(string, batchFile);
		    (void) fputc((int) '\n', batchFile);
		    if (ignore)
			(void) fputs("@if errorlevel 1 clerr.exe\n",
				     batchFile);
		    else
			(void) fputs("@if errorlevel 1 goto end\n",
				     batchFile);
#else
		    if (!silent)
			{
			    (void) fputc((int) '\n', _out);
			    (void) fputs(string, _out);
			    (void) fputc((int) '\n', _out);
			}
		    if ((val = MySystem(string, ignore)) && !ignore)
			Exit(val);
#endif
		}
	}
}

#define Token_if int
#define TI_DRIF 0
#define TI_DRFI 1
#define TI_DRELSE 2
#define TI_SEMICOLON 3
#define TI_OTHER 4

/*
 * check_token_if :
 *   checks to see if the current input string
 *   is an $IF, $ELSE, $FI, etc. expression.
 */
/***************************************************************************
*
* Function:  check_token_if
* Purpose:
* Arguments: s - 
* Returns:
* See also:
*
****************************************************************************
*/
FS1(Token_if, check_token_if,
    char *, s)
{
    s = eat_space(s);
    if (SUBSTRCMP("$IF ", s) || SUBSTRCMP("$IF\t", s) ||
	SUBSTRCMP("$IF(", s) || SUBSTRCMP("$IF\n", s) || !strcmp("$IF", s))
	return TI_DRIF;
    else if (SUBSTRCMP("$ELSE ", s) || SUBSTRCMP("$ELSE\t", s) ||
	     SUBSTRCMP("$ELSE\n", s) || !strcmp("$ELSE", s))
	return TI_DRELSE;
    else if (SUBSTRCMP("$FI ", s) || SUBSTRCMP("$FI\t", s) ||
	     SUBSTRCMP("$FI\n", s) || !strcmp("$FI", s))
	return TI_DRFI;
    else if (*s == ';')
	return TI_SEMICOLON;
    else 
	return TI_OTHER;
}

#define operation int
#define OP_NOTEQUALS 0
#define OP_EQUALS 1
#define OP_UNDEFINED 2

/*
 * evaluate_if : 
 *   Evaluates IF expressions of form:
 *       $IF ( arg1 == arg2 ) 
 *       $IF ( arg1 != arg2 ) 
 *   This handles very simple IF expressions now.  Modify
 *   it to handle more complex ones if they are needed.
 *   Basically does a strcmp (arg1, arg2).
 */
/***************************************************************************
*
* Function:  evaluate_if
* Purpose:
* Arguments: str - 
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, evaluate_if,
    char *, str)
{
    BOOL truth = TRUE;
    operation op = OP_UNDEFINED;
    char * arg1 = NULL, * arg2 = NULL;
    char * cp = NULL, *s = NULL;
    int i=0, len = 0, brkflg = 0;
    
    s = str;
    s = eat_space (s);
    if (s[0] == '$' && s[1] == 'I' && s[2] == 'F')
	{
	    /* $IF */
	    s = eat_space (s+3);        /* skip space after $IF */
	    if (s[0] == '(')
		{
		    s = eat_space (s+1);        /* skip space after ( */
		    if ((s[0] == '=' || s[0] == '!') && s[1] == '=')
			{
			    /* no arg1 */
			    arg1 = NULL;
			}
		    else
			{
			    /* compute strlen arg1 */
			    for (cp = s; *cp != '\0'; cp++)
				{
				    switch (*cp)
					{
					case ' ':
					case '\t':
					case '=':
					case '!':
					case ')':
					    brkflg = 1; /* exit for loop */
					    break;
					default:
					    if (!isprint((UCHAR) *cp))
						{
						    brkflg = 1;
						    len = 0;
						    (void) fprintf (_err,
								    "Error: Unprintable char [%x] in \"%s\"\n",
								    (unsigned)
								    *cp,
								    str);
						    return FALSE;
						}
					    else
						{
						    len++;
						    /* strlen (arg1) */
						}
					    break;
					}
				    if (brkflg)
					break;
				}
			    if (len == 0)
				{
				    (void) fprintf (_err,
						    "Error: $IF - bad argument \"%s\"\n",
						    str);
				    return FALSE;
				}
			    arg1 = (char *) Allocate((unsigned) (len + 1));
			    for (i=0, cp=s; i<len; i++, cp++)
				arg1[i] = *cp;
			    arg1[i] = '\0';
			    len = 0;
			    s = eat_space (cp);
			    /* set operation */
			    if (s[0] == '=' && s[1] == '=')
				{
				    op = OP_EQUALS;
				    s = &s[2];
				}
			    else if (s[0] == '!' && s[1] == '=')
				{
				    op = OP_NOTEQUALS;
				    s = &s[2];
				}
			    else
				{
				    op = OP_UNDEFINED;
				    (void) fprintf (_err,
						    "Error : Undefined operator - \"%s\"\n",
						    str);
				    return FALSE;
				}
			    s = eat_space(s);
			    /* compute strlen arg2 */
			    for (cp = s, len=0, brkflg=0; 
				 *cp != '\0'; cp++)
				{
				    switch (*cp)
					{
					case ' ':
					case '\t':
					case '=':
					case '!':
					case ')':
					    brkflg = 1; /* exit for loop */
					    break;
					default:
					    if (!isprint((UCHAR) *cp))
						{
						    brkflg = 1;
						    len = 0;
						    (void) fprintf (_err,
								    "Error: Unprintable char [%x] in \"%s\"\n",
								    (unsigned)
								    *cp,
								    str);
						    return FALSE;
						}
					    else
						{
						    len++;
						    /* strlen (arg2) */
						}
					    break;
					}
				    if (brkflg)
					break;
				}
			    if (len == 0)
				{
				    (void) fprintf (_err,
						    "Error: $IF - bad argument \"%s\"\n",
						    str);
				    return FALSE;
				}
			    arg2 = (char *) Allocate((unsigned) (len + 1));
			    for (i=0; i<len; i++, s++)
				arg2[i] = *s;
			    arg2[i] = '\0';
			    s = eat_space(s);
			    /* check for end ')' */
			    if (*s != ')')
				{
				    (void) fprintf (_err,
						    "Error: missing end ')' - %s\n",
						    str);
				    Exit(1);
				}
			    s++;
			    s = eat_space(s);
			    if (*s != '\0')
				{
				    (void) fprintf (_err, 
						    "Warning: text after ')' ignored - %s\n", 
						    str);
				}
			    /* compare arg1 arg2 */
			    switch (op)
				{
				case OP_EQUALS:
				    if (strcmp (arg1, arg2)) return FALSE;
				    break;
				case OP_NOTEQUALS:
				    if (!strcmp (arg1, arg2)) return FALSE;
				    break;
				default:
				    (void) fprintf (_err,
						    "Error: bad operator - \"%s\"\n",
						    str);
				    Exit(1);
				}
			}
		}
	    else
		{
		    (void) fprintf (_err,
				    "Error: $IF - bad argument \"%s\"\n", s);
		    return FALSE;
		}
	}
    return truth;
}

#define If_states int
#define IF_EVAL 0
#define IF_ENTER 1
#define IF_EXECUTE 2
#define DENEST 3
#define SKIP_TILL_FI 4
#define SKIP_TILL_ELSE_FI 5
#define ELSE_ENTER 6
#define ELSE_EXECUTE 7
#define IF_FI_DONE 8

/*
 * handle_if :
 *   Handles complex IF ELSE FI expressions.
 *   It's basically a state machine to handle the correct
 *   parsing of the IF ELSE construct.
 */
/***************************************************************************
*
* Function:  handle_if
* Purpose:
* Arguments: mhead - 
*            str - 
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, handle_if,
    Macro *, mhead,
    char *, str)
{
    /* at this point, s should point to the first "$IF..." */    
    char * s = str;
    char * command;
    int nest = 0, loc_nest = 0;
    If_states state = IF_EVAL;
    
    for (;;)
	{
	    switch (state)
		{
		case IF_EVAL:
		    if (evaluate_if(expandMacro (mhead, s)) == TRUE)
			state = IF_ENTER;
		    else
			state = SKIP_TILL_ELSE_FI;
		    break;
		case IF_ENTER:
		    nest++;     /* IF nesting level */
		    str = s = ReadLine();
		    switch (check_token_if(s))
			{
			case TI_DRIF:
			    state = IF_EVAL;
			    break;
			case TI_DRFI:
			    state = DENEST;
			    break;
			case TI_DRELSE:
			    state = SKIP_TILL_FI;
			    break;
			case TI_SEMICOLON:
			    (void) fprintf (_err, "Error: unexpected ';'\n");
			    Exit(1);
			    break;
			case TI_OTHER:
			default:
			    state = IF_EXECUTE;
			    break;
			}
		    break;
		case IF_EXECUTE:
		    command = expandMacro (mhead, s);
		    system_command (command);
		    str = s = ReadLine();
		    switch (check_token_if(s))
			{
			case TI_DRIF:
			    state = IF_EVAL;
			    break;
			case TI_DRFI:
			    state = DENEST;
			    break;
			case TI_DRELSE:
			    state = SKIP_TILL_FI;
			    break;
			case TI_SEMICOLON:
			    (void) fprintf (_err, "Error: unexpected ';'\n");
			    Exit(1);
			    break;
			case TI_OTHER:
			default:
			    state = IF_EXECUTE;
			    break;
			}
		    break;
		case DENEST:
		    nest--;
		    if (nest <= 0)
			{
			    nest = 0;
			    state = IF_FI_DONE; break;
			}
		    else
			{
			    str = s = ReadLine();
			    switch (check_token_if(s))
				{
				case TI_DRIF:
				    state = IF_EVAL;
				    break;
				case TI_DRFI:
				    state = DENEST;
				    break;
				case TI_DRELSE:
				    state = SKIP_TILL_FI;
				    break;
				case TI_SEMICOLON:
				    (void) fprintf (_err,
						    "Error: unexpected ';'\n");
				    Exit(1);
				    break;
				case TI_OTHER:
				default:
				    state = IF_EXECUTE;
				    break;
				}
			}
		    break;
		case SKIP_TILL_FI:
		    loc_nest = 0;
		    while (state == SKIP_TILL_FI)
			{
			    str = s = ReadLine();
			    switch (check_token_if(s))
				{
				case TI_DRIF:
				    loc_nest++;
				    break;
				case TI_DRFI:
				    if (loc_nest == 0)
					state = DENEST;
				    /* found proper FI */
				    else
					loc_nest--;
				    break;
				case TI_DRELSE:
				case TI_OTHER:
				default:
				    /* skip */
				    break;
				case TI_SEMICOLON:
				    (void) fprintf (_err, "Error: unexpected ';'\n");
				    Exit(1);
				    break;
				}
			}
		    break;
		case SKIP_TILL_ELSE_FI:
		    loc_nest = 0;
		    while (state == SKIP_TILL_ELSE_FI)
			{
			    str = s = ReadLine();
			    switch (check_token_if(s))
				{
				case TI_DRIF:
				    loc_nest++;
				    break;
				case TI_DRFI:
				    if (loc_nest == 0)
					state = DENEST; /* found proper FI */
				    else
					loc_nest--;
				    break;
				case TI_DRELSE:
				    if (loc_nest == 0) 
					state = ELSE_ENTER;
				    break;
				case TI_OTHER:
				default:
				    /* skip */
				    break;
				case TI_SEMICOLON:
				    (void) fprintf (_err,
						    "Error: unexpected ';'\n");
				    Exit(1);
				    break;
				}
			    ;
			}
		    break;
		case ELSE_ENTER:
		    s = eat_space (s);
		    s = eat_space(s + strlen ("$ELSE"));
		    if (*s != '\0')
			(void) fprintf (_err, 
					"Warning - text after $ELSE ignored - %s\n",
					str);
		    nest++;
		    str = s = ReadLine();
		    switch (check_token_if(s))
			{
			case TI_DRIF:
			    state = IF_EVAL;
			    break;
			case TI_DRFI:
			    state = DENEST;
			    break;
			case TI_DRELSE:
			    (void) fprintf (_err,
					    "Error : unexpected $ELSE\n");
			    Exit(1);
			    break;
			case TI_SEMICOLON:
			    (void) fprintf (_err,
					    "Error: unexpected ';'\n");
			    Exit(1);
			    break;
			case TI_OTHER:
			default:
			    state = ELSE_EXECUTE;
			    break;
			}
		    break;
		case ELSE_EXECUTE:
		    command = expandMacro (mhead, s);
		    system_command (command);
		    str = s = ReadLine();
		    switch (check_token_if(s))
			{
			case TI_DRIF:
			    state = IF_EVAL;
			    break;
			case TI_DRFI:
			    state = DENEST;
			    break;
			case TI_DRELSE:
			    (void) fprintf (_err,
					    "Error : unexpected $ELSE\n");
			    Exit(1);
			    break;
			case TI_SEMICOLON:
			    (void) fprintf (_err, "Error: unexpected ';'\n");
			    Exit(1);
			    break;
			case TI_OTHER:
			default:
			    state = ELSE_EXECUTE;
			    break;
			}
		    break;
		case IF_FI_DONE:
#if 0
		    s = eat_space (s);
		    s = eat_space(s + strlen ("$FI"));
		    if (*s != '\0')
			(void) fprintf (_err, 
					"Warning - text after $FI ignored - %s\n", 
					str);
#endif
		    return;
		}
	}
}

/*
 * handle_command - used to execute commands, and
 *      to take care of the nested $IF, $ELSE, $FI construct.
 */
/***************************************************************************
*
* Function:  handle_command
* Purpose:
* Arguments: mhead - 
*            s - 
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, handle_command,
    Macro *, mhead,
    char *, s)
{
    char *command = NULL;

    s = eat_space (s);
    switch (check_token_if(s))
	{
	case TI_DRIF:
	    handle_if (mhead, s);
	    break;
	case TI_DRELSE:
	    (void) fprintf (_err, "Error: Unexpected $ELSE found\n");
	    Exit(1);
	    break;
	case TI_DRFI:
	    (void) fprintf (_err, "Error: Unexpected $FI found\n");
	    Exit(1);
	    break;
	case TI_SEMICOLON:
	    (void) fprintf (_err, "Error: Unexpected $FI found\n");
	    Exit(1);
	    break;
	case TI_OTHER:
	default:
	    command = expandMacro (mhead, s);
	    system_command (command);
	    break;
	}
}

/*
 * handle_target:
 *   attempts to match any possible target with
 *   the current target indicated by the prodfile 
 *   input stream.  If a potential user target is
 *   found that matches the current target in the
 *   prodfile, the correct prodfile actions are executed
 *   for the target.  
 *
 *   If no potential user target matches, the pointer in
 *   the prodfile is moved ahead to the ';' that delimits
 *   the end of a target.
 */
/***************************************************************************
*
* Function:  handle_target
* Purpose:
* Arguments: thead - 
*            mhead - 
*            s - 
* Returns:
* See also:
*
****************************************************************************
*/
FE3(void, handle_target,
    Macro *, thead,
    Macro *, mhead,
    char *, s)
{
    char *cp = NULL, *q, *r;
    Macro * t;
    unsigned long rewind_pos = 0L;
    BOOL matched = FALSE;

    cp = strchr (s, (int) ':');
    if ((*s != '$') || (*(s+1) != '*') || (*(s+2) != '.') || (cp == NULL)
	|| (thead == NULL) )
	{
	    while ((s = ReadLine()))
		if (*s == ';') return;  /* bulletproofing */
	}
    else
	{
	    (void) strcpy(dupBuf, s + 3);
	    cp = dupBuf + (unsigned) (cp - (s + 3));
	    *cp = '\0';
	    cp = eat_space(cp + 1);
	    /*
	     *  Rewinding:  Save the current position in the prodfile
	     *  in case it should be exec'd for different physical
	     *  targets with same postfix.      (e.g. targets 1.gph 2.gph etc)
	     */
	    rewind_pos = posOfStream;
	    for (t = thead; t != NULL; t = t -> next)
		{
		    cuTaBaseName = NULL;
		    if (*(t->name))
			{
			    for (q = t->name; *q; q++)
				;
			    if (!*dupBuf)
				{
				    q--;
				    if (*q == '.')
					{
					    *q = '\0';
					    cuTaBaseName =
						stringAlloc(t->name);
					    *q = '.';
					}
				}
			    else
				{
				    for (r = dupBuf; *r; r++)
					;
				    do
					{
					    q--;
					    r--;
					} while ((*q == *r) &&
						 (q != t->name) &&
						 (r != dupBuf));
				    if ((*q == *r) && (q != t->name) &&
					(*(q - 1) == '.'))
					{
					    q--;
					    *q = '\0';
					    cuTaBaseName =
						stringAlloc(t->name);
					    *q = '.';
					}
				}
			}
		    if (cuTaBaseName)
			{
			    /* match */
			    t->matched = TRUE;
			    cuTaSuffix = dupBuf;
			    if (matched)
				{
				    /* Reposition for multiple targets */
				    endOfStream = 0;
				    posOfStream = rewind_pos;
				    if (((unsigned long) -1L) ==
					MyFskTell(prodStream, rewind_pos,
						  (int) SEEK_SET))
					Troba0();
				}
			    else
				matched = TRUE;
			    handle_command (mhead, cp);
			    while ((s = ReadLine()))
				{
				    s = eat_space (s);
				    if (*s == ';') 
					break;  /* bulletproofing */
				    else
					{
					    handle_command (mhead, s);
					}
				}
			    Deallocate((char *) cuTaBaseName);
			}
		}
	    /* If no potential targets matched, skip ahead to ';' */
	    if (!matched)
		{
		    while ((s = ReadLine()) )
			if (*s == ';') return;  /* bulletproofing */
		}
	}
}
