/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prpexpr3.c
*
* Part of: PROD's program prpp
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See prpexpr.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "output.h"
#include "place.h"
#include "prpp.h"
#include "lexer.h"
#include "macro.h"
#include "parse.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "net.h"
#include "prec.h"
#include "pplst.h"
#include "prpexpr.h"
#include "prpshrt.h"

/***************************************************************************
*
* Function:  ForRUntil
* Purpose:   This routine reduces an "until" formula. The reduction is
*            based on the rules given in the CAV '93 paper of Peled.
* Arguments: formula - the original formula
* Returns:   the result of the reduction
* See also:  ForRImplies, ForRExpr
*
****************************************************************************
*/
FS1(FORMULA *, ForRUntil,
    FORMULA *, formula)
{
    FORMULA *fp1;
    FORMULA *fp2;
    FORMULA *fp3;
    FORMULA *fp4;
    
    if (FSNOC_I(formula))
	Troba2("ForRUntil: formula 0x%lx", FSNOC_U(formula));
    if (errorFlag != NO_ERROR)
	{
	    ForDracsid(formula);
	    return(NULL);
	}
 START: ;
    if (!(formula->uf.of.sub2f))
	{
	    ForDracsid(formula);
	    return(NULL);
	}
    if ((!(formula->uf.of.sub1f)) ||
	((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	 !(formula->uf.of.sub2f->uf.of.sub1f)) ||
	((formula->uf.of.sub2f->operator == FO_UNTIL) &&
	 (formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	 (!(formula->uf.of.sub1f->uf.of.sub1f)) && formula->uf.of.sub2f->uf.of.sub1f &&
	 (formula->uf.of.sub2f->uf.of.sub1f->operator == FO_IMPLIES) &&
	 !(formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f)))
	{
	    fp1 = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    return(fp1);
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub1f->uf.of.sub2f)) &&
	formula->uf.of.sub1f->uf.of.sub1f &&
	(formula->uf.of.sub1f->uf.of.sub1f->operator == FO_IMPLIES))
	{
	    fp1 = formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub1f = NULL;
	    fp2 = formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub2f = NULL;
	    fp3 = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    fp2 = ForXua(FO_IMPLIES, ForXua(FO_UNTIL,
					    ForXua(FO_IMPLIES, fp2,
						   (FORMULA *) NULL),
					    ForYpoc(fp3)),
			 (FORMULA *) NULL);
	    return(ForXua(FO_IMPLIES,
			  ForXua(FO_IMPLIES,
				 ForXua(FO_UNTIL, fp1, fp3), fp2),
			  (FORMULA *) NULL));
	}
    if ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	formula->uf.of.sub2f->uf.of.sub2f)
	{
	    fp1 = formula->uf.of.sub1f;
	    formula->uf.of.sub1f = NULL;
	    fp2 = formula->uf.of.sub2f->uf.of.sub1f;
	    formula->uf.of.sub2f->uf.of.sub1f = NULL;
	    fp3 = formula->uf.of.sub2f->uf.of.sub2f;
	    formula->uf.of.sub2f->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    fp3 = ForXua(FO_UNTIL, ForYpoc(fp1), fp3);
	    return(ForXua(FO_IMPLIES,
			  ForXua(FO_IMPLIES,
				 ForXua(FO_UNTIL, fp1,
					ForXua(FO_IMPLIES, fp2,
					       (FORMULA *) NULL)),
				 (FORMULA *) NULL),
			  fp3));
	}
    if ((formula->uf.of.sub1f->operator != FO_IMPLIES) ||
	formula->uf.of.sub1f->uf.of.sub1f ||
	(formula->uf.of.sub2f->operator != FO_IMPLIES) ||
	formula->uf.of.sub2f->uf.of.sub2f)
	goto CONT;
    fp3 = formula->uf.of.sub2f->uf.of.sub1f;
    if ((fp3->operator != FO_IMPLIES) || (!(fp3->uf.of.sub1f)) ||
	(!(fp3->uf.of.sub2f)) || (fp3->uf.of.sub1f->operator != FO_IMPLIES) ||
	fp3->uf.of.sub1f->uf.of.sub2f || (fp3->uf.of.sub2f->operator != FO_UNTIL) ||
	(!(fp3->uf.of.sub2f->uf.of.sub1f)) ||
	(fp3->uf.of.sub2f->uf.of.sub1f->operator != FO_IMPLIES) ||
	fp3->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f)
	goto CONT;
    fp2 = fp3->uf.of.sub2f->uf.of.sub2f;
    fp4 = fp3;
    fp3 = fp3->uf.of.sub1f->uf.of.sub1f;
    if ((fp3->operator != FO_UNTIL) || (!(fp3->uf.of.sub1f)) ||
	(fp3->uf.of.sub1f->operator != FO_IMPLIES) ||
	fp3->uf.of.sub1f->uf.of.sub1f)
	goto CONT;
    fp4->uf.of.sub2f->uf.of.sub2f = NULL;
    fp1 = fp3->uf.of.sub2f;
    fp3->uf.of.sub2f = NULL;
    ForDracsid(formula);
    return(ForXua(FO_IMPLIES,
		  ForXua(FO_IMPLIES,
			 ForXua(FO_UNTIL,
				ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
					(FORMULA *) NULL),
				ForXua(FO_IMPLIES,
				       ForXua(FO_UNTIL,
					      ForSnoc(FO_IMPLIES,
						      (FORMULA *) NULL,
						      (FORMULA *) NULL),
					      fp1),
				       (FORMULA *) NULL)),
			 ForXua(FO_IMPLIES,
				ForXua(FO_UNTIL,
				       ForSnoc(FO_IMPLIES,
					       (FORMULA *) NULL,
					       (FORMULA *) NULL),
				       ForXua(FO_IMPLIES,
					      ForXua(FO_UNTIL,
						     ForSnoc(FO_IMPLIES,
							     (FORMULA *) NULL,
							     (FORMULA *) NULL),
						     fp2),
					      (FORMULA *) NULL)),
				(FORMULA *) NULL)),
		  (FORMULA *) NULL));
 CONT: ;
    if (ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f) ||
	((formula->uf.of.sub1f->operator == FO_UNTIL) &&
	 ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub2f)))
	{
	    fp1 = formula->uf.of.sub1f;
	    formula->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    return(fp1);
	}
    if ((formula->uf.of.sub2f->operator == FO_UNTIL) &&
	 ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f))
	{
	    fp1 = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    return(fp1);
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub1f->uf.of.sub2f)) &&
	 ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub1f))
	{
	    ForDracsid(formula->uf.of.sub1f->uf.of.sub1f);
	    formula->uf.of.sub1f->uf.of.sub1f = NULL;
	    goto START;
	}
    if ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub2f->uf.of.sub2f)) &&
	 ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f))
	{
	    ForDracsid(formula->uf.of.sub1f);
	    formula->uf.of.sub1f = ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
					(FORMULA *) NULL);
	    goto START;
	}
    if ((formula->uf.of.sub2f->operator != FO_IMPLIES) ||
	formula->uf.of.sub2f->uf.of.sub2f || (!(formula->uf.of.sub2f->uf.of.sub1f)) ||
	(formula->uf.of.sub2f->uf.of.sub1f->operator != FO_IMPLIES))
	return(formula);
    if (ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f))
	{
	    fp1 = formula->uf.of.sub1f;
	    formula->uf.of.sub1f = NULL;
	    fp2 = formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    fp1 = ForXua(FO_UNTIL, ForYpoc(fp2),
			 ForXua(FO_IMPLIES, fp1, (FORMULA *) NULL));
	    return(ForXua(FO_IMPLIES,
			  ForXua(FO_IMPLIES,
				 ForXua(FO_UNTIL,
					ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
						(FORMULA *) NULL),
					ForXua(FO_IMPLIES, fp2,
					       (FORMULA *) NULL)),
				 fp1),
			  (FORMULA *) NULL));
	}
    if (formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f &&
	(formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f->uf.of.sub2f)) &&
	ForIdentic(formula->uf.of.sub1f,
		   formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f->uf.of.sub1f))
	{
	    fp1 = formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f = NULL;
	    fp2 = formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    fp1 = ForXua(FO_UNTIL,
			 ForXua(FO_IMPLIES, ForYpoc(fp2), (FORMULA *) NULL),
			 fp1);
	    return(ForXua(FO_IMPLIES,
			  ForXua(FO_IMPLIES,
				 ForXua(FO_UNTIL,
					ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
						(FORMULA *) NULL),
					fp2),
				 fp1),
			  (FORMULA *) NULL));
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub1f->uf.of.sub2f)) &&
	ForIdentic(formula->uf.of.sub1f->uf.of.sub1f,
		   formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f))
	{
	    fp1 = formula->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub1f->uf.of.sub1f = NULL;
	    fp2 = formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    fp1 = ForXua(FO_UNTIL,
			 ForXua(FO_IMPLIES, ForYpoc(fp2), (FORMULA *) NULL),
			 fp1);
	    return(ForXua(FO_IMPLIES,
			  ForXua(FO_IMPLIES,
				 ForXua(FO_UNTIL,
					ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
						(FORMULA *) NULL),
					fp2),
				 fp1),
			  (FORMULA *) NULL));
	}
    return(formula);
}

/***************************************************************************
*
* Function:  ForRImplies
* Purpose:   This routine reduces an implication formula.
* Arguments: formula - the original formula
* Returns:   the result of the reduction
* See also:  ForRUntil, ForRExpr
*
****************************************************************************
*/
FS1(FORMULA *, ForRImplies,
    FORMULA *, formula)
{
    FORMULA *fp;

    if (FSNOC_I(formula))
	Troba2("ForRImplies: formula 0x%lx", FSNOC_U(formula));
    if (errorFlag != NO_ERROR)
	{
	    ForDracsid(formula);
	    return(NULL);
	}
 START: ;
    if (!(formula->uf.of.sub1f))
	{
	    if (formula->uf.of.sub2f)
		{
		    ForDracsid(formula->uf.of.sub2f);
		    formula->uf.of.sub2f = NULL;
		}
	    return(formula);
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	!(formula->uf.of.sub1f->uf.of.sub1f))
	{
	    fp = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    return(fp);
	}
    if (!(formula->uf.of.sub2f))
	{
	    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
		!(formula->uf.of.sub1f->uf.of.sub2f))
		{
		    fp = formula->uf.of.sub1f->uf.of.sub1f;
		    formula->uf.of.sub1f->uf.of.sub1f = NULL;
		    ForDracsid(formula);
		    return(fp);
		}
	    return(formula);
	}
    if ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	!(formula->uf.of.sub2f->uf.of.sub1f))
	{
	    ForDracsid(formula->uf.of.sub1f);
	    formula->uf.of.sub1f = NULL;
	    ForDracsid(formula->uf.of.sub2f);
	    formula->uf.of.sub2f = NULL;
	    return(formula);
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub1f->uf.of.sub2f)) &&
	!(formula->uf.of.sub2f->uf.of.sub2f))
	{
	    fp = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    fp->uf.of.sub2f = formula->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub1f->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    formula = fp;
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	formula->uf.of.sub2f->uf.of.sub2f &&
	(formula->uf.of.sub2f->uf.of.sub2f->operator == FO_IMPLIES) &&
	(!(formula->uf.of.sub1f->uf.of.sub2f)) &&
	!(formula->uf.of.sub2f->uf.of.sub2f->uf.of.sub2f))
	{
	    fp = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    fp->uf.of.sub2f->uf.of.sub2f = formula->uf.of.sub1f->uf.of.sub1f;
	    formula->uf.of.sub1f->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    formula = fp;
	    formula->uf.of.sub2f = ForRImplies(formula->uf.of.sub2f);
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	!(formula->uf.of.sub1f->uf.of.sub2f))
	{
	    fp = formula->uf.of.sub2f;
	    while (fp)
		{
		    if (ForIdentic(fp, formula->uf.of.sub1f))
			{
			    ForDracsid(formula->uf.of.sub1f);
			    formula->uf.of.sub1f = NULL;
			    ForDracsid(formula->uf.of.sub2f);
			    formula->uf.of.sub2f = NULL;
			    return(formula);
			}
		    if (fp->operator != FO_IMPLIES)
			break;
		    if (ForIdentic(fp->uf.of.sub1f, formula->uf.of.sub1f->uf.of.sub1f))
			{
			    ForDracsid(formula->uf.of.sub1f);
			    formula->uf.of.sub1f = NULL;
			    ForDracsid(formula->uf.of.sub2f);
			    formula->uf.of.sub2f = NULL;
			    return(formula);
			}
		    fp = fp->uf.of.sub2f;
		}
	}
    else
	{
	    fp = formula->uf.of.sub2f;
	    while (fp)
		{
		    if (ForIdentic(fp, formula->uf.of.sub1f))
			{
			    ForDracsid(formula->uf.of.sub1f);
			    formula->uf.of.sub1f = NULL;
			    ForDracsid(formula->uf.of.sub2f);
			    formula->uf.of.sub2f = NULL;
			    return(formula);
			}
		    if (fp->operator != FO_IMPLIES)
			break;
		    if (fp->uf.of.sub1f &&
			(fp->uf.of.sub1f->operator == FO_IMPLIES) &&
			(!(fp->uf.of.sub1f->uf.of.sub2f)) &&
			ForIdentic(fp->uf.of.sub1f->uf.of.sub1f, formula->uf.of.sub1f))
			{
			    ForDracsid(formula->uf.of.sub1f);
			    formula->uf.of.sub1f = NULL;
			    ForDracsid(formula->uf.of.sub2f);
			    formula->uf.of.sub2f = NULL;
			    return(formula);
			}
		    fp = fp->uf.of.sub2f;
		}
	}
    if (((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	 (ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f) ||
	  (formula->uf.of.sub2f->uf.of.sub2f &&
	   (formula->uf.of.sub2f->uf.of.sub2f->operator == FO_IMPLIES) &&
	   ForIdentic(formula->uf.of.sub1f,
		      formula->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f)) ||
	  ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	   (ForIdentic(formula->uf.of.sub1f->uf.of.sub1f,
		       formula->uf.of.sub2f->uf.of.sub2f) ||
	    ForIdentic(formula->uf.of.sub1f->uf.of.sub2f,
		       formula->uf.of.sub2f->uf.of.sub1f))))) ||
	((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	 (ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub1f) ||
	  (formula->uf.of.sub1f->uf.of.sub2f &&
	   (formula->uf.of.sub1f->uf.of.sub2f->operator == FO_IMPLIES) &&
	   ForIdentic(formula->uf.of.sub2f,
		      formula->uf.of.sub1f->uf.of.sub2f->uf.of.sub1f)))))
	{
	    fp = formula->uf.of.sub2f;
	    formula->uf.of.sub2f = NULL;
	    ForDracsid(formula);
	    return(fp);
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub2f) ||
	 ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	  ForIdentic(formula->uf.of.sub1f->uf.of.sub2f,
		     formula->uf.of.sub2f->uf.of.sub2f))))
	{
	    ForDracsid(formula->uf.of.sub1f->uf.of.sub2f);
	    formula->uf.of.sub1f->uf.of.sub2f = NULL;
	    formula->uf.of.sub1f = ForRImplies(formula->uf.of.sub1f);
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	formula->uf.of.sub1f->uf.of.sub1f &&
	(formula->uf.of.sub1f->uf.of.sub1f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub2f))
	{
	    ForDracsid(formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub2f);
	    formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub2f = NULL;
	    formula->uf.of.sub1f->uf.of.sub1f =
		ForRImplies(formula->uf.of.sub1f->uf.of.sub1f);
	    formula->uf.of.sub1f = ForRImplies(formula->uf.of.sub1f);
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	formula->uf.of.sub1f->uf.of.sub2f &&
	(formula->uf.of.sub1f->uf.of.sub2f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub2f->uf.of.sub2f))
	{
	    ForDracsid(formula->uf.of.sub1f->uf.of.sub2f->uf.of.sub2f);
	    formula->uf.of.sub1f->uf.of.sub2f->uf.of.sub2f = NULL;
	    formula->uf.of.sub1f->uf.of.sub2f =
		ForRImplies(formula->uf.of.sub1f->uf.of.sub2f);
	    formula->uf.of.sub1f = ForRImplies(formula->uf.of.sub1f);
	    goto START;
	}
    if ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	formula->uf.of.sub2f->uf.of.sub1f &&
	(formula->uf.of.sub2f->uf.of.sub1f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub1f))
	{
	    fp = formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f = NULL;
	    ForDracsid(formula->uf.of.sub2f->uf.of.sub1f);
	    formula->uf.of.sub2f->uf.of.sub1f = fp;
	    formula->uf.of.sub2f = ForRImplies(formula->uf.of.sub2f);
	    goto START;
	}
    if ((formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	formula->uf.of.sub2f->uf.of.sub1f &&
	(formula->uf.of.sub2f->uf.of.sub1f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f->uf.of.sub2f))
	{
	    fp = formula->uf.of.sub2f->uf.of.sub2f;
	    formula->uf.of.sub2f->uf.of.sub2f = NULL;
	    ForDracsid(formula->uf.of.sub2f);
	    formula->uf.of.sub2f = fp;
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	(formula->uf.of.sub2f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub1f->uf.of.sub1f, formula->uf.of.sub2f->uf.of.sub1f))
	{
	    fp = formula->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub1f->uf.of.sub2f = NULL;
	    ForDracsid(formula->uf.of.sub1f);
	    formula->uf.of.sub1f = fp;
	    goto START;
	}
    if ((formula->uf.of.sub1f->operator == FO_IMPLIES) &&
	formula->uf.of.sub1f->uf.of.sub1f &&
	(formula->uf.of.sub1f->uf.of.sub1f->operator == FO_IMPLIES) &&
	ForIdentic(formula->uf.of.sub2f, formula->uf.of.sub1f->uf.of.sub1f->uf.of.sub1f))
	{
	    fp = formula->uf.of.sub1f->uf.of.sub2f;
	    formula->uf.of.sub1f->uf.of.sub2f = NULL;
	    ForDracsid(formula->uf.of.sub1f);
	    formula->uf.of.sub1f = fp;
	    goto START;
	}
    return(formula);
}

/***************************************************************************
*
* Function:  ForRExpr
* Purpose:   This routine reduces an expression formula.
* Arguments: formula - the original formula
* Returns:   the result of the reduction
* See also:  ForRUntil, ForRImplies
*
****************************************************************************
*/
FS1(FORMULA *, ForRExpr,
    FORMULA *, formula)
{
    unsigned i;

    if (FSNOC_I(formula))
	Troba2("ForRExpr: formula 0x%lx", FSNOC_U(formula));
    if (errorFlag != NO_ERROR)
	{
	    ForDracsid(formula);
	    return(NULL);
	}
    i = XprReduce(formula->uf.expression);
    if (errorFlag != NO_ERROR)
	{
	    ForDracsid(formula);
	    return(NULL);
	}
    if ((i != RE_CONSTANT) || !binaryMode)
	return(formula);
    i = (!XprEval(formula->uf.expression)) ? 0 : 1;
    ForDracsid(formula);
    return((i == 0) ? NULL : ForSnoc(FO_IMPLIES, (FORMULA *) NULL,
				     (FORMULA *) NULL));
}

FORMF forPoTable[SZ_FO] = { ForRUntil, ForRImplies, ForRExpr };
PP_TOKEN *prpExpNow;
char *prpExpPtr;

/***************************************************************************
*
* Function:  PrpCaret
* Purpose:   If the user has written a syntactically incorrect formula
*            or integer expression or marking expression, this routine
*            repeats the stuff and marks with with "^" a part where a
*            syntax error was found.
* Arguments: count - the position of the last "^"
*            textLen - the number of "^" needed
* Returns:
* See also:  PrtCaret (comline.c)
*
****************************************************************************
*/
FE2(void, PrpCaret,
    unsigned, count,
    unsigned, textLen)
{
    PP_TOKEN *token;
    VOID *cursor;

    cursor = LstGet(prpExpList);
    LSTFOR (prpExpList, token, PP_TOKEN)
	Q(token->token, _err);
    LstSet(prpExpList, cursor);
    (void) fputc((int) '\n', _err);
    while (count-- != textLen)
	(void) fputc((int) ' ', _err);
    while (textLen-- != 0) 
	(void) fputc((int) '^', _err);
    (void) fputc((int) '\n', _err);
    (void) fflush(_err);
}

/***************************************************************************
*
* Function:  PrpExpCh
* Purpose:   This routine substitutes spaces for newlines in a list
*            containing input data.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, PrpExpCh)
{
    PP_TOKEN *token;
    char *s;
    VOID *cursor;

    cursor = LstGet(prpExpList);
    LSTFOR (prpExpList, token, PP_TOKEN)
	for (s = token->token; *s != '\0'; s++)
	    if (*s == '\n')
		*s = ' ';
    LstSet(prpExpList, cursor);
}

/***************************************************************************
*
* Function:  PrpExpInput
* Purpose:   This routine reads input from a token list.
* Arguments:
* Returns:   the character being in turn
* See also:
*
****************************************************************************
*/
FE0(int, PrpExpInput)
{
    if (!prpExpNow)
	return((int) EOF);
    while (*prpExpPtr == '\0')
	{
	    prpExpNow = (PP_TOKEN *) LstStep(prpExpList);
	    if (!prpExpNow)
		return((int) '\n');
	    prpExpPtr = prpExpNow->token;
	}
    return((int) (*prpExpPtr++));
}

/***************************************************************************
*
* Function:  ReadVerify
* Purpose:   This routine reads a formula that the user wants to verify.
*            The actual parsing of the formula is postponed until
*            the whole net description has been read.
* Arguments:
* Returns:
* See also:  BuTra
*
****************************************************************************
*/
FE0(void, ReadVerify)
{
    PP_TOKEN *token;
    PLACE *place;
    TRANSITION *tran;
    
    token = (PP_TOKEN *) NULL;
    prpExpList = (LIST *) NULL;
    prpVerList = (LIST *) NULL;
    if (verify)
	{
	    error("#verify is not allowed more than once");
	    goto recover;
	}
    if (monName)
	{
	    error("#verify is possible\nonly when there is no tester");
	    goto recover;
	}
#if 0
    if (!LstEmpty(PrecLst))
	{
	    error("#verify is possible\nonly when there is no #prec");
	    goto recover;
	}
#endif
    LSTFOR (PlaceLst, place, PLACE)
	if (!strcmp(PlaGName(place), "t_e_s_t_e_r"))
	    {
		error("#verify is possible\nonly when there is no place t_e_s_t_e_r");
		goto recover;
	    }
    LSTFOR (TransLst, tran, TRANSITION)
	if ((!strcmp(TraGName(tran), "i_n_i_t_i_a_l_i_z_e")) ||
	    !strcmp(TraGName(tran), "d_u_m_m_y"))
	    {
		(void) sprintf(errStr,
			       "#verify is possible\nonly when there is no transition %s",
			       TraGName(tran));
		error(errStr);
		goto recover;
	    }
    prpExpList = LstCreate(&ppTokTools);
    prpVerList = prpExpList;
    (void) LstInsFirst(ppLsLs, (char *) prpExpList);
    token = AppPrs(PPSEMICOLON, FALSE); 
    prpExpList = (LIST *) LstDelFirst(ppLsLs, FALSE);
    if (prpExpList != prpVerList)
	Troba1("Internal error in ReadVerify");
    TOKEN_MATCH(token, PPSEMICOLON, "\";\" in #verify");
    if ((((LIST *) NULL) == prpExpList) || LstEmpty(prpExpList))
	{
	    error("Expected formula in #verify, got nothing");
	    goto recover;
	}
    PpTokDiscard(token);
    PrpExpCh();
    verify = TRUE;
    return;
 recover: ;
    if (((PP_TOKEN *) NULL) != token)
	PpTokDiscard(token);
    if (((LIST *) NULL) == prpExpList)
	{
	    prpExpList = LstCreate(&ppTokTools);
	    (void) LstInsFirst(ppLsLs, (char *) prpExpList);
	    token = AppPrs(PPSEMICOLON, FALSE); 
	    if (((LIST *) LstDelFirst(ppLsLs, FALSE)) != prpExpList)
		Troba1("Internal error in ReadVerify");
	    if (token)
		PpTokDiscard(token);
	}
    LstDestroy(prpExpList, TRUE);
    prpExpList = (LIST *) NULL;
}

/***************************************************************************
*
* Function:  PrpExpr
* Purpose:   This routine reads an integer or marking expression.
* Arguments: marking - if a marking is wanted, the value is put here
*            placeName - a name of a place if a marking is wanted,
*                        and NULL otherwise
*            placeFlag - possible extra information if a marking is wanted
*            list - a possible token list containing the expression
*            value - if an integer number is wanted, the value is put here
* Returns:
* See also:
*
****************************************************************************
*/
FE5(void, PrpExpr,
    MARKING **, marking,
    char *, placeName,
    unsigned, placeFlag,
    LIST **, list,
    DOMAIN *, value)
{
    PP_TOKEN *token;
    
    token = (PP_TOKEN *) NULL;
    errorFlag = NO_ERROR;
    if (marking)
	{
	    if ((((LIST **) NULL) != list) && (((LIST *) NULL) != *list))
		prpExpList = *list;
	    else
		{
		    prpExpList = LstCreate(&ppTokTools);
		    (void) LstInsFirst(ppLsLs, (char *) prpExpList);
		    token = AppPrs(PPCLOSING_PARENTHESIS, FALSE); 
		    if (((LIST *) LstDelFirst(ppLsLs, FALSE)) != prpExpList)
			Troba1("Internal error in PrpExpr");
		    TOKEN_MATCH(token, PPCLOSING_PARENTHESIS,
				((void)
				 sprintf(errStr,
					 "end \")\" of %s of %s %s%s%s",
					 (placeFlag & 1)
					 ? ((placeFlag & 2) ? "mk"
					    : ((placeFlag & 4) ? "lo" : "hi"))
					 : ((placeFlag & 2) ? "reject"
					    : ((placeFlag & 4) ? "deadlock"
					       : ((placeFlag & 8) ? "livelock"
						  : "infinite"))),
					 (placeFlag & 1) ? "place" : "tester",
					 (placeFlag & 1) ? "" : "\"",
					 placeName,
					 (placeFlag & 1) ? "" : "\""),
				 errStr));
		    PpTokDiscard(token);
		}
	    *marking = NULL;
	}
    else if ((((LIST **) NULL) != list) && (((LIST *) NULL) != *list)
	     && !LstEmpty(*list))
	prpExpList = *list;
    else
	{
	    error("Expected integer valued expression, got nothing");
	    errorFlag = XPR_ERROR;
	    if ((((LIST **) NULL) != list) && (((LIST *) NULL) != *list))
		{
		    LstDestroy(*list, TRUE);
		    *list = (LIST *) NULL;
		}
	    return;
	}
    PrpExpCh();
    prpExpNow = (PP_TOKEN *) LstBegin(prpExpList);
    if (prpExpNow)
	{
	    prpExpPtr = prpExpNow->token;
	    binaryMode = FALSE;
	    if (marking)
		LtlParse((FORMULA **) NULL, (DOMAIN *) NULL, marking);
	    else
		LtlParse((FORMULA **) NULL, value, (MARKING **) NULL);
	}
    else
	*marking = MarCreate();
    if (((LIST *) NULL) != prpExpList)
	{
	    LstDestroy(prpExpList, TRUE);
	    prpExpList = (LIST *) NULL;
	}
    return;
 recover: ;
    if (((PP_TOKEN *) NULL) != token)
	PpTokDiscard(token);
    if (((LIST *) NULL) != prpExpList)
	{
	    LstDestroy(prpExpList, TRUE);
	    prpExpList = (LIST *) NULL;
	}
    errorFlag = CML_ERROR;
}

FE2(void, BuchiExt,
    LIST *, buchi,
    LIST *, atomList)
{
    FORMULA *atom = NULL;
    FILE *file = NULL;
    BCHST *bchst = NULL;
    BCHST *tshcb = NULL;
    unsigned long mask;
    unsigned i;
    unsigned k;
    BOOL found = FALSE;

    LSTFOR (atomList, atom, FORMULA)
	{
	    if (FO_EXPR != atom->operator)
		Troba1("Internal error in BuchiExt");
	    if ((EO_MNEQUAL != atom->uf.expression->operator) ||
		(MO_PLACE != atom->uf.expression->ue.me.msub1->operator) ?
		((MO_PLACE != atom->uf.expression->ue.me.msub2->operator) ||
		 (MO_CONSTANT != atom->uf.expression->ue.me.msub1->operator) ||
		 !MarEmpty(atom->uf.expression->ue.me.msub1->um.marking)) :
		((MO_CONSTANT != atom->uf.expression->ue.me.msub2->operator) ||
		 !MarEmpty(atom->uf.expression->ue.me.msub2->um.marking)))
		{
		    error("Some atomic formula prevents the program\nfrom printing a Buchi automaton into a text file.");
		    return;
		}
	}
    if (!(file = MyFopen(extBuchi, "w", (BOOL *) NULL)))
	return;
    Q("never {\nnonacc_init:\n  if\n", file);
    for (k = 0; k < 3; k++)
	{
	    LSTFOR (buchi, bchst, BCHST)
		{
		    if (BchstInitial(bchst) ? (2 == k) : (2 != k))
			continue;
		    if (0 == k)
			Q("  /* like ", file);
		    if (BchstAccepting(bchst))
			Q("accept", file);
		    else
			Q("nonacc", file);
		    P(file, "_%lx_", (unsigned long) bchst);
		    if (BchstInitial(bchst))
			Q("start", file);
		    else
			Q("conti", file);
		    if (0 == k)
			Q(" */\n", file);
		    else
			Q(":\n  if\n", file);
		    LSTFOR (BchstGSucc(bchst), tshcb, BCHST)
			{
			    Q("  :: (", file);
			    for (atom = (FORMULA *) LstBegin(atomList),
				 mask = 2L, found = FALSE; atom;
				 atom = (FORMULA *) LstStep(atomList),
				 mask <<= 2)
				{
				    if ((BchstGVal(bchst) & mask) == 0L)
					continue;
				    if (found)
					Q(" && ", file);
				    found = TRUE;
				    if ((BchstGVal(bchst) & (mask >> 1))
					== 0L)
					Q("(!", file);
				    else
					Q("  ", file);
				    if (atom->uf.expression->ue.me.msub1->operator
					== MO_PLACE)
					i = atom->uf.expression->ue.me.msub1->um.place;
				    else
					i = atom->uf.expression->ue.me.msub2->um.place;
				    P(file, "(%s)",
				      PlaGName((PLACE *)
					       LstNth(PlaceLst, i + 1)));
				    if ((BchstGVal(bchst) & (mask >> 1))
					== 0L)
					R(')', file);
				    else
					R(' ', file);
				}
			    if (!found)
				R('1', file);
			    Q(") -> goto ", file);
			    if (BchstAccepting(tshcb))
				Q("accept", file);
			    else
				Q("nonacc", file);
			    P(file, "_%lx_", (unsigned long) tshcb);
			    if (BchstInitial(tshcb))
				Q("start", file);
			    else
				Q("conti", file);
			    R('\n', file);
			}
		    if (0 != k)
			Q("  fi;\n", file);
		}
	    if (0 == k)
		Q("  fi;\n", file);
	}
    Q("}\n", file);
    MyFclose(file);
}

static LIST *buBuchi;

/***************************************************************************
*
* Function:  BuTra
* Purpose:   This routine "adds" to a high level net a Buchi automaton
*            corresponding to the negation of user's formula.
* Arguments:
* Returns:   the automaton
* See also:  ForBuchi
*
****************************************************************************
*/
FE0(LIST *, BuTra)
{
    FORMULA *formula = NULL;
    FORMULA *fp;
    char *str;
    unsigned long flag;
    unsigned long mask;
    unsigned long atMask;
    unsigned i;
    LIST *forList;
    LIST *atomList;
    TUPLE *tuple;
    MON *mon;
    TRANSITION *tran;
    VARIABLE *var;
    BCHST *bchst0;

    prpExpList = prpVerList;
    prpExpNow = (PP_TOKEN *) LstBegin(prpExpList);
    prpExpPtr = prpExpNow->token;
    errorFlag = NO_ERROR;
    binaryMode = TRUE;
    LtlParse(&formula, (DOMAIN *) NULL, (MARKING **) NULL);
    if (((LIST *) NULL) != prpExpList)
	{
	    LstDestroy(prpExpList, TRUE);
	    prpExpList = (LIST *) NULL;
	}
    prpVerList = (LIST *) NULL;
    if (errorFlag != NO_ERROR)
	{
	    Exit(1);
	}
    if (!formula)
	{
	    Q("\nThe formula to be verified is equivalent to 0 and is thus false\n", _out);
	    (void) fflush(_out);
	    Exit(1);
	}
    if ((formula->operator == FO_IMPLIES) && !(formula->uf.of.sub1f))
	{
	    Q("\nThe formula to be verified is equivalent to 1 and is thus true\n", _out);
	    (void) fflush(_out);
	    Exit(1);
	}
    if ((formula->operator == FO_IMPLIES) && !(formula->uf.of.sub2f))
	{
	    fp = formula->uf.of.sub1f;
	    formula->uf.of.sub1f = NULL;
	    ForDracsid(formula);
	    formula = fp;
	}
    else
	formula = ForSnoc(FO_IMPLIES, formula, (FORMULA *) NULL);
    atomList = NULL;
    forList = LstCreate(&forSloot);
    ForClosure(formula, forList, atomList);
    atomList = LstCopy(forList, TRUE);
    LstDestroy(forList, FALSE);
    forList = (LIST *) NULL;
    if (LstEmpty(atomList))
	Troba1("Internal error in BuTra");
    atMask = ((unsigned long) 1L) << (2 * (LstSize(atomList) - 1));
    buBuchi = ForBuchi(formula, atomList, atMask);
    ForDracsid(formula);
    if (LstEmpty(buBuchi))
	{
	    Q("\nThe formula to be verified is equivalent to 1 and is thus true\n", _out);
	    (void) fflush(_out);
	    Exit(1);
	}
    if (extBuchi)
	BuchiExt(buBuchi, atomList);
    flag = 0L;
    for (fp = (FORMULA *) LstBegin(atomList), mask = 1L; fp;
	 fp = (FORMULA *) LstStep(atomList), mask <<= 2)
	if (XprEval(fp->uf.expression))
	    flag |= mask;
    i = 0;
    LSTFOR (buBuchi, bchst0, BCHST)
	{
	    if (!BchstInitial(bchst0))
		continue;
	    for (mask = atMask;
		 mask && (((BchstGVal(bchst0) & mask) == (flag & mask)) ||
			   !(BchstGVal(bchst0) & (mask << 1))); mask >>= 2)
		;
	    if (mask)
		{
		    BchstCInitial(bchst0);
		    i = 1;
		}
	}
    if (i)
	BuchiSimplify(buBuchi);
#ifdef DEBUG
    if (i)
	{
	    Q("LAST SIMPLIFICATION\n", F);
	    LSTFOR (buBuchi, bchst0, BCHST)
		BchstDump(bchst0, (unsigned) 0);
	}
#endif
    if (LstEmpty(buBuchi))
	{
	    Q("\nThe formula to be verified is true\nas can be seen from the initial marking\n",
	      _out);
	    (void) fflush(_out);
	    Exit(1);
	}
    monName = StrCopy("t_e_s_t_e_r");
    monPlace = PlaCons(monName, LstCreate(&tupTools),
		       LstCreate(&tupTools), MarCreate());
    PlaSArity(monPlace, (unsigned) 1);
    tuple = TupCons((unsigned) 1);
    TupSField(tuple, (unsigned) 0,
              (DOMAIN) (BchstGVis((BCHST *) LstLast(buBuchi)) + 1L));
    MarInsLast(PlaGMarking(monPlace), MultupCons((unsigned long) 1L, tuple));
    (void) LstInsLast(PlaceLst, (char *) monPlace);
    LSTFOR (buBuchi, bchst0, BCHST)
	{
	    if (!BchstAccepting(bchst0))
		continue;
	    mon = MonCons((DOMAIN) BchstGVis(bchst0));
	    (void) BalTrInsert(monitors, (char *) mon,
			       (char *) MonAState(mon), TRUE);
	    MonSInfinite(mon);
	}
    (void) LstInsLast(TransLst, (char *)
		      TraCons(StrCopy("i_n_i_t_i_a_l_i_z_e")));
    TraSKind((TRANSITION *) LstLast(TransLst), TK_INITIAL);
    (void) LstInsLast(TransLst, (char *) TraCons(StrCopy("d_u_m_m_y")));
    TraSKind((TRANSITION *) LstLast(TransLst), TK_DUMMY);
    LSTFOR (TransLst, tran, TRANSITION)
	{
	    (void) LstInsLast(TraGInputs(tran), (char *)
			      ArcCons(monPlace, LstCreate(&sytTools)));
	    (void) LstInsLast(ArcGSymsum((ARC *)
					 LstLast(TraGInputs(tran))),
			      (char *) SytCons((unsigned) 1));
	    SytSMulti((SYMMULTUP *)
		      LstLast(ArcGSymsum((ARC *)
					 LstLast(TraGInputs(tran)))),
		      (unsigned long) 1L);
	    if (TraGKind(tran) == TK_INITIAL)
		{
		    str = (char *) Allocate((unsigned) 16);
		    (void) sprintf(str, DOMFORMAT,
                                   TupGField(tuple, (unsigned) 0));
		    SytRSymbol((SYMMULTUP *)
			       LstLast(ArcGSymsum((ARC *)
						  LstLast(TraGInputs(tran)))),
			       (unsigned) 0, SymCons(SK_CONSTANT, str));
		}
	    else
		{
		    var = VarCons((char *) NULL);
		    VarSIsIn(var);
		    (void) LstInsLast(TraGVariables(tran), (char *) var);
		    SytRSymbol((SYMMULTUP *)
			       LstLast(ArcGSymsum((ARC *)
						  LstLast(TraGInputs(tran)))),
			       (unsigned) 0, SymCons(SK_VARIABLE,
						     (char *) var));
		}
	    var = VarCons((char *) NULL);
	    VarSIsOut(var);
	    (void) LstInsLast(TraGVariables(tran), (char *) var);
	    (void) LstInsLast(TraGOutputs(tran), (char *)
			      ArcCons(monPlace, LstCreate(&sytTools)));
	    (void) LstInsLast(ArcGSymsum((ARC *)
					 LstLast(TraGOutputs(tran))),
			      (char *) SytCons((unsigned) 1));
	    SytSMulti((SYMMULTUP *)
		      LstLast(ArcGSymsum((ARC *)
					 LstLast(TraGOutputs(tran)))),
		      (unsigned long) 1L);
	    SytRSymbol((SYMMULTUP *)
		       LstLast(ArcGSymsum((ARC *)
					  LstLast(TraGOutputs(tran)))),
		       (unsigned) 0, SymCons(SK_VARIABLE, (char *) var));
	}
    return(atomList);
}

/***************************************************************************
*
* Function:  OutBuchi
* Purpose:   This routine writes to a binary file, datFile, some
*            information about a constructed Buchi automaton.
* Arguments: pos - indicates the position in the file 
* Returns:
* See also:  BuTra, pplst_dat (pplst.c)
*
****************************************************************************
*/
FE1(void, OutBuchi,
    unsigned long *, pos)
{
    unsigned sLen;
    DATATYPE *buffer;
    BCHST *bchst;
    BCHST *tshcb;
    unsigned j;
    register unsigned i, len;
    register UDOMAIN value;
    register DATATYPE *pointer;

    sLen = 1 + (unsigned) BchstGVis((BCHST *) LstLast(buBuchi));
    WriteUns(&sLen, 1, datFile);
    (*pos) += (unsigned long) sizeof(unsigned);
    LSTFOR (buBuchi, bchst, BCHST)
	{
	    len = 0;
	    i = LstSize(BchstGSucc(bchst));
	    IncLen(i, len);
	    LSTFOR (BchstGSucc(bchst), tshcb, BCHST)
		{
		    value = (UDOMAIN) BchstGVis(tshcb);
		    value = Encode((DOMAIN) value);
		    IncLen(value, len);
		}
	    sLen = len;
	    WriteUns(&sLen, 1, datFile);
	    (*pos) += (unsigned long) sizeof(unsigned);
	    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
	    pointer = buffer;
	    i = LstSize(BchstGSucc(bchst));
	    PutValue(i, pointer);
	    LSTFOR (BchstGSucc(bchst), tshcb, BCHST)
		{
		    value = (UDOMAIN) BchstGVis(tshcb);
		    value = Encode((DOMAIN) value);
		    PutValue(value, pointer);
		}
	    WriteData(buffer, sLen, datFile);
	    (*pos) += (unsigned long) (sLen * sizeof(DATATYPE));
	    Deallocate((char *) buffer);
	}
    len = 0;
    j = 0;
    LSTFOR (buBuchi, bchst, BCHST)
	if (BchstInitial(bchst))
	    j++;
    i = j;
    IncLen(i, len);
    LSTFOR (buBuchi, bchst, BCHST)
	{
	    if (!BchstInitial(bchst))
		continue;
	    value = (UDOMAIN) BchstGVis(bchst);
	    value = Encode((DOMAIN) value);
	    IncLen(value, len);
	}
    sLen = len;
    WriteUns(&sLen, 1, datFile);
    (*pos) += (unsigned long) sizeof(unsigned);
    buffer = (DATATYPE *) AllocArray(sLen, sizeof(DATATYPE));
    pointer = buffer;
    i = j;
    PutValue(i, pointer);
    LSTFOR (buBuchi, bchst, BCHST)
	{
	    if (!BchstInitial(bchst))
		continue;
	    value = (UDOMAIN) BchstGVis(bchst);
	    value = Encode((DOMAIN) value);
	    PutValue(value, pointer);
	}
    WriteData(buffer, sLen, datFile);
    (*pos) += (unsigned long) (sLen * sizeof(DATATYPE));
    Deallocate((char *) buffer);
    LSTFOR (buBuchi, bchst, BCHST)
	{
	    WriteLng(BchstAVal(bchst), 1, datFile);
	    (*pos) += (unsigned long) sizeof(unsigned long);
	}
    LstDestroy(buBuchi, TRUE);
    buBuchi = (LIST *) NULL;
}
