/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prpp.c
*
* Part of: PROD's program prpp
*
* Authors: Lasse Anderson, Robert Jurgens, Heikki Suonsivu,
*          and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The "prpp" program reads the net description and transforms the
*     net information so that a C compiler and a linker can produce
*     an executable reachability graph generator program. Some auxiliary
*     files needed by other programs of PROD are also produced.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "output.h"
#include "parse.h"
#include "macro.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "net.h"
#include "prec.h"
#include "pplst.h"
#include "cmdnm.h"
#include "prpexpr.h"
#include "prpp.h"

char *numStr = NULL;
char *srcPrefix = NULL;
char *cName = NULL;
FILE *prpOutput = NULL;
LIST *ppTokList = NULL;
unsigned long nOutpLines;
BOOL placeLimits;
BOOL preserve_comments;
BOOL ansi;
BOOL line_directives;
unsigned parsing_control;
BOOL may_have_lost_line;
BOOL arity_complaints;
#ifdef TRAIN
BOOL train;
#endif
BOOL prpMemCalled = FALSE;
UDOMAIN unfMaxDiff;
FILE *defsfile = NULL;
FILE *enmFile = NULL;
static FILE *preFile = NULL;
static int i;
static char *argument = NULL, *name = NULL, *preName = NULL, *body = NULL;
static char *fileName = NULL, *us = NULL;
static LIST *prpDefList = NULL, *prpKndList = NULL;
static unsigned long unfMaxL;

/***************************************************************************
*
* Function:  Misusage
* Purpose:   This routine handles most of the cases where the command line
*            of the program is incorrect.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, Misusage)
{
    MsgExit("[-A] [-a] [-C] [-Dname[=defstr]]... [-Iname]...\n[-L] [-N] [-R] [-t]name [-Uname]... [-uinteger]",
	    (int) 1);
}

FE0(void, PrpMemInit)
{
    _memClose = (VOF) PrpMemClose;
    if (((LIST *) NULL) == ppTokList)
	ppTokList = LstCreate(&ppTokTools);
    if (((LIST *) NULL) == prpDefList)
	prpDefList = LstCreate(&strTools);
    if (((LIST *) NULL) == prpKndList)
	prpKndList = LstCreate(&strTools);
    macro_init();
    ppio_init();
    pplst_init();
    if (((LIST *) NULL) == ppLsLs)
	ppLsLs = LstCreate(&lstTools);
    if (((char *) NULL) == numStr)
	numStr = (char *) Allocate(64);
    if (((char *) NULL) == errStr)
	errStr = (char *) Allocate(1024);
}

FE0(void, PrpMemClose)
{
    LIST *list;

    pplst_close();
    ppio_close();
    macro_close();
    clear_output_file();
    PrsIncl((LIST **) NULL);
    delete_stack();
    LtlClean();
    ClFclose(prpOutput);
    ClFclose(preFile);
    ClFclose(defsfile);
    ClFclose(enmFile);
    OwnClose(&datFile);
    ClDeallocate(cName);
    ClDeallocate(errStr);
    ClDeallocate(extBuchi);
    ClDeallocate(name);
    ClDeallocate(numStr);
    ClDeallocate(preName);
    ClDeallocate(srcPrefix);
    LstDestroy(prpDefList, FALSE);
    LstDestroy(prpKndList, FALSE);
    LstDestroy(ppLsLs, TRUE);
    if (((LIST *) NULL) != ppTokList)
	{
	    list = ppTokList;
	    ppTokList = (LIST *) NULL;
	    LstDestroy(list, TRUE);
	}
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments: argc - the number of command line arguments given to the
*                   program
*            argv - an array containing the arguments
* Returns:   0 if everything succeeds, and something else otherwise
*            (PROD has no systematic convention for non-zero exits.)
* See also:
*
****************************************************************************
*/
MAIN2()
{
    int argc = 0;
    char **argv = NULL;

#ifndef UNIX
    ClockOn();
#endif
    SetIO();
    ClearErrno();
    SetExMex();
    baseName = StrCopy("prpp");
#ifdef ARGV0
    myName = CmdNm(_argv_[0]);
#else
    myName = baseName;
#endif
#ifndef UNIX
    StoreCmd(_argc_, myName, (char **) _argv_);
#endif
    ExpArgs(_argc_, (char **) _argv_, &argc, &argv);
    error_flag = 0;
    placeLimits = TRUE;
    preserve_comments = FALSE;
    ansi = FALSE;
    line_directives = FALSE;
    parsing_control = 0;
    may_have_lost_line = FALSE;
    arity_complaints = FALSE;
#ifdef TRAIN
    train = FALSE;
#endif
    unfMaxDiff = (UDOMAIN) (-1);
    fileName = NULL;
    us = NULL;
    PrpMemInit();
    for (i = 1; i < argc; i++)
	{
	    argument = argv[i];
	    if (*argument == '-')
		{
		    argument++;
		    if ((*argument == '\0') &&
			(i + 1 == argc))
			Misusage();
		    if (*argument == '\0')
			{
			    argument = argv[i + 1];
			    i++;
			}
		    while (*argument != '\0')
			{
			    if (*argument == 'A')
				ansi = TRUE;
			    else if (*argument == 'a')
				arity_complaints = TRUE;
#ifdef TRAIN
			    else if (*argument == 'B')
				{
				    arity_complaints = TRUE;
				    train = TRUE;
				}
#endif
			    else if (*argument == 'C')
				preserve_comments = TRUE;
			    else if (*argument == 'D')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsLast(prpDefList, argument);
				    (void) LstInsLast(prpKndList, "D");
				    break;
				}
			    else if (*argument == 'I')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsFirst(include_path,
						       StrCopy(argument));
				    break;
				}
			    else if (*argument == 'L')
				line_directives = TRUE;
			    else if (*argument == 'N')
				extBuchi = StrCopy("N");
                            else if (*argument == 'R')
                                placeLimits = FALSE;
			    else if (*argument == 't')
				{
				    argument++;
				    if ((fileName != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    fileName = argument;
				    break;
				}
			    else if (*argument == 'U')
				{
				    argument++;
				    if ((*argument == '\0') && 
					(i + 1 == argc))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    (void) LstInsLast(prpDefList, argument);
				    (void) LstInsLast(prpKndList, "U");
				    break;
				}
			    else if (*argument == 'u')
				{
				    argument++;
				    if ((us != NULL) ||
					((*argument == '\0') && 
					 (i + 1 == argc)))
					Misusage();
				    if (*argument == '\0')
					{
					    argument = argv[i + 1];
					    i++;
					}
				    us = argument;
				    if (ScanLdu(us, &unfMaxL) != 1)
					Misusage();
				    unfMaxDiff = (UDOMAIN) unfMaxL;
				    break;
				}
			    else
				Misusage();
			    argument++;
			}
		}
	    else if (fileName == NULL)
		fileName = argument;
	    else
		Misusage();
	}
    if (fileName == NULL)
	Misusage();
    if (us && !placeLimits)
        {
            (void)
		fprintf
		(_err,
		 "\n%s: options -R and -u are mutually exclusive\n",
		 myName);
            (void) fflush(_err);
            Exit(1);
        }
    i = (int) strlen(fileName);
    if ((i > 4) && (strcmp(fileName + i - 4, ".net") == 0))
	fileName[i - 4] = '\0';
    if (LstEmpty(prpDefList))
	{
	    LstDestroy(prpDefList, FALSE);
	    LstDestroy(prpKndList, FALSE);
	    preName = NULL;
	}
    else
	{
	    preName = FileName(fileName, "pre");
	    preFile = MyFopen(preName, "w", (BOOL *) NULL);
	    if (!preFile)
		{
		    *errStr = '\0';
		    panic(errStr);
		}
	    argument = (char *) LstBegin(prpKndList);
	    for (name = (char *) LstBegin(prpDefList); name != NULL;
		 name = (char *) LstStep(prpDefList))
		{
		    if (*argument == 'D')
			{
			    body = name;
			    if ((*body == '\0') || (*body == '='))
				{
				    error("Command line definition without name:");
				    (void) sprintf(errStr, "-D%s", name);
				    error(errStr);
				    continue;
				}
			    if ((*body != '_') && (*body != '$') &&
				!isalpha((UCHAR) *body))
				{
				    error("Erroneous command line definition:");
				    (void) sprintf(errStr, "-D%s", name);
				    error(errStr);
				    continue;
				}
			    for (body++;
				 (*body == '_') || (*body == '$') ||
				 isalnum((UCHAR) *body);
				 body++)
				;
			    if ((*body != '\0') && (*body != '='))
				{
				    error("Erroneous command line definition:");
				    (void) sprintf(errStr, "-D%s", name);
				    error(errStr);
				    continue;
				}
			    if (*body == '=')
				*body++ = '\0';
			    else
				body = "1";
			    if (*body != '\0')
				(void) fprintf(preFile, "#define %s %s\n",
					       name, body);
			    else
				(void) fprintf(preFile, "#define %s\n",
					       name);
			}
		    else
			(void) fprintf(preFile, "#undef %s\n", name);
		    argument = (char *) LstStep(prpKndList);
		}
	    LstDestroy(prpDefList, FALSE);
	    LstDestroy(prpKndList, FALSE);
	    ClFclose(preFile);
	}
    if (extBuchi)
	{
	    ClDeallocate(extBuchi);
	    extBuchi = FileName(fileName, "nev");
	}
    name = FileName(fileName, "def");
    defsfile = MyFopen(name, "w", (BOOL *) NULL);
    if (!defsfile)
	{
	    *errStr = '\0';
	    panic(errStr);
	}
    ClDeallocate(name);
    i = (int) strlen(fileName);
    srcPrefix = (char *) Allocate(6 + (unsigned) i);
    cName = (char *) Allocate(16 + (unsigned) i);
    (void) sprintf(srcPrefix, "%s.src%c", fileName, DIRCHAR);
    (void) sprintf(cName, "%s_enum.h", srcPrefix);
    enmFile = MyFopen(cName, "w", (BOOL *) NULL);
    if (!enmFile)
	{
	    *errStr = '\0';
	    panic(errStr);
	}
    (void) sprintf(cName, "%s_.h", srcPrefix);
    prpOutput = MyFopen(cName, "w", (BOOL *) NULL);
    if (!prpOutput)
	{
	    *errStr = '\0';
	    panic(errStr);
	}
    nOutpLines = 0L;
    if (preName)
	{
	    ppio_push_file(&preName, FALSE);
	    OutPrs();
	    ppio_pop_file();
	}
    name = FileName(fileName, "net");
    ppio_push_file(&name, FALSE);
    insert_line_tag();
    OutPrs();
    ppio_pop_file();
    ClFclose(defsfile);
    ClFclose(enmFile);
    if (line_directives)
	{
            (void) fprintf(prpOutput, "\n#line %lu \"", nOutpLines + 3L);
	    for (us = cName; *us; us++)
		{
		    (void) fputc((int) *us, prpOutput);
		    if (*us == '\\')
			(void) fputc((int) *us, prpOutput);
		}
	    (void) fputs("\"\n", prpOutput);
            nOutpLines = nOutpLines + 2L;
	}
    pplst_gen();
    ClDeallocate(srcPrefix);
    ClDeallocate(cName);
    name = FileName(fileName, "dat");
    if (NotBinaryFile(datFile = CreateWrite(name)))
	{
	    Exit(1);
	}
    ClDeallocate(name);
    pplst_dat();
    OwnClose(&datFile);
    Exit(error_flag);
    return(error_flag);
}
