/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptandor2.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptandor.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptstable.h"

static unsigned long ptTokCount;
static unsigned long ptArcWeight;

/***************************************************************************
*
* Function:  TrnNdPrIn
* Purpose:   This is an auxiliary routine of TrnAndOr.
* Arguments: trn - a transition
* Returns:
* See also:  TrnAndOr
*
****************************************************************************
*/
FE1(void, TrnNdPrIn,
    TRN *, trn)
{
    switch (TrnGWhere(trn))
        {
        case 1:
            TrnSWhere(trn, (unsigned) 0);
            goto L1;
        case 2:
            TrnSWhere(trn, (unsigned) 0);
            goto L2;
        case 3:
            TrnSWhere(trn, (unsigned) 0);
            goto L3;
        case 4:
            TrnSWhere(trn, (unsigned) 0);
            goto L4;
        case 5:
            TrnSWhere(trn, (unsigned) 0);
            goto L5;
        }
    LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
	{
	    ptPlc = TrArcGPlc(ptTrArc);
	    MkTAStrong(trn, ptTrArc);
	    if (PlcGPurOut(ptPlc))
		LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
		    {
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGPurOut(ptPlc)));
                                TrnSWhere(trn, (unsigned) 1);
                                return;
L1: ;
				ptTrArc = (TRARC *)
				    LstThis(TrnGPurIn(trn));
				ptPlc = TrArcGPlc(ptTrArc);
                                LstSet(PlcGPurOut(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGPurOut(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			StartRule(TrArcGStrong(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    ptTokCount = PlcGCount(ptPlc);
	    if (PlcGMosOut(ptPlc))
		LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
		    {
			ptTrn = PlDrcGTrn(ptPlDrc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGMosOut(ptPlc)));
                                TrnSWhere(trn, (unsigned) 2);
                                return;
L2: ;
				ptTrArc = (TRARC *)
				    LstThis(TrnGPurIn(trn));
				ptPlc = TrArcGPlc(ptTrArc);
				ptTokCount = PlcGCount(ptPlc);
                                LstSet(PlcGMosOut(ptPlc), TrnGCursor(trn));
				ptPlDrc = (PLDRC *)
				    LstThis(PlcGMosOut(ptPlc));
				ptTrn = PlDrcGTrn(ptPlDrc);
			    }
			StartRule(TrArcGStrong(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
			if (
#ifdef SMODELS
			    (!PtLongPlay()) &&
#endif
			    (PlDrcGOut(ptPlDrc) > ptTokCount))
			    continue;
			StartRule(TrArcGWeak(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc, PlDrcGOut(ptPlDrc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    ptLimit = PlcGCount(ptPlc) - TrArcGWeight(ptTrArc);
	    if (PlcGNeutral(ptPlc))
		LSTFOR (PlcGNeutral(ptPlc), ptPlArc, PLARC)
		    {
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGNeutral(ptPlc)));
                                TrnSWhere(trn, (unsigned) 3);
                                return;
L3: ;
				ptTrArc = (TRARC *)
				    LstThis(TrnGPurIn(trn));
				ptPlc = TrArcGPlc(ptTrArc);
				ptTokCount = PlcGCount(ptPlc);
				ptLimit = PlcGCount(ptPlc)
				    - TrArcGWeight(ptTrArc);
                                LstSet(PlcGNeutral(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGNeutral(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			if (
#ifdef SMODELS
			    PtLongPlay() ||
#endif
			    (PlArcGWeight(ptPlArc) > ptLimit))
			    {
				StartRule(TrArcGStrong(ptTrArc));
#ifdef SMODELS
				if (PtLongPlay())
				    {
					ContRule(TrnGActive(trn));
					ContRule
					    (IneqGLt
					     (MkPlcIneq
					      (ptPlc,
					       PlArcGWeight(ptPlArc)
					       + TrArcGWeight
					       (ptTrArc))));
				    }
#endif
				ContRule(TrnGGamma(ptTrn));
				EndRule();
			    }
			if (
#ifdef SMODELS
			    (!PtLongPlay()) &&
#endif
			    (PlArcGWeight(ptPlArc) > ptTokCount))
			    continue;
			StartRule(TrArcGWeak(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc,
				       PlArcGWeight(ptPlArc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    if (PlcGMosIn(ptPlc))
		LSTFOR (PlcGMosIn(ptPlc), ptPlDrc, PLDRC)
		    {
			ptTrn = PlDrcGTrn(ptPlDrc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGMosIn(ptPlc)));
                                TrnSWhere(trn, (unsigned) 4);
                                return;
L4: ;
				ptTrArc = (TRARC *)
				    LstThis(TrnGPurIn(trn));
				ptPlc = TrArcGPlc(ptTrArc);
				ptTokCount = PlcGCount(ptPlc);
				ptLimit = PlcGCount(ptPlc)
				    - TrArcGWeight(ptTrArc);
                                LstSet(PlcGMosIn(ptPlc), TrnGCursor(trn));
				ptPlDrc = (PLDRC *)
				    LstThis(PlcGMosIn(ptPlc));
				ptTrn = PlDrcGTrn(ptPlDrc);
			    }
			if (
#ifdef SMODELS
			    PtLongPlay() ||
#endif
			    (PlDrcGOut(ptPlDrc) > ptLimit))
			    {
				StartRule(TrArcGStrong(ptTrArc));
#ifdef SMODELS
				if (PtLongPlay())
				    {
					ContRule(TrnGActive(trn));
					ContRule
					    (IneqGLt
					     (MkPlcIneq
					      (ptPlc,
					       PlDrcGOut(ptPlDrc)
					       + TrArcGWeight
					       (ptTrArc))));
				    }
#endif
				ContRule(TrnGGamma(ptTrn));
				EndRule();
			    }
			if (
#ifdef SMODELS
			    (!PtLongPlay()) &&
#endif
			    (PlDrcGOut(ptPlDrc) > ptTokCount))
			    continue;
			StartRule(TrArcGWeak(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc,
				       PlDrcGOut(ptPlDrc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    if (PlcGPurIn(ptPlc))
		LSTFOR (PlcGPurIn(ptPlc), ptPlArc, PLARC)
		    {
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGPurIn(ptPlc)));
                                TrnSWhere(trn, (unsigned) 5);
                                return;
L5: ;
				ptTrArc = (TRARC *)
				    LstThis(TrnGPurIn(trn));
				ptPlc = TrArcGPlc(ptTrArc);
                                LstSet(PlcGPurIn(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGPurIn(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			StartRule(TrArcGWeak(ptTrArc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	}
}

/***************************************************************************
*
* Function:  TrnNdMsIn
* Purpose:   This is an auxiliary routine of TrnAndOr.
* Arguments: trn - a transition
* Returns:
* See also:  TrnAndOr
*
****************************************************************************
*/
FE1(void, TrnNdMsIn,
    TRN *, trn)
{
    switch (TrnGWhere(trn))
        {
        case 1:
            TrnSWhere(trn, (unsigned) 0);
            goto L1;
        case 2:
            TrnSWhere(trn, (unsigned) 0);
            goto L2;
        case 3:
            TrnSWhere(trn, (unsigned) 0);
            goto L3;
        case 4:
            TrnSWhere(trn, (unsigned) 0);
            goto L4;
        case 5:
            TrnSWhere(trn, (unsigned) 0);
            goto L5;
        }
    LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
	{
	    ptPlc = TrDrcGPlc(ptTrDrc);
	    MkTDStrong(trn, ptTrDrc);
	    if (PlcGPurOut(ptPlc))
		LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
		    {
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGPurOut(ptPlc)));
                                TrnSWhere(trn, (unsigned) 1);
                                return;
L1: ;
				ptTrDrc = (TRDRC *)
				    LstThis(TrnGMosIn(trn));
				ptPlc = TrDrcGPlc(ptTrDrc);
                                LstSet(PlcGPurOut(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGPurOut(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			StartRule(TrDrcGStrong(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    ptTokCount = PlcGCount(ptPlc);
	    ptArcWeight = TrDrcGOut(ptTrDrc);
	    if (PlcGMosOut(ptPlc))
		LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
		    {
			ptTrn = PlDrcGTrn(ptPlDrc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGMosOut(ptPlc)));
                                TrnSWhere(trn, (unsigned) 2);
                                return;
L2: ;
				ptTrDrc = (TRDRC *)
				    LstThis(TrnGMosIn(trn));
				ptPlc = TrDrcGPlc(ptTrDrc);
				ptTokCount = PlcGCount(ptPlc);
				ptArcWeight = TrDrcGOut(ptTrDrc);
                                LstSet(PlcGMosOut(ptPlc), TrnGCursor(trn));
				ptPlDrc = (PLDRC *)
				    LstThis(PlcGMosOut(ptPlc));
				ptTrn = PlDrcGTrn(ptPlDrc);
			    }
			StartRule(TrDrcGStrong(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
			if ((PlDrcGIn(ptPlDrc) <= ptArcWeight) ||
			     (
#ifdef SMODELS
				 (!PtLongPlay()) &&
#endif
				 (PlDrcGOut(ptPlDrc) > ptTokCount)))
			    continue;
			StartRule(TrDrcGWeak(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc,
				       PlDrcGOut(ptPlDrc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    ptLimit = PlcGCount(ptPlc)
		- TrDrcGIn(ptTrDrc) + TrDrcGOut(ptTrDrc);
	    if (PlcGNeutral(ptPlc))
		LSTFOR (PlcGNeutral(ptPlc), ptPlArc, PLARC)
		    {
			if (
#ifdef SMODELS
			    (!PtLongPlay()) &&
#endif
			    ((PlArcGWeight(ptPlArc) <= ptLimit) &&
			     (PlArcGWeight(ptPlArc) <= ptArcWeight)))
			    continue;
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGNeutral(ptPlc)));
                                TrnSWhere(trn, (unsigned) 3);
                                return;
L3: ;
				ptTrDrc = (TRDRC *)
				    LstThis(TrnGMosIn(trn));
				ptPlc = TrDrcGPlc(ptTrDrc);
				ptTokCount = PlcGCount(ptPlc);
				ptArcWeight = TrDrcGOut(ptTrDrc);
				ptLimit = PlcGCount(ptPlc)
				    - TrDrcGIn(ptTrDrc)
					+ TrDrcGOut(ptTrDrc);
                                LstSet(PlcGNeutral(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGNeutral(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			if (
#ifdef SMODELS
			    PtLongPlay() ||
#endif
			    (PlArcGWeight(ptPlArc) > ptLimit))
			    {
				StartRule(TrDrcGStrong(ptTrDrc));
#ifdef SMODELS
				if (PtLongPlay())
				    {
					ContRule(TrnGActive(trn));
					ContRule
					    (IneqGLt
					     (MkPlcIneq
					      (ptPlc,
					       PlArcGWeight(ptPlArc)
					       + TrDrcGIn(ptTrDrc)
					       - TrDrcGOut
					       (ptTrDrc))));
				    }
#endif
				ContRule(TrnGGamma(ptTrn));
				EndRule();
			    }
			if ((PlArcGWeight(ptPlArc) <= ptArcWeight) ||
			    (
#ifdef SMODELS
				(!PtLongPlay()) &&
#endif
				(PlArcGWeight(ptPlArc) > ptTokCount)))
			    continue;
			StartRule(TrDrcGWeak(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc,
				       PlArcGWeight(ptPlArc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    if (PlcGMosIn(ptPlc))
		LSTFOR (PlcGMosIn(ptPlc), ptPlDrc, PLDRC)
		    {
			ptTrn = PlDrcGTrn(ptPlDrc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGMosIn(ptPlc)));
                                TrnSWhere(trn, (unsigned) 4);
                                return;
L4: ;
				ptTrDrc = (TRDRC *)
				    LstThis(TrnGMosIn(trn));
				ptPlc = TrDrcGPlc(ptTrDrc);
				ptTokCount = PlcGCount(ptPlc);
				ptArcWeight = TrDrcGOut(ptTrDrc);
				ptLimit = PlcGCount(ptPlc)
				    - TrDrcGIn(ptTrDrc)
					+ TrDrcGOut(ptTrDrc);
                                LstSet(PlcGMosIn(ptPlc), TrnGCursor(trn));
				ptPlDrc = (PLDRC *)
				    LstThis(PlcGMosIn(ptPlc));
				ptTrn = PlDrcGTrn(ptPlDrc);
			    }
			if (
#ifdef SMODELS
			    PtLongPlay() ||
#endif
			    (PlDrcGOut(ptPlDrc) > ptLimit))
			    {
				StartRule(TrDrcGStrong(ptTrDrc));
#ifdef SMODELS
				if (PtLongPlay())
				    {
					ContRule(TrnGActive(trn));
					ContRule
					    (IneqGLt
					     (MkPlcIneq
					      (ptPlc,
					       PlDrcGOut(ptPlDrc)
					       + TrDrcGIn(ptTrDrc)
					       - TrDrcGOut
					       (ptTrDrc))));
				    }
#endif
				ContRule(TrnGGamma(ptTrn));
				EndRule();
			    }
			if (
#ifdef SMODELS
			    (!PtLongPlay()) &&
#endif
			    (PlDrcGOut(ptPlDrc) > ptTokCount))
			    continue;
			StartRule(TrDrcGWeak(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
				ContRule
				    (IneqGGe
				     (MkPlcIneq
				      (ptPlc,
				       PlDrcGOut(ptPlDrc))));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    if (PlcGPurIn(ptPlc))
		LSTFOR (PlcGPurIn(ptPlc), ptPlArc, PLARC)
		    {
			ptTrn = PlArcGTrn(ptPlArc);
			if (!TrnOld(ptTrn))
			    {
                                TrnSCursor(trn, LstGet(PlcGPurIn(ptPlc)));
                                TrnSWhere(trn, (unsigned) 5);
                                return;
L5: ;
				ptTrDrc = (TRDRC *)
				    LstThis(TrnGMosIn(trn));
				ptPlc = TrDrcGPlc(ptTrDrc);
                                LstSet(PlcGPurIn(ptPlc), TrnGCursor(trn));
				ptPlArc = (PLARC *)
				    LstThis(PlcGPurIn(ptPlc));
				ptTrn = PlArcGTrn(ptPlArc);
			    }
			StartRule(TrDrcGWeak(ptTrDrc));
#ifdef SMODELS
			if (PtLongPlay())
			    {
				ContRule(TrnGActive(trn));
			    }
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	}
}

/***************************************************************************
*
* Function:  TrnOrPromote
* Purpose:   This is an auxiliary routine of TrnAndOr.
* Arguments: trn - a transition
* Returns:
* See also:  TrnAndOr
*
****************************************************************************
*/
FE1(void, TrnOrPromote,
    TRN *, trn)
{
    switch (TrnGWhere(trn))
        {
        case 1:
            TrnSWhere(trn, (unsigned) 0);
            goto L1;
        case 2:
            TrnSWhere(trn, (unsigned) 0);
            goto L2;
        case 3:
            TrnSWhere(trn, (unsigned) 0);
            goto L3;
        }
    LSTFOR (TrnGPromoted(trn), ptTrProm, TRPROM)
	{
	    if (TrnAsleep(TrPromGTrn(ptTrProm)))
		continue;
	    TrnPutToSleep(TrPromGTrn(ptTrProm));
	    if (!TrnGReRho(TrPromGTrn(ptTrProm)))
		MkTrnReRho(TrPromGTrn(ptTrProm));
	    LSTFOR (TrnGLower(TrPromGTrn(ptTrProm)), ptTrn, TRN)
		    {
			if (!TrnOld(ptTrn))
			    {
				TrnSCursor(trn,
					   LstGet
					   (TrnGLower(TrPromGTrn(ptTrProm))));
				TrnSWhere(trn, (unsigned) 1);
				return;
L1: ;
				ptTrProm = (TRPROM *)
				    LstThis(TrnGPromoted(trn));
				LstSet(TrnGLower(TrPromGTrn(ptTrProm)),
				       TrnGCursor(trn));
				ptTrn = (TRN *)
				    LstThis(TrnGLower(TrPromGTrn(ptTrProm)));
			    }
			StartRule(TrnGReRho(TrPromGTrn(ptTrProm)));
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	}
    TrnForward(trn);
    LSTFOR (TrnGPromoted(trn), ptTrProm, TRPROM)
	{
	    StartRule(TrnGGamma(trn));
	    if (TrPromGArcs(ptTrProm))
		LSTFOR (TrPromGArcs(ptTrProm), ptOtArc, OTARC)
		    {
			ptDefer = OtArcGDefer(ptOtArc);
			ptPlc = DeferGPlc(ptDefer);
			if (PlcGCount(ptPlc) < OtArcGWeight(ptOtArc))
			    goto ISOR;
		    }
	    ContRule(TrnGReRho(TrPromGTrn(ptTrProm)));
	    EndRule();
	    continue;
ISOR: ;
	    if (!TrPromGAtom(ptTrProm))
		MkTrPromAtom(trn, ptTrProm);
	    ContRule(TrPromGAtom(ptTrProm));
	    EndRule();
	    StartRule(TrPromGAtom(ptTrProm));
	    ContRule(TrnGReRho(TrPromGTrn(ptTrProm)));
	    LSTFOR (TrPromGArcs(ptTrProm), ptOtArc, OTARC)
		    {
			ptDefer = OtArcGDefer(ptOtArc);
			ptPlc = DeferGPlc(ptDefer);
			if (PlcGCount(ptPlc) < OtArcGWeight(ptOtArc))
			    {
				DeferCFlag(ptDefer);
				if (!DeferGAtom(ptDefer))
				    MkDeferAtom(trn, ptDefer);
				ContRule(DeferGAtom(ptDefer));
			    }
		    }
	    EndRule();
	    LSTFOR (TrPromGArcs(ptTrProm), ptOtArc, OTARC)
		    {
			ptDefer = OtArcGDefer(ptOtArc);
			ptPlc = DeferGPlc(ptDefer);
			if ((PlcGCount(ptPlc) >= OtArcGWeight(ptOtArc)) ||
			    DeferGFlag(ptDefer))
			    continue;
			DeferSFlag(ptDefer);
			if (PlcGPurIn(ptPlc))
			    LSTFOR (PlcGPurIn(ptPlc), ptPlArc, PLARC)
				{
				    ptTrn = PlArcGTrn(ptPlArc);
				    if (!TrnOld(ptTrn))
					{
					    TrnSCursor(trn,
						       LstGet
						       (PlcGPurIn(ptPlc)));
					    TrnBackward(trn);
					    TrnSWhere(trn, (unsigned) 2);
					    return;
L2: ;
					    TrnForward(trn);
					    ptTrProm = (TRPROM *)
						LstThis(TrnGPromoted(trn));
					    ptOtArc = (OTARC *)
						LstThis(TrPromGArcs(ptTrProm));
					    ptDefer = OtArcGDefer(ptOtArc);
					    ptPlc = DeferGPlc(ptDefer);
					    LstSet(PlcGPurIn(ptPlc),
						   TrnGCursor(trn));
					    ptPlArc = (PLARC *)
						LstThis(PlcGPurIn(ptPlc));
					    ptTrn = PlArcGTrn(ptPlArc);
					}
				    StartRule(DeferGAtom(ptDefer));
				    ContRule(TrnGGamma(ptTrn));
				    EndRule();
				}
			ptLimit = PlcGCount(ptPlc);
			if (PlcGMosIn(ptPlc))
			    LSTFOR (PlcGMosIn(ptPlc), ptPlDrc, PLDRC)
				{
				    if (PlDrcGOut(ptPlDrc) > ptLimit)
					continue;
				    ptTrn = PlDrcGTrn(ptPlDrc);
				    if (!TrnOld(ptTrn))
					{
					    TrnSCursor(trn,
						       LstGet
						       (PlcGMosIn(ptPlc)));
					    TrnBackward(trn);
					    TrnSWhere(trn, (unsigned) 3);
					    return;
L3: ;
					    TrnForward(trn);
					    ptTrProm = (TRPROM *)
						LstThis(TrnGPromoted(trn));
					    ptOtArc = (OTARC *)
						LstThis(TrPromGArcs(ptTrProm));
					    ptDefer = OtArcGDefer(ptOtArc);
					    ptPlc = DeferGPlc(ptDefer);
					    ptLimit = PlcGCount(ptPlc);
					    LstSet(PlcGMosIn(ptPlc),
						   TrnGCursor(trn));
					    ptPlDrc = (PLDRC *)
						LstThis(PlcGMosIn(ptPlc));
					    ptTrn = PlDrcGTrn(ptPlDrc);
					}
				    StartRule(DeferGAtom(ptDefer));
				    ContRule(TrnGGamma(ptTrn));
				    EndRule();
				}
		    }
	}
    TrnBackward(trn);
}
