/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptelem2.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This is an auxiliary module of ptnet.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"
#include "error.h"
#include "cmdnm.h"

/* All these Cons, Copy, Discard, Dump, KeyCmp, Compare, Tools and
 * balTools are trivial and similar to those explained in detail
 * in modules such as gphutil.c. */
 
#ifdef DEBUG
FE2(void, PtrDump,
    char *, ptr,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "pointer 0x%lx\n", (unsigned long) ptr);
    (void) fflush(_err);
}
#endif

FE2(int, PtrCompare,
    char *, ptr1,
    char *, ptr2)
{
    return((ptr1 == ptr2) ? 0 : ((ptr1 < ptr2) ? -1 : 1));
}

TOOLS ptrTools =
{ (CFUNC) NULL, (VOF) NULL
#ifdef DEBUG
, PtrDump
#endif
 };

BALTOOLS ptrBalTools =
{ (CFUNC) NULL, (VOF) NULL
#ifdef DEBUG
, PtrDump
#endif
, PtrCompare, (FOOL) NULL };

FE0(PLC *, PlcCons)
{
    PLC *plc;

    plc = PlcAlloc();
    PlcSCount(plc, (unsigned long) 0L);
    PlcCKey(plc);
    PlcCPurIn(plc);
    PlcCPurOut(plc);
    PlcCNeutral(plc);
    PlcCMosIn(plc);
    PlcCMosOut(plc);
    if (PtDelAlg())
	{
	    PlcCGamma(plc);
#ifdef SMODELS
	    if (PtSmodels())
		{
		    PlcCOutMax(plc);
		    PlcCIneqs(plc);
		}
#endif
	}
    PlcCFlag(plc);
    return(plc);
}

FE1(PLC *, PlcCopy,
    PLC *, plc)
{
    PLC *copy;

    copy = PlcAlloc();
    PlcSCount(copy, PlcGCount(plc));
    PlcSFlag(copy, PlcGFlag(plc));
    if (PlcGKey(plc))
	PlcSKey(copy, DataCopy(PlcGKey(plc)));
    else
	PlcCKey(copy);
    if (PlcBuchi(plc))
	PlcZNext(copy, LstCopy(PlcGNext(plc), FALSE));
    else
	{
	    if (PlcGPurIn(plc))
		PlcZPurIn(copy, LstCopy(PlcGPurIn(plc), TRUE));
	    else
		PlcCPurIn(copy);
	    if (PlcGPurOut(plc))
		PlcZPurOut(copy, LstCopy(PlcGPurOut(plc), TRUE));
	    else
		PlcCPurOut(copy);
	    if (PlcGNeutral(plc))
		PlcZNeutral(copy, LstCopy(PlcGNeutral(plc), TRUE));
	    else
		PlcCNeutral(copy);
	    if (PlcGMosIn(plc))
		PlcZMosIn(copy, LstCopy(PlcGMosIn(plc), TRUE));
	    else
		PlcCMosIn(copy);
	    if (PlcGMosOut(plc))
		PlcZMosOut(copy, LstCopy(PlcGMosOut(plc), TRUE));
	    else
		PlcCMosOut(copy);
	}
    if (PtDelAlg())
	{
	    PlcCGamma(copy);
#ifdef SMODELS
	    if (PtSmodels())
		{
		    PlcSOutMax(copy, PlcGOutMax(plc));
		    PlcCIneqs(copy);
		}
#endif
	}
    else if (PtSymm())
	PlcSInd(copy, PlcGInd(plc));
    return(copy);
}

FE1(void, PlcDracsid,
    PLC *, plc)
{
    if (plc)
	{
	    if (PlcGKey(plc))
		PlcDKey(plc);
	    if (PlcBuchi(plc))
		LstDestroy(PlcGNext(plc), FALSE);
	    else
		{
		    if (PlcGPurIn(plc))
			LstDestroy(PlcGPurIn(plc), TRUE);
		    if (PlcGPurOut(plc))
			LstDestroy(PlcGPurOut(plc), TRUE);
		    if (PlcGNeutral(plc))
			LstDestroy(PlcGNeutral(plc), TRUE);
		    if (PlcGMosIn(plc))
			LstDestroy(PlcGMosIn(plc), TRUE);
		    if (PlcGMosOut(plc))
			LstDestroy(PlcGMosOut(plc), TRUE);
		}
#ifdef SMODELS
	    if (PtSmodels() && PlcGIneqs(plc))
		BalTrDestroy(PlcGIneqs(plc), TRUE);
#endif
	    Deallocate((char *) plc);
	}
}

#ifdef DEBUG
FE2(void, PlcDump,
    PLC *, plc,
    unsigned, depth)
{
    unsigned p = 0;
    TUPLE t_u_p_l_e;
    TUPLE *tuple;
    PLC *pl2c;

    DumpIndent(depth);
    if (PtSymm() && !PtDelAlg())
	(void) fprintf(_err, "PLC 0x%lx count %lu flag %u ind %lu\n",
		       (unsigned long) plc, PlcGCount(plc), PlcGFlag(plc),
		       PlcGInd(plc));
    else
	(void) fprintf(_err, "PLC 0x%lx count %lu flag %u\n",
		       (unsigned long) plc, PlcGCount(plc), PlcGFlag(plc));
    DumpIndent(depth + 1);
    if (PlcGKey(plc))
	{
	    tuple = &t_u_p_l_e;
	    TupSArity(tuple, (unsigned) 0);
	    TupSArray(tuple, (DOMAIN *) NULL);
	    TupDecode(&p, PlcGKey(plc), tuple);
	    (void) fprintf(_err, "place %u\n", p);
	    TupDump(tuple, depth + 1);
	    if (TupGArity(tuple))
		Deallocate((char *) TupGArray(tuple));
	}
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (PlcGGamma(plc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("gamma\n", _err);
		    MotaDump((MOTA *) PlcGGamma(plc), depth + 2);
		}
	}
    DumpIndent(depth + 1);
    if (PlcBuchi(plc))
	{
	    (void) fputs("next\n", _err);
	    LSTFOR (PlcGNext(plc), pl2c, PLC)
		{
		    DumpIndent(depth + 2);
		    (void) fprintf(_err, "PLC 0x%lx\n", (unsigned long) pl2c);
		}
	}
    else
	{
	    (void) fputs("pureIn\n", _err);
	    if (PlcGPurIn(plc))
		LstDump(PlcGPurIn(plc), depth + 2);
	    DumpIndent(depth + 1);
	    (void) fputs("pureOut\n", _err);
	    if (PlcGPurOut(plc))
		LstDump(PlcGPurOut(plc), depth + 2);
	    DumpIndent(depth + 1);
	    (void) fputs("neutral\n", _err);
	    if (PlcGNeutral(plc))
		LstDump(PlcGNeutral(plc), depth + 2);
	    DumpIndent(depth + 1);
	    (void) fputs("mostIn\n", _err);
	    if (PlcGMosIn(plc))
		LstDump(PlcGMosIn(plc), depth + 2);
	    DumpIndent(depth + 1);
	    (void) fputs("mostOut\n", _err);
	    if (PlcGMosOut(plc))
		LstDump(PlcGMosOut(plc), depth + 2);
	}
    DumpIndent(depth);
    (void) fputs("End of PLC\n", _err);
    (void) fflush(_err);
}
#endif

FE2(int, PlcKeyCmp,
    PLC *, plc,
    DATATYPE *, key)
{
    return(BTupComp(PlcGKey(plc), key));
}

TOOLS plcTools =
{ (CFUNC) PlcCopy, PlcDracsid
#ifdef DEBUG
, PlcDump
#endif
 };

BALTOOLS plcBalTools =
{ (CFUNC) PlcCopy, PlcDracsid
#ifdef DEBUG
, PlcDump
#endif
, PlcKeyCmp, (FOOL) NULL };

FE0(TRN *, TrnCons)
{
    TRN *trn;

    trn = TrnAlloc();
    TrnCKey(trn);
    TrnCNumber(trn);
    TrnCFlag(trn);
    TrnSWhere(trn, (unsigned) 0);
    TrnCCursor(trn);
    TrnCDown(trn);
    TrnCBlocks(trn);
    TrnCPurIn(trn);
    TrnCPurOut(trn);
    TrnCNeutral(trn);
    TrnCMosIn(trn);
    TrnCMosOut(trn);
    TrnCHigher(trn);
    TrnCLower(trn);
    TrnCPromoted(trn);
    TrnCDefers(trn);
    if (PtDelAlg())
	TrnCAtoms(trn);
    else if (PtStub())
	{
	    TrnSVal(trn, (unsigned long) 0L);
	    TrnSDupVal(trn, (unsigned long) 0L);
            TrnSMin(trn, (unsigned long) 0L);
	}
    return(trn);
}

FE1(TRN *, TrnCopy,
    TRN *, trn)
{
    TRN *copy;
    unsigned i;

    copy = TrnAlloc();
    if (TrnGKey(trn))
	TrnSKey(copy, DataCopy(TrnGKey(trn)));
    else
	TrnCKey(copy);
    TrnSNumber(copy, TrnGNumber(trn));
    TrnSFlag(copy, TrnGFlag(trn));
    TrnSWhere(trn, TrnGWhere(trn));
    TrnCCursor(trn);
    TrnCDown(trn);
    if (TrnBlocks(trn))
        {
            for (i = 0; TrnGBlock(trn, i) != (unsigned) -1; i++)
                ;
            TrnABlocks(copy, i + 1);
            while (TRUE)
                {
                    TrnSBlock(copy, i, TrnGBlock(trn, i));
                    if (!i)
                        break;
                    i--;
                }
        }
    else
        TrnCBlocks(trn);
    if (TrnGPurIn(trn))
	TrnZPurIn(copy, LstCopy(TrnGPurIn(trn), TRUE));
    else
	TrnCPurIn(copy);
    if (TrnGPurOut(trn))
	TrnZPurOut(copy, LstCopy(TrnGPurOut(trn), TRUE));
    else
	TrnCPurOut(copy);
    if (TrnGNeutral(trn))
	TrnZNeutral(copy, LstCopy(TrnGNeutral(trn), TRUE));
    else
	TrnCNeutral(copy);
    if (TrnGMosIn(trn))
	TrnZMosIn(copy, LstCopy(TrnGMosIn(trn), TRUE));
    else
	TrnCMosIn(copy);
    if (TrnGMosOut(trn))
	TrnZMosOut(copy, LstCopy(TrnGMosOut(trn), TRUE));
    else
	TrnCMosOut(copy);
    if (TrnGHigher(trn))
	TrnZHigher(copy, LstCopy(TrnGHigher(trn), FALSE));
    else
	TrnCHigher(copy);
    if (TrnGLower(trn))
	TrnZLower(copy, LstCopy(TrnGLower(trn), FALSE));
    else
	TrnCLower(copy);
    if (TrnGPromoted(trn))
	TrnZPromoted(copy, LstCopy(TrnGPromoted(trn), TRUE));
    else
	TrnCPromoted(copy);
    if (TrnGDefers(trn))
	TrnZDefers(copy, LstCopy(TrnGDefers(trn), TRUE));
    else
	TrnCDefers(copy);
    if (PtDelAlg())
	TrnCAtoms(copy);
    else if (PtStub())
	{
	    TrnSVal(copy, TrnGVal(trn));
            TrnSDupVal(copy, TrnGDupVal(trn));
            TrnSMin(copy, TrnGMin(trn));
	}
    return(copy);
}

FE1(void, TrnDracsid,
    TRN *, trn)
{
    if (trn)
	{
	    if (TrnGKey(trn))
		TrnDKey(trn);
	    if (TrnBlocks(trn))
		TrnDBlocks(trn);
	    if (TrnGPurIn(trn))
		LstDestroy(TrnGPurIn(trn), TRUE);
	    if (TrnGPurOut(trn))
		LstDestroy(TrnGPurOut(trn), TRUE);
	    if (TrnGNeutral(trn))
		LstDestroy(TrnGNeutral(trn), TRUE);
	    if (TrnGMosIn(trn))
		LstDestroy(TrnGMosIn(trn), TRUE);
	    if (TrnGMosOut(trn))
		LstDestroy(TrnGMosOut(trn), TRUE);
	    if (TrnGHigher(trn))
		LstDestroy(TrnGHigher(trn), FALSE);
	    if (TrnGLower(trn))
		LstDestroy(TrnGLower(trn), FALSE);
	    if (TrnGPromoted(trn))
		LstDestroy(TrnGPromoted(trn), TRUE);
	    if (TrnGDefers(trn))
		LstDestroy(TrnGDefers(trn), TRUE);
	    Deallocate((char *) trn);
	}
}

#ifdef DEBUG
FE2(void, TrnDump,
    TRN *, trn,
    unsigned, depth)
{
    TRN *rtrn;
    DOMAIN *instance;
#ifdef WEIRD
    DATATYPE *key;
#endif
    unsigned t = 0;
    unsigned p = 0;
    unsigned i;

    DumpIndent(depth);
    (void) fprintf(_err, "TRN 0x%lx number %lu flag %lu", (unsigned long) trn,
		   TrnGNumber(trn), TrnGFlag(trn));
    if (PtStub() && !PtDelAlg())
        (void) fprintf(_err, " val %lu dupVal %lu min %lu",
                       TrnGVal(trn), TrnGDupVal(trn), TrnGMin(trn));
    (void) fputc((int) '\n', _err);
    if (TrnGKey(trn))
	{
	    DumpIndent(depth + 1);
#ifdef WEIRD
	    key = TrnGKey(trn);
#endif
	    instance = InsDecode(&t, &p,
#ifdef WEIRD
				 &key
#else
				 TrnGKey(trn)
#endif
				 );
	    (void) fprintf(_err, "transition %u priority %u\n", t, p);
	    DumpIndent(depth + 1);
	    TrnDispLabel(trn, _err);
	    (void) fputc((int) '\n', _err);
	    DumpIndent(depth + 1);
	    (void) fputs("instance", _err);
	    for (i = 0; i < IngGProfile(t); i++)
		{
		    (void) fputs(" ", _err);
		    (void) fprintf(_err, DOMFORMAT, instance[i]);
		}
	    (void) fputc((int) '\n', _err);
	    Deallocate((char *) instance);
	}
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (TrnGGamma(trn))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("gamma\n", _err);
		    MotaDump((MOTA *) TrnGGamma(trn), depth + 2);
		}
	    if (TrnGKappa(trn))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("kappa\n", _err);
		    MotaDump((MOTA *) TrnGKappa(trn), depth + 2);
		}
	    if (TrnGDiRho(trn))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("diRho\n", _err);
		    MotaDump((MOTA *) TrnGDiRho(trn), depth + 2);
		}
	    if (TrnGEnRho(trn))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("enRho\n", _err);
		    MotaDump((MOTA *) TrnGEnRho(trn), depth + 2);
		}
	    if (TrnGReRho(trn))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("reRho\n", _err);
		    MotaDump((MOTA *) TrnGReRho(trn), depth + 2);
		}
	}
    if (TrnBlocks(trn))
        {
            DumpIndent(depth + 1);
            (void) fputs("blocks", _err);
            for (i = 0; TrnGBlock(trn, i) != (unsigned) -1; i++)
                (void) fprintf(_err, " %u", TrnGBlock(trn, i));
	    (void) fputc((int) '\n', _err);
        }
    DumpIndent(depth + 1);
    (void) fputs("pureIn\n", _err);
    if (TrnGPurIn(trn))
	LstDump(TrnGPurIn(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("pureOut\n", _err);
    if (TrnGPurOut(trn))
	LstDump(TrnGPurOut(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("neutral\n", _err);
    if (TrnGNeutral(trn))
	LstDump(TrnGNeutral(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("mostIn\n", _err);
    if (TrnGMosIn(trn))
	LstDump(TrnGMosIn(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("mostOut\n", _err);
    if (TrnGMosOut(trn))
	LstDump(TrnGMosOut(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("higher\n", _err);
    if (TrnGHigher(trn))
	LSTFOR (TrnGHigher(trn), rtrn, TRN)
	    {
		DumpIndent(depth + 2);
		(void) fprintf(_err, "TRN 0x%lx\n", (unsigned long) rtrn);
	    }
    DumpIndent(depth + 1);
    (void) fputs("lower\n", _err);
    if (TrnGLower(trn))
	LSTFOR (TrnGLower(trn), rtrn, TRN)
	    {
		DumpIndent(depth + 2);
		(void) fprintf(_err, "TRN 0x%lx\n", (unsigned long) rtrn);
	    }
    DumpIndent(depth + 1);
    (void) fputs("promoted\n", _err);
    if (TrnGPromoted(trn))
	LstDump(TrnGPromoted(trn), depth + 2);
    DumpIndent(depth + 1);
    (void) fputs("defers\n", _err);
    if (TrnGDefers(trn))
	LstDump(TrnGDefers(trn), depth + 2);
    DumpIndent(depth);
    (void) fputs("End of TRN\n", _err);
    (void) fflush(_err);
}
#endif

FE2(int, TrnKeyCmp,
    TRN *, trn,
    DATATYPE *, key)
{
    return(InsComp(TrnGKey(trn), key));
}

TOOLS trnTools =
{ (CFUNC) TrnCopy, TrnDracsid
#ifdef DEBUG
, TrnDump
#endif
 };

BALTOOLS trnBalTools =
{ (CFUNC) TrnCopy, TrnDracsid
#ifdef DEBUG
, TrnDump
#endif
, TrnKeyCmp, (FOOL) NULL };

FE2(PLARC *, PlArcCons,
    TRN *, trn,
    unsigned long, weight)
{
    PLARC *plArc;

    plArc = PlArcAlloc();
    PlArcSTrn(plArc, trn);
    PlArcSWeight(plArc, weight);
    return(plArc);
}

FE1(PLARC *, PlArcCopy,
    PLARC *, plArc)
{
    PLARC *copy;

    copy = PlArcAlloc();
    PlArcSTrn(copy, PlArcGTrn(plArc));
    PlArcSWeight(copy, PlArcGWeight(plArc));
    return(copy);
}

#ifdef DEBUG
FE2(void, PlArcDump,
    PLARC *, plArc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "PLARC weight %lu trn 0x%lx\n",
		   PlArcGWeight(plArc), (unsigned long) PlArcGTrn(plArc));
    (void) fflush(_err);
}
#endif

FE2(int, PlArcCompare,
    PLARC *, plArc1,
    PLARC *, plArc2)
{
    if (PlArcGWeight(plArc1) < PlArcGWeight(plArc2))
	return(-1);
    if (PlArcGWeight(plArc1) > PlArcGWeight(plArc2))
	return(1);
    return(0);
}

TOOLS plArcTools =
{ (CFUNC) PlArcCopy, Deallocate
#ifdef DEBUG
, PlArcDump
#endif
 };

BALTOOLS plArcBalTools =
{ (CFUNC) PlArcCopy, Deallocate
#ifdef DEBUG
, PlArcDump
#endif
, PlArcCompare, (FOOL) NULL };

FE3(PLDRC *, PlDrcCons,
    TRN *, trn,
    unsigned long, in,
    unsigned long, out)
{
    PLDRC *plDrc;

    plDrc = PlDrcAlloc();
    PlDrcSTrn(plDrc, trn);
    PlDrcSIn(plDrc, in);
    PlDrcSOut(plDrc, out);
    return(plDrc);
}

FE1(PLDRC *, PlDrcCopy,
    PLDRC *, plDrc)
{
    PLDRC *copy;

    copy = PlDrcAlloc();
    PlDrcSTrn(copy, PlDrcGTrn(plDrc));
    PlDrcSIn(copy, PlDrcGIn(plDrc));
    PlDrcSOut(copy, PlDrcGOut(plDrc));
    return(copy);
}

#ifdef DEBUG
FE2(void, PlDrcDump,
    PLDRC *, plDrc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "PLDRC in %lu out %lu trn 0x%lx\n",
		   PlDrcGIn(plDrc), PlDrcGOut(plDrc),
		   (unsigned long) PlDrcGTrn(plDrc));
    (void) fflush(_err);
}
#endif

FE2(int, PlDrcCompare,
    PLDRC *, plDrc1,
    PLDRC *, plDrc2)
{
    if (PlDrcGIn(plDrc1) < PlDrcGIn(plDrc2))
	return(-1);
    if (PlDrcGIn(plDrc1) > PlDrcGIn(plDrc2))
	return(1);
    if (PlDrcGOut(plDrc1) < PlDrcGOut(plDrc2))
	return(-1);
    if (PlDrcGOut(plDrc1) > PlDrcGOut(plDrc2))
	return(1);
    return(0);
}

TOOLS plDrcTools =
{ (CFUNC) PlDrcCopy, Deallocate
#ifdef DEBUG
, PlDrcDump
#endif
 };

BALTOOLS plDrcBalTools =
{ (CFUNC) PlDrcCopy, Deallocate
#ifdef DEBUG
, PlDrcDump
#endif
, PlDrcCompare, (FOOL) NULL };

FE2(TRARC *, TrArcCons,
    PLC *, plc,
    unsigned long, weight)
{
    TRARC *trArc;

    trArc = TrArcAlloc();
    TrArcSPlc(trArc, plc);
    TrArcSWeight(trArc, weight);
    TrArcCAtoms(trArc);
    return(trArc);
}

FE1(TRARC *, TrArcCopy,
    TRARC *, trArc)
{
    TRARC *copy;

    copy = TrArcAlloc();
    TrArcSPlc(copy, TrArcGPlc(trArc));
    TrArcSWeight(copy, TrArcGWeight(trArc));
    TrArcCAtoms(copy);
    return(copy);
}

#ifdef DEBUG
FE2(void, TrArcDump,
    TRARC *, trArc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "TRARC weight %lu trn 0x%lx\n",
		   TrArcGWeight(trArc), (unsigned long) TrArcGPlc(trArc));
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (TrArcGOder(trArc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("oder\n", _err);
		    MotaDump((MOTA *) TrArcGOder(trArc), depth + 2);
		}
	    if (TrArcGStrong(trArc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("strong\n", _err);
		    MotaDump((MOTA *) TrArcGStrong(trArc), depth + 2);
		}
	    if (TrArcGWeak(trArc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("weak\n", _err);
		    MotaDump((MOTA *) TrArcGWeak(trArc), depth + 2);
		}
	}
    (void) fflush(_err);
}
#endif

FE2(int, TrArcCompare,
    TRARC *, trArc1,
    TRARC *, trArc2)
{
    if (TrArcGWeight(trArc1) < TrArcGWeight(trArc2))
	return(-1);
    if (TrArcGWeight(trArc1) > TrArcGWeight(trArc2))
	return(1);
    return(0);
}

TOOLS trArcTools =
{ (CFUNC) TrArcCopy, Deallocate
#ifdef DEBUG
, TrArcDump
#endif
 };

BALTOOLS trArcBalTools =
{ (CFUNC) TrArcCopy, Deallocate
#ifdef DEBUG
, TrArcDump
#endif
, TrArcCompare, (FOOL) NULL };

FE3(TRDRC *, TrDrcCons,
    PLC *, plc,
    unsigned long, in,
    unsigned long, out)
{
    TRDRC *trDrc;

    trDrc = TrDrcAlloc();
    TrDrcSPlc(trDrc, plc);
    TrDrcSIn(trDrc, in);
    TrDrcSOut(trDrc, out);
    TrDrcCAtoms(trDrc);
    return(trDrc);
}

FE1(TRDRC *, TrDrcCopy,
    TRDRC *, trDrc)
{
    TRDRC *copy;

    copy = TrDrcAlloc();
    TrDrcSPlc(copy, TrDrcGPlc(trDrc));
    TrDrcSIn(copy, TrDrcGIn(trDrc));
    TrDrcSOut(copy, TrDrcGOut(trDrc));
    TrDrcCAtoms(copy);
    return(copy);
}

#ifdef DEBUG
FE2(void, TrDrcDump,
    TRDRC *, trDrc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "TRDRC in %lu out %lu trn 0x%lx\n",
		   TrDrcGIn(trDrc), TrDrcGOut(trDrc),
		   (unsigned long) TrDrcGPlc(trDrc));
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (TrDrcGOder(trDrc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("oder\n", _err);
		    MotaDump((MOTA *) TrDrcGOder(trDrc), depth + 2);
		}
	    if (TrDrcGStrong(trDrc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("strong\n", _err);
		    MotaDump((MOTA *) TrDrcGStrong(trDrc), depth + 2);
		}
	    if (TrDrcGWeak(trDrc))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("weak\n", _err);
		    MotaDump((MOTA *) TrDrcGWeak(trDrc), depth + 2);
		}
	}
    (void) fflush(_err);
}
#endif

FE2(int, TrDrcCompare,
    TRDRC *, trDrc1,
    TRDRC *, trDrc2)
{
    if (TrDrcGIn(trDrc1) < TrDrcGIn(trDrc2))
	return(-1);
    if (TrDrcGIn(trDrc1) > TrDrcGIn(trDrc2))
	return(1);
    if (TrDrcGOut(trDrc1) < TrDrcGOut(trDrc2))
	return(-1);
    if (TrDrcGOut(trDrc1) > TrDrcGOut(trDrc2))
	return(1);
    return(0);
}

TOOLS trDrcTools =
{ (CFUNC) TrDrcCopy, Deallocate
#ifdef DEBUG
, TrDrcDump
#endif
 };

BALTOOLS trDrcBalTools =
{ (CFUNC) TrDrcCopy, Deallocate
#ifdef DEBUG
, TrDrcDump
#endif
, TrDrcCompare, (FOOL) NULL };

FE1(TRPROM *, TrPromCons,
    TRN *, trn)
{
    TRPROM *trProm;

    trProm = TrPromAlloc();
    TrPromSTrn(trProm, trn);
    TrPromCAtom(trProm);
    TrPromCArcs(trProm);
    return(trProm);
}

FE1(TRPROM *, TrPromCopy,
    TRPROM *, trProm)
{
    TRPROM *copy;

    copy = TrPromAlloc();
    TrPromSTrn(copy, TrPromGTrn(trProm));
    TrPromCAtom(copy);
    if (TrPromGArcs(trProm))
	TrPromZArcs(copy, LstCopy(TrPromGArcs(trProm), TRUE));
    else
	TrPromCArcs(copy);
    return(copy);
}

FE1(void, TrPromDracsid,
    TRPROM *, trProm)
{
    if (trProm)
	{
	    LstDestroy(TrPromGArcs(trProm), TRUE);
	    Deallocate((char *) trProm);
	}
}

#ifdef DEBUG
FE2(void, TrPromDump,
    TRPROM *, trProm,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "TRPROM trn 0x%lx\n",
		   (unsigned long) TrPromGTrn(trProm));
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (TrPromGAtom(trProm))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("atom\n", _err);
		    MotaDump((MOTA *) TrPromGAtom(trProm), depth + 2);
		}
	}
    DumpIndent(depth + 1);
    (void) fputs("arcs\n", _err);
    if (TrPromGArcs(trProm))
	LstDump(TrPromGArcs(trProm), depth + 2);
    (void) fflush(_err);
}
#endif

TOOLS trPromTools =
{ (CFUNC) TrPromCopy, TrPromDracsid
#ifdef DEBUG
, TrPromDump
#endif
 };

FE1(DEFER *, DeferCons,
    PLC *, plc)
{
    DEFER *defer;

    defer = DeferAlloc();
    DeferSPlc(defer, plc);
    DeferCFlag(defer);
    DeferCAtom(defer);
    return(defer);
}

FE1(DEFER *, DeferCopy,
    DEFER *, defer)
{
    DEFER *copy;

    copy = DeferAlloc();
    DeferSPlc(copy, DeferGPlc(defer));
    DeferCFlag(copy);
    DeferCAtom(copy);
    return(copy);
}

#ifdef DEBUG
FE2(void, DeferDump,
    DEFER *, defer,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "DEFER plc 0x%lx flag %u\n",
		   (unsigned long) DeferGPlc(defer), DeferGFlag(defer));
    if (
#ifdef SMODELS
	(!PtSmodels()) &&
#endif
	PtDelAlg())
	{
	    if (DeferGAtom(defer))
		{
		    DumpIndent(depth + 1);
		    (void) fputs("atom\n", _err);
		    MotaDump((MOTA *) DeferGAtom(defer), depth + 2);
		}
	}
    (void) fflush(_err);
}
#endif

FE2(int, DeferKeyCmp,
    DEFER *, defer,
    PLC *, plc)
{
    return((DeferGPlc(defer) < plc) ? -1 :
	   ((DeferGPlc(defer) > plc) ? 1 : 0));
}

TOOLS deferTools =
{ (CFUNC) DeferCopy, Deallocate
#ifdef DEBUG
, DeferDump
#endif
 };

BALTOOLS deferBalTools =
{ (CFUNC) DeferCopy, Deallocate
#ifdef DEBUG
, DeferDump
#endif
, DeferKeyCmp, (FOOL) NULL };

FE2(OTARC *, OtArcCons,
    DEFER *, defer,
    unsigned long, weight)
{
    OTARC *otArc;

    otArc = OtArcAlloc();
    OtArcSDefer(otArc, defer);
    OtArcSWeight(otArc, weight);
    return(otArc);
}

FE1(OTARC *, OtArcCopy,
    OTARC *, otArc)
{
    OTARC *copy;

    copy = OtArcAlloc();
    OtArcSDefer(copy, OtArcGDefer(otArc));
    OtArcSWeight(copy, OtArcGWeight(otArc));
    return(copy);
}

#ifdef DEBUG
FE2(void, OtArcDump,
    OTARC *, otArc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "OTARC weight %lu\n", OtArcGWeight(otArc));
    DumpIndent(depth + 1);
    (void) fputs("defer\n", _err);
    DeferDump(OtArcGDefer(otArc), depth + 2);
}
#endif

TOOLS otArcTools =
{ (CFUNC) OtArcCopy, Deallocate
#ifdef DEBUG
, OtArcDump
#endif
 };

static STATE *ptState;
static TUPLE *ptTuple;
static unsigned ptPlInd;

/***************************************************************************
*
* Function:  PlcState
* Purpose:   This is an auxiliary routine of PtState
* Arguments: plc - a place
* Returns:
* See also:  PtState
*
****************************************************************************
*/
FS1(BOOL, PlcState,
    PLC *, plc)
{
    if (PlcGCount(plc) != 0L)
	{
	    ptTuple = TupAlloc();
	    TupDecode(&ptPlInd, PlcGKey(plc), ptTuple);
	    MarInsLast(StaGMarking(ptState, ptPlInd),
		       MultupCons(PlcGCount(plc), ptTuple));
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtState
* Purpose:   This routine constructs a high level marking of a net
*            from a low level marking.
* Arguments:
* Returns:   the high level marking
* See also:
*
****************************************************************************
*/
FE0(STATE *, PtState)
{
    ptState = StaAlloc();
    for (ptPlInd = 0; ptPlInd < MagGNofPlaces(); ptPlInd++)
	StaIMarking(ptState, ptPlInd, MarCreate());
    (void) PtrBlApply(TblPlc(), PlcState);
    return(ptState);
}
