/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptelem3.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This is an auxiliary module of ptnet.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"
#include "error.h"
#include "cmdnm.h"

/***************************************************************************
*
* Function:  PtInit
* Purpose:   This routine initializes the data structures needed in
*            unfolding and graph generation according to information
*            read from a binary file.
* Arguments: file - the file
* Returns:
* See also:  pplst_dat (pplst.c), BaseInit (encode.c),
*            InitialState (encode.c), PtBaseInit
*
****************************************************************************
*/
FE1(void, PtInit,
    BINARYFILE, file)
{
    unsigned p = 0;
    unsigned q;
    unsigned r;
    unsigned len = 0;
    DATATYPE *buffer;
    register DATATYPE *pointer;
    register unsigned i;
    register unsigned long count;
    register UDOMAIN value;
    DOMAIN monState;
    TUPLE *tuple;
    PLC *plc, *pl2c;

    PtBaseInit();
    while (TRUE)
	{
	    ReadUns(&len, 1, file);
	    if (len == (unsigned) -1)
		break;
	    buffer = (DATATYPE *) AllocArray(len + 1, sizeof(DATATYPE));
	    ReadData(buffer, len, file);
	    pointer = buffer;
	    GetValue(unsigned, i, pointer);
	    if (AriAt(pointer))
		{
		    AriPass(pointer);
		    GetValue(unsigned, i, pointer);
		}
	    else
		i = MagGProfile(i);
	    if (i)
		{
		    while (i--)
			SkipValue(pointer);
		    if (CntAt(pointer))
			{
			    *pointer = (DATATYPE) 0;
			    CntPass(pointer);
			    GetValue(unsigned long, count, pointer);
			    PlcSCount(GetPlc(buffer), count + 2L);
			}
		    else
			PlcSCount(GetPlc(buffer), (unsigned long) 1L);
		}
	    else
		{
		    GetValue(unsigned long, count, pointer);
		    PlcSCount(GetPlc(buffer), count + 1L);
		}
	    Deallocate((char *) buffer);
	}
    if (!OnTheFly())
	return;
    tuple = TupCons((unsigned) 1);
    while (TRUE)
	{
	    ReadUns(&len, 1, file);
	    if (len == (unsigned) -1)
		break;
	    SetOnTheFly();
	    buffer = (DATATYPE *) AllocArray(len + 1, sizeof(DATATYPE));
	    ReadData(buffer, len, file);
	    pointer = buffer;
	    GetValue(UDOMAIN, value, pointer);
	    monState = Decode((DOMAIN) value);
	    TupSField(tuple, (unsigned) 0, monState);
	    (void) TupEncode(MonPlInd(), tuple, EngABuffer(), EngABufSize());
	    plc = GetPlc(EngBuffer());
	    GetValue(unsigned, i, pointer);
	    if (i & 1)
		{
		    PlcSReject(plc);
		    SetReject();
		}
	    if (i & 2)
		{
		    PlcSDeadlock(plc);
		    SetStabFail();
		}
	    if (i & 4)
		{
		    PlcSLivelock(plc);
		    SetDivergence();
		}
	    if (i & 8)
		{
		    PlcSInfinite(plc);
		    SetInfinite();
		}
	    if (PlcGCount(plc))
		PtSFlyPlc(plc);
	    Deallocate((char *) buffer);
	}
    ReadUns(&p, 1, file);
    if (!p)
	{
	    TupDiscard(tuple);
	    return;
	}
    SetBuchi();
    for (q = 0; q < p; q++)
	{
	    TupSField(tuple, (unsigned) 0, (DOMAIN) (q + 1));
	    (void) TupEncode(MonPlInd(), tuple, EngABuffer(), EngABufSize());
	    plc = GetPlc(EngBuffer());
	    PlcSBuchi(plc);
	    PlcSNext(plc);
	    if (PlcGCount(plc))
		PtSFlyPlc(plc);
	    ReadUns(&len, 1, file);
	    buffer = (DATATYPE *) AllocArray(len + 1, sizeof(DATATYPE));
	    ReadData(buffer, len, file);
	    pointer = buffer;
	    GetValue(unsigned, i, pointer);
	    r = i;
	    while (r--)
		{
		    GetValue(UDOMAIN, value, pointer);
		    monState = Decode((DOMAIN) value);
		    TupSField(tuple, (unsigned) 0, monState);
		    (void) TupEncode(MonPlInd(), tuple,
				     EngABuffer(), EngABufSize());
		    pl2c = GetPlc(EngBuffer());
		    (void) LstInsLast(PlcGNext(plc), (char *) pl2c);
		}
	    Deallocate((char *) buffer);
	}
    TupDiscard(tuple);
    p--;
    _z_ = (unsigned long *) AllocArray(p, sizeof(unsigned long));
    for (q = 0; q < p; q++)
	ReadLng(_z_ + q, 1, file);
}

FE0(void, PtBaseInit)
{
    unsigned class;

    _nTrVisible = 0L;
    for (class = 0; class < 4; class++)
	_ptSlpSet[class] = NULL;
    _plcBalTree = BalTrCreate(&plcBalTools);
    _trnBalTree = (BALTREE **) AllocArray(IngGNofClasses(), sizeof(BALTREE *));
    for (class = 0; class < IngGNofClasses(); class++)
	_trnBalTree[class] = BalTrCreate(&trnBalTools);
}

FE1(void, SlpDis,
    unsigned, i)
{
    unsigned j;

    for (j = 0; j < 4; j++)
	{
	    if ((j != i) && (_ptSlpSet[j] == _ptSlpSet[i]))
		_ptSlpSet[j] = (BALTREE *) NULL;
	}
    BalTrDestroy(_ptSlpSet[i], TRUE);
}

FE0(void, PtClose)
{
    unsigned class;

    ClDelegate();
    SymmClose();
    for (class = 0; class < 4; class++)
	SlpDis(class);
    PtrBlDestroy(_tblPlc, TRUE);
    PtrBlDestroy(_dupTblPlc, TRUE);
    BalTrDestroy(_plcBalTree, TRUE);
    PtrBlDestroy(visibles, FALSE);
    PtrBlDestroy(candidates, FALSE);
    ClDeallocate(candInd);
    if (_visBlock)
	{
	    for (class = 0; _visBlock[class]; class++)
		LstDestroy(_visBlock[class], FALSE);
	    ClDeallocate(_visBlock);
	}
    ClDeallocate(_visColour);
    PtrBlDestroy(_tblTrn, TRUE);
    if (_trnBalTree)
	{
	    for (class = 0; class < IngGNofClasses(); class++)
		BalTrDestroy(_trnBalTree[class], TRUE);
	    ClDeallocate(_trnBalTree);
	}
}

/***************************************************************************
*
* Function:  PutInArc
* Purpose:   This routine introduces an input arc for a transition.
* Arguments: trn - the transition
*            weight - the weight of the arc
*            plIndex - the high level place corresponding to the source
*                      of the arc
*            tuple - the tuple corresponding to the source of the arc
* Returns:
* See also:  PutOutArc
*
****************************************************************************
*/
FE4(void, PutInArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
{
    PLC *plc;
    TRARC *trArc;

    (void) TupEncode(plIndex, tuple, EngABuffer(), EngABufSize());
    plc = GetPlc(EngBuffer());
    if ((!PlcGPurOut(plc)) || LstEmpty(PlcGPurOut(plc)) ||
	(PlArcGTrn((PLARC *) LstFirst(PlcGPurOut(plc))) != trn))
	{
	    if (!PlcGPurOut(plc))
		PlcSPurOut(plc);
	    (void) LstInsFirst(PlcGPurOut(plc), (char *)
			       PlArcCons(trn, weight));
	    if (!TrnGPurIn(trn))
		TrnSPurIn(trn);
	    (void) LstInsLast(TrnGPurIn(trn), (char *)
			      TrArcCons(plc, weight));
	    return;
	}
    PlArcIWeight((PLARC *) LstFirst(PlcGPurOut(plc)), weight);
    LSTFOR (TrnGPurIn(trn), trArc, TRARC)
	{
	    if (TrArcGPlc(trArc) == plc)
		break;
	}
    TrArcIWeight(trArc, weight);
}

/***************************************************************************
*
* Function:  PutOutArc
* Purpose:   This routine introduces an input arc for a transition.
* Arguments: trn - the transition
*            weight - the weight of the arc
*            plIndex - the high level place corresponding to the source
*                      of the arc
*            tuple - the tuple corresponding to the source of the arc
* Returns:
* See also:  PutInArc
*
****************************************************************************
*/
FE4(void, PutOutArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
{
    PLC *plc;
    TRARC *trArc;

    (void) TupEncode(plIndex, tuple, EngABuffer(), EngABufSize());
    plc = GetPlc(EngBuffer());
    if ((!PlcGPurIn(plc)) || LstEmpty(PlcGPurIn(plc)) ||
	(PlArcGTrn((PLARC *) LstFirst(PlcGPurIn(plc))) != trn))
	{
	    if (!PlcGPurIn(plc))
		PlcSPurIn(plc);
	    (void) LstInsFirst(PlcGPurIn(plc), (char *)
			       PlArcCons(trn, weight));
	    if (!TrnGPurOut(trn))
		TrnSPurOut(trn);
	    (void) LstInsLast(TrnGPurOut(trn), (char *)
			      TrArcCons(plc, weight));
	    return;
	}
    PlArcIWeight((PLARC *) LstFirst(PlcGPurIn(plc)), weight);
    LSTFOR (TrnGPurOut(trn), trArc, TRARC)
	{
	    if (TrArcGPlc(trArc) == plc)
		break;
	}
    TrArcIWeight(trArc, weight);
}

/***************************************************************************
*
* Function:  TrnReform
* Purpose:   This routine "rationalizes" the presentation of the arcs
*            connected to a transition.
* Arguments: trn - the transition
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, TrnReform,
    TRN *, trn)
{
    TRARC *trInArc, *trOutArc;
    PLC *plc;
    VOID *inCursor, *outCursor;
    unsigned long inWeight, outWeight;

    if ((!TrnGPurIn(trn)) || (!TrnGPurOut(trn)))
	return;
    inCursor = NULL;
    LSTFOR (TrnGPurIn(trn), trInArc, TRARC)
	{
	    plc = TrArcGPlc(trInArc);
	    if ((!PlcGPurIn(plc)) || LstEmpty(PlcGPurIn(plc)) ||
		(PlArcGTrn((PLARC *) LstFirst(PlcGPurIn(plc))) != trn))
		{
		    inCursor = LstGet(TrnGPurIn(trn));
		    continue;
		}
	    inWeight = TrArcGWeight(trInArc);
	    outCursor = NULL;
	    LSTFOR (TrnGPurOut(trn), trOutArc, TRARC)
		{
		    if (TrArcGPlc(trOutArc) == plc)
			break;
		    outCursor = LstGet(TrnGPurOut(trn));
		}
	    outWeight = TrArcGWeight(trOutArc);
	    LstSet(TrnGPurIn(trn), inCursor);
	    (void) LstDelNext(TrnGPurIn(trn), TRUE);
	    if (LstEmpty(TrnGPurIn(trn)))
		TrnDPurIn(trn);
	    (void) LstDelFirst(PlcGPurOut(plc), TRUE);
	    if (LstEmpty(PlcGPurOut(plc)))
		PlcDPurOut(plc);
	    LstSet(TrnGPurOut(trn), outCursor);
	    if (inWeight == outWeight)
		{
		    if (!TrnGNeutral(trn))
			TrnSNeutral(trn);
		    (void) LstInsLast(TrnGNeutral(trn), (char *)
				      LstDelNext(TrnGPurOut(trn), FALSE));
		    if (!PlcGNeutral(plc))
			PlcSNeutral(plc);
		    (void) LstInsLast(PlcGNeutral(plc), (char *)
				      LstDelFirst(PlcGPurIn(plc), FALSE));
		}
	    else
		{
		    (void) LstDelNext(TrnGPurOut(trn), TRUE);
		    (void) LstDelFirst(PlcGPurIn(plc), TRUE);
		}
	    if (LstEmpty(TrnGPurOut(trn)))
		TrnDPurOut(trn);
	    if (LstEmpty(PlcGPurIn(plc)))
		PlcDPurIn(plc);
	    if (inWeight < outWeight)
		{
		    if (!TrnGMosOut(trn))
			TrnSMosOut(trn);
		    (void) LstInsLast(TrnGMosOut(trn), (char *)
				      TrDrcCons(plc, inWeight, outWeight));
		    if (!PlcGMosIn(plc))
			PlcSMosIn(plc);
		    (void) LstInsLast(PlcGMosIn(plc), (char *)
				      PlDrcCons(trn, outWeight, inWeight));
		}
	    else if (inWeight > outWeight)
		{
		    if (!TrnGMosIn(trn))
			TrnSMosIn(trn);
		    (void) LstInsLast(TrnGMosIn(trn), (char *)
				      TrDrcCons(plc, inWeight, outWeight));
		    if (!PlcGMosOut(plc))
			PlcSMosOut(plc);
		    (void) LstInsLast(PlcGMosOut(plc), (char *)
				      PlDrcCons(trn, outWeight, inWeight));
		}
	    if ((!TrnGPurIn(trn)) || (!TrnGPurOut(trn)))
		return;
	    inCursor = LstGet(TrnGPurIn(trn));
	}
}

static unsigned long fillInd;

/***************************************************************************
*
* Function:  PlcToTbl
* Purpose:   This is an auxiliary routine of PlcFill.
* Arguments: plc - a place
* Returns:
* See also:  PlcFill
*
****************************************************************************
*/
FS1(BOOL, PlcToTbl,
    PLC *, plc)
{
    if (PtSymm())
	PlcSInd(plc, fillInd);
#ifndef DEBUG
    if ((!Pep()) && (!Buchi()) && ((!PtSymm()) || !PtVerbose()))
	PlcDKey(plc);
#endif
    PtrBlPut(TblPlc(), (char *) plc, fillInd);
    fillInd++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  PlcFill
* Purpose:   This routine moves places from a tree to a segmented table.
* Arguments:
* Returns:
* See also:  TrnFill
*
****************************************************************************
*/
FE0(void, PlcFill)
{
    BOOL delAlg;

    _tblPlc = PtrBlCreate(BalTrGElCount(PlcBalTree()), &plcTools);
    fillInd = 0L;
    (void) BalTrApply(PlcBalTree(), PlcToTbl);
    if (PtSymm())
	{
	    delAlg = PtDelAlg();
	    if (delAlg)
		ClPtDelAlg();
	    _dupTblPlc = PtrBlCopy(_tblPlc, TRUE);
	    if (delAlg)
		SetPtDelAlg();
	}
    BalTrDestroy(PlcBalTree(), FALSE);
}

/***************************************************************************
*
* Function:  TrnToTbl
* Purpose:   This is an auxiliary routine of TrnFill.
* Arguments: trn - a transition
* Returns:
* See also:  TrnFill
*
****************************************************************************
*/
FS1(BOOL, TrnToTbl,
    TRN *, trn)
{
#ifndef DEBUG
    TrnDKey(trn);
#endif
    PtrBlPut(TblTrn(), (char *) trn, fillInd);
    TrnSNumber(trn, fillInd);
    fillInd++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnFill
* Purpose:   This routine moves places from a tree to a segmented table.
* Arguments:
* Returns:
* See also:  PlcFill
*
****************************************************************************
*/
FE0(void, TrnFill)
{
    unsigned class;

    for (class = 0, fillInd = 0L; class < IngGNofClasses(); class++)
	fillInd += BalTrGElCount(TrnBalTree(class));
    _tblTrn = PtrBlCreate(fillInd, &trnTools);
    for (class = 0, fillInd = 0L; class < IngGNofClasses(); class++)
	{
	    (void) BalTrApply(TrnBalTree(class), TrnToTbl);
	    BalTrDestroy(TrnBalTree(class), FALSE);
	}
    ClDeallocate(_trnBalTree);
}

static BINARYFILE ptFile = NonBinaryFile();

/***************************************************************************
*
* Function:  PlcOutput
* Purpose:   This is an auxiliary routine of PtOutput.
* Arguments: plc - a place
* Returns:
* See also:  PtOutput
*
****************************************************************************
*/
FS1(BOOL, PlcOutput,
    PLC *, plc)
{
    unsigned len;

    len = DataLen(PlcGKey(plc));
    WriteUns(&len, 1, ptFile);
    WriteData(PlcGKey(plc), len, ptFile);
    return(FALSE);
}

static unsigned trKeyLen;

/***************************************************************************
*
* Function:  TrnOutput
* Purpose:   This is an auxuiliary routine of PtOutput.
* Arguments: trn - a transition
* Returns:
* See also:  PtOutput
*
****************************************************************************
*/
FS1(BOOL, TrnOutput,
    TRN *, trn)
{
    trKeyLen = DataLen(TrnGKey(trn));
    WriteUns(&trKeyLen, 1, ptFile);
    WriteData(TrnGKey(trn), trKeyLen, ptFile);
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtOutput
* Purpose:   This routine writes to a binary file the folding needed
*            for interpreting the reachability graph of an unfolded net.
* Arguments: file - the file
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, PtOutput,
    BINARYFILE, file)
{
    unsigned class;
    unsigned long a;

    ptFile = file;
    a = BalTrGElCount(PlcBalTree());
    WriteLng(&a, 1, ptFile);
    (void) BalTrApply(PlcBalTree(), PlcOutput);
    a = 0L;
    for (class = 0; class < IngGNofClasses(); class++)
	a += BalTrGElCount(TrnBalTree(class));
    WriteLng(&a, 1, ptFile);
    for (class = 0; class < IngGNofClasses(); class++)
	(void) BalTrApply(TrnBalTree(class), TrnOutput);
}

/***************************************************************************
*
* Function:  TrAraOutput
* Purpose:   This is an auxiliary routine of PtAraOutput.
* Arguments: trn - a transition
* Returns:
* See also:  PtAraOutput
*
****************************************************************************
*/
FS1(BOOL, TrAraOutput,
    TRN *, trn)
{
    trNumber++;
    if (TrnVisible(trn))
	InsOutput(trNumber, TrnGKey(trn), AraFile(), FALSE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtAraOutput
* Purpose:   This routine writes the action declarations to an ARA LTS.
*            An action corresponds to a transition instance.
* Arguments:
* Returns:
* See also:  LTSBegin, LTSEnd, TrnFire (ptelem.c), InsOutput (encode.c),
*            AraStep (inst.c)
*
****************************************************************************
*/
FE0(void, PtAraOutput)
{
    unsigned class;

    trNumber = 0L;
    for (class = 0; class < IngGNofClasses(); class++)
	(void) BalTrApply(TrnBalTree(class), TrAraOutput);
    (void) fprintf(AraFile(), "END_SYMBOLS\nARCS\n");
}

static unsigned long plNumber;

FS1(BOOL, PlcPepOutput,
    PLC *, plc)
{
    plNumber++;
    (void) fprintf(PepFile(), "%lu\"", plNumber);
    PlKeyOut(PlcGKey(plc), PepFile(), TRUE);
    (void) fprintf(PepFile(), "\"0@0M%lum%luk-1\n",
		   PlcGCount(plc), PlcGCount(plc));
#ifndef DEBUG
    if ((!Buchi()) && ((!PtSymm()) || !PtVerbose()))
	PlcDKey(plc);
#endif
    return(FALSE);
}

FS1(BOOL, PlcPepIncrease,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;

    plNumber++;
    if (!PlcBuchi(plc))
	{
	    if (PlcGPurIn(plc))
		LSTFOR (PlcGPurIn(plc), plArc, PLARC)
		    (void) fprintf(PepFile(), "%lu<%luw%lu\n",
				   TrnGNumber(PlArcGTrn(plArc)) + 1L,
				   plNumber, PlArcGWeight(plArc));
	    if (PlcGNeutral(plc))
		LSTFOR (PlcGNeutral(plc), plArc, PLARC)
		    (void) fprintf(PepFile(), "%lu<%luw%lu\n",
				   TrnGNumber(PlArcGTrn(plArc)) + 1L,
				   plNumber, PlArcGWeight(plArc));
	    if (PlcGMosIn(plc))
		LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
		    (void) fprintf(PepFile(), "%lu<%luw%lu\n",
				   TrnGNumber(PlDrcGTrn(plDrc)) + 1L,
				   plNumber, PlDrcGIn(plDrc));
	    if (PlcGMosOut(plc))
		LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
		    (void) fprintf(PepFile(), "%lu<%luw%lu\n",
				   TrnGNumber(PlDrcGTrn(plDrc)) + 1L,
				   plNumber, PlDrcGIn(plDrc));
	}
    return(FALSE);
}

FS1(BOOL, PlcPepDecrease,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;

    plNumber++;
    if (!PlcBuchi(plc))
	{
	    if (PlcGPurOut(plc))
		LSTFOR (PlcGPurOut(plc), plArc, PLARC)
		    (void) fprintf(PepFile(), "%lu>%luw%lu\n", plNumber,
				   TrnGNumber(PlArcGTrn(plArc)) + 1L,
				   PlArcGWeight(plArc));
	    if (PlcGNeutral(plc))
		LSTFOR (PlcGNeutral(plc), plArc, PLARC)
		    (void) fprintf(PepFile(), "%lu>%luw%lu\n", plNumber,
				   TrnGNumber(PlArcGTrn(plArc)) + 1L,
				   PlArcGWeight(plArc));
	    if (PlcGMosIn(plc))
		LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
		    (void) fprintf(PepFile(), "%lu>%luw%lu\n", plNumber,
				   TrnGNumber(PlDrcGTrn(plDrc)) + 1L,
				   PlDrcGOut(plDrc));
	    if (PlcGMosOut(plc))
		LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
		    (void) fprintf(PepFile(), "%lu>%luw%lu\n", plNumber,
				   TrnGNumber(PlDrcGTrn(plDrc)) + 1L,
				   PlDrcGOut(plDrc));
	}
    return(FALSE);
}

FS1(BOOL, TrPepOutput,
    TRN *, trn)
{
    TrnSNumber(trn, trNumber);
    trNumber++;
    InsOutput(trNumber, TrnGKey(trn), PepFile(), TRUE);
    return(FALSE);
}

FE1(void, PtPepOutput,
    char *, name)
{
    char *str;
    unsigned class;
    FILE *fp;

    str = FileName(name,
#ifdef UNIX
		   "ll_net"
#else
		   "pep"
#endif
		   );
    fp = MyFopen(str, "w", (BOOL *) NULL);
    ClDeallocate(str);
    if (!fp)
	Exit(12);
    SetPepFile(fp);
    (void) fputs("PEP\nPetriBox\nFORMAT_N\nPL\n", PepFile());
    plNumber = 0L;
    (void) PtrBlApply(TblPlc(), PlcPepOutput);
    (void) fputs("TR\n", PepFile());
    trNumber = 0L;
    for (class = 0; class < IngGNofClasses(); class++)
	(void) BalTrApply(TrnBalTree(class), TrPepOutput);
    (void) fputs("TP\n", PepFile());
    plNumber = 0L;
    (void) PtrBlApply(TblPlc(), PlcPepIncrease);
    (void) fputs("PT\n", PepFile());
    plNumber = 0L;
    (void) PtrBlApply(TblPlc(), PlcPepDecrease);
    MyFclose(PepFile());
}

/***************************************************************************
*
* Function:  PlcClear
* Purpose:   This routine sets a place to have no token.
* Arguments: plc - the place
* Returns:
* See also:
*
****************************************************************************
*/
FS1(BOOL, PlcClear,
    PLC *, plc)
{
    PlcSCount(plc, 0L);
    return(FALSE);
}

static unsigned plcBufUsed;
static DATATYPE *plcBuffer;

/***************************************************************************
*
* Function:  PlcBfUsed
* Purpose:   This is an auxiliary routine of PlcEncode.
* Arguments: plc - a place
* Returns:
* See also:  PlcEncode
*
****************************************************************************
*/
FS1(BOOL, PlcBfUsed,
    PLC *, plc)
{
    if (PlcGCount(plc) != 0L)
	{
	    register unsigned long a;
	    register unsigned len;

	    len = 0;
	    a = plNumber;
	    IncLen(a, len);
	    if (PlcGCount(plc) != 1L)
		{
		    len++;
		    a = PlcGCount(plc) - 2L;
		    IncLen(a, len);
		}
	    plcBufUsed += len;
	}
    plNumber++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  PlcWrite
* Purpose:   This is an auxiliary routine of PlcEncode.
* Arguments: plc - a place
* Returns:
* See also:  PlcEncode
*
****************************************************************************
*/
FS1(BOOL, PlcWrite,
    PLC *, plc)
{
    if (PlcGCount(plc) != 0L)
	{
	    register DATATYPE *pointer;
	    register unsigned long a;

	    if (PlcMonitor(plc))
		PtSFlyPlc(plc);
	    pointer = plcBuffer;
	    a = plNumber;
	    PutValue(a, pointer);
	    if (PlcGCount(plc) != 1L)
		{
		    CntCode(pointer);
		    a = PlcGCount(plc) - 2L;
		    PutValue(a, pointer);
		}
	    plcBuffer = pointer;
	}
    plNumber++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  PlcIniFlag
* Purpose:   This routine cleans the flag of a place so that the
*            construction of an and/or-graph can be started in the
*            deletion algorithm of the stubborn set method.
* Arguments: plc - the place
* Returns:
* See also:  TrnPreAO (ptnet2.c)
*
****************************************************************************
*/
FE1(BOOL, PlcIniFlag,
    PLC *, plc)
{
    PlcReset(plc);
#ifdef SMODELS
    if (PtSmodels())
	{
	    PlcCGamma(plc);
	    PlcCIneqs(plc);
	}
    if (PtLongPlay())
	{
	    PlcCOutMax(plc);
	    if (PlcGPurOut(plc))
		{
		    LSTFOR (PlcGPurOut(plc), ptPlArc, PLARC)
			{
			    if (PlArcGWeight(ptPlArc) > PlcGOutMax(plc))
				PlcSOutMax(plc, PlArcGWeight(ptPlArc));
			}
		}
	    if (PlcGNeutral(plc))
		{
		    LSTFOR (PlcGNeutral(plc), ptPlArc, PLARC)
			{
			    if (PlArcGWeight(ptPlArc) > PlcGOutMax(plc))
				PlcSOutMax(plc, PlArcGWeight(ptPlArc));
			}
		}
	    if (PlcGMosOut(plc))
		{
		    LSTFOR (PlcGMosOut(plc), ptPlDrc, PLDRC)
			{
			    if (PlDrcGOut(ptPlDrc) > PlcGOutMax(plc))
				PlcSOutMax(plc, PlDrcGOut(ptPlDrc));
			}
		}
	    if (PlcGMosIn(plc))
		{
		    LSTFOR (PlcGMosIn(plc), ptPlDrc, PLDRC)
			{
			    if (PlDrcGOut(ptPlDrc) > PlcGOutMax(plc))
				PlcSOutMax(plc, PlDrcGOut(ptPlDrc));
			}
		}
	}
#endif
    return(FALSE);
}

/***************************************************************************
*
* Function:  PlcEncode
* Purpose:   This routine encodes a marking to a buffer.
* Arguments: tbl - a table of all places (either TblPLc() or DupTblPlc())
*            *buffer - the buffer (extended if needed)
*            *size - the size of the buffer
* Returns:   the length of the encoding
* See also:
*
****************************************************************************
*/
FE3(unsigned, PlcEncode,
    PTRBL *, tbl,
    DATATYPE **, buffer,
    unsigned *, size)
{
    PtSFlyPlc((PLC *) NULL);
    plcBufUsed = 0;
    plNumber = 0L;
    (void) PtrBlApply(tbl, PlcBfUsed);
    if (plcBufUsed >= *size)
	{
	    *size = (plcBufUsed >> 1) + plcBufUsed + 1;
	    Deallocate((char *) *buffer);
	    *buffer = (DATATYPE *) AllocArray(*size, sizeof(DATATYPE));
	}
    plcBuffer = *buffer;
    plNumber = 0L;
    (void) PtrBlApply(tbl, PlcWrite);
    *plcBuffer = (DATATYPE) 0;
    return(plcBufUsed);
}

static PLC *decPlc;

/***************************************************************************
*
* Function:  PlcDecode
* Purpose:   This routine decodes an encoded marking. TblPlc() then has
*            the marking.
* Arguments: buffer - this contains the encoding
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, PlcDecode,
    DATATYPE *, buffer)
{
    register DATATYPE *pointer;
    register unsigned long a;

    PtSFlyPlc((PLC *) NULL);
    (void) PtrBlApply(TblPlc(), PlcClear);
    pointer = buffer;
    while (*pointer)
	{
	    GetValue(unsigned long, a, pointer);
	    if (a >= PtrBlGCount(TblPlc()))
		Troba3("Illegal data for PlcDecode: %lu >= %lu",
		       a, PtrBlGCount(TblPlc()));
	    decPlc = (PLC *) PtrBlGet(TblPlc(), a);
	    if (CntAt(pointer))
		{
		    CntPass(pointer);
		    GetValue(unsigned long, a, pointer);
		    PlcSCount(decPlc, a + 2L);
		}
	    else
		{
		    PlcSCount(decPlc, (unsigned long) 1L);
		    if (PlcMonitor(decPlc))
			PtSFlyPlc(decPlc);
		}
	}
}

/***************************************************************************
*
* Function:  TrnPrec
* Purpose:   This routine is needed in the unfolding phase if there are
*            more than one precedence class.
* Arguments: trn - a transition
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BOOL, TrnPrec,
    TRN *, trn)
{
    if (FdEqual(TrnGKey(LowTrn()), TrnGKey(trn)))
	{
	    if (!TrnGHigher(LowTrn()))
		TrnSHigher(LowTrn());
	    (void) LstInsLast(TrnGHigher(LowTrn()), (char *) trn);
	    if (!TrnGLower(trn))
		TrnSLower(trn);
	    (void) LstInsLast(TrnGLower(trn), (char *) LowTrn());
	}
    return(FALSE);
}

static UNS32 ptTerm;

/***************************************************************************
*
* Function:  PtSRandom
* Purpose:   This routine sets the seed value of the pseudo-random
*            number generator of PROD.
* Arguments: seedVal - the seed value
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, PtSRandom,
    UNS32, seedVal)
{
    ptTerm = seedVal;
}

/***************************************************************************
*
* Function:  PtTerm
* Purpose:   This routine returns the current value in the sequence
*            from which pseudo-random numbers are calculated.
* Arguments:
* Returns:   the value described above
* See also:
*
****************************************************************************
*/
FE0(UNS32, PtTerm)
{
    return(ptTerm);
}

/***************************************************************************
*
* Function:  PtRandom
* Purpose:   This routine returns a pseudo-random number
* Arguments: domSize - any number less than domSize is an acceptable
*                      pseudo-random number
* Returns:   an acceptable pseudo-random number if domSize is not 0,
*            and 0 otherwise
* See also:
*
****************************************************************************
*/
FE1(UNS32, PtRandom,
    UNS32, domSize)
{
    register UNS32 p0, p1, q0, q1, r;

    ptTerm = ptTerm * 1103515245 + 12345;
    p0 = ptTerm & 65535;
    p1 = ptTerm >> 16;
    q0 = domSize & 65535;
    q1 = domSize >> 16;
    r = p0 * q1 + p1 * q0;
    if (((UNS32) (p0 * q1)) <= (UNS32) (MAX32 - p1 * q0))
	return((((UNS32) (p0 * q0)) <= (UNS32) (MAX32 - ((r << 16) & MAX32)))
	       ? ((r >> 16) + p1 * q1)
	       : ((r >> 16) + p1 * q1 + 1L));
    else
	return((((UNS32) (p0 * q0)) <= (UNS32) (MAX32 - ((r << 16) & MAX32)))
	       ? ((r >> 16) + p1 * q1 + 65536)
	       : ((r >> 16) + p1 * q1 + 65537));
}
