/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet.h
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptandor.c, ptelem,c, ptnet.c and ptvisit.c.
*
* Version history:
*     $Log$
***************************************************************************/

typedef struct MOTA
{
    unsigned _flag;
    unsigned long _counter;
    unsigned long _dupCounter;
    VOID *_cursor;
    struct MOTA *_down;
    LIST *_preds;
#ifdef DEBUG
    char *_label;
#endif
} MOTA;

#define MotaAlloc() ((MOTA *) Allocate(sizeof(MOTA)))
DE1(void, MotaDracsid,
    MOTA *, mota)
#define MotaDiscard(mota) (MotaDracsid(mota), (mota) = (MOTA *) NULL)
#ifdef DEBUG
DE2(void, MotaDump,
    MOTA *, mota,
    unsigned, depth)
#endif
extern TOOLS motaTools;
DE1(void, MotaClear,
    MOTA *, mota)
#define MotaGFlag(mota) ((mota)->_flag)
#define MotaCleFlag(mota) ((mota)->_flag = 0)
#define MotaSatisfied(mota) ((mota)->_flag & 1)
#define MotaFeed(mota) ((mota)->_flag |= 1)
#define MotaImpoverish(mota) ((mota)->_flag &= ~1)
#define MotaThirsty(mota) ((mota)->_flag & 2)
#define MotaSalt(mota) ((mota)->_flag |= 2)
#define MotaWater(mota) ((mota)->_flag &= ~2)
#define MotaAsleep(mota) ((mota)->_flag & 4)
#define MotaPutToSleep(mota) ((mota)->_flag |= 4)
#define MotaWake(mota) ((mota)->_flag &= ~4)
#define MotaOld(mota) ((mota)->_flag & 8)
#define MotaEnter(mota) ((mota)->_flag |= 8)
#define MotaLeave(mota) ((mota)->_flag &= ~8)
#define MotaEnabled(mota) ((mota)->_flag & 16)
#define MotaAccept(mota) ((mota)->_flag |= 16)
#define MotaReject(mota) ((mota)->_flag &= ~16)
#define MotaIsProtected(mota) ((mota)->_flag & 32)
#define MotaProtect(mota) ((mota)->_flag |= 32)
#define MotaUnprotect(mota) ((mota)->_flag &= ~32)
#define MotaSomewhere(mota) ((mota)->_flag & 64)
#define MotaPutSomewhere(mota) ((mota)->_flag |= 64)
#define MotaPutNowhere(mota) ((mota)->_flag &= ~64)
#define IsOrArity(n) (((unsigned long) 1L) < (unsigned long) (n))
#define MotaCounter(mota) ((mota)->_counter)
#define MotaDupCounter(mota) ((mota)->_dupCounter)
#define MotaSetCounter(mota,n) \
((mota)->_dupCounter = (mota)->_counter = (n))
#define MotaCleCounter(mota) MotaSetCounter(mota, (unsigned long) 0L)
#define MotaResCounter(mota) ((mota)->_counter = (mota)->_dupCounter)
#define MotaDecCounter(mota) ((mota)->_counter--)
#define MotaIncCounter(mota) ((mota)->_counter++)
#define MotaIsOr(mota) IsOrArity(MotaDupCounter(mota))
#define MotaGCursor(mota) ((mota)->_cursor)
#define MotaSetCursor(mota,cursor) ((mota)->_cursor = (cursor))
#define MotaCleCursor(mota) MotaSetCursor(mota, (VOID *) NULL)
#define MotaGDown(mota) ((mota)->_down)
#define MotaSetDown(mota,down) ((mota)->_down = (down))
#define MotaCleDown(mota) MotaSetDown(mota, (MOTA *) NULL)
#define MotaGPreds(mota) ((mota)->_preds)
#define MotaNilPreds(mota) ((mota)->_preds = (LIST *) NULL)
#define MotaIniPreds(mota) ((mota)->_preds = LstCreate(&motaTools))
#define MotaDesPreds(mota) LstDestroy((mota)->_preds, FALSE)
#ifdef DEBUG
#define MotaGLabel(mota) ((mota)->_label)
#define MotaSetLabel(mota,s) ((mota)->_label = (char *) (s))
#define MotaNilLabel(mota) MotaSetLabel(mota, NULL)
#define MotaDesLabel(mota) ClDeallocate((mota)->_label)
#endif

/* The elements of PLC are called low level places. A low level place
 * is a place in a place/transition net and corresponds to a pair
 * of a high level place (a place of a predicate/transition net) and
 * a tuple. */
typedef struct PLC
{
    DATATYPE *_key;        /* an encoding of the high level correspondent */
    unsigned _flag;
    unsigned long _count;  /* the number of tokens */
    union
	{
	    LIST *_pureIn; /* arcs from {t | W(t,p) > 0 = W(p,t)} */
	    LIST *_next;   /* if the place corresponds to a state of a
                            * Buchi automaton, then the place is isolated
                            * and next is a list of places corresponding
                            * to the successor states */
	} _lp;
    LIST *_pureOut; /* arcs to {t | W(t,p) = 0 < W(p,t)} */
    LIST *_neutral; /* arcs from and to {t | W(t,p) = W(p,t) > 0} */
    LIST *_mostIn;  /* arcs from and to {t | W(t,p) > W(p,t) > 0} */
    LIST *_mostOut; /* arcs from and to {t | 0 < W(t,p) < W(p,t)} */
    union
	{
	    unsigned long _ind; /* _ind indicates the number of the place
				 * but is used only during the computation
                                 * of place symmetries in the symmetry
                                 * method (it is thus possible to use
                                 * the symmetry method together with
                                 * the deletion algorithm of the stubborn
                                 * set method) */
#ifdef SMODELS
	    struct
		{
		    unsigned long _outMax;
		    VOID *_gamma;
		    BALTREE *_ineqs;
		} _sp;
#else
	    VOID *_gamma;
#endif
	} _up;
} PLC;

extern TOOLS ptrTools;
extern BALTOOLS ptrBalTools;

#define PlcAlloc() ((PLC *) Allocate(sizeof(PLC)))
extern TOOLS plcTools;
extern BALTOOLS plcBalTools;

#ifdef SMODELS
typedef struct INEQ
{
    unsigned long _card;
    VOID *_ge;
    VOID *_lt;
} INEQ;

#define IneqAlloc() ((INEQ *) Allocate(sizeof(INEQ)))
extern TOOLS ineqTools;
extern BALTOOLS ineqBalTools;

#define IneqGCard(ineq) ((ineq)->_card)
#define IneqCCard(ineq) ((ineq)->_card = (unsigned long) 0L)
#define IneqSCard(ineq,card) ((ineq)->_card = (card))
#define IneqACard(ineq) (&((ineq)->_card))
#define IneqGGe(ineq) ((ineq)->_ge)
#define IneqCGe(ineq) ((ineq)->_ge = (VOID *) NULL)
#define IneqSGe(ineq,ge) ((ineq)->_ge = (ge))
#define IneqAGe(ineq) (&((ineq)->_ge))
#define IneqGLt(ineq) ((ineq)->_lt)
#define IneqCLt(ineq) ((ineq)->_lt = (VOID *) NULL)
#define IneqSLt(ineq,lt) ((ineq)->_lt = (lt))
#define IneqALt(ineq) (&((ineq)->_lt))

DE1(INEQ *, IneqCons,
    unsigned long, card)
DE1(INEQ *, IneqCopy,
    INEQ *, ineq)
#ifdef DEBUG
DE2(void, IneqDump,
    INEQ *, ineq,
    unsigned, depth)
#endif
DE2(int, IneqKeyCmp,
    INEQ *, ineq,
    unsigned long *, key)
#endif

/* The elements of TRN are called low level transitions. A low level
 * transition is a transition in a place/transition net and corresponds
 * to an instance of a high level transition (a transition in a
 * predicate/transition net). */
typedef struct TRN
{
    DATATYPE *_key; /* an encoding of the high level correspondent */
    unsigned long _number; /* the order number of the low level transition
                            * (the numbering starts from 0) */
    unsigned long _flag;
    unsigned _where;
    VOID *_cursor;
    struct TRN *_down;
    unsigned *_block;
    LIST *_pureIn;  /* arcs from {p | W(p,t) > 0 = W(t,p)} */
    LIST *_pureOut; /* arcs to {p | W(p,t) = 0 < W(t,p)} */
    LIST *_neutral; /* arcs from and to {p | W(p,t) = W(t,p) > 0} */
    LIST *_mostIn;  /* arcs from and to {p | W(p,t) > W(t,p) > 0} */
    LIST *_mostOut; /* arcs from and to {p | 0 < W(p,t) < W(t,p)} */
    LIST *_higher; /* if there is more than one precedence class,
		    * then _higher is a list of those transitions
		    * that have a priority over this transition */
    LIST *_lower;  /* the inverse of _higher */
    LIST *_promoted;
    LIST *_defers;
    union
	{
	    union
		{
		    struct
			{
			    unsigned long _val; /* in the incremental
                                                 * algorithm of the stubborn
                                                 * set method, _val is
                                                 * a number associated with
                                                 * a transition vertex in an
                                                 * implicit dependency graph */
                            unsigned long _dupVal; /* _dupVal is used for
                                                    * storing and restoring
                                                    * of _val */
                            unsigned long _min;
			} _n;
		    struct
			{
			    VOID *_gamma;
			    VOID *_kappa;
			    VOID *_diRho;
			    VOID *_enRho;
			    VOID *_reRho;
#ifdef SMODELS
			    VOID *_plain;
			    VOID *_active;
			    VOID *_toStack;
#endif
			} _a;
		} _vt;
	} _ut;
} TRN;

#define TrnAlloc() ((TRN *) Allocate(sizeof(TRN)))
extern TOOLS trnTools;
extern BALTOOLS trnBalTools;

/* An element of PLARC is an arc or a zero-effect self-loop
 * connected to a place. */
typedef struct PLARC
{
    TRN *_trn;
    unsigned long _weight;
} PLARC;

#define PlArcAlloc() ((PLARC *) Allocate(sizeof(PLARC)))
extern TOOLS plArcTools;
extern BALTOOLS plArcBalTools;

/* An element of PLDRC is a non-zero-effect self-loop connected to a place. */
typedef struct PLDRC
{
    TRN *_trn;
    unsigned long _in;  /* W(t,p) */
    unsigned long _out; /* W(p,t) */
} PLDRC;

#define PlDrcAlloc() ((PLDRC *) Allocate(sizeof(PLDRC)))
extern TOOLS plDrcTools;
extern BALTOOLS plDrcBalTools;

/* An element of TRARC is an arc or a zero-effect self-loop connected
 * to a transition. */
typedef struct TRARC
{
    PLC *_plc;
    unsigned long _weight;
    VOID *_oder;
    VOID *_strong;
    VOID *_weak;
#ifdef SMODELS
    VOID *_gamma;
#endif
} TRARC;

#define TrArcAlloc() ((TRARC *) Allocate(sizeof(TRARC)))
extern TOOLS trArcTools;
extern BALTOOLS trArcBalTools;

/* An element of PLDRC is a non-zero-effect self-loop connected
 * to a transition. */
typedef struct TRDRC
{
    PLC *_plc;
    unsigned long _in;  /* W(p,t) */
    unsigned long _out; /* W(t,p) */
    VOID *_oder;
    VOID *_strong;
    VOID *_weak;
#ifdef SMODELS
    VOID *_gamma;
#endif
} TRDRC;

#define TrDrcAlloc() ((TRDRC *) Allocate(sizeof(TRDRC)))
extern TOOLS trDrcTools;
extern BALTOOLS trDrcBalTools;

typedef struct TRPROM
{
    TRN *_trn;
    VOID *_atom;
    LIST *_arcs;
} TRPROM;

#define TrPromAlloc() ((TRPROM *) Allocate(sizeof(TRPROM)))
extern TOOLS trPromTools;

typedef struct DEFER
{
    PLC *_plc;
    unsigned _flag;
    VOID *_atom;
} DEFER;

#define DeferAlloc() ((DEFER *) Allocate(sizeof(DEFER)))
extern TOOLS deferTools;
extern BALTOOLS deferBalTools;

typedef struct OTARC
{
    DEFER *_defer;
    unsigned long _weight;
} OTARC;

#define OtArcAlloc() ((OTARC *) Allocate(sizeof(OTARC)))
extern TOOLS otArcTools;

#define PlcGKey(plc) ((plc)->_key)
#define PlcAKey(plc) (&((plc)->_key))
#define PlcSKey(plc,key) ((plc)->_key = (key))
#define PlcCKey(plc) ((plc)->_key = (DATATYPE *) NULL)
#define PlcDKey(plc) ClDeallocate((plc)->_key)
#define PlcGInd(plc) ((plc)->_up._ind)
#define PlcAInd(plc) (&((plc)->_up._ind))
#define PlcSInd(plc,ind) ((plc)->_up._ind = (ind))
#define PlcGCount(plc) ((plc)->_count)
#define PlcACount(plc) (&((plc)->_count))
#define PlcSCount(plc,count) ((plc)->_count = (count))
#define PlcICount(plc,count) ((plc)->_count += (count))
#define PlcDCount(plc,count) ((plc)->_count -= (count))
#define PlcGPurIn(plc) ((plc)->_lp._pureIn)
#define PlcZPurIn(plc,list) ((plc)->_lp._pureIn = (list))
#define PlcSPurIn(plc) ((plc)->_lp._pureIn = LstCreate(&plArcTools))
#define PlcCPurIn(plc) ((plc)->_lp._pureIn = (LIST *) NULL)
#define PlcDPurIn(plc) LstDestroy((plc)->_lp._pureIn, TRUE)
#define PlcGNext(plc) ((plc)->_lp._next)
#define PlcZNext(plc,list) ((plc)->_lp._next = (list))
#define PlcSNext(plc) ((plc)->_lp._next = LstCreate(&plcTools))
#define PlcCNext(plc) ((plc)->_lp._next = (LIST *) NULL)
#define PlcDNext(plc) LstDestroy((plc)->_lp._next, TRUE)
#define PlcGPurOut(plc) ((plc)->_pureOut)
#define PlcZPurOut(plc,list) ((plc)->_pureOut = (list))
#define PlcSPurOut(plc) ((plc)->_pureOut = LstCreate(&plArcTools))
#define PlcCPurOut(plc) ((plc)->_pureOut = (LIST *) NULL)
#define PlcDPurOut(plc) LstDestroy((plc)->_pureOut, TRUE)
#define PlcGNeutral(plc) ((plc)->_neutral)
#define PlcZNeutral(plc,list) ((plc)->_neutral = (list))
#define PlcSNeutral(plc) ((plc)->_neutral = LstCreate(&plArcTools))
#define PlcCNeutral(plc) ((plc)->_neutral = (LIST *) NULL)
#define PlcGMosIn(plc) ((plc)->_mostIn)
#define PlcZMosIn(plc,list) ((plc)->_mostIn = (list))
#define PlcSMosIn(plc) ((plc)->_mostIn = LstCreate(&plDrcTools))
#define PlcCMosIn(plc) ((plc)->_mostIn = (LIST *) NULL)
#define PlcGMosOut(plc) ((plc)->_mostOut)
#define PlcZMosOut(plc,list) ((plc)->_mostOut = (list))
#define PlcSMosOut(plc) ((plc)->_mostOut = LstCreate(&plDrcTools))
#define PlcCMosOut(plc) ((plc)->_mostOut = (LIST *) NULL)
#ifdef SMODELS
#define PlcGOutMax(plc) ((plc)->_up._sp._outMax)
#define PlcSOutMax(plc,outMax) ((plc)->_up._sp._outMax = (outMax))
#define PlcCOutMax(plc) ((plc)->_up._sp._outMax = (unsigned long) 0L)
#define PlcGGamma(plc) ((plc)->_up._sp._gamma)
#define PlcSGamma(plc,gamma) ((plc)->_up._sp._gamma = (gamma))
#define PlcCGamma(plc) ((plc)->_up._sp._gamma = (VOID *) NULL)
#define PlcGIneqs(plc) ((plc)->_up._sp._ineqs)
#define PlcZIneqs(plc,tree) ((plc)->_up._sp._ineqs = (tree))
#define PlcSIneqs(plc) ((plc)->_up._sp._ineqs = BalTrCreate(&ineqBalTools))
#define PlcCIneqs(plc) ((plc)->_up._sp._ineqs = (BALTREE *) NULL)
#define PlcDIneqs(plc) BalTrDestroy((plc)->_up._sp._ineqs, TRUE)
#else
#define PlcGGamma(plc) ((plc)->_up._gamma)
#define PlcSGamma(plc,gamma) ((plc)->_up._gamma = (gamma))
#define PlcCGamma(plc) ((plc)->_up._gamma = (VOID *) NULL)
#endif
#define PlcGFlag(plc) ((plc)->_flag)
#define PlcSFlag(plc,flag) ((plc)->_flag = (flag))
#define PlcCFlag(plc) ((plc)->_flag = 0)
#define PlcSatisfied(plc) ((plc)->_flag & 1)
#define PlcFeed(plc) ((plc)->_flag |= 1)
#define PlcImpoverish(plc) ((plc)->_flag &= ~1)
#define PlcThirsty(plc) ((plc)->_flag & 2)
#define PlcSalt(plc) ((plc)->_flag |= 2)
#define PlcWater(plc) ((plc)->_flag &= ~2)
#define PlcReset(plc) ((plc)->_flag &= ~3)
#define PlcReject(plc) ((plc)->_flag & 4)
#define PlcSReject(plc) ((plc)->_flag |= 4)
#define PlcDeadlock(plc) ((plc)->_flag & 8)
#define PlcSDeadlock(plc) ((plc)->_flag |= 8)
#define PlcLivelock(plc) ((plc)->_flag & 16)
#define PlcSLivelock(plc) ((plc)->_flag |= 16)
#define PlcInfinite(plc) ((plc)->_flag & 32)
#define PlcSInfinite(plc) ((plc)->_flag |= 32)
#define PlcBuchi(plc) ((plc)->_flag & 64)
#define PlcSBuchi(plc) ((plc)->_flag |= 64)
#define PlcMonitor(plc) ((plc)->_flag & 124)

extern TRN *_dummyTrn;
#define DummyTrn() _dummyTrn
#define SetDummyTrn(trn) (_dummyTrn = (trn))

#define TrnGKey(trn) ((trn)->_key)
#define TrnAKey(trn) (&((trn)->_key))
#define TrnSKey(trn,key) ((trn)->_key = (key))
#define TrnCKey(trn) ((trn)->_key = (DATATYPE *) NULL)
#define TrnDKey(trn) ClDeallocate((trn)->_key)
#define TrnGNumber(trn) ((trn)->_number)
#define TrnANumber(trn) (&((trn)->_number))
#define TrnSNumber(trn,number) ((trn)->_number = (number))
#define TrnCNumber(trn) \
((trn)->_number = ((unsigned long) 0L))
#define TrnGFlag(trn) \
((trn)->_flag)
#define TrnSFlag(trn,flag) \
((trn)->_flag = (flag))
#define TrnCFlag(trn) \
((trn)->_flag = ((unsigned long) 0L))
#define TrnPlcEnabled(trn) \
((trn)->_flag & ((unsigned long) 1L))
#define TrnPlcAccept(trn) \
((trn)->_flag |= ((unsigned long) 1L))
#define TrnPlcReject(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 1L))
#define TrnPriEnabled(trn) \
((trn)->_flag & ((unsigned long) 2L))
#define TrnPriAccept(trn) \
((trn)->_flag |= ((unsigned long) 2L))
#define TrnPriReject(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 2L))
#define TrnEnabled(trn) \
(((trn)->_flag & ((unsigned long) 3L)) == ((unsigned long) 3L))
#define TrnSatisfied(trn) \
((trn)->_flag & ((unsigned long) 4L))
#define TrnFeed(trn) \
((trn)->_flag |= ((unsigned long) 4L))
#define TrnImpoverish(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 4L))
#define TrnThirsty(trn) \
((trn)->_flag & ((unsigned long) 8L))
#define TrnSalt(trn) \
((trn)->_flag |= ((unsigned long) 8L))
#define TrnWater(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 8L))
#define TrnAsleep(trn) \
((trn)->_flag & ((unsigned long) 16L))
#define TrnPutToSleep(trn) \
((trn)->_flag |= ((unsigned long) 16L))
#define TrnWake(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 16L))
#define TrnOld(trn) \
((trn)->_flag & ((unsigned long) 32L))
#define TrnEnter(trn) \
((trn)->_flag |= ((unsigned long) 32L))
#define TrnLeave(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 32L))
#define TrnIsDiRho(trn) \
((trn)->_flag & ((unsigned long) 64L))
#define TrnAckDiRho(trn) \
((trn)->_flag |= ((unsigned long) 64L))
#define TrnRejDiRho(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 64L))
#define TrnReset(trn) \
((trn)->_flag &= (unsigned long) ~((unsigned long) 127L))
#define TrnVisible(trn) \
((trn)->_flag & ((unsigned long) 128L))
#define TrnSVisible(trn) \
(_nTrVisible++, (trn)->_flag |= ((unsigned long) 128L))
#define TrnInitial(trn) \
((trn)->_flag & ((unsigned long) 256L))
#define TrnSInitial(trn) \
((trn)->_flag |= ((unsigned long) 256L))
#define TrnDummy(trn) \
((trn)->_flag & ((unsigned long) 512L))
#define TrnSDummy(trn) \
(SetDummyTrn(trn), (trn)->_flag |= ((unsigned long) 512L))
#define TrnIsGamma(trn) \
(((trn)->_flag & ((unsigned long) 768L)) == ((unsigned long) 0L))
#define TrnFlagStore(trn) \
((trn)->_flag = ((trn)->_flag & ((unsigned long) 1023L)) | \
(((trn)->_flag & ((unsigned long) 127L)) << 10))
#define TrnFlagRestore(trn) \
((trn)->_flag = ((trn)->_flag & ((unsigned long) 896L)) | \
(((trn)->_flag & ((unsigned long) 130048L)) >> 10))
#define TrnGWhere(trn) ((trn)->_where)
#define TrnSWhere(trn,where) ((trn)->_where = (where))
#define TrnGCursor(trn) ((trn)->_cursor)
#define TrnSCursor(trn,cursor) ((trn)->_cursor = (cursor))
#define TrnCCursor(trn) ((trn)->_cursor = (VOID *) NULL)
#define TrnGDown(trn) ((trn)->_down)
#define TrnSDown(trn,down) ((trn)->_down = (down))
#define TrnCDown(trn) ((trn)->_down = (TRN *) NULL)
#define TrnBlocks(trn) ((trn)->_block)
#define TrnCBlocks(trn) ((trn)->_block = (unsigned *) NULL)
#define TrnABlocks(trn,n) \
((trn)->_block = (unsigned *) AllocArray(n, sizeof(unsigned)))
#define TrnSBlock(trn,i,v) ((trn)->_block[i] = (v))
#define TrnGBlock(trn,i) ((trn)->_block[i])
#define TrnDBlocks(trn) ClDeallocate((trn)->_block)
#define TrnGPurIn(trn) ((trn)->_pureIn)
#define TrnZPurIn(trn,list) ((trn)->_pureIn = (list))
#define TrnSPurIn(trn) ((trn)->_pureIn = LstCreate(&trArcTools))
#define TrnCPurIn(trn) ((trn)->_pureIn = (LIST *) NULL)
#define TrnDPurIn(trn) LstDestroy((trn)->_pureIn, TRUE)
#define TrnGPurOut(trn) ((trn)->_pureOut)
#define TrnZPurOut(trn,list) ((trn)->_pureOut = (list))
#define TrnSPurOut(trn) ((trn)->_pureOut = LstCreate(&trArcTools))
#define TrnCPurOut(trn) ((trn)->_pureOut = (LIST *) NULL)
#define TrnDPurOut(trn) LstDestroy((trn)->_pureOut, TRUE)
#define TrnGNeutral(trn) ((trn)->_neutral)
#define TrnZNeutral(trn,list) ((trn)->_neutral = (list))
#define TrnSNeutral(trn) ((trn)->_neutral = LstCreate(&trArcTools))
#define TrnCNeutral(trn) ((trn)->_neutral = (LIST *) NULL)
#define TrnGMosIn(trn) ((trn)->_mostIn)
#define TrnZMosIn(trn,list) ((trn)->_mostIn = (list))
#define TrnSMosIn(trn) ((trn)->_mostIn = LstCreate(&trDrcTools))
#define TrnCMosIn(trn) ((trn)->_mostIn = (LIST *) NULL)
#define TrnGMosOut(trn) ((trn)->_mostOut)
#define TrnZMosOut(trn,list) ((trn)->_mostOut = (list))
#define TrnSMosOut(trn) ((trn)->_mostOut = LstCreate(&trDrcTools))
#define TrnCMosOut(trn) ((trn)->_mostOut = (LIST *) NULL)
#define TrnGHigher(trn) ((trn)->_higher)
#define TrnZHigher(trn,list) ((trn)->_higher = (list))
#define TrnSHigher(trn) ((trn)->_higher = LstCreate(&trnTools))
#define TrnCHigher(trn) ((trn)->_higher = (LIST *) NULL)
#define TrnGLower(trn) ((trn)->_lower)
#define TrnZLower(trn,list) ((trn)->_lower = (list))
#define TrnSLower(trn) ((trn)->_lower = LstCreate(&trnTools))
#define TrnCLower(trn) ((trn)->_lower = (LIST *) NULL)
#define TrnGPromoted(trn) ((trn)->_promoted)
#define TrnZPromoted(trn,balTree) ((trn)->_promoted = (balTree))
#define TrnSPromoted(trn) ((trn)->_promoted = LstCreate(&trPromTools))
#define TrnCPromoted(trn) ((trn)->_promoted = (LIST *) NULL)
#define TrnGDefers(trn) ((trn)->_defers)
#define TrnZDefers(trn,balTree) ((trn)->_defers = (balTree))
#define TrnSDefers(trn) ((trn)->_defers = LstCreate(&deferTools))
#define TrnCDefers(trn) ((trn)->_defers = (LIST *) NULL)
#define TrnGVal(trn) ((trn)->_ut._vt._n._val)
#define TrnSVal(trn,val) ((trn)->_ut._vt._n._val = (val))
#define TrnGDupVal(trn) ((trn)->_ut._vt._n._dupVal)
#define TrnSDupVal(trn,val) ((trn)->_ut._vt._n._dupVal = (val))
#define TrnGMin(trn) ((trn)->_ut._vt._n._min)
#define TrnSMin(trn,min) ((trn)->_ut._vt._n._min = (min))
#ifdef SMODELS
#define TrnIsPlain(trn) (TrnIsGamma(trn) && (PtLongPlay() || TrnEnabled(trn)))
#else
#define TrnIsPlain(trn) (TrnIsGamma(trn) && TrnEnabled(trn))
#endif
#define TrnGGamma(trn) ((trn)->_ut._vt._a._gamma)
#define TrnSGamma(trn,gamma) ((trn)->_ut._vt._a._gamma = (gamma))
#define TrnCGamma(trn) ((trn)->_ut._vt._a._gamma = (VOID *) NULL)
#define TrnIsKappa(trn) \
(TrnIsPlain(trn) && (((!OnTheFly()) && !PtCFFD()) || !TrnVisible(trn)))
#define TrnGKappa(trn) ((trn)->_ut._vt._a._kappa)
#define TrnSKappa(trn,kappa) ((trn)->_ut._vt._a._kappa = (kappa))
#define TrnCKappa(trn) ((trn)->_ut._vt._a._kappa = (VOID *) NULL)
#define TrnIsToStack(trn) \
(TrnIsPlain(trn) && \
(OnTheFly() || PtCFFD()) && (Blocked() || !TrnVisible(trn)))
#define TrnGDiRho(trn) ((trn)->_ut._vt._a._diRho)
#define TrnSDiRho(trn,diRho) ((trn)->_ut._vt._a._diRho = (diRho))
#define TrnCDiRho(trn) ((trn)->_ut._vt._a._diRho = (VOID *) NULL)
#ifdef SMODELS
#define TrnIsEnRho(trn) \
(TrnIsGamma(trn) && (PtLongPlay() || TrnPlcEnabled(trn)) && \
TrnGLower(trn) && !LstEmpty(TrnGLower(trn)))
#else
#define TrnIsEnRho(trn) \
(TrnIsGamma(trn) && TrnPlcEnabled(trn) && \
TrnGLower(trn) && !LstEmpty(TrnGLower(trn)))
#endif
#define TrnGEnRho(trn) ((trn)->_ut._vt._a._enRho)
#define TrnSEnRho(trn,enRho) ((trn)->_ut._vt._a._enRho = (enRho))
#define TrnCEnRho(trn) ((trn)->_ut._vt._a._enRho = (VOID *) NULL)
#define TrnIsReRho(trn) \
(TrnIsGamma(trn) && TrnGLower(trn) && !LstEmpty(TrnGLower(trn)))
#define TrnGReRho(trn) ((trn)->_ut._vt._a._reRho)
#define TrnSReRho(trn,reRho) ((trn)->_ut._vt._a._reRho = (reRho))
#define TrnCReRho(trn) ((trn)->_ut._vt._a._reRho = (VOID *) NULL)
#ifdef SMODELS
#define TrnGPlain(trn) ((trn)->_ut._vt._a._plain)
#define TrnSPlain(trn,plain) ((trn)->_ut._vt._a._plain = (plain))
#define TrnCPlain(trn) ((trn)->_ut._vt._a._plain = (VOID *) NULL)
#define TrnGActive(trn) ((trn)->_ut._vt._a._active)
#define TrnSActive(trn,active) ((trn)->_ut._vt._a._active = (active))
#define TrnCActive(trn) ((trn)->_ut._vt._a._active = (VOID *) NULL)
#define TrnGToStack(trn) ((trn)->_ut._vt._a._toStack)
#define TrnSToStack(trn,toStack) ((trn)->_ut._vt._a._toStack = (toStack))
#define TrnCToStack(trn) ((trn)->_ut._vt._a._toStack = (VOID *) NULL)
#define TrnCAtoms(trn) \
(TrnCGamma(trn), TrnCKappa(trn), TrnCDiRho(trn), TrnCEnRho(trn), \
TrnCReRho(trn), TrnCPlain(trn), TrnCActive(trn), TrnCToStack(trn))
#else
#define TrnCAtoms(trn) \
(TrnCGamma(trn), TrnCKappa(trn), TrnCDiRho(trn), TrnCEnRho(trn), \
TrnCReRho(trn))
#endif

#define PlArcGTrn(plArc) ((plArc)->_trn)
#define PlArcSTrn(plArc,trn) ((plArc)->_trn = (trn))
#define PlArcGWeight(plArc) ((plArc)->_weight)
#define PlArcAWeight(plArc) (&((plArc)->_weight))
#define PlArcSWeight(plArc,weight) ((plArc)->_weight = (weight))
#define PlArcIWeight(plArc,weight) ((plArc)->_weight += (weight))

#define PlDrcGTrn(plDrc) ((plDrc)->_trn)
#define PlDrcSTrn(plDrc,trn) ((plDrc)->_trn = (trn))
#define PlDrcGIn(plDrc) ((plDrc)->_in)
#define PlDrcAIn(plDrc) (&((plDrc)->_in))
#define PlDrcSIn(plDrc,in) ((plDrc)->_in = (in))
#define PlDrcGOut(plDrc) ((plDrc)->_out)
#define PlDrcAOut(plDrc) (&((plDrc)->_out))
#define PlDrcSOut(plDrc,out) ((plDrc)->_out = (out))

#define TrArcGPlc(trArc) ((trArc)->_plc)
#define TrArcSPlc(trArc,plc) ((trArc)->_plc = (plc))
#define TrArcGWeight(trArc) ((trArc)->_weight)
#define TrArcAWeight(trArc) (&((trArc)->_weight))
#define TrArcSWeight(trArc,weight) ((trArc)->_weight = (weight))
#define TrArcIWeight(trArc,weight) ((trArc)->_weight += (weight))
#define TrArcGOder(trArc) ((trArc)->_oder)
#define TrArcSOder(trArc,oder) ((trArc)->_oder = (oder))
#define TrArcCOder(trArc) ((trArc)->_oder = (VOID *) NULL)
#define TrArcGStrong(trArc) ((trArc)->_strong)
#define TrArcSStrong(trArc,strong) ((trArc)->_strong = (strong))
#define TrArcCStrong(trArc) ((trArc)->_strong = (VOID *) NULL)
#define TrArcGWeak(trArc) ((trArc)->_weak)
#define TrArcSWeak(trArc,weak) ((trArc)->_weak = (weak))
#define TrArcCWeak(trArc) ((trArc)->_weak = (VOID *) NULL)
#ifdef SMODELS
#define TrArcGGamma(trArc) ((trArc)->_gamma)
#define TrArcSGamma(trArc,gamma) ((trArc)->_gamma = (gamma))
#define TrArcCGamma(trArc) ((trArc)->_gamma = (VOID *) NULL)
#define TrArcCAtoms(trArc) \
(TrArcCOder(trArc), TrArcCStrong(trArc), TrArcCWeak(trArc), TrArcCGamma(trArc))
#else
#define TrArcCAtoms(trArc) \
(TrArcCOder(trArc), TrArcCStrong(trArc), TrArcCWeak(trArc))
#endif

#define TrDrcGPlc(trDrc) ((trDrc)->_plc)
#define TrDrcSPlc(trDrc,plc) ((trDrc)->_plc = (plc))
#define TrDrcGIn(trDrc) ((trDrc)->_in)
#define TrDrcAIn(trDrc) (&((trDrc)->_in))
#define TrDrcSIn(trDrc,in) ((trDrc)->_in = (in))
#define TrDrcGOut(trDrc) ((trDrc)->_out)
#define TrDrcAOut(trDrc) (&((trDrc)->_out))
#define TrDrcSOut(trDrc,out) ((trDrc)->_out = (out))
#define TrDrcGOder(trDrc) ((trDrc)->_oder)
#define TrDrcSOder(trDrc,oder) ((trDrc)->_oder = (oder))
#define TrDrcCOder(trDrc) ((trDrc)->_oder = (VOID *) NULL)
#define TrDrcGStrong(trDrc) ((trDrc)->_strong)
#define TrDrcSStrong(trDrc,strong) ((trDrc)->_strong = (strong))
#define TrDrcCStrong(trDrc) ((trDrc)->_strong = (VOID *) NULL)
#define TrDrcGWeak(trDrc) ((trDrc)->_weak)
#define TrDrcSWeak(trDrc,weak) ((trDrc)->_weak = (weak))
#define TrDrcCWeak(trDrc) ((trDrc)->_weak = (VOID *) NULL)
#ifdef SMODELS
#define TrDrcGGamma(trDrc) ((trDrc)->_gamma)
#define TrDrcSGamma(trDrc,gamma) ((trDrc)->_gamma = (gamma))
#define TrDrcCGamma(trDrc) ((trDrc)->_gamma = (VOID *) NULL)
#define TrDrcCAtoms(trDrc) \
(TrDrcCOder(trDrc), TrDrcCStrong(trDrc), TrDrcCWeak(trDrc), TrDrcCGamma(trDrc))
#else
#define TrDrcCAtoms(trDrc) \
(TrDrcCOder(trDrc), TrDrcCStrong(trDrc), TrDrcCWeak(trDrc))
#endif

#define TrPromGTrn(trProm) ((trProm)->_trn)
#define TrPromSTrn(trProm,trn) ((trProm)->_trn = (trn))
#define TrPromGAtom(trProm) ((trProm)->_atom)
#define TrPromSAtom(trProm,atom) ((trProm)->_atom = (atom))
#define TrPromCAtom(trProm) ((trProm)->_atom = (VOID *) NULL)
#define TrPromGArcs(trProm) ((trProm)->_arcs)
#define TrPromZArcs(trProm,arcs) ((trProm)->_arcs = (arcs))
#define TrPromSArcs(trProm) ((trProm)->_arcs = LstCreate(&otArcTools))
#define TrPromCArcs(trProm) ((trProm)->_arcs = (LIST *) NULL)

#define DeferGPlc(defer) ((defer)->_plc)
#define DeferSPlc(defer,plc) ((defer)->_plc = (plc))
#define DeferGFlag(defer) ((defer)->_flag)
#define DeferSFlag(defer) ((defer)->_flag = 1)
#define DeferCFlag(defer) ((defer)->_flag = 0)
#define DeferGAtom(defer) ((defer)->_atom)
#define DeferSAtom(defer,atom) ((defer)->_atom = (atom))
#define DeferCAtom(defer) ((defer)->_atom = (VOID *) NULL)

#define OtArcGDefer(otArc) ((otArc)->_defer)
#define OtArcSDefer(otArc,defer) ((otArc)->_defer = (defer))
#define OtArcGWeight(otArc) ((otArc)->_weight)
#define OtArcSWeight(otArc,weight) ((otArc)->_weight = (weight))

extern char *araArcName;
extern unsigned long _nTrVisible;
#define NTrVisible() _nTrVisible
extern VOID *kappa;
extern VOID *kappakey;
extern VOID *theta;
extern LIST *globMota;
extern BALTREE *curTrns;
extern BALTREE *curDefers;
extern PTRBL *_tblPlc;
#define TblPlc() _tblPlc
extern PTRBL *_dupTblPlc;
#define DupTblPlc() _dupTblPlc
extern PTRBL *_tblTrn;
#define TblTrn() _tblTrn
extern BALTREE *_plcBalTree;
#define PlcBalTree() _plcBalTree
extern BALTREE **_trnBalTree;
#define TrnBalTree(class) _trnBalTree[class]
extern TRN *_lowTrn;
#define LowTrn() _lowTrn
#define SetLowTrn(trn) (_lowTrn = (trn))
extern BOOL _ptSy;
#define PtSymm() _ptSy
#define SetPtSymm() (_ptSy = TRUE)
#define ClPtSymm() (_ptSy = FALSE)
extern unsigned _ptVerbose;
#define PtVerbose() _ptVerbose
#define SetPtVerbose() (_ptVerbose |= 1)
#define PtSuperVerbose() (_ptVerbose & 2)
#define SetPtSuperVerbose() (_ptVerbose |= 2)
#define ClPtVerbose() (_ptVerbose = 0)
extern BOOL _ptStub;
#define PtStub() _ptStub
#define SetPtStub() (_ptStub = TRUE)
#define ClPtStub() (_ptStub = FALSE)
extern BOOL _ptCFFD;
#define PtCFFD() _ptCFFD
#define SetPtCFFD() (_ptCFFD = TRUE)
#define ClPtCFFD() (_ptCFFD = FALSE)
extern BOOL _ptDefVis;
#define PtDefVis() _ptDefVis
#define SetPtDefVis() (_ptDefVis = TRUE)
#define ClPtDefVis() (_ptDefVis = FALSE)
extern BOOL _ptSleep;
#define PtSleep() _ptSleep
#define SetPtSleep() (_ptSleep = TRUE)
#define ClPtSleep() (_ptSleep = FALSE)
extern BOOL _plRnd;
#define PlRnd() _plRnd
#define SetPlRnd() (_plRnd = TRUE)
#define ResetPlRnd() (_plRnd = FALSE)
extern BOOL _ptLazy;
#define PtLazy() _ptLazy
#define SetPtLazy() (_ptLazy = TRUE)
#define ClPtLazy() (_ptLazy = FALSE)
extern BOOL _trRnd;
#define TrRnd() _trRnd
#define SetTrRnd() (_trRnd = TRUE)
#define ResetTrRnd() (_trRnd = FALSE)
extern BOOL _ptDelAlg;
#define PtDelAlg() _ptDelAlg
#define SetPtDelAlg() (_ptDelAlg = TRUE)
#define ClPtDelAlg() (_ptDelAlg = FALSE)
extern BOOL _ptTryAgain;
#define PtTryAgain() _ptTryAgain
#define SetPtTryAgain() (_ptTryAgain = TRUE)
#define ClPtTryAgain() (_ptTryAgain = FALSE)
extern LIST *_ptList;
#define PtList() _ptList
#define PtSList() (_ptList = LstCreate(&trnTools))
#define PtDList() LstDestroy(_ptList, FALSE)
extern unsigned long _ptSgCount;
#define PtSgCount() _ptSgCount
#define PtCSgCount() (_ptSgCount = 0L)
#define PtISgCount() (_ptSgCount++)
extern unsigned long _ptECnt;
#define PtGECnt() _ptECnt
#define PtCECnt() (_ptECnt = 0L)
#define PtIECnt() (_ptECnt++)
extern unsigned long _ptInvisCnt;
#define PtGInvisCnt() _ptInvisCnt
#define PtCInvisCnt() (_ptInvisCnt = 0L)
#define PtIInvisCnt() (_ptInvisCnt++)
extern unsigned long _ptNow;
#define PtNow() _ptNow
#define PtCNow() (_ptNow = 0L)
#define PtINow() (_ptNow++)
extern unsigned long _ptBest;
#define PtBest() _ptBest
#define PtSetBest(b) (_ptBest = (b))
extern unsigned long _ptSzBest;
#define PtSzBest() _ptSzBest
#define PtSzSBest(sz) (_ptSzBest = (sz))
extern unsigned long _ptCounter;
#define PtSCounter(c) (_ptCounter = (c))
#define PtDCounter() (_ptCounter--)
extern TRN *_ptKingTrn;
#define PtKingTrn() _ptKingTrn
#define PtSKingTrn(trn) (_ptKingTrn = (trn))
extern BOOL _ptRecurs;
#define PtRecurs() _ptRecurs
#define SetPtRecurs() (_ptRecurs = TRUE)
#define ClPtRecurs() (_ptRecurs = FALSE)
extern BALTREE *_ptSlpSet[4];
#define PtSlpSet() (_ptSlpSet[0])
#define PtASlpSet() (&(_ptSlpSet[0]))
#define PtSSlpSet(slpSet) (_ptSlpSet[0] = (slpSet))
#define PtDSlpSet() SlpDis((unsigned) 0)
#define Pt2SlpSet() (_ptSlpSet[1])
#define PtA2SlpSet() (&(_ptSlpSet[1]))
#define PtS2SlpSet(slpSet) (_ptSlpSet[1] = (slpSet))
#define PtD2SlpSet() SlpDis((unsigned) 1)
#define Pt3SlpSet() (_ptSlpSet[2])
#define PtA3SlpSet() (&(_ptSlpSet[2]))
#define PtS3SlpSet(slpSet) (_ptSlpSet[2] = (slpSet))
#define PtD3SlpSet() SlpDis((unsigned) 2)
#define Pt4SlpSet() (_ptSlpSet[3])
#define PtA4SlpSet() (&(_ptSlpSet[3]))
#define PtS4SlpSet(slpSet) (_ptSlpSet[3] = (slpSet))
#define PtD4SlpSet() SlpDis((unsigned) 3)
extern PLC *_ptFlyPlc;
#define PtFlyPlc() _ptFlyPlc
#define PtSFlyPlc(plc) (_ptFlyPlc = (plc))
extern BOOL ptDoIt;
extern BOOL ptBranching;
extern TRN *ptFirst;
extern PLC *buPlc;
extern FOOL _buchiOK;

DE0(void, FndEnbl)
DE0(void, PtGphUpdate)
DE2(void, PtC1,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC2,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC3,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC4,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE0(void, PtOnTheFly)
DE0(void, PtInfinite)
DE0(void, PtIncrementalAlg)
#define PtDeletionAlg PtDelegate
DE0(void, PtFollowers)
DE0(void, PtExpand)
DE1(BOOL, TrnChkEnbl,
    TRN *, trn)
DE1(BOOL, TrnAntiVisit,
    TRN *, trn)
DE1(BOOL, TrnStub,
    TRN *, trn)
DE1(void, TrnNaps,
    TRN *, trn)
DE1(void, TrnSpan,
    TRN *, trn)
DE1(BOOL, TrnVisStub,
    TRN *, trn)
DE1(void, TrnOA,
    TRN *, trn)
DE1(void, TrnAO,
    TRN *, trn)
DE1(BOOL, TrnAndOr,
    TRN *, trn)
#ifdef DEBUG
DE2(void, PtrDump,
    char *, ptr,
    unsigned, depth)
#endif
DE2(int, PtrCompare,
    char *, ptr1,
    char *, ptr2)
DE0(PLC *, PlcCons)
DE1(PLC *, PlcCopy,
    PLC *, plc)
DE1(void, PlcDracsid,
    PLC *, plc)
#define PlcDiscard(plc) (PlcDracsid(plc), (plc) = (PLC *) NULL)
#ifdef DEBUG
DE2(void, PlcDump,
    PLC *, plc,
    unsigned, depth)
#endif
DE2(int, PlcKeyCmp,
    PLC *, plc,
    DATATYPE *, key)
DE0(TRN *, TrnCons)
DE1(TRN *, TrnCopy,
    TRN *, trn)
DE1(void, TrnDracsid,
    TRN *, trn)
#define TrnDiscard(trn) (TrnDracsid(trn), (trn) = (TRN *) NULL)
#ifdef DEBUG
DE2(void, TrnDump,
    TRN *, trn,
    unsigned, depth)
#endif
DE2(int, TrnKeyCmp,
    TRN *, trn,
    DATATYPE *, key)
DE2(PLARC *, PlArcCons,
    TRN *, trn,
    unsigned long, weight)
DE1(PLARC *, PlArcCopy,
    PLARC *, plArc)
#ifdef DEBUG
DE2(void, PlArcDump,
    PLARC *, plArc,
    unsigned, depth)
#endif
DE2(int, PlArcCompare,
    PLARC *, plArc1,
    PLARC *, plArc2)
DE3(PLDRC *, PlDrcCons,
    TRN *, trn,
    unsigned long, in,
    unsigned long, out)
DE1(PLDRC *, PlDrcCopy,
    PLDRC *, plDrc)
#ifdef DEBUG
DE2(void, PlDrcDump,
    PLDRC *, plDrc,
    unsigned, depth)
#endif
DE2(int, PlDrcCompare,
    PLDRC *, plDrc1,
    PLDRC *, plDrc2)
DE2(TRARC *, TrArcCons,
    PLC *, plc,
    unsigned long, weight)
DE1(TRARC *, TrArcCopy,
    TRARC *, trArc)
#ifdef DEBUG
DE2(void, TrArcDump,
    TRARC *, trArc,
    unsigned, depth)
#endif
DE2(int, TrArcCompare,
    TRARC *, trArc1,
    TRARC *, trArc2)
DE3(TRDRC *, TrDrcCons,
    PLC *, plc,
    unsigned long, in,
    unsigned long, out)
DE1(TRDRC *, TrDrcCopy,
    TRDRC *, trDrc)
#ifdef DEBUG
DE2(void, TrDrcDump,
    TRDRC *, trDrc,
    unsigned, depth)
#endif
DE2(int, TrDrcCompare,
    TRDRC *, trDrc1,
    TRDRC *, trDrc2)
DE1(TRPROM *, TrPromCons,
    TRN *, trn)
DE1(TRPROM *, TrPromCopy,
    TRPROM *, trProm)
DE1(void, TrPromDracsid,
    TRPROM *, trProm)
#define TrPromDiscard(trProm) \
(TrPromDracsid(trProm), (trProm) = (TRPROM *) NULL)
#ifdef DEBUG
DE2(void, TrPromDump,
    TRPROM *, trProm,
    unsigned, depth)
#endif
DE1(DEFER *, DeferCons,
    PLC *, plc)
DE1(DEFER *, DeferCopy,
    DEFER *, defer)
#ifdef DEBUG
DE2(void, DeferDump,
    DEFER *, defer,
    unsigned, depth)
#endif
DE2(int, DeferKeyCmp,
    DEFER *, defer,
    PLC *, plc)
DE2(OTARC *, OtArcCons,
    DEFER *, defer,
    unsigned long, weight)
DE1(OTARC *, OtArcCopy,
    OTARC *, otArc)
#ifdef DEBUG
DE2(void, OtArcDump,
    OTARC *, otArc,
    unsigned, depth)
#endif
DE4(void, PutInArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
DE4(void, PutOutArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
DE1(void, TrnReform,
    TRN *, trn)
DE0(void, PlcFill)
DE0(void, TrnFill)
DE1(void, PtOutput,
    BINARYFILE, file)
DE0(void, PtAraOutput)
DE1(void, PtPepOutput,
    char *, name)
DE1(BOOL, PlcIniFlag,
    PLC *, plc)
DE3(unsigned, PlcEncode,
    PTRBL *, tbl,
    DATATYPE **, buffer,
    unsigned *, size)
DE1(void, PlcDecode,
    DATATYPE *, buffer)
DE1(BOOL, TrnPrec,
    TRN *, trn)
DE1(void, PtInit,
    BINARYFILE, file)
DE0(void, PtBaseInit)
DE1(void, SlpDis,
    unsigned, i)
DE0(void, PtClose)
DE1(void, PtSRandom,
    UNS32, seedVal)
DE0(UNS32, PtTerm)
DE1(UNS32, PtRandom,
    UNS32, domSize)
DE2(void, LTSBegin,
    char *, name,
    BOOL, separate)
DE0(void, LTSEnd)
DE0(STATE *, PtState)
DE1(PLC *, GetPlc,
    DATATYPE *, key)
DE3(TRN *, GetTrn,
    unsigned, class,
    DATATYPE *, key,
    BOOL *, isNew)
DE1(void, TrnForward,
    TRN *, trn)
DE1(void, TrnBackward,
    TRN *, trn)
DE1(void, TrnFire,
    TRN *, trn)
DE1(void, TrnLookStack,
    TRN *, trn)

extern TRN *ptInvFirst;
extern unsigned long ptM;
extern MOTA *ptMota;
extern TRN *ptTrn;
extern TRN *pt2Trn;
extern PLC *ptPlc;
extern TRARC *ptTrArc;
extern TRDRC *ptTrDrc;
extern TRPROM *ptTrProm;
extern DEFER *ptDefer;
extern OTARC *ptOtArc;
extern PLARC *ptPlArc;
extern PLDRC *ptPlDrc;
extern TRN *ptVisible;
extern TRN *ptInvisible;
extern unsigned long ptSz;
extern unsigned long ptV;
extern unsigned long ptLimit;
extern unsigned long trNumber;

DE1(BOOL, TrnPlcEC,
    TRN *, trn)
DE1(BOOL, TrnPriEC,
    TRN *, trn)
DE2(BOOL, AfterEnbl,
    TRN *, trn1,
    TRN *, trn2)
DE1(BOOL, TrnBest,
    TRN *, trn)
DE1(BOOL, TrnReCheck,
    TRN *, trn)
DE1(BOOL, PtNewSleep,
    unsigned long *, a)
DE1(BOOL, PtOldSleep,
    unsigned long *, a)
DE1(BOOL, TrnChkRe,
    TRN *, trn)
DE1(BOOL, TrnWorseVis,
    TRN *, trn)
DE1(BOOL, MotaCntCand,
    MOTA *, mota)
DE1(void, MotaYrd,
    MOTA *, mota)
DE1(void, MotaDry,
    MOTA *, mota)
DE1(BOOL, MotaTryDel,
    MOTA *, mota)
DE1(void, MotaLecnac,
    MOTA *, mota)
DE1(void, MotaCancel,
    MOTA *, mota)
DE1(void, MotaTewTas,
    MOTA *, mota)
DE1(void, MotaSatWet,
    MOTA *, mota)
DE1(BOOL, TrnPreAO,
    TRN *, trn)
DE1(BOOL, TrnDelCheck,
    TRN *, trn)
DE1(BOOL, TrnHowDelCheck,
    TRN *, trn)
DE1(BOOL, TrnVisCheck,
    TRN *, trn)
#ifdef SMODELS
DE1(BOOL, TrnSmoCheck,
    TRN *, trn)
DE1(BOOL, TrnHowSmoCheck,
    TRN *, trn)
DE1(BOOL, TrnVisSmoCheck,
    TRN *, trn)
#endif
DE1(BOOL, TrnRedInc,
    TRN *, trn)
DE0(void, PtInside)
DE0(void, PtActual)
DE0(void, PtInner)
DE0(void, PtBuchi)
DE0(void, PtProCFFD)
DE0(void, PtRedInc)
DE1(BOOL, MotaBackup,
    MOTA *, mota)
DE1(BOOL, MotaRehab,
    MOTA *, mota)
DE1(BOOL, MotaRemember,
    MOTA *, mota)
DE1(BOOL, TrnCollect,
    TRN *, trn)
DE0(void, Collect)
DE1(BOOL, TrnCntV,
    TRN *, trn)
DE1(BOOL, TrnVCollect,
    TRN *, trn)
DE0(void, VCollect)
DE0(void, DelThem)
DE1(BOOL, TrnDsPrIn,
    TRN *, trn)
DE1(BOOL, TrnDsNeutral,
    TRN *, trn)
DE1(BOOL, TrnDsMsIn,
    TRN *, trn)
DE1(BOOL, TrnDsMsOut,
    TRN *, trn)
DE1(void, TrnEnNeutral,
    TRN *, trn)
DE1(void, TrnEnMsOut,
    TRN *, trn)
DE1(void, TrnEnPrIn,
    TRN *, trn)
DE1(void, TrnEnMsIn,
    TRN *, trn)
DE1(void, TrnTivis,
    TRN *, trn)
DE1(void, TrnVisit,
    TRN *, trn)
DE1(BOOL, TrnDvPrIn,
    TRN *, trn)
DE1(BOOL, TrnDvNeutral,
    TRN *, trn)
DE1(BOOL, TrnDvMsIn,
    TRN *, trn)
DE1(BOOL, TrnDvMsOut,
    TRN *, trn)
DE1(void, TrnEvNeutral,
    TRN *, trn)
DE1(void, TrnEvMsOut,
    TRN *, trn)
DE1(void, TrnEvPrIn,
    TRN *, trn)
DE1(void, TrnEvMsIn,
    TRN *, trn)
DE1(void, TrnNdPrIn,
    TRN *, trn)
DE1(void, TrnNdMsIn,
    TRN *, trn)
DE1(void, TrnOrPromote,
    TRN *, trn)
DE1(void, TrnOrPrIn,
    TRN *, trn)
DE1(void, TrnOrNeutral,
    TRN *, trn)
DE1(void, TrnOrMsIn,
    TRN *, trn)
DE1(void, TrnOrMsOut,
    TRN *, trn)

extern LIST **_visBlock;
extern char *_visColour;
#define ABlocks(n) (_visBlock = (LIST **) AllocArray(n, sizeof(LIST *)), \
_visColour = (char *) Allocate(n))
#define SBlock(i) (_visBlock[i] = LstCreate(&trnTools), _visColour[i] = 'y')
#define CBlock(i) (_visBlock[i] = (LIST *) NULL, _visColour[i] = '\0')
#define Block(i) (_visBlock[i])
#define Yellow(i) ('y' == _visColour[i])
#define Orange(i) ('o' == _visColour[i])
#define Blue(i) ('b' == _visColour[i])
#define SetYellow(i) (_visColour[i] = 'y')
#define SetOrange(i) (_visColour[i] = 'o')
#define SetBlue(i) (_visColour[i] = 'b')

extern BOOL _blocked;
#define Blocked() _blocked
#define SetBlocked() (_blocked = TRUE)
#define ClBlocked() (_blocked = FALSE)

extern PTRBL *visibles;
extern PTRBL *candidates;
extern unsigned long *candInd;

#ifdef SMODELS
extern BOOL _ptSmodels;
#define PtSmodels() _ptSmodels
#define SetPtSmodels() (_ptSmodels = TRUE)
#define ClPtSmodels() (_ptSmodels = FALSE)
extern BOOL _ptLongPlay;
#define PtLongPlay() _ptLongPlay
#define SetPtLongPlay() (_ptLongPlay = TRUE)
#define ClPtLongPlay() (_ptLongPlay = FALSE)
extern INEQ *ptIneq;
#endif
DE4(void, AuxDefer,
    TRN *, trn,
    TRPROM *, trProm,
    PLC *, plc,
    unsigned long, weight)
DE2(void, TrnAuxProm,
    TRN *, trn1,
    TRN *, trn2)
DE1(BOOL, TrnPromote,
    TRN *, trn)
DE0(void, ClDelegate)
DE0(void, PtDelegate)
DE1(void, MkPlcGamma,
    PLC *, plc)
#ifdef SMODELS
DE2(INEQ *, MkPlcIneq,
    PLC *, plc,
    unsigned long, card)
#endif
DE1(void, MkTrnGamma,
    TRN *, trn)
#ifdef DEBUG
DE2(void, TrnDispLabel,
    TRN *, trn,
    FILE *, file)
#endif
DE1(void, MkTrnKappa,
    TRN *, trn)
DE1(void, MkTrnDiRho,
    TRN *, trn)
DE1(void, MkTrnEnRho,
    TRN *, trn)
DE1(void, MkTrnReRho,
    TRN *, trn)
#ifdef SMODELS
DE1(void, MkTrnPlain,
    TRN *, trn)
DE1(void, MkTrnActive,
    TRN *, trn)
DE1(void, MkTrnToStack,
    TRN *, trn)
DE2(void, MkTAGamma,
    TRN *, trn,
    TRARC *, trArc)
DE2(void, MkTDGamma,
    TRN *, trn,
    TRDRC *, trDrc)
#endif
DE2(void, MkTAStrong,
    TRN *, trn,
    TRARC *, trArc)
DE2(void, MkTDStrong,
    TRN *, trn,
    TRDRC *, trDrc)
DE2(void, MkTrPromAtom,
    TRN *, trn,
    TRPROM *, trProm)
DE2(void, MkDeferAtom,
    TRN *, trn,
    DEFER *, defer)
DE2(void, SimpleRule,
    VOID *, head,
    VOID *, tail)
#ifdef SMODELS
DE2(void, ReverseRules,
    VOID *, atom1,
    VOID *, atom2)
#endif
DE0(VOID *, FreshAtom)
#ifdef DEBUG
DE1(VOID *, NmAtom,
    char *, name)
#endif
DE1(void, StartRule,
    VOID *, atom)
DE1(void, ContRule,
    VOID *, atom)
DE0(void, EndRule)
