/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet2.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptnet.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"

FE1(BOOL, MotaCntCand,
    MOTA *, mota)
{
    if (!mota)
	return(TRUE);
    if (MotaEnabled(mota) &&
	(!MotaSatisfied(mota)) && (!MotaThirsty(mota)))
	ptM++;
    return(FALSE);
}

static BOOL illegal = FALSE;

FE1(void, MotaYrd,
    MOTA *, mota)
{
    if (MotaSomewhere(mota))
	{
	    MotaPutNowhere(mota);
	    goto CONT;
	}
    MotaSalt(mota);
    if (MotaGPreds(mota))
	LSTFOR (MotaGPreds(mota), ptMota, MOTA)
	    if ((!MotaSatisfied(ptMota)) && !MotaThirsty(ptMota))
		{
		    if (MotaIsOr(ptMota))
			{
			    MotaDecCounter(ptMota);
			    if (MotaCounter(ptMota) >= MotaDupCounter(ptMota))
				Troba1("Counter failure in MotaYrd");
			}
		    if ((!MotaIsOr(ptMota)) || !MotaCounter(ptMota))
			{
			    if (MotaIsProtected(ptMota))
				illegal = TRUE;
			    MotaSetCursor(mota, LstGet(MotaGPreds(mota)));
			    MotaPutSomewhere(mota);
			    return;
CONT: ;
			    LstSet(MotaGPreds(mota), MotaGCursor(mota));
			}
		}
}

FE1(void, MotaDry,
    MOTA *, mota)
{
    while (TRUE)
        {
	    MotaYrd(mota);
            if (MotaSomewhere(mota))
                {
                    MotaSetDown(ptMota, mota);
                    mota = ptMota;
                }
            else if (MotaGDown(mota))
                {
                    ptMota = MotaGDown(mota);
                    MotaCleDown(mota);
                    mota = ptMota;
                }
            else
                return;
        }
}

FE1(BOOL, MotaTryDel,
    MOTA *, mota)
{
    if (!mota)
	return(TRUE);
    if (MotaIsProtected(mota) || MotaSatisfied(mota) ||
	MotaThirsty(mota) || MotaAsleep(mota))
        return(FALSE);
    if (ptM-- != 0L)
	return(FALSE);
    MotaPutToSleep(mota);
    MotaDry(mota);
    return(TRUE);
}

FE1(void, MotaLecnac,
    MOTA *, mota)
{
    if (MotaSomewhere(mota))
	{
	    MotaPutNowhere(mota);
	    goto CONT;
	}
    MotaWater(mota);
    if (MotaGPreds(mota))
	LSTFOR (MotaGPreds(mota), ptMota, MOTA)
	    if (!MotaSatisfied(ptMota))
		{
		    if (MotaIsOr(ptMota))
			{
			    if (MotaCounter(ptMota) >= MotaDupCounter(ptMota))
				Troba1("Counter failure in MotaLecnac");
			    MotaIncCounter(ptMota);
			}
		    if (MotaThirsty(ptMota))
			{
			    MotaSetCursor(mota, LstGet(MotaGPreds(mota)));
			    MotaPutSomewhere(mota);
			    return;
CONT: ;
			    LstSet(MotaGPreds(mota), MotaGCursor(mota));
			}
		}
}

FE1(void, MotaCancel,
    MOTA *, mota)
{
    while (TRUE)
        {
	    MotaLecnac(mota);
            if (MotaSomewhere(mota))
                {
                    MotaSetDown(ptMota, mota);
                    mota = ptMota;
                }
            else if (MotaGDown(mota))
                {
                    ptMota = MotaGDown(mota);
                    MotaCleDown(mota);
                    mota = ptMota;
                }
            else
                return;
        }
}

FE1(void, MotaTewTas,
    MOTA *, mota)
{
    if (MotaSomewhere(mota))
        {
            MotaPutNowhere(mota);
            goto CONT;
        }
    MotaFeed(mota);
    MotaWater(mota);
    if (MotaGPreds(mota))
	LSTFOR (MotaGPreds(mota), ptMota, MOTA)
	    if (MotaThirsty(ptMota) && !MotaSatisfied(ptMota))
		{
		    MotaSetCursor(mota, LstGet(MotaGPreds(mota)));
		    MotaPutSomewhere(mota);
		    return;
CONT: ;
		    LstSet(MotaGPreds(mota), MotaGCursor(mota));
		}
}

FE1(void, MotaSatWet,
    MOTA *, mota)
{
    while (TRUE)
        {
            MotaTewTas(mota);
            if (MotaSomewhere(mota))
                {
                    MotaSetDown(ptMota, mota);
                    mota = ptMota;
                }
            else if (MotaGDown(mota))
                {
                    ptMota = MotaGDown(mota);
                    MotaCleDown(mota);
                    mota = ptMota;
                }
            else
                return;
        }
}

/***************************************************************************
*
* Function:  TrnPreAO
* Purpose:   This routine cleans the flag of a transition so that the
*            construction of an and/or-graph can be started in the
*            deletion algorithm of the stubborn set method.
* Arguments: trn - the transition
* Returns:
* See also:  PlcIniFlag (ptelem3.c)
*
****************************************************************************
*/
FE1(BOOL, TrnPreAO,
    TRN *, trn)
{
    TRN *rtrn;

#ifdef SMODELS
    if (PtSmodels())
	{
	    TrnCAtoms(trn);
	    if (TrnGPurIn(trn))
		LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
		    TrArcCAtoms(ptTrArc);
	    if (TrnGMosIn(trn))
		LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
		    TrDrcCAtoms(ptTrDrc);
	    if (TrnGPromoted(trn))
		LSTFOR (TrnGPromoted(trn), ptTrProm, TRPROM)
		    TrPromCAtom(ptTrProm);
	    if (TrnGDefers(trn))
		LSTFOR (TrnGDefers(trn), ptDefer, DEFER)
		    DeferCAtom(ptDefer);
	}
#endif
    TrnImpoverish(trn);
    TrnWake(trn);
    TrnWater(trn);
    TrnLeave(trn);
    TrnRejDiRho(trn);
    if ((
#ifdef SMODELS
	    PtLongPlay() ||
#endif
	    !TrnPlcEnabled(trn)) &&
	TrnIsGamma(trn) && TrnGLower(trn))
	LSTFOR (TrnGLower(trn), rtrn, TRN)
	    if (TrnIsKappa(rtrn))
		{
		    TrnAckDiRho(trn);
		    break;
		}
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnDelCheck
* Purpose:   This routine fires a transition if it is enabled in the
*            stubborn set computed by the deletion algorithm.
* Arguments: trn - the transition
* Returns:
* See also:  TrnFire, TrnBest, TrnPrecCheck, TrnReCheck, PtOldSleep,
*            TrnChkRe
*
****************************************************************************
*/
FE1(BOOL, TrnDelCheck,
    TRN *, trn)
{
    if (!TrnEnabled(trn))
	return(FALSE);
    if (TrnIsGamma(trn))
	{
	    if (!TrnGGamma(trn))
		Troba1("Not TrnGGamma in TrnDelCheck");
	    else if (MotaSatisfied((MOTA *) TrnGGamma(trn)))
		return(FALSE);
	}
    TrnFire(trn);
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnHowDelCheck
* Purpose:   This routine simulates TrnDelCheck in order to know if
*            more than one transition is to be fired.
* Arguments: trn - a transition
* Returns:
* See also:  TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, TrnHowDelCheck,
    TRN *, trn)
{
    if (!TrnEnabled(trn))
	return(FALSE);
    if (TrnIsGamma(trn))
	{
	    if (!TrnGGamma(trn))
		Troba1("No TrnGGamma in TrnHowDelCheck");
	    else if (MotaSatisfied((MOTA *) TrnGGamma(trn)))
		return(FALSE);
	}
    if (ptFirst)
	{
	    ptBranching = TRUE;
	    return(TRUE);
	}
    else
	ptFirst = trn;
    return(FALSE);
}

FE1(BOOL, TrnVisCheck,
    TRN *, trn)
{
    if (TrnVisible(trn) && TrnIsGamma(trn))
	{
	    if (!TrnGGamma(trn))
		Troba1("No TrnGGamma in TrnVisCheck");
	    else if (MotaSatisfied((MOTA *) TrnGGamma(trn)))
		{
		    SetPtRecurs();
		    return(TRUE);
		}
	}
    return(FALSE);
}

#ifdef SMODELS
FE1(BOOL, TrnSmoCheck,
    TRN *, trn)
{
    if (TrnEnabled(trn) && !TrnSatisfied(trn))
	TrnFire(trn);
    return(FALSE);
}

FE1(BOOL, TrnHowSmoCheck,
    TRN *, trn)
{
    if (TrnEnabled(trn) && !TrnSatisfied(trn))
	{
	    if (ptFirst)
		{
		    ptBranching = TRUE;
		    return(TRUE);
		}
	    else
		ptFirst = trn;
	}
    return(FALSE);
}

FE1(BOOL, TrnVisSmoCheck,
    TRN *, trn)
{
    if (TrnVisible(trn) && TrnSatisfied(trn))
	{
	    SetPtRecurs();
	    return(TRUE);
	}
    return(FALSE);
}
#endif

/* The following routines before DelThem are auxiliary routines of DelThem. */

FE1(BOOL, MotaBackup,
    MOTA *, mota)
{
    if (!mota)
	return(TRUE);
    if (MotaEnabled(mota) && !MotaSatisfied(mota))
	MotaEnter(mota);
    else
	MotaLeave(mota);
    return(FALSE);
}

FE1(BOOL, MotaRehab,
    MOTA *, mota)
{
    MotaImpoverish(mota);
    MotaWater(mota);
    MotaWake(mota);
    MotaUnprotect(mota);
    MotaResCounter(mota);
    return(FALSE);
}

FE1(BOOL, MotaRemember,
    MOTA *, mota)
{
    if (MotaOld(mota))
	MotaImpoverish(mota);
    else
	MotaFeed(mota);
    return(FALSE);
}

static MOTA *ptDone;       /* a candidate to be deleted in
			    * the deletion algorithm of the stubborn
			    * set method */
static unsigned maxExhaustive;

static TRN *prevIsGamma = NULL;
static TRN *prevNotGamma = NULL;

FE1(BOOL, TrnCollect,
    TRN *, trn)
{
    if (!TrnEnabled(trn))
	return(FALSE);
    if (TrnIsGamma(trn))
	{
	    if (!TrnGGamma(trn))
		Troba1("No TrnGGamma in TrnCollect");
	    else if (MotaSatisfied((MOTA *) TrnGGamma(trn)))
		return(FALSE);
	    prevIsGamma = trn;
	    PtrBlPut(candidates, (char *) TrnGGamma(trn), ptM);
	    ptM++;
	}
    else
	prevNotGamma = trn;
    return(FALSE);
}

FE0(void, Collect)
{
    if (!candidates)
	{
	    candidates = PtrBlCreate(PtrBlGCount(TblTrn()) + 1L, &motaTools);
	    maxExhaustive = 5;
	    candInd = (unsigned long *)
		AllocArray(maxExhaustive - 2, sizeof(unsigned long));
	}
    prevIsGamma = (TRN *) NULL;
    prevNotGamma = (TRN *) NULL;
    ptM = 0L;
    (void) PtrBlApply(TblTrn(), TrnCollect);
    PtrBlPut(candidates, (char *) NULL, ptM);
    if (prevIsGamma && prevNotGamma)
	{
#ifdef DEBUG
	    TrnDump(prevIsGamma, (unsigned) 0);
	    TrnDump(prevNotGamma, (unsigned) 0);
#endif	    
	    Troba1("Collect failed");
	}
}

FE1(BOOL, TrnCntV,
    TRN *, trn)
{
    if (TrnVisible(trn))
	ptM++;
    return(FALSE);
}

FE1(BOOL, TrnVCollect,
    TRN *, trn)
{
    if (TrnVisible(trn))
	{
	    PtrBlPut(visibles, (char *) trn, ptM);
	    ptM++;
	}
    return(FALSE);
}

FE0(void, VCollect)
{
    ptM = 0L;
    (void) PtrBlApply(TblTrn(), TrnCntV);
    visibles = PtrBlCreate(ptM + 1L, &trnTools);
    ptM = 0L;
    (void) PtrBlApply(TblTrn(), TrnVCollect);
    PtrBlPut(visibles, (char *) NULL, ptM);
}

/***************************************************************************
*
* Function:  DelThem
* Purpose:   This routine is the heart of the deletion algorithm of the
*            stubborn set method.
* Arguments:
* Returns:
* See also:  PtDeletionAlg
*
****************************************************************************
*/
FE0(void, DelThem)
{
    unsigned long nCand;
    unsigned long nDeCand;
    unsigned many;
    unsigned ceiling;
    unsigned i;
    MOTA *mota;

    Collect();
    if (prevNotGamma)
	{
	    ptBranching = TRUE;
	    return;
	}
    nDeCand = nCand = PtGECnt();
    MotaProtect((MOTA *) kappa);
    while ((nDeCand != 0L) && (nCand > (unsigned long) 1L))
	{
	    if (TrRnd())
		ptM = (unsigned long) PtRandom((UNS32) nDeCand);
	    else
		ptM = 0L;
	    illegal = FALSE;
	    ptDone = *((MOTA **) PtrBlApply(candidates, MotaTryDel));
	    if (illegal)
		{
		    nDeCand--;
		    MotaCancel(ptDone);
		}
	    else
		{
#ifdef DEBUG
#if 0
		    LstDump(globMota, (unsigned) 0);
#endif
#endif
		    ptM = 0L;
		    (void) PtrBlApply(candidates, MotaCntCand);
		    nDeCand -= nCand - ptM;
		    nCand = ptM;
		    MotaSatWet(ptDone);
		}
	}
#ifdef DEBUG
#if 0
    LstDump(globMota, (unsigned) 0);
#endif
#endif
    if (nCand == PtGECnt())
	ptBranching = TRUE;
    if (PtTryAgain() && (nCand > (unsigned long) 1L) &&	(nCand < PtGECnt()))
	{
	    if (PtGECnt() > (unsigned long) maxExhaustive)
		ceiling = 1;
	    else
		ceiling = ((unsigned) nCand) - 1;
	    LSTFOR (globMota, mota, MOTA)
		MotaBackup(mota);
	    many = 1;
	    for (candInd[0] = 0L;
		 MotaOld((MOTA *) PtrBlGet(candidates, candInd[0]));
		 candInd[0]++)
		;
	    do
		{
		    LSTFOR (globMota, mota, MOTA)
			MotaRehab(mota);
		    MotaProtect((MOTA *) kappa);
		    for (i = many; i--; )
			{
			    mota = (MOTA *) PtrBlGet(candidates,
						     candInd[i]);
			    MotaPutToSleep(mota);
			    MotaProtect(mota);
			}
		    for (ptM = 0L, illegal = FALSE;
			 (ptDone = *((MOTA **)
				     PtrBlApply(candidates, MotaTryDel)));
			 ptM = 0L, illegal = FALSE)
			{
			    if (illegal)
				MotaCancel(ptDone);
			    else
				MotaSatWet(ptDone);
			}
		    ptM = 0L;
		    (void) PtrBlApply(candidates, MotaCntCand);
		    nCand = ptM;
		    if (nCand > (unsigned long) many)
			{
			    if (nCand <= (unsigned long) ceiling)
				{
				    ceiling = ((unsigned) nCand) - 1;
				    (void) PtrBlApply(candidates, MotaBackup);
				}
			    do
				{
				    if (candInd[many - 1] < PtGECnt() - 1L)
					candInd[many - 1]++;
				    else if (candInd[0] <
					     PtGECnt() - (unsigned long) many)
					{
					    for (i = many - 2;
						 candInd[i + 1] ==
						 candInd[i] + 1L;
						 i--)
						;
					    candInd[i]++;
					    for (; i < many - 1; i++)
						candInd[i + 1] =
						    candInd[i] + 1L;
					}
				    else
					{
					    many++;
					    if (many <= ceiling)
						for (i = many; i--; )
						    candInd[i] =
							(unsigned long) i;
					    else
						LSTFOR (globMota, mota, MOTA)
						    MotaRemember(mota);
					}
				    if (many <= ceiling)
					for (i = many;
					     i &&
						 MotaOld((MOTA *)
							 PtrBlGet
							 (candidates,
							  candInd[i - 1]));
					     i--)
					    ;
				} while ((many <= ceiling) && i);
			}
		} while ((many <= ceiling) &&
			 (nCand > (unsigned long) many));
	}
}
