/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet3.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptnet.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"

/***************************************************************************
*
* Function:  PtOnTheFly
* Purpose:   This on-the-fly verification routine searches for errors
*            of the form specified by the livelock monitor states,
*            reject states and deadlock monitor states of the tester.
* Arguments:
* Returns:
* See also:  PtC1, PtC2, PtC3, PtC4, PtInfinite, PtBuchi,
*            DoOnTheFly (inst.c)
*
****************************************************************************
*/
FE0(void, PtOnTheFly)
{
    nowPath = thePath;
    ClInfNow();
    if (PtFlyPlc() && PlcLivelock(PtFlyPlc()))
	SetDivNow();
    else
	ClDivNow();
    while (TRUE)
	{
	    if ((!PthLast(thePath))
		||
		((Infinite()
		  ?
		  GphFresh(theGraph, PthEnd(thePath), (unsigned long) 1L)
		  :
		  (GphAwCount(theGraph, PthEnd(thePath)) ==
		   (unsigned long) -1L))
		 &&
		 ((PtCFFD() && !DivNow())
		  ? GphAwHot(theGraph,
			     StpGAwAddr(PthLast(thePath)))
		  : GphAwInvisible(theGraph,
				   StpGAwAddr(PthLast(thePath))))))
		{
		    if (Infinite())
			GphConsume(theGraph, PthEnd(thePath),
				   (unsigned long) 1L);
		    if (Infinite() &&
			((theAwCount =
			  GphAwCount(theGraph, PthEnd(thePath)))
			 != (unsigned long) -1L))
			{
			    if ((theAwCount != 0L) && DivNow())
				{
				    GphFlFirst(theGraph, PthEnd(thePath),
					       &theAwFirst, &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    ptM = theAwCount;
				    while (TRUE)
					{
					    if (((theNdAddr == PthEnd(thePath))
						 ||
						 BalTrElement(theNodeTree,
							      (char *)
							      &theNdAddr))
						&&
						GphAwInvisible(theGraph,
							       theAwAddr))
						PtC3(theAwAddr, theNdAddr);
					    if (--ptM == 0L)
						break;
					    GphFlNext(theGraph, &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    if (PthLast(thePath))
				PlcDecode(GphNode(theGraph, PthEnd(thePath)));
			    if (PtSleep())
				{
				    if (Pt2SlpSet())
					{
					    theAwCount =
						GphAwCount(theGraph,
							   PthEnd(thePath));
					    if (theAwCount != 0L)
						GphFlFirst(theGraph,
							   PthEnd(thePath),
							   &theAwFirst,
							   (unsigned long *)
							   NULL);
					    PtS3SlpSet(BalTrCreate(&lngBalTools));
					    (void) BalTrApply(Pt2SlpSet(),
							      PtOldSleep);
					    PtDSlpSet();
					    PtSSlpSet(Pt3SlpSet());
					}
				    else
					{
					    PtS3SlpSet(BalTrCopy(PtSlpSet(),
								 TRUE));
					    PtExpand();
					    PtD3SlpSet();
					}
				    PtGphUpdate();
				    PtDSlpSet();
				    if (Pt2SlpSet())
					PtD2SlpSet();
				}
			    else if (PtCFFD())
				{
				    FndEnbl();
				    if ((PtGECnt() <= 1L) ||
					(PtGInvisCnt() == 0L))
					{
					    (void) PtrBlApply(TblTrn(),
							      TrnReCheck);
					    PtGphUpdate();
					}
				    else
					{
					    if (PtDelAlg())
						PtDeletionAlg();
					    else
						PtIncrementalAlg();
					}
				}
			    else
				{
				    FndEnbl();
				    if ((PtGECnt() <= 1L) || (!PtStub())
#ifdef SMODELS
					|| (PtSmodels() && OnTheFly() &&
					    (PtGInvisCnt() == 0L))
#endif
					)
					{
					    (void) PtrBlApply(TblTrn(),
							      TrnReCheck);
					    PtGphUpdate();
					}
				    else if (PtDelAlg())
					PtDeletionAlg();
				    else
					PtIncrementalAlg();
				}
			}
		    if ((theAwCount != 0L) &&
			(DivNow() || (PtCFFD() && PtRecurs())))
			{
			    PthInsLast(thePath, theAwFirst, theNdFirst);
			    (void) BalTrInsert(theNodeTree, (char *)
					       StpAWorld(PthLast(thePath)),
					       (char *)
					       StpAWorld(PthLast(thePath)),
					       TRUE);
			    continue;
			}
		}
	    if (!PthLast(thePath))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(thePath)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(thePath)),
			  PthAEnd(thePath));
	    else
		{
		    (void) BalTrDelete(theNodeTree, (char *) NULL,
				       (char *) StpAWorld(PthLast(thePath)),
				       FALSE);
		    PthDelLast(thePath);
		    if (!PthLast(thePath))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  PtInside
* Purpose:   This is an auxiliary routine for PtInfinite.
* Arguments:
* Returns:
* See also:  PtInfinite
*
****************************************************************************
*/
FE0(void, PtInside)
{
    nowPath = the3Path;
    PlcDecode(GphNode(theGraph, StpGWorld(PthLast(the2Path))));
    if ((!PtFlyPlc()) || !PlcInfinite(PtFlyPlc()))
	return;
    if (GphAwInvisible(theGraph, StpGAwAddr(PthLast(the2Path))))
	return;
    PthSEnd(the3Path, StpGWorld(PthLast(the2Path)));
    if (PthEnd(the2Path) == StpGWorld(PthLast(the2Path)))
	PtC4(StpGAwAddr(PthLast(the2Path)), PthEnd(the2Path));
    PthInsLast(the3Path, StpGAwAddr(PthLast(the2Path)), PthEnd(the2Path));
    while (TRUE)
	{
	    if (GphFresh(theGraph, PthEnd(the3Path), (unsigned long) 4L))
		{
		    GphConsume(theGraph, PthEnd(the3Path), (unsigned long) 4L);
		    if (PtCFFD() &&
			GphFresh(theGraph, PthEnd(the3Path),
				 (unsigned long) 1L))
			{
			    PthSEnd(thePath, PthEnd(the3Path));
			    PlcDecode(GphNode(theGraph, PthEnd(thePath)));
			    PtOnTheFly();
			    nowPath = the3Path;
			    ClDivNow();
			    SetInfNow();
			}
		    if ((theAwCount = GphAwCount(theGraph, PthEnd(the3Path)))
			!= (unsigned long) -1L)
			{
			    if (theAwCount != 0L)
				{
				    GphFlFirst(theGraph, PthEnd(the3Path),
					       &theAwFirst, &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    ptM = theAwCount;
				    while (TRUE)
					{
					    if (theNdAddr ==
						StpGWorld(PthLast(the2Path)))
						PtC4(theAwAddr, theNdAddr);
					    if (--ptM == 0L)
						break;
					    GphFlNext(theGraph, &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    PlcDecode(GphNode(theGraph, PthEnd(the3Path)));
			    FndEnbl();
			    if ((PtGECnt() <= 1L) || (!PtStub())
#ifdef SMODELS
				|| (PtSmodels() && (OnTheFly() || PtCFFD()) &&
				    (PtGInvisCnt() == 0L))
#endif
				)
				{
				    (void) PtrBlApply(TblTrn(), TrnReCheck);
				    PtGphUpdate();
				}
			    else if (PtDelAlg())
				PtDeletionAlg();
			    else
				PtIncrementalAlg();
			}
		    if (theAwCount != 0L)
			{
			    PthInsLast(the3Path, theAwFirst, theNdFirst);
			    continue;
			}
		}
	    if (!PthLast(the3Path))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(the3Path)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(the3Path)),
			  PthAEnd(the3Path));
	    else
		{
		    PthDelLast(the3Path);
		    if (!PthLast(the3Path))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  PtInfinite
* Purpose:   This on-the-fly verification routine searches for errors
*            of the form specified by the infinite path monitor states,
*            reject states and deadlock monitor states of the tester.
* Arguments:
* Returns:
* See also:  PtC1, PtC2, PtC3, PtC4, PtOnTheFly, PtInside, PtBuchi,
*            DoInfinite (inst.c)
*
****************************************************************************
*/
FE0(void, PtInfinite)
{
    nowPath = the2Path;
    ClDivNow();
    ClInfNow();
    if (GphFresh(theGraph, PthEnd(the2Path), (unsigned long) 2L))
	{
	    GphConsume(theGraph, PthEnd(the2Path), (unsigned long) 2L);
	    if (PtCFFD() &&
		GphFresh(theGraph, PthEnd(the2Path), (unsigned long) 1L))
		{
		    PthSEnd(thePath, PthEnd(the2Path));
		    PlcDecode(GphNode(theGraph, PthEnd(thePath)));
		    PtOnTheFly();
		    nowPath = the2Path;
		    ClDivNow();
		}
	    if ((theAwCount = GphAwCount(theGraph, PthEnd(the2Path)))
		!= (unsigned long) -1L)
		{
		    if (theAwCount != 0L)
			GphFlFirst(theGraph, PthEnd(the2Path),
				   &theAwFirst, &theNdFirst);
		}
	    else
		{
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    PlcDecode(GphNode(theGraph, PthEnd(the2Path)));
		    FndEnbl();
		    if ((PtGECnt() <= 1L) || (!PtStub())
#ifdef SMODELS
			|| (PtSmodels() && (OnTheFly() || PtCFFD()) &&
			    (PtGInvisCnt() == 0L))
#endif
			)
			{
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			}
		    else if (PtDelAlg())
			PtDeletionAlg();
		    else
			PtIncrementalAlg();
		}
	    if (theAwCount != 0L)
		{
		    PthInsLast(the2Path, theAwFirst, theNdFirst);
		    return;
		}
	    else if (!PthLast(the2Path))
		{
		    PthSEnd(the2Path, (unsigned long) -1L);
		    return;
		}
	}
    else
	{
	    SetInfNow();
	    PtInside();
	    ClInfNow();
	}
    if (FreeExpansion()
	?
	(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) != 0L)
	:
	(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) + 1L
	 < GphAwCount(theGraph, StpGWorld(PthLast(the2Path)))))
	{
	    GphFlNext(theGraph, StpAAwAddr(PthLast(the2Path)),
		      PthAEnd(the2Path));
	    return;
	}
    PthDelLast(the2Path);
    if (!PthLast(the2Path))
	PthSEnd(the2Path, (unsigned long) -1L);
}

/***************************************************************************
*
* Function:  PtActual
* Purpose:   This is an auxiliary routine for PtBuchi.
* Arguments:
* Returns:
* See also:  PtBuchi
*
****************************************************************************
*/
FE0(void, PtActual)
{
    register DATATYPE *pointer;
    register unsigned long a = 0L;
    DATATYPE *buffer;
    BOOL invisible;
    unsigned len;
    unsigned awLen;
    unsigned long awAddr;

    if (GphAwCount(theGraph, PthEnd(nowPath)) != (unsigned long) -1L)
	return;
    PlcDecode(GphNode(theGraph, PthEnd(nowPath)));
    buPlc = PtFlyPlc();
    PlcSCount(buPlc, (unsigned long) 0L);
    (void) PlcEncode(TblPlc(), EngABuffer(), EngABufSize());
    PthSEnd(thePath,
	    GphFndNode(the2Graph, EngBuffer(), (unsigned long *) NULL));
    if (PthEnd(thePath) == (unsigned long) -1L)
	Troba1("Internal error in PtActual");
    if ((theAwCount = GphAwCount(the2Graph, PthEnd(thePath)))
	!= (unsigned long) -1L)
	{
	    if (theAwCount == 0L)
		Troba1("Internal error in PtActual");
	    goto BUCHI;
	}
    while (TRUE)
	{
	    if ((GphAwCount(the2Graph, PthEnd(thePath)) == (unsigned long) -1L)
		&&
		((!PthLast(thePath))
		 ||
		 GphAwHot(the2Graph, StpGAwAddr(PthLast(thePath)))))
		{
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    if (PthLast(thePath))
			PlcDecode(GphNode(the2Graph, PthEnd(thePath)));
		    FndEnbl();
		    if ((PtGECnt() <= 1L) || (PtGInvisCnt() == 0L))
			{
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			}
		    else
			{
			    if (PtDelAlg())
				PtDeletionAlg();
			    else
				PtIncrementalAlg();
			}
		    if (theAwCount == 0L)
			Troba1("Internal error in PtActual");
		    if (PtRecurs())
			{
			    PthInsLast(thePath, theAwFirst, theNdFirst);
			    (void) BalTrInsert(theNodeTree, (char *)
					       StpAWorld(PthLast(thePath)),
					       (char *)
					       StpAWorld(PthLast(thePath)),
					       TRUE);
			    continue;
			}
		}
	    if (!PthLast(thePath))
		break;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(the2Graph, StpGAwAddr(PthLast(thePath))) != 0L)
		:
		(GphAwOrdinal(the2Graph, StpGAwAddr(PthLast(thePath))) + 1L
		 < GphAwCount(the2Graph, StpGWorld(PthLast(thePath)))))
		GphFlNext(the2Graph, StpAAwAddr(PthLast(thePath)),
			  PthAEnd(thePath));
	    else
		{
		    (void) BalTrDelete(theNodeTree, (char *) NULL,
				       (char *) StpAWorld(PthLast(thePath)),
				       FALSE);
		    PthDelLast(thePath);
		    if (!PthLast(thePath))
			break;
		}
	}
    theAwCount = GphAwCount(the2Graph, PthEnd(thePath));
 BUCHI: ;
    ptM = theAwCount;
    theAwCount = 0L;
    theAwFirst = (unsigned long) -1L;
    GphFlFirst(the2Graph, PthEnd(thePath), &theAwAddr, &theNdAddr);
    do
	{
	    PlcDecode(GphNode(the2Graph, theNdAddr));
	    invisible = GphAwInvisible(the2Graph, theAwAddr);
	    buffer = GphArrow(the2Graph, theAwAddr);
	    awLen = DataLen(buffer);
	    if (Ara())
		{
		    pointer = buffer;
		    GetValue(unsigned long, a, pointer);
		    a = invisible ? 0L : (a + 1L);
		    if (a > _araMaxAct)
			Troba3("Illegal data for PtActual: %lu > %lu",
			       a, _araMaxAct);
		}
	    LSTFOR (PlcGNext(buPlc), ptPlc, PLC)
		{
		    PlcSCount(ptPlc, (unsigned long) 1L);
		    theEvalState = PtState();
		    if ((*_buchiOK)())
			{
			    StaDiscard(theEvalState);
			    len = PlcEncode(TblPlc(), EngABuffer(),
					    EngABufSize());
			    theNdAddr = GphPutNode(theGraph, EngBuffer(),
						   len, &theBack);
			    awAddr = GphPutArrow(theGraph, buffer, awLen,
						 theAwCount,
						 PthEnd(nowPath), theNdAddr,
						 theAwFirst, theBack,
						 invisible);
			    if ((theAwCount == 0L) || FreeExpansion())
				{
				    theAwFirst = awAddr;
				    theNdFirst = theNdAddr;
				}
			    theAwCount++;
			    if (Ara())
				{
				    (void) fprintf(AraArcFile(), "%lu %lu %lu\n",
						   GphOrdinal(theGraph,
							      PthEnd(nowPath)),
						   GphOrdinal(theGraph, theNdAddr),
						   a);
				}
			}
		    else
			StaDiscard(theEvalState);
		    PlcSCount(ptPlc, (unsigned long) 0L);
		}
	    ptM--;
	    if (ptM != 0L)
		GphFlNext(the2Graph, &theAwAddr, &theNdAddr);
	} while (ptM != 0L);
    ClPtStub();
    PtGphUpdate();
    SetPtStub();
    PlcDecode(GphNode(theGraph, PthEnd(nowPath)));
}

#ifdef DEBUG
FS2(void, PtPthDump,
    PATH *, path,
    GRAPH *, graph)
{
    STEP *step;

    LSTNEST(PthGSteps(path), step, STEP)
	;
    LSTUNNEST (PthGSteps(path), step, STEP)
	(void) fprintf(_err, "%lu [%lu> ",
		       GphOrdinal(graph, StpGWorld(step)),
		       GphAwOrdinal(graph, StpGAwAddr(step)));
    (void) fprintf(_err, "%lu\n", GphOrdinal(graph, PthEnd(path)));
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  PtInner
* Purpose:   This is an auxiliary routine for PtBuchi.
* Arguments:
* Returns:
* See also:  PtBuchi
*
****************************************************************************
*/
FE0(void, PtInner)
{
    nowPath = the3Path;
    PlcDecode(GphNode(theGraph, PthEnd(the2Path)));
    if (!PlcInfinite(PtFlyPlc()))
	return;
    PthSEnd(the3Path, PthEnd(the2Path));
#ifdef DEBUG
    (void) fputs("the2Path\n", _err);
    PtPthDump(the2Path, theGraph);
#endif
    while (TRUE)
	{
	    if (GphFresh(theGraph, PthEnd(the3Path), (unsigned long) 4L))
		{
		    GphConsume(theGraph, PthEnd(the3Path), (unsigned long) 4L);
		    if (PtStub())
			PtActual();
#ifdef DEBUG
		    (void) fputs("the3Path\n", _err);
		    PtPthDump(the3Path, theGraph);
#endif
		    if ((theAwCount = GphAwCount(theGraph, PthEnd(the3Path)))
			!= (unsigned long) -1L)
			{
			    if (theAwCount != 0L)
				{
				    GphFlFirst(theGraph, PthEnd(the3Path),
					       &theAwFirst, &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    ptM = theAwCount;
				    while (TRUE)
					{
#ifdef DEBUG
					    (void)
						fprintf(_err, "[%lu> %lu\n",
							GphAwOrdinal(theGraph,
								     theAwAddr
								     ),
							GphOrdinal(theGraph,
								   theNdAddr));
					    (void) fflush(_err);
#endif
					    if (theNdAddr == PthEnd(the2Path))
						PtC4(theAwAddr, theNdAddr);
					    if (--ptM == 0L)
						break;
					    GphFlNext(theGraph, &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else if (PtStub())
			Troba1("Internal error in PtInner");
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    PlcDecode(GphNode(theGraph, PthEnd(the3Path)));
			    buPlc = PtFlyPlc();
			    FndEnbl();
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			}
		    if (theAwCount != 0L)
			{
			    PthInsLast(the3Path, theAwFirst, theNdFirst);
			    continue;
			}
		}
	    if (!PthLast(the3Path))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(the3Path)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(the3Path)),
			  PthAEnd(the3Path));
	    else
		{
		    PthDelLast(the3Path);
		    if (!PthLast(the3Path))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  PtBuchi
* Purpose:   This on-the-fly verification routine checks if the so far
*            computed intersection of the reachability graph and the
*            Buchi automaton (the latter corresponding to the negation of
*            the formula given by the user) accepts any word.
* Arguments:
* Returns:
* See also:  PtC1, PtC2, PtC3, PtC4, PtOnTheFly, PtInfinite, PtActual,
*            PtInner
*
****************************************************************************
*/
FE0(void, PtBuchi)
{
    nowPath = the2Path;
    ClDivNow();
    ClInfNow();
    if (GphFresh(theGraph, PthEnd(the2Path), (unsigned long) 2L))
	{
	    GphConsume(theGraph, PthEnd(the2Path), (unsigned long) 2L);
	    if (PtStub() && (PthEnd(the2Path) != GphStart(theGraph)))
		PtActual();
	    if ((theAwCount = GphAwCount(theGraph, PthEnd(the2Path)))
		!= (unsigned long) -1L)
		{
		    if (theAwCount != 0L)
			GphFlFirst(theGraph, PthEnd(the2Path),
				   &theAwFirst, &theNdFirst);
		}
	    else
		{
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    PlcDecode(GphNode(theGraph, PthEnd(the2Path)));
		    buPlc = PtFlyPlc();
		    FndEnbl();
		    if (PtStub())
			{
			    ClPtStub();
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			    SetPtStub();
			}
		    else
			{
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			}
		}
	    if (theAwCount != 0L)
		{
		    PthInsLast(the2Path, theAwFirst, theNdFirst);
		    if (GphFresh(theGraph, PthEnd(the2Path),
				 (unsigned long) 2L))
			return;
		}
	}
    while (TRUE)
	{
	    SetInfNow();
	    PtInner();
	    ClInfNow();
	    nowPath = the2Path;
	    if (PthLast(the2Path)
		&&
		(FreeExpansion()
		 ?
		 (GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) != 0L)
		 :
		 (GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) + 1L
		  < GphAwCount(theGraph, StpGWorld(PthLast(the2Path))))))
		{
		    GphFlNext(theGraph, StpAAwAddr(PthLast(the2Path)),
			      PthAEnd(the2Path));
		    if (GphFresh(theGraph, PthEnd(the2Path),
				 (unsigned long) 2L))
			return;
		}
	    else
		{
		    if (PthLast(the2Path))
			PthDelLast(the2Path);
		    else
			PthSEnd(the2Path, (unsigned long) -1L);
		    return;
		}
	}
}

/***************************************************************************
*
* Function:  PtProCFFD
* Purpose:   This is the main routine of the CFFD-preserving version of
*            the stubborn set method when there is no on-the-fly
*            verification.
* Arguments:
* Returns:
* See also:  PtOnTheFly, PtInfinite, PtBuchi, PtFollowers
*
****************************************************************************
*/
FE0(void, PtProCFFD)
{
    nowPath = thePath;
    while (TRUE)
	{
	    if ((!PthLast(thePath))
		||
		((GphAwCount(theGraph, PthEnd(thePath))
		  == (unsigned long) -1L)
		 &&
		 GphAwHot(theGraph, StpGAwAddr(PthLast(thePath)))))
		{
		    if (PthLast(thePath))
			PlcDecode(GphNode(theGraph, PthEnd(thePath)));
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    FndEnbl();
		    if ((PtGECnt() <= 1L) || (PtGInvisCnt() == 0L))
			{
			    (void) PtrBlApply(TblTrn(), TrnReCheck);
			    PtGphUpdate();
			}
		    else
			{
			    ClPtRecurs();
			    if (PtDelAlg())
				PtDeletionAlg();
			    else
				PtIncrementalAlg();
			    if (PtRecurs())
				{
				    PthInsLast(thePath, theAwFirst,
					       theNdFirst);
				    (void)
					BalTrInsert(theNodeTree,
						    (char *)
						    StpAWorld(PthLast(thePath)),
						    (char *)
						    StpAWorld(PthLast(thePath)),
						    TRUE);
				    continue;
				}
			}
		}
	    if (!PthLast(thePath))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(thePath)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(thePath)),
			  PthAEnd(thePath));
	    else
		{
		    (void) BalTrDelete(theNodeTree, (char *) NULL,
				       (char *) StpAWorld(PthLast(thePath)),
				       FALSE);
		    PthDelLast(thePath);
		    if (!PthLast(thePath))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  PtRedInc
* Purpose:   In the CFFD-preserving version of the incremental algorithm
*            of the stubborn set method, this routine computes a new
*            stubborn set having all visible transitions if the preliminary
*            stubborn set is not known to contain them and firing the
*            enabled transitions of the preliminary set would close
*            an invisible loop. The new stubborn set is then final.
* Arguments:
* Returns:
* See also:  TrnLookStack, TrnRedInc
*
****************************************************************************
*/
FE0(void, PtRedInc)
{
    SetPtRecurs();
    (void) PtrBlApply(TblTrn(), TrnRedInc);
    if (!PtRecurs())
	{
	    (void) PtrBlApply(TblTrn(), TrnAntiVisit);
	    if (PtGInvisCnt() > 1L)
		{
		    PtSKingTrn((TRN *) NULL);
		    (void) PtrBlApply(TblTrn(), TrnVisStub);
		    if (!PtKingTrn())
			{
			    PtCSgCount();
			    PtCNow();
			    PtSetBest(0L);
			    PtSzSBest((unsigned long) -1L);
			    PtSList();
			    (void) PtrBlApply(TblTrn(), TrnStub);
			    PtDList();
			}
		}
	    else
		PtSetBest(0L);
	}
}

/***************************************************************************
*
* Function:  PtIncrementalAlg
* Purpose:   This routine contains the incremental algorithm
*            of the stubborn set method.
* Arguments:
* Returns:
* See also:  PtDeletionAlg
*
****************************************************************************
*/
FE0(void, PtIncrementalAlg)
{
    (void) PtrBlApply(TblTrn(), TrnAntiVisit);
    PtCSgCount();
    PtCNow();
    PtSetBest(0L);
    PtSzSBest((unsigned long) -1L);
    if (PtLazy())
	{
	    if (TrRnd())
		{
		    if (OnTheFly())
			PtSCounter((unsigned long)
				   PtRandom((UNS32) PtGInvisCnt()));
		    else
			PtSCounter((unsigned long)
				   PtRandom((UNS32) PtGECnt()));
		}
	    else
		PtSCounter(0L);
	}
    if (OnTheFly() && !PtCFFD())
	{
	    if ((!PtFlyPlc()) || !PlcLivelock(PtFlyPlc()))
		{
		    if (PtGInvisCnt() != 0L)
			(void) PtrBlApply(TblTrn(), TrnVisStub);
		}
	    else if (PtGInvisCnt() == 1L)
		TrnSpan(ptInvFirst);
	    else if (PtGInvisCnt() != 0L)
		{
		    PtSList();
		    PtSKingTrn((TRN *) NULL);
		    (void) PtrBlApply(TblTrn(), TrnStub);
		    PtDList();
		    if (PtBest() == 0L)
			{
			    (void) PtrBlApply(TblTrn(), TrnAntiVisit);
			    TrnSpan(PtKingTrn());
			}
		}
	}
    else if (!PtCFFD())
	{
	    PtSList();
	    (void) PtrBlApply(TblTrn(), TrnStub);
	    PtDList();
	}
    else
	{
	    PtSList();
	    (void) PtrBlApply(TblTrn(), TrnStub);
	    PtDList();
	    if (PtBest() == 0L)
		{
		    (void) PtrBlApply(TblTrn(), TrnAntiVisit);
		    PtSKingTrn((TRN *) NULL);
		    (void) PtrBlApply(TblTrn(), TrnVisStub);
		    if (!PtKingTrn())
			{
			    PtCSgCount();
			    PtCNow();
			    PtSetBest(0L);
			    PtSzSBest((unsigned long) -1L);
			    PtSList();
			    (void) PtrBlApply(TblTrn(), TrnStub);
			    PtDList();
			}
		}
	    else if (PtSzBest() == PtGECnt())
		ptBranching = TRUE;
	    else if (((!OnTheFly()) || (!PtFlyPlc()) ||
		      !PlcLivelock(PtFlyPlc())) &&
		     PtrBlApply(TblTrn(), TrnWorseVis))
		PtRedInc();
	}
    if (ptDoIt)
	{
	    (void) PtrBlApply(TblTrn(), TrnBest);
	    PtGphUpdate();
	}
}
