/***************************************************************************
*
* Copyright (c) 1999 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet4.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptnet.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptstable.h"

#ifdef SMODELS
FE1(INEQ *, IneqCons,
    unsigned long, card)
{
    INEQ *ineq;

    ineq = IneqAlloc();
    IneqSCard(ineq, card);
    IneqCGe(ineq);
    IneqCLt(ineq);
    return(ineq);
}

FE1(INEQ *, IneqCopy,
    INEQ *, ineq)
{
    INEQ *copy;

    copy = IneqAlloc();
    IneqSCard(copy, IneqGCard(ineq));
    IneqSGe(copy, IneqGGe(ineq));
    IneqSLt(copy, IneqGLt(ineq));
    return(copy);
}

#ifdef DEBUG
FE2(void, IneqDump,
    INEQ *, ineq,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "INEQ card %lu ge 0x%lx lt 0x%lx\n",
		   IneqGCard(ineq), (unsigned long) IneqGGe(ineq),
		   (unsigned long) IneqGLt(ineq));
    (void) fflush(_err);
}
#endif

FE2(int, IneqKeyCmp,
    INEQ *, ineq,
    unsigned long *, key)
{
    if (IneqGCard(ineq) < *key)
	return(-1);
    if (IneqGCard(ineq) > *key)
	return(1);
    return(0);
}

TOOLS ineqTools =
{ (CFUNC) IneqCopy, Deallocate
#ifdef DEBUG
, IneqDump
#endif
 };

BALTOOLS ineqBalTools =
{ (CFUNC) IneqCopy, Deallocate
#ifdef DEBUG
, IneqDump
#endif
, IneqKeyCmp, (FOOL) NULL };
#endif

FS1(BOOL, TrnPriKeyRules,
    TRN *, trn)
{
    if (!TrnIsDiRho(trn))
	return(FALSE);
    if (!TrnGDiRho(trn))
	Troba1("No TrnGDiRho in TrnPriKeyRules");
    StartRule(TrnGDiRho(trn));
    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGCount(ptPlc) >= TrArcGWeight(ptTrArc))
		    continue;
		if (!PlcGGamma(ptPlc))
		    MkPlcGamma(ptPlc);
		ContRule(PlcGGamma(ptPlc));
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGCount(ptPlc) >= TrArcGWeight(ptTrArc))
		    continue;
		if (!PlcGGamma(ptPlc))
		    MkPlcGamma(ptPlc);
		ContRule(PlcGGamma(ptPlc));
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGCount(ptPlc) >= TrDrcGIn(ptTrDrc))
		    continue;
		if (!PlcGGamma(ptPlc))
		    MkPlcGamma(ptPlc);
		ContRule(PlcGGamma(ptPlc));
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGCount(ptPlc) >= TrDrcGIn(ptTrDrc))
		    continue;
		if (!PlcGGamma(ptPlc))
		    MkPlcGamma(ptPlc);
		ContRule(PlcGGamma(ptPlc));
	    }
    EndRule();
    return(FALSE);
}

FS2(void, TrnDecrRules,
    TRN *, trn1,
    TRN *, trn2)
{
    if (!TrnSatisfied(trn2))
	{
	    TrnFeed(trn2);
	    if (!TrnIsGamma(trn2))
		return;
	    if (!TrnGGamma(trn2))
		Troba1("No TrnGGamma in TrnDecrRules");
	    if (TrnIsKappa(trn1))
		{
		    if (!TrnGKappa(trn1))
			Troba1("No TrnGKappa in TrnDecrRules");
		    StartRule(TrnGKappa(trn1));
#ifdef SMODELS
		    if (PtLongPlay() &&
			TrnGActive(trn1))
			ContRule(TrnGActive(trn1));
#endif
		    ContRule(TrnGGamma(trn2));
		    EndRule();
		}
	    if (TrnIsEnRho(trn1))
		{
		    if (!TrnGEnRho(trn1))
			Troba1("No TrnGEnRho in TrnDecrRules");
		    StartRule(TrnGEnRho(trn1));
		    ContRule(TrnGGamma(trn2));
		    EndRule();
		}
	}
}

FS1(BOOL, TrnKeyRules,
    TRN *, trn)
{
    TRN *rtrn;

    if ((!TrnIsKappa(trn)) && !TrnIsEnRho(trn))
	return(FALSE);
    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnImpoverish(PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnImpoverish(PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnImpoverish(PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnImpoverish(PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnImpoverish(PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnImpoverish(PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnImpoverish(PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnImpoverish(PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnDecrRules(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnDecrRules(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnDecrRules(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnDecrRules(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnDecrRules(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnDecrRules(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnDecrRules(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnDecrRules(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (!TrnIsKappa(trn))
	return(FALSE);
    if (!TrnGKappa(trn))
	Troba1("No TrnGKappa in TrnKeyRules");
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnKeyRules");
#ifdef SMODELS
    if (PtLongPlay() && TrnGActive(trn))
	{
	    StartRule(TrnGKappa(trn));
	    NegContRule(TrnGActive(trn));
	    EndRule();
	}
#endif
    if (!TrnSatisfied(trn))
	{
	    TrnFeed(trn);
	    StartRule(TrnGKappa(trn));
#ifdef SMODELS
	    if (PtLongPlay() && TrnGActive(trn))
		ContRule(TrnGActive(trn));
#endif
	    ContRule(TrnGGamma(trn));
	    EndRule();
	}
    if (TrnGHigher(trn))
	LSTFOR (TrnGHigher(trn), rtrn, TRN)
	    {
		if (!TrnGDiRho(rtrn))
		    MkTrnDiRho(rtrn);
		StartRule(TrnGKappa(trn));
		ContRule(TrnGDiRho(rtrn));
		EndRule();
	    }
    return(FALSE);
}

FS1(BOOL, TrnRootRule,
    TRN *, trn)
{
    if (!TrnIsKappa(trn))
	return(FALSE);
    if (!TrnGKappa(trn))
	Troba1("No TrnGKappa in TrnRootRule");
    ContRule(TrnGKappa(trn));
    return(FALSE);
}

#ifdef SMODELS
FS1(BOOL, TrnMinRule,
    TRN *, trn)
{
    if (!TrnIsPlain(trn))
	return(FALSE);
    if (!TrnGPlain(trn))
	Troba1("No TrnGPlain in TrnMinRule");
    ContMinRule(TrnGPlain(trn));
    return(FALSE);
}
#endif

FS1(BOOL, TrnVisInner,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (!TrnIsGamma(trn))
	return(FALSE);
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnVisInner");
    StartRule(theta);
    ContRule(TrnGGamma(trn));
    EndRule();
    return(FALSE);
}

FS1(BOOL, TrnVisRules,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (!TrnIsGamma(trn))
	return(FALSE);
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnVisRules");
    if (
#ifdef SMODELS
	(!PtLongPlay()) &&
#endif
	(TrnThirsty(trn) || !TrnEnabled(trn)))
	return(FALSE);
    StartRule(TrnGGamma(trn));
#ifdef SMODELS
    if (PtLongPlay())
	{
	    if (TrnGActive(trn))
		ContRule(TrnGActive(trn));
	}
#endif
    ContRule(theta);
    EndRule();
    return(FALSE);
}

FS1(BOOL, TrnProRules,
    TRN *, trn)
{
    if (!TrnIsGamma(trn))
	return(FALSE);
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnProRules");
    if (
#ifdef SMODELS
	(!PtLongPlay()) &&
#endif
	!TrnThirsty(trn))
	return(FALSE);
    StartRule(TrnGGamma(trn));
#ifdef SMODELS
    if (PtLongPlay())
	{
	    if (TrnGActive(trn))
		ContRule(TrnGActive(trn));
	    if (TrnGToStack(trn))
		ContRule(TrnGToStack(trn));
	}
#endif
    ContRule(theta);
    EndRule();
    return(FALSE);
}

#ifdef SMODELS
FS1(BOOL, TrnMakePoor,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (!TrnIsGamma(trn))
	return(FALSE);
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnMakePoor");
    TrnImpoverish(trn);
    return(FALSE);
}
#endif

static unsigned blkInd;

FS1(BOOL, TrnBlockRules,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (!TrnIsGamma(trn))
	return(FALSE);
    if (!TrnGGamma(trn))
	Troba1("No TrnGGamma in TrnBlockRules");
    if (
#ifdef SMODELS
	(!PtLongPlay()) &&
#endif
	(TrnThirsty(trn) || !TrnEnabled(trn)))
	return(FALSE);
#ifdef SMODELS
    if (PtLongPlay())
	(void) PtrBlApply(visibles, TrnMakePoor);
    else
	{
#endif
	    for (blkInd = 0; TrnGBlock(trn, blkInd) != (unsigned) -1; blkInd++)
		LSTFOR (Block(TrnGBlock(trn, blkInd)), ptTrn, TRN)
		    {
			if (TrnGGamma(ptTrn))
			    TrnImpoverish(ptTrn);
		    }
#ifdef SMODELS
	}
#endif
    TrnFeed(trn);
    for (blkInd = 0; TrnGBlock(trn, blkInd) != (unsigned) -1; blkInd++)
	LSTFOR (Block(TrnGBlock(trn, blkInd)), ptTrn, TRN)
	    {
		if (TrnGGamma(ptTrn) && !TrnSatisfied(ptTrn))
		    {
			TrnFeed(ptTrn);
			StartRule(TrnGGamma(trn));
#ifdef SMODELS
			if (PtLongPlay() && TrnGActive(trn))
			    ContRule(TrnGActive(trn));
#endif
			ContRule(TrnGGamma(ptTrn));
			EndRule();
		    }
	    }
    return(FALSE);
}

#ifdef SMODELS
FS1(BOOL, TrnStable,
    TRN *, trn)
{
    if (TrnEnabled(trn))
	{
	    if ((!TrnGPlain(trn)) || Stable(TrnGPlain(trn)))
	        {
		    TrnImpoverish(trn);
		    ptM++;
		}
	    else
		TrnFeed(trn);
	}
    return(FALSE);
}
#endif

#ifdef DEBUG
static char *atomName = NULL;
#endif

FS0(void, MkABK)
{
#ifdef SMODELS
    VOID *alpha;
#endif

    if (
#ifdef SMODELS
	PtSmodels() ||
#endif
	!kappa)
	{
#ifdef DEBUG
	    (void) strcpy(atomName, "kappa");
	    kappa = NmAtom(atomName);
#else
	    kappa = FreshAtom();
#endif
#ifdef SMODELS
	    if (PtSmodels())
		{
#ifdef DEBUG
		    (void) strcpy(atomName, "alpha");
		    alpha = NmAtom(atomName);
#else
		    alpha = FreshAtom();
#endif
		    StartRule(alpha);
		    ContRule(kappa);
		    NegContRule(alpha);
		    EndRule();
		}
#endif
	}
    if ((OnTheFly() || PtCFFD()) &&
	(
#ifdef SMODELS
	    PtSmodels() ||
#endif
	    !theta))
	{
#ifdef DEBUG
	    (void) strcpy(atomName, "theta");
	    theta = NmAtom(atomName);
#else
	    theta = FreshAtom();
#endif
	}
    if (OnTheFly() && (!PtCFFD()) && (!Blocked()) &&
	((!PtFlyPlc()) || !PlcLivelock(PtFlyPlc())))
	{
	    if (
#ifdef SMODELS
		PtSmodels() ||
#endif
		!kappakey)
		{
#ifdef DEBUG
		    (void) strcpy(atomName, "kappakey");
		    kappakey = NmAtom(atomName);
#else
		    kappakey = FreshAtom();
#endif
		}
	    StartRule(kappa);
	    ContRule(theta);
	    EndRule();
	    StartRule(kappa);
	    ContRule(kappakey);
	    EndRule();
	    StartRule(kappakey);
	}
    else
	StartRule(kappa);
}

#ifdef SMODELS
FS1(BOOL, IneqExtra,
    INEQ *, ineq)
{
    ReverseRules(IneqGGe(ineq), IneqGLt(ineq));
    if (ptIneq)
	{
	    SimpleRule(IneqGGe(ptIneq), IneqGGe(ineq));
	    SimpleRule(IneqGLt(ineq), IneqGLt(ptIneq));
	}
    ptIneq = ineq;
    return(FALSE);
}

FS1(BOOL, PlcIneqExtra,
    PLC *, plc)
{
    if (PlcGIneqs(plc))
        {
	    ptIneq = NULL;
	    (void) BalTrApply(PlcGIneqs(plc), IneqExtra);
	}
    return(FALSE);
}

FS1(BOOL, PlcSmoSet,
    PLC *, plc)
{
    if (PlcGIneqs(plc))
        {
	    if ((ptIneq = (INEQ *)
                 BalTrLtEqPrefix(PlcGIneqs(plc),
                                 (char *) PlcACount(plc), FALSE)))
	        {
	            SetPosCompute(IneqGGe(ptIneq));
	            if ((ptIneq = (INEQ *)
		         BalTrNext(PlcGIneqs(plc), (char *) ptIneq,
			           (char *) IneqACard(ptIneq), (char *) NULL)))
		        SetNegCompute(IneqGGe(ptIneq));
		}
            else if ((ptIneq = (INEQ *) BalTrFirst(PlcGIneqs(plc))))
	        SetNegCompute(IneqGGe(ptIneq));
            else
		Troba1("Internal error in PlcSmoSet");
	}
    return(FALSE);
}

FS1(BOOL, PlcSmoRes,
    PLC *, plc)
{
    if (PlcGIneqs(plc))
        {
	    if ((ptIneq = (INEQ *)
                 BalTrLtEqPrefix(PlcGIneqs(plc),
                                 (char *) PlcACount(plc), FALSE)))
	        {
	            ResPosCompute(IneqGGe(ptIneq));
	            if ((ptIneq = (INEQ *)
		         BalTrNext(PlcGIneqs(plc), (char *) ptIneq,
			           (char *) IneqACard(ptIneq), (char *) NULL)))
		        ResNegCompute(IneqGGe(ptIneq));
		}
            else if ((ptIneq = (INEQ *) BalTrFirst(PlcGIneqs(plc))))
	        ResNegCompute(IneqGGe(ptIneq));
            else
		Troba1("Internal error in PlcSmoRes");
	}
    return(FALSE);
}

FS1(BOOL, TrnSmoSet,
    TRN *, trn)
{
    if (TrnGToStack(trn) && TrnEnabled(trn))
	{
	    TrnWater(trn);
	    SetPtRecurs();
	    TrnLookStack(trn);
	    if (PtRecurs())
		SetNegCompute(TrnGToStack(trn));
	    else
		{
		    TrnSalt(trn);
		    SetPosCompute(TrnGToStack(trn));
		}
	}
    return(FALSE);
}

FS1(BOOL, TrnSmoRes,
    TRN *, trn)
{
    if (TrnGToStack(trn) && TrnEnabled(trn))
	{
	    if (TrnThirsty(trn))
		ResPosCompute(TrnGToStack(trn));
	    else
		ResNegCompute(TrnGToStack(trn));
	}
    return(FALSE);
}
#endif

FE4(void, AuxDefer,
    TRN *, trn,
    TRPROM *, trProm,
    PLC *, plc,
    unsigned long, weight)
{
    DEFER *defer;

    if ((((BOOL) !TrnGDefers(trn)) != (BOOL) !curDefers) ||
	(TrPromGArcs(trProm) && !curDefers))
	Troba1("Incompatibility in AuxDefer");
    if (!curDefers)
	{
	    curDefers = BalTrCreate(&deferBalTools);
	    defer = (DEFER *) NULL;
	}
    else
	defer = (DEFER *) BalTrElement(curDefers, (char *) plc);
    if (!defer)
	{
	    defer = DeferCons(plc);
	    if (BalTrInsert(curDefers, (char *) defer, (char *) plc, TRUE))
		Troba1("Failure in AuxDefer");
	    if (!TrnGDefers(trn))
		TrnSDefers(trn);
	    (void) LstInsLast(TrnGDefers(trn), (char *) defer);
	}
    if (!TrPromGArcs(trProm))
	TrPromSArcs(trProm);
    (void) LstInsLast(TrPromGArcs(trProm), (char *)
		      OtArcCons(defer, weight));
}

FE2(void, TrnAuxProm,
    TRN *, trn1,
    TRN *, trn2)
{
    TRPROM *trProm;
    TRARC *trArc;
    TRDRC *trDrc;

    if ((!TrnGLower(trn2)) || LstEmpty(TrnGLower(trn2)))
	return;
    if (((BOOL) !TrnGPromoted(trn1)) != (BOOL) !curTrns)
	Troba1("Internal error in TrnAuxProm");
    if (!curTrns)
	{
	    curTrns = BalTrCreate(&ptrBalTools);
	    TrnSPromoted(trn1);
	}
    else if (BalTrInsert(curTrns, (char *) trn2, (char *) trn2, TRUE))
	return;
    trProm = TrPromCons(trn2);
    (void) LstInsLast(TrnGPromoted(trn1), (char *) trProm);
    if (TrnGPurIn(trn2))
	LSTFOR (TrnGPurIn(trn2), trArc, TRARC)
	    AuxDefer(trn1, trProm, TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGNeutral(trn2))
	LSTFOR (TrnGNeutral(trn2), trArc, TRARC)
	    AuxDefer(trn1, trProm, TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGMosIn(trn2))
	LSTFOR (TrnGMosIn(trn2), trDrc, TRDRC)
	    AuxDefer(trn1, trProm, TrDrcGPlc(trDrc), TrDrcGIn(trDrc));
    if (TrnGMosOut(trn2))
	LSTFOR (TrnGMosOut(trn2), trDrc, TRDRC)
	    AuxDefer(trn1, trProm, TrDrcGPlc(trDrc), TrDrcGIn(trDrc));
}

FE1(BOOL, TrnPromote,
    TRN *, trn)
{
    if (TrnGPurOut(trn))
	LSTFOR (TrnGPurOut(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnAuxProm(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnAuxProm(trn, PlDrcGTrn(ptPlDrc));
		if (PlcGNeutral(ptPlc))
		    LSTFOR (PlcGNeutral(ptPlc), ptPlArc, PLARC)
			TrnAuxProm(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosIn(ptPlc))
		    LSTFOR (PlcGMosIn(ptPlc), ptPlDrc, PLDRC)
			TrnAuxProm(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			TrnAuxProm(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			TrnAuxProm(trn, PlDrcGTrn(ptPlDrc));
		if (PlcGNeutral(ptPlc))
		    LSTFOR (PlcGNeutral(ptPlc), ptPlArc, PLARC)
			TrnAuxProm(trn, PlArcGTrn(ptPlArc));
		if (PlcGMosIn(ptPlc))
		    LSTFOR (PlcGMosIn(ptPlc), ptPlDrc, PLDRC)
			TrnAuxProm(trn, PlDrcGTrn(ptPlDrc));
	    }
    if (((BALTREE *) NULL) != curTrns)
	BalTrDestroy(curTrns, FALSE);
    if (((BALTREE *) NULL) != curDefers)
	BalTrDestroy(curDefers, FALSE);
    return(FALSE);
}

#ifdef SMODELS
static BOOL constructed = FALSE;
#endif

FE0(void, ClDelegate)
{
#ifdef SMODELS
    if (PtLongPlay() && constructed)
	DelApi();
#endif
    LstDestroy(globMota, TRUE);
#ifdef DEBUG
    ClDeallocate(atomName);
#endif
}

FE0(void, PtDelegate)
{
    MOTA *mota;

#ifdef DEBUG
    if (!atomName)
	atomName = (char *) Allocate(8192);
#endif
#ifdef SMODELS
    if ((!PtLongPlay()) || !constructed)
	{
#endif
	    (void) PtrBlApply(TblTrn(), TrnPreAO);
	    (void) PtrBlApply(TblPlc(), PlcIniFlag);
	    if ((1 != IngGNofClasses()) &&
		(
#ifdef SMODELS
		    PtSmodels() ? (!constructed) :
#endif
		    !globMota))
		{
		    curTrns = (BALTREE *) NULL;
		    curDefers = (BALTREE *) NULL;
		    (void) PtrBlApply(TblTrn(), TrnPromote);
		}
#ifdef SMODELS
	    if (PtSmodels())
		MkApi();
	    else
#endif
		if (!globMota)
		    globMota = LstCreate(&motaTools);
		else
		    LSTFOR (globMota, mota, MOTA)
			MotaClear(mota);
	    (void) PtrBlApply(TblTrn(), TrnAndOr);
	    MkABK();
	    (void) PtrBlApply(TblTrn(), TrnRootRule);
	    EndRule();
	    (void) PtrBlApply(TblTrn(), TrnKeyRules);
	    if (1 != IngGNofClasses())
		(void) PtrBlApply(TblTrn(), TrnPriKeyRules);
	    if (OnTheFly() || PtCFFD())
		{
		    if (!visibles)
			VCollect();
		    (void) PtrBlApply(visibles, TrnVisInner);
		    if (Blocked())
			(void) PtrBlApply(visibles, TrnBlockRules);
		    else
			(void) PtrBlApply(visibles, TrnVisRules);
		    (void) PtrBlApply(TblTrn(), TrnProRules);
		}
#ifdef SMODELS
	    if (PtLongPlay())
		(void) PtrBlApply(TblPlc(), PlcIneqExtra);
	    if (PtSmodels())
		{
		    StartMinRule();
		    (void) PtrBlApply(TblTrn(), TrnMinRule);
		    EndRule();
		}
	}
    if (PtSmodels())
	{
	    if (PtLongPlay())
		{
		    (void) PtrBlApply(TblPlc(), PlcSmoSet);
		    if (OnTheFly() || PtCFFD())
			(void) PtrBlApply(TblTrn(), TrnSmoSet);
		}
	    if ((!PtLongPlay()) || !constructed)
		{
		    DoneApi();
		    constructed = TRUE;
		    SmoCalcInit();
		}
	    if (!MkModel())
		Troba1("Internal MkModel error in PtDelegate");
	    do
		{
#ifdef DEBUG
		    ShowModel();
#endif
		    ptM = (unsigned long) 0L;
		    (void) PtrBlApply(TblTrn(), TrnStable);
		    if (((unsigned long) 0L) == ptM)
			Troba1("Internal TrnStable error in PtDelegate");
		} while ((((unsigned long) 1L) != ptM) && MkModel());
	    if (PtLongPlay())
		{
		    SmoRevert();
		    (void) PtrBlApply(TblPlc(), PlcSmoRes);
		    if (OnTheFly() || PtCFFD())
			(void) PtrBlApply(TblTrn(), TrnSmoRes);
		}
	    else
		DelApi();
#ifdef DEBUG
	    CutVerbose(_out);
#endif
	    if (PtRecurs())
		{
		    ClPtRecurs();
		    PtrBlApply(TblTrn(), TrnVisSmoCheck);
		}
	    if (ptDoIt)
		{
		    (void) PtrBlApply(TblTrn(), TrnSmoCheck);
		    PtGphUpdate();
		}
	}
    else
	{
#endif
            DelThem();
#ifdef DEBUG
	    CutVerbose(_err);
#endif
	    if (PtRecurs())
		{
		    ClPtRecurs();
		    PtrBlApply(TblTrn(), TrnVisCheck);
		}
	    if (ptDoIt)
		{
		    (void) PtrBlApply(TblTrn(), TrnDelCheck);
		    PtGphUpdate();
		}
#ifdef SMODELS
	}
#endif
}

FE1(void, MkPlcGamma,
    PLC *, plc)
{
    if (PlcGGamma(plc))
	Troba1("Internal error in MkPlcGamma");
#ifdef DEBUG
    (void) strcpy(atomName, "gamma__");
    PlKeySp(PlcGKey(plc), atomName);
    PlcSGamma(plc, NmAtom(atomName));
#else
    PlcSGamma(plc, FreshAtom());
#endif
}

#ifdef SMODELS
FE2(INEQ *, MkPlcIneq,
    PLC *, plc,
    unsigned long, card)
{
    INEQ *ineq;
#ifdef DEBUG
    char *s;
#endif
    
    if (((unsigned long) 0L) == card)
	Troba1("Internal error in MkPlcIneq");
    if (PlcGIneqs(plc))
	{
	    if ((ineq = (INEQ *) BalTrElement(PlcGIneqs(plc), (char *) &card)))
		return(ineq);
	}
    else
	PlcSIneqs(plc);
    ineq = IneqCons(card);
    (void) BalTrInsert(PlcGIneqs(plc), (char *) ineq, (char *) &card, TRUE);
#ifdef DEBUG
    (void) strcpy(atomName, "mk__");
    PlKeySp(PlcGKey(plc), atomName);
    (void) strcat(atomName, "__");
    s = atomName + strlen(atomName);
    (void) sprintf(s, "ge__%lu", card);
    IneqSGe(ineq, NmAtom(atomName));
    s[0] = 'l';
    s[1] = 't';
    IneqSLt(ineq, NmAtom(atomName));
#else
    IneqSGe(ineq, FreshAtom());
    IneqSLt(ineq, FreshAtom());
#endif
    return(ineq);
}
#endif

FE1(void, MkTrnGamma,
    TRN *, trn)
{
    if (TrnGGamma(trn) || !TrnIsGamma(trn))
	Troba1("Internal error in MkTrnGamma");
#ifdef DEBUG
    (void) strcpy(atomName, "gamma__");
    InsSp(TrnGKey(trn), atomName);
    TrnSGamma(trn, NmAtom(atomName));
#else
    TrnSGamma(trn, FreshAtom());
#endif
}

#ifdef DEBUG
FE2(void, TrnDispLabel,
    TRN *, trn,
    FILE *, file)
{
    if (!atomName)
	atomName = (char *) Allocate(8192);
    *atomName = '\0';
    InsSp(TrnGKey(trn), atomName);
    (void) fputs(atomName, file);
}
#endif

#ifdef SMODELS
FE1(void, MkTrnPlain,
    TRN *, trn)
{
    if (TrnGPlain(trn) || !TrnIsPlain(trn))
	Troba1("Internal error in MkTrnPlain");
#ifdef DEBUG
    (void) strcpy(atomName, "plain__");
    InsSp(TrnGKey(trn), atomName);
    TrnSPlain(trn, NmAtom(atomName));
#else
    TrnSPlain(trn, FreshAtom());
#endif
}
#endif

FE1(void, MkTrnKappa,
    TRN *, trn)
{
    if (TrnGKappa(trn) || !TrnIsKappa(trn))
	Troba1("Internal error in MkTrnKappa");
#ifdef DEBUG
    (void) strcpy(atomName, "kappa__");
    InsSp(TrnGKey(trn), atomName);
    TrnSKappa(trn, NmAtom(atomName));
#else
    TrnSKappa(trn, FreshAtom());
#endif
}

#ifdef SMODELS
FE1(void, MkTrnActive,
    TRN *, trn)
{
    if (TrnGActive(trn))
	Troba1("Internal error in MkTrnActive");
#ifdef DEBUG
    (void) strcpy(atomName, "active__");
    InsSp(TrnGKey(trn), atomName);
    TrnSActive(trn, NmAtom(atomName));
#else
    TrnSActive(trn, FreshAtom());
#endif
}

FE1(void, MkTrnToStack,
    TRN *, trn)
{
    VOID *offStack;

    if (TrnGToStack(trn) || !TrnIsToStack(trn))
	Troba1("Internal error in MkTrnToStack");
#ifdef DEBUG
    (void) strcpy(atomName, "tostack__");
    InsSp(TrnGKey(trn), atomName);
    TrnSToStack(trn, NmAtom(atomName));
#else
    TrnSToStack(trn, FreshAtom());
#endif
    if (PtSmodels())
	{
#ifdef DEBUG
	    (void) strcpy(atomName, "offstack__");
	    InsSp(TrnGKey(trn), atomName);
	    offStack = NmAtom(atomName);
#else
	    offStack = FreshAtom();
#endif
	    if (TrnGActive(trn))
		{
		    StartRule(TrnGToStack(trn));
		    ContRule(TrnGActive(trn));
		    NegContRule(offStack);
		    EndRule();
		    StartRule(offStack);
		    ContRule(TrnGActive(trn));
		    NegContRule(TrnGToStack(trn));
		    EndRule();
		}
	    else
		ReverseRules(TrnGToStack(trn), offStack);
	}
}
#endif

FE1(void, MkTrnDiRho,
    TRN *, trn)
{
    if (TrnGDiRho(trn) || !TrnIsDiRho(trn))
	Troba1("Internal error in MkTrnDiRho");
#ifdef DEBUG
    (void) strcpy(atomName, "dirho__");
    InsSp(TrnGKey(trn), atomName);
    TrnSDiRho(trn, NmAtom(atomName));
#else
    TrnSDiRho(trn, FreshAtom());
#endif
}

FE1(void, MkTrnEnRho,
    TRN *, trn)
{
    if (TrnGEnRho(trn) || !TrnIsEnRho(trn))
	Troba1("Internal error in MkTrnEnRho");
#ifdef DEBUG
    (void) strcpy(atomName, "enrho__");
    InsSp(TrnGKey(trn), atomName);
    TrnSEnRho(trn, NmAtom(atomName));
#else
    TrnSEnRho(trn, FreshAtom());
#endif
}

FE1(void, MkTrnReRho,
    TRN *, trn)
{
    if (TrnGReRho(trn) || !TrnIsReRho(trn))
	Troba1("Internal error in MkTrnReRho");
#ifdef DEBUG
    (void) strcpy(atomName, "rerho__");
    InsSp(TrnGKey(trn), atomName);
    TrnSReRho(trn, NmAtom(atomName));
#else
    TrnSReRho(trn, FreshAtom());
#endif
}

#ifdef SMODELS
FE2(void, MkTAGamma,
    TRN *, trn,
    TRARC *, trArc)
{
    if (TrArcGGamma(trArc) || !TrnIsGamma(trn))
	Troba1("Internal error in MkTAGamma");
#ifdef DEBUG
    (void) strcpy(atomName, "gamma__");
    InsSp(TrnGKey(trn), atomName);
    (void) strcat(atomName, "____");
    PlKeySp(PlcGKey(TrArcGPlc(trArc)), atomName);
    TrArcSGamma(trArc, NmAtom(atomName));
#else
    TrArcSGamma(trArc, FreshAtom());
#endif
}

FE2(void, MkTDGamma,
    TRN *, trn,
    TRDRC *, trDrc)
{
    if (TrDrcGGamma(trDrc) || !TrnIsGamma(trn))
	Troba1("Internal error in MkTDGamma");
#ifdef DEBUG
    (void) strcpy(atomName, "gamma__");
    InsSp(TrnGKey(trn), atomName);
    (void) strcat(atomName, "____");
    PlKeySp(PlcGKey(TrDrcGPlc(trDrc)), atomName);
    TrDrcSGamma(trDrc, NmAtom(atomName));
#else
    TrDrcSGamma(trDrc, FreshAtom());
#endif
}
#endif

FE2(void, MkTAStrong,
    TRN *, trn,
    TRARC *, trArc)
{
#ifdef DEBUG
    PLC *plc;
#endif

    if ((!TrnIsPlain(trn)) ||
	(((BOOL) !TrArcGOder(trArc)) != (BOOL) !TrArcGStrong(trArc)) ||
	(((BOOL) !TrArcGOder(trArc)) != (BOOL) !TrArcGWeak(trArc)))
	Troba1("Internal error in MkTAStrong");
    if (!TrArcGOder(trArc))
	{
#ifdef DEBUG
	    plc = TrArcGPlc(trArc);
	    (void) strcpy(atomName, "oder__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrArcSOder(trArc, NmAtom(atomName));
	    (void) strcpy(atomName, "strong__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrArcSStrong(trArc, NmAtom(atomName));
	    (void) strcpy(atomName, "weak__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrArcSWeak(trArc, NmAtom(atomName));
#else
	    TrArcSOder(trArc, FreshAtom());
	    TrArcSStrong(trArc, FreshAtom());
	    TrArcSWeak(trArc, FreshAtom());
#endif
	}
    StartRule(TrnGGamma(trn));
#ifdef SMODELS
    if (PtLongPlay())
	ContRule(TrnGActive(trn));
#endif
    ContRule(TrArcGOder(trArc));
    EndRule();
    StartRule(TrArcGOder(trArc));
#ifdef SMODELS
    if (PtLongPlay())
	ContRule(TrnGActive(trn));
#endif
    ContRule(TrArcGStrong(trArc));
    ContRule(TrArcGWeak(trArc));
    EndRule();
}

FE2(void, MkTDStrong,
    TRN *, trn,
    TRDRC *, trDrc)
{
#ifdef DEBUG
    PLC *plc;
#endif

    if ((!TrnIsPlain(trn)) ||
	(((BOOL) !TrDrcGOder(trDrc)) != (BOOL) !TrDrcGStrong(trDrc)) ||
	(((BOOL) !TrDrcGOder(trDrc)) != (BOOL) !TrDrcGWeak(trDrc)))
	Troba1("Internal error in MkTDStrong");
    if (!TrDrcGOder(trDrc))
	{
#ifdef DEBUG
	    plc = TrDrcGPlc(trDrc);
	    (void) strcpy(atomName, "oder__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrDrcSOder(trDrc, NmAtom(atomName));
	    (void) strcpy(atomName, "strong__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrDrcSStrong(trDrc, NmAtom(atomName));
	    (void) strcpy(atomName, "weak__");
	    InsSp(TrnGKey(trn), atomName);
	    (void) strcat(atomName, "____");
	    PlKeySp(PlcGKey(plc), atomName);
	    TrDrcSWeak(trDrc, NmAtom(atomName));
#else
	    TrDrcSOder(trDrc, FreshAtom());
	    TrDrcSStrong(trDrc, FreshAtom());
	    TrDrcSWeak(trDrc, FreshAtom());
#endif
	}
    StartRule(TrnGGamma(trn));
#ifdef SMODELS
    if (PtLongPlay())
	ContRule(TrnGActive(trn));
#endif
    ContRule(TrDrcGOder(trDrc));
    EndRule();
    StartRule(TrDrcGOder(trDrc));
#ifdef SMODELS
    if (PtLongPlay())
	ContRule(TrnGActive(trn));
#endif
    ContRule(TrDrcGStrong(trDrc));
    ContRule(TrDrcGWeak(trDrc));
    EndRule();
}

FE2(void, MkTrPromAtom,
    TRN *, trn,
    TRPROM *, trProm)
{
    if (TrPromGAtom(trProm) || !TrnIsPlain(trn))
	Troba1("Internal error in MkTrPromAtom");
#ifdef DEBUG
    (void) strcpy(atomName, "prom__");
    InsSp(TrnGKey(trn), atomName);
    (void) strcat(atomName, "____");
    InsSp(TrnGKey(TrPromGTrn(trProm)), atomName);
    TrPromSAtom(trProm, NmAtom(atomName));
#else
    TrPromSAtom(trProm, FreshAtom());
#endif
}

FE2(void, MkDeferAtom,
    TRN *, trn,
    DEFER *, defer)
{
    if (DeferGAtom(defer) || !TrnIsPlain(trn))
	Troba1("Internal error in MkDeferAtom");
#ifdef DEBUG
    (void) strcpy(atomName, "defer__");
    InsSp(TrnGKey(trn), atomName);
    (void) strcat(atomName, "____");
    PlKeySp(PlcGKey(DeferGPlc(defer)), atomName);
    DeferSAtom(defer, NmAtom(atomName));
#else
    DeferSAtom(defer, FreshAtom());
#endif
}

FE2(void, SimpleRule,
    VOID *, head,
    VOID *, tail)
{
    StartRule(head);
    ContRule(tail);
    EndRule();
}

#ifdef SMODELS
FE2(void, ReverseRules,
    VOID *, atom1,
    VOID *, atom2)
{
    StartRule(atom1);
    NegContRule(atom2);
    EndRule();
    StartRule(atom2);
    NegContRule(atom1);
    EndRule();
}
#endif

FE0(VOID *, FreshAtom)
{
    MOTA *mota;

#ifdef SMODELS
    if (PtSmodels())
	return(SmoFreshAtom());
#endif
    mota = MotaAlloc();
    MotaCleFlag(mota);
    MotaCleCounter(mota);
    MotaCleCursor(mota);
    MotaCleDown(mota);
    MotaNilPreds(mota);
#ifdef DEBUG
    MotaNilLabel(mota);
#endif
    (void) LstInsLast(globMota, (char *) mota);
    return((VOID *) mota);
}

#ifdef DEBUG
FE1(VOID *, NmAtom,
    char *, name)
{
    MOTA *mota;

#ifdef SMODELS
    if (PtSmodels())
	return(SmoNmAtom(name));
#endif
    mota = (MOTA *) FreshAtom();
    MotaSetLabel(mota, StrCopy(name));
    return((VOID *) mota);
}
#endif

FE1(void, MotaDracsid,
    MOTA *, mota)
{
    if (mota)
	{
	    MotaDesPreds(mota);
#ifdef DEBUG
	    MotaDesLabel(mota);
#endif
	    Deallocate((char *) mota);
	}
}

FS1(void, MotaPrint,
    MOTA *, mota)
{
    (void) fprintf(_err, "0x%lx", (unsigned long) mota);
    if (mota)
	{
	    (void) fprintf(_err, ":%u:%lu:%lu", MotaGFlag(mota),
			   MotaCounter(mota), MotaDupCounter(mota));
#ifdef DEBUG
	    if (MotaGLabel(mota))
		(void) fprintf(_err, ":\"%s\"", MotaGLabel(mota));
#endif
	}
}

#ifdef DEBUG
FE2(void, MotaDump,
    MOTA *, mota,
    unsigned, depth)
{
    MOTA *pred;

    DumpIndent(depth);
    (void) fputs("MOTA ", _err);
    MotaPrint(mota);
    (void) fputc((int) '\n', _err);
    if (MotaGPreds(mota))
	LSTFOR (MotaGPreds(mota), pred, MOTA)
	    {
		DumpIndent(depth + 1);
		(void) fputs("pred ", _err);
		MotaPrint(pred);
		(void) fputc((int) '\n', _err);
	    }
    (void) fflush(_err);
}
#endif

TOOLS motaTools =
{ (CFUNC) NULL, MotaDracsid
#ifdef DEBUG
, MotaDump
#endif
 };

FE1(void, MotaClear,
    MOTA *, mota)
{
    MotaCleFlag(mota);
    MotaCleCounter(mota);
    MotaCleCursor(mota);
    MotaCleDown(mota);
    MotaDesPreds(mota);
}

static MOTA *ruleHead = NULL;
static unsigned long nRuleTail = 0L;

FE1(void, StartRule,
    VOID *, atom)
{
    MOTA *mota;

#ifdef SMODELS
    if (PtSmodels())
	{
	    SmoStartRule(atom);
	    return;
	}
#endif
    mota = (MOTA *) atom;
    if ((!mota) || MotaIsOr(mota) || nRuleTail || ruleHead)
	{
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fputs("\nStartRule(", _err);
	    MotaPrint(mota);
	    (void) fputs("), h ", _err);
	    MotaPrint(ruleHead);
	    (void) fprintf(_err, ", n %lu\n", nRuleTail);
	    (void) fflush(_err);
	    Troba0();
	}
    ruleHead = mota;
    nRuleTail = 0L;
#ifdef DEBUG
    (void) fputs(MotaGLabel(mota), _err);
    (void) fputs(" :- ", _err);
    (void) fflush(_err);
#endif
}

FE1(void, ContRule,
    VOID *, atom)
{
    MOTA *mota;

#ifdef SMODELS
    if (PtSmodels())
	{
	    SmoContRule(atom);
	    return;
	}
#endif
    mota = (MOTA *) atom;
    nRuleTail++;
    if ((!ruleHead) || (!mota) || (!nRuleTail) ||
	(nRuleTail > (unsigned long) ~nRuleTail) ||
	(mota == ruleHead) || MotaIsOr(ruleHead) ||
	(IsOrArity(nRuleTail) && MotaCounter(ruleHead)))
	{
	    nRuleTail--;
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fputs("\nContRule(", _err);
	    MotaPrint(mota);
	    (void) fputs("), h ", _err);
	    MotaPrint(ruleHead);
	    (void) fprintf(_err, ", n %lu\n", nRuleTail);
	    (void) fflush(_err);
	    Troba0();
	}
    if (!MotaGPreds(mota))
	MotaIniPreds(mota);
    (void) LstInsLast(MotaGPreds(mota), (char *) ruleHead);
#ifdef DEBUG
    if (nRuleTail > (unsigned long) 1L)
	(void) fputs(", ", _err); 
    (void) fputs(MotaGLabel(mota), _err);
    (void) fflush(_err);
#endif
}

FE0(void, EndRule)
{
#ifdef SMODELS
    if (PtSmodels())
	{
	    SmoEndRule();
	    return;
	}
#endif
    if ((!ruleHead) || MotaIsOr(ruleHead) || (!nRuleTail) ||
	(nRuleTail > (unsigned long) ~nRuleTail) ||
	(IsOrArity(nRuleTail) && MotaCounter(ruleHead)))
	{
	    if  (_out != _err)
		{
		    (void) fputc((int) '\n', _out);
		    (void) fflush(_out);
		}
	    (void) fputs("\nEndRule(), h ", _err);
	    MotaPrint(ruleHead);
	    (void) fprintf(_err, ", n %lu\n", nRuleTail);
	    (void) fflush(_err);
	    Troba0();
	}
    if (!MotaCounter(ruleHead))
	MotaSetCounter(ruleHead, nRuleTail);
    ruleHead = NULL;
    nRuleTail = 0L;
#ifdef DEBUG
    (void) fputs(".\n", _err);
    (void) fflush(_err);
#endif
}
