/***************************************************************************
*
* Copyright (c) 1999 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptstable.cc
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     An interface from PROD to Smodels.
*
* Version history:
*     $Log$
***************************************************************************/

#ifdef SMODELS

#include <iostream>
#include "smodels.h"
#include "api.h"
#include "atomrule.h"
#ifdef FE0
#undef FE0
#endif
#ifdef DE0
#undef DE0
#endif
#ifdef FE1
#undef FE1
#endif
#ifdef DE1
#undef DE1
#endif
#ifdef FE2
#undef FE2
#endif
#ifdef DE2
#undef DE2
#endif
#ifdef VOID
#undef VOID
#endif
#define FE0(k0,n0) extern "C" k0 n0()
#define DE0(k0,n0) FE0(k0,n0);
#define FE1(k0,n0,k1,n1) extern "C" k0 n0(k1 n1)
#define DE1(k0,n0,k1,n1) FE1(k0,n0,k1,n1);
#define FE2(k0,n0,k1,n1,k2,n2) extern "C" k0 n0(k1 n1,k2 n2)
#define DE2(k0,n0,k1,n1,k2,n2) FE2(k0,n0,k1,n1,k2,n2);
#define VOID void
#include "ptstable.h"

using namespace std;

static Smodels *smodels;
static Api *api;
#ifdef DEBUG
static int tailStarted = 0;
static int minRule = 0;
#endif

FE0(void, MkApi)
{
    smodels = new Smodels;
    api = new Api(&(smodels->program));
}

FE0(VOID *, SmoFreshAtom)
{
    return((VOID *) (api->new_atom()));
}

FE1(void, SmoStartRule,
    VOID *, atom)
{
    api->begin_rule(BASICRULE);
    api->add_head((Atom *) atom);
#ifdef DEBUG
    cout << ((Atom *) atom)->atom_name() << " :- ";
#endif
}


FE0(void, StartMinRule)
{
    api->begin_rule(OPTIMIZERULE);
#ifdef DEBUG
    minRule = 1;
    cout << "minimize { ";
#endif
}

FE1(void, SmoContRule,
    VOID *, atom)
{
    api->add_body((Atom *) atom, true);
#ifdef DEBUG
    if (tailStarted)
        cout << ", ";
    else
        tailStarted = 1;
    cout << ((Atom *) atom)->atom_name();
#endif
}

FE1(void, NegContRule,
    VOID *, atom)
{
    api->add_body((Atom *) atom, false);
#ifdef DEBUG
    if (tailStarted)
        cout << ", ";
    else
        tailStarted = 1;
    cout << "not " << ((Atom *) atom)->atom_name();
#endif
}

FE1(void, ContMinRule,
    VOID *, atom)
{
    api->add_body((Atom *) atom, true, 1);
#ifdef DEBUG
    if (tailStarted)
        cout << ", ";
    else
        tailStarted = 1;
    cout << ((Atom *) atom)->atom_name() << " = 1";
#endif
}

FE0(void, SmoEndRule)
{
    api->end_rule();
#ifdef DEBUG
    if (minRule)
        {
	    cout << " }";
	    minRule = 0;
	}
    else
        cout << '.';
    cout << endl;
    tailStarted = 0;
#endif
}

FE0(void, DoneApi)
{
    api->done();
}

FE1(void, SetPosCompute,
    VOID *, atom)
{
    api->set_compute((Atom *) atom, true);
}

FE1(void, SetNegCompute,
    VOID *, atom)
{
    api->set_compute((Atom *) atom, false);
}

FE1(void, ResPosCompute,
    VOID *, atom)
{
    api->reset_compute((Atom *) atom, true);
}

FE1(void, ResNegCompute,
    VOID *, atom)
{
    api->reset_compute((Atom *) atom, false);
}

FE0(void, SmoCalcInit)
{
    smodels->init();
}

FE0(void, SmoRevert)
{
    smodels->revert();
}

FE0(int, MkModel)
{
    if (smodels->model(true, false))
	return(1);
    return(0);
}

FE1(int, Stable,
    VOID *, atom)
{
    if (((Atom *) atom)->Bpos)
	return(1);
    return(0);
}

FE0(void, DelApi)
{
    delete api;
    delete smodels;
}

#ifdef DEBUG

FE1(VOID *, SmoNmAtom,
    char *, name)
{
    VOID *atom;

    atom = SmoFreshAtom();
    api->set_name((Atom *) atom, name);
    /* cout << name << " :- " << name << '.' << endl; */
    return(atom);
}

FE0(void, ShowProg)
{
    smodels->program.print();
}

FE0(void, ShowModel)
{
    smodels->printAnswer();
}

#endif

#endif
