/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  t.c
*
* Part of: PROD's program t
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The "t" program transforms a file written in the "tpl language"
*     into a corresponding file in the C language. The best way to learn
*     the tpl language is to compare the gen.tpl file of PROD to the gen.c
*     file produced by "t < gen.tpl > gen.c". (If you want an exact
*     description of the tpl language, just read the below code.)
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"

#ifdef STRFMT
#undef STRFMT
#endif
#define STRFMT "%s"

/***************************************************************************
*
* Function:  tOutput
* Purpose:   This routine transforms a line or a partial line written
*            in the tpl language into its correspondent in the C language.
* Arguments: s - a line or a partial line written in the tpl language
*            newLine - if this is 0, s is not a line but a partial line
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, tOutput,
    char *, s,
    char *, buffer,
    int, newline)
{
    LIST *args;
    char *t;
    char c;
    BOOL started;
    
    if (!*s)
	{
	    if (newline)
                (void)
		    fputs
		    ("((void) fputc((int) '\\n', f), nOutpLines++);",
		     _out);
	    return;
	}
    (void) fputs("((void) fprintf(f, \"", _out);
    args = LstCreate(&strTools);
    started = TRUE;
    while ((*s) && (*s != '$') && (*s != '#') &&
           (*s != '%') && (*s != '@'))
	{
	    if (*s == '"')
		(void) fputs("\\\"", _out);
	    else if (*s == '\\')
		(void) fputs("\\\\", _out);
	    else
		(void) fputc((int) *s, _out);
	    s++;
	}
    while (*s)
	{
            c = *s++;
            switch (c)
		{
		case '$':
                    if (*s == c) 
			{
                            (void) fputc((int) c, _out);
			    s++;
			}
		    else 
			{
			    (void) fputs(STRFMT, _out);
			    for (t = buffer; (*s) && (*s != '$');
				 *t++ = *s++)
				;
			    *t = '\0';
			    if (*s)
				s++;
			    (void) LstInsLast(args, StrCopy(buffer));
			}
		    break;
                case '#':
		case '%':
                case '@':
                    if (*s == c)
			{
                            (void) fputc((int) c, _out);
			    s++;
			}
		    else
			{
			    (void) fputs(STRFMT, _out);
			    if (newline)
				{
                                    for (t = s; (*t) && (*t != c); t++)
					;
				    if ((!(t[0])) || !(t[1]))
					(void) fputs("\\n", _out);
				}
			    (void) fputc((int) '"', _out);
			    LSTFOR (args, t, char)
				    {
					(void) fputs(", ", _out);
					(void) fputs(t, _out);
				    }
			    LstDestroy(args, TRUE);
			    started = FALSE;
                            (void) fputs(", ", _out);
                            (void) fputs(('#' == c) ? "Ulong"
					 : (('%' == c) ? "Uns" : "Dom"),
					 _out);
                            (void) fputs("Str(", _out);
                            for (; (*s) && (*s != c);
				 (void) fputc((int) *s++, _out))
				;
                            (void) fputs("))", _out);
			    if (*s)
				s++;
			    if (*s)
				{
				    (void) fputs(", (void) fprintf(f, \"",
						 _out);
				    args = LstCreate(&strTools);
				    started = TRUE;
				}
			}
		    break;
		default:
		    ;
		}
            while ((*s) && (*s != '$') && (*s != '#') &&
                   (*s != '%') && (*s != '@'))
		{
		    if (*s == '"')
			(void) fputs("\\\"", _out);
		    else if (*s == '\\')
			(void) fputs("\\\\", _out);
		    else
			(void) fputc((int) *s, _out);
		    s++;
		}
	}
    if (started)
	{
	    if (newline)
		(void) fputs("\\n\"", _out);
	    else
		(void) fputc((int) '"', _out);
	    LSTFOR (args, t, char)
		    {
			(void) fputs(", ", _out);
			(void) fputs(t, _out);
		    }
	    LstDestroy(args, TRUE);
	    (void) fputc((int) ')', _out);
	}
    if (newline)
        (void) fputs(", nOutpLines++", _out);
    (void) fputs(");", _out);
}

/***************************************************************************
*
* Function:  main
* Purpose:   See the "Description:" in the beginning of this file.
* Arguments:
* Returns:   0
* See also:
*
****************************************************************************
*/
MAIN0()
{
    char *line;
    char *buffer;
    char *s;
    unsigned n;
    unsigned p;
    BOOL rdStat;
    
    SetIO();
    DynamInit();
    n = (unsigned) BUFSIZ;
    p = n;
    line = (char *) Allocate(n);
    buffer = (char *) Allocate(n);
    while (TRUE)
	{
	    rdStat = BruteReadln(&line, &n, _in);
	    if ((!rdStat) && !*line)
		break;
	    for (s = line; (*s) && (*s != '\n'); s++)
		;
	    *s = '\0';
	    for (s = line; (*s > '\0') && (*s <= ' ');
		 (void) fputc((int) *s++, _out))
		;
	    if ((s[0] != '@') || ((s[1] != '@') && (s[1] != '+')))
		{
		    (void) fputs(s, _out);
		    (void) fputc((int) '\n', _out);
		}
	    else
		{
		    if (p != n)
			{
			    p = n;
			    Deallocate(buffer);
			    buffer = (char *) Allocate(n);
			}
		    tOutput(s + 2, buffer, s[1] == '@');
		    (void) fputc((int) '\n', _out);
		}
	    if (!rdStat)
		break;
	}
    Deallocate(line);
    Deallocate(buffer);
    DynamClose();
    Exit(0);
    return(0);
}
