/***************************************************************************
*
* Copyright (c) 1989 -- 2006
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  trans2.c
*
* Part of: PROD's program prpp
*
* Authors: Johannes Helander and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See trans.c.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "place.h"
#include "prpp.h"
#include "lexer.h"
#include "macro.h"
#include "output.h"
#include "parse.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "net.h"
#include "pplst.h"
#include "cmdnm.h"
#include "prpexpr.h"
#include "trans.h"

/***************************************************************************
*
* Function:  trans_var_im_expr
* Purpose:
* Arguments: tran - 
*            place -
*            isin -
* Returns:
* See also:
*
****************************************************************************
*/
FE3(void, trans_var_im_expr,
    TRANSITION *, tran,
    PLACE *, place,
    BOOL, isin)
{
    PP_TOKEN *token;
    PP_TOKEN *nekot;
    PP_TOKEN *dToken;
    LIST *list;
    SYMMULTUP *symmultup;
    SYMMULTUP *syt;
    SYMBOL *cffSym;
    unsigned long multi = 0L;
    ARC *arc;
    LIST *symsum;
    LIST *symlist;
    VARIABLE *var;
    DOMAIN value;
    VOID *cursor;

    dToken = (PP_TOKEN *) NULL;
    value = MINDOMAIN;
    if (isin)
        {
            LSTFOR (TraGInputs(tran), arc, ARC)
                if (ArcGPlace(arc) == place)
                    break;
        }
    else
        {
            LSTFOR (TraGOutputs(tran), arc, ARC)
                if (ArcGPlace(arc) == place)
                    break;
        }
    symsum = place ? (arc ? ArcGSymsum(arc) : LstCreate(&sytTools)) : NULL;
    for (;;)
	{
	    cffSym = NULL;
	    list = LstCreate(&ppTokTools);
	    (void) LstInsFirst(ppLsLs, (char *) list);
	    dToken = AppPrs(PPTUPLE_LEFT, FALSE);
	    if (((LIST *) LstDelFirst(ppLsLs, FALSE)) != list)
		Troba1("Internal error in trans_var_im_expr");
	    TOKEN_MATCH(dToken, PPTUPLE_LEFT,
			((void)
			 sprintf(errStr,
				 "\"<.\" in \"%s block\" of transition %s",
				 isin ? "in" : "out", TraGName(tran)),
			 errStr));
	    PpTokDiscard(dToken);
	    LSTFOR (list, nekot, PP_TOKEN)
		{
		    if ((nekot->type != PPSPACE) &&
			(nekot->type != PPCOMMENT) &&
			(nekot->type != PPNEWLINE))
			break;
		}
	    if (!nekot)
		{
		    multi = 1L;
		    LstDestroy(list, TRUE);
		    goto CONT;
		}
	    LSTFOR (list, nekot, PP_TOKEN)
		{
		    if (nekot->type == PPNAME)
			break;
		}
	    if (!nekot)
		{
		    PrpExpr((MARKING **) NULL, (char *) NULL, (unsigned) 0,
			    &list, &value);
		    if (errorFlag != NO_ERROR)
			goto recover;
		    multi = (unsigned long) value;
		    goto CONT;
		}
	    symlist = LstCreate(&symTools);
	    LSTFOR (list, nekot, PP_TOKEN)
		{
		    if (nekot->type == PPNAME)
			{
			    cursor = LstGet(list);
			    token = (PP_TOKEN *) LstStep(list);
			    while (token && ((token->type == PPSPACE) ||
					     (token->type == PPNEWLINE) ||
					     (token->type == PPCOMMENT)))
				token = (PP_TOKEN *) LstStep(list);
			    LstSet(list, cursor);
			    if (token &&
				((token->type == PPOPENING_PARENTHESIS) ||
				 (token->type == PPOPENING_BRACKET) ||
				 (token->type == PPOPENING_SQUARE_BRACKET) ||
				 (token->type == PPTUPLE_LEFT)))
				{
				    (void)
					LstInsLast(symlist, (char *)
						   SymCons(SK_CONSTANT,
							   StrCopy(nekot->token)));
				    continue;
				}
			    var = TrVarLookup(nekot->token, tran);
			    if (!var)
				{
				    var = VarCons(StrCopy(nekot->token));
				    VarSIsCoeff(var);
				    if (isin)
					VarSIsIn(var);
				    else
					VarSIsOut(var);
				    (void)
					LstInsLast(TraGVariables(tran),
						   (char *) var);
				}
			    else
				{
				    VarSIsCoeff(var);
				    if (isin)
					VarSIsIn(var);
				    else if (!VarIsIn(var))
					VarSIsOut(var);
				}
			    (void) LstInsLast(symlist, (char *)
					      SymCons(SK_VARIABLE,
						      (char *) var));
			}
		    else
			(void) LstInsLast(symlist, (char *) 
					  SymCons(SK_CONSTANT,
						  StrCopy(nekot->token)));
		}
	    LstDestroy(list, TRUE);
	    cffSym = SymCons(SK_OPERATION, (char *) symlist);
	CONT: ;
	    symmultup = trans_expression(tran, isin,
					 (cffSym || (multi != 0L))
					 ? place : (PLACE *) NULL);
	    if (symmultup)
		{
		    if (cffSym)
			{
			    SytSCffSym(symmultup, cffSym);
			    (void) LstInsLast(symsum, (char *) symmultup);
			}
                    else if (SytGArity(symmultup))
			{
			    SytSMulti(symmultup, multi);
			    (void) LstInsLast(symsum, (char *)
					      symmultup);
			}
		    else
			{
			    LSTFOR (symsum, syt, SYMMULTUP)
                                if ((!SytGArity(syt)) && !SytGCffSym(syt))
				    break;
			    if (syt)
				{
				    SytIMulti(syt, multi);
				    SytDiscard(symmultup);
				}
			    else
				{
				    SytSMulti(symmultup, multi);
				    (void) LstInsLast(symsum, (char *)
						      symmultup);
				}
			}
		}
	    else if (cffSym)
		SymDiscard(cffSym);
	    dToken = gTSkEolAndSpace();  
	    if (!dToken)
		{
		    (void) sprintf(errStr,
				   "Transition %s: unexpected end of file",
				   TraGName(tran));
		    panic(errStr);
		}
	    switch (dToken->type)
		{
		case PPPLUS:
		    break;
		case PPCLOSING_BRACKET:
		    (void) sprintf(errStr,
				   "Transition %s: \"%s block\" missing \";\" before end \"}\"",
				   TraGName(tran), isin ? "in" : "out");
		    warning(errStr);
		    unget_token(dToken);
		    /* intentional abuse of switch */ 
		case PPSEMICOLON:
                    if (place && !arc)
			{
                            if (LstEmpty(symsum))
                                LstDestroy(symsum, TRUE);
                            else if (isin)
				(void) TraAttInput(tran, place, symsum);
			    else
				(void) TraAttOutput(tran, place, symsum);
			}
		    PpTokDiscard(dToken);
		    return;
		default:
		    TOKEN_MATCH(dToken, PPNEVERMATCH,
				((void) sprintf(errStr,
						"\"+\" or \";\" in \"%s block\" of transition %s",
						isin ? "in" : "out",
						TraGName(tran)),
				 errStr));
		}
	    PpTokDiscard(dToken);
	}
 recover: ;
    if (dToken)
	PpTokDiscard(dToken);
    error_flag = 1;
    if (place && !arc)
	LstDestroy(symsum, TRUE);
}

/***************************************************************************
*
* Function:  trans_in_out
* Purpose:
* Arguments: tran - 
*            isin - 
* Returns:
* See also:
*
****************************************************************************
*/
FE2(BOOL, trans_in_out,
    TRANSITION *, tran,
    BOOL, isin)
{
    PP_TOKEN *token;
    char *name;
    PLACE *place;
    
    name = (char *) NULL;
    token = gTSkEolAndSpace();
    TOKEN_MATCH(token, PPOPENING_BRACKET,
		((void) sprintf(errStr, "\"{\" after \"%s\" in transition %s",
				isin ? "in" : "out", TraGName(tran)),
		 errStr));
    PpTokDiscard(token);
    for (token = gTSkEolAndSpace(); 
	 token && (token->type != PPCLOSING_BRACKET); 
	 token = gTSkEolAndSpace())
	{
	    TOKEN_MATCH(token, PPNAME,
			((void) sprintf(errStr,
					"place name in \"%s block\" of transition %s",
					isin ? "in" : "out", TraGName(tran)),
			 errStr));
	    name = StrCopy(token->token);
	    PpTokDiscard(token);
	    token = gTSkEolAndSpace();
	    TOKEN_MATCH(token, PPCOLON,
			((void) sprintf(errStr,
					"\":\" after \"%s\" in \"%s block\" of transition %s",
					name, isin ? "in" : "out",
					TraGName(tran)),
			 errStr));
	    PpTokDiscard(token);
	    LSTFOR (PlaceLst, place, PLACE)
		if (!strcmp(name, PlaGName(place)))
		    break;
            if (!place)
		{
		    (void) sprintf(errStr, "Place %s not defined", name);
		    error(errStr);
		}
	    Deallocate(name);
	    name = (char *) NULL;
            trans_var_im_expr(tran, place, isin);
	}
    if (token)
	PpTokDiscard(token);
    return(TRUE);
  recover: ;
    if (((PP_TOKEN *) NULL) != token)
	PpTokDiscard(token);
    if (name)
	Deallocate(name);
    return(FALSE);
}

/***************************************************************************
*
* Function:  trans_gate
* Purpose:
* Arguments: tran - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BOOL, trans_gate,
    TRANSITION *, tran)
{
    PP_TOKEN *token;
    PP_TOKEN *nekot;
    PP_TOKEN *dToken;
    LIST *list;
    LIST *symlist;
    VARIABLE *var;
    VOID *cursor;

    symlist = LstCreate(&symTools);
    list = LstCreate(&ppTokTools);
    (void) LstInsFirst(ppLsLs, (char *) list);
    dToken = AppPrs(PPSEMICOLON, FALSE);
    if (((LIST *) LstDelFirst(ppLsLs, FALSE)) != list)
	Troba1("Internal error in trans_gate");
    TOKEN_MATCH(dToken, PPSEMICOLON,
		((void) sprintf(errStr, "end \";\" of a gate of transition %s",
				TraGName(tran)), errStr));
    PpTokDiscard(dToken);
    LSTFOR (list, nekot, PP_TOKEN)
	{
	    if (nekot->type == PPNAME)
		{
		    cursor = LstGet(list);
		    token = (PP_TOKEN *) LstStep(list);
		    while (token && ((token->type == PPSPACE) ||
				     (token->type == PPNEWLINE) ||
				     (token->type == PPCOMMENT)))
			token = (PP_TOKEN *) LstStep(list);
		    LstSet(list, cursor);
		    if (token && ((token->type == PPOPENING_PARENTHESIS) ||
				  (token->type == PPOPENING_BRACKET) ||
				  (token->type == PPOPENING_SQUARE_BRACKET) ||
				  (token->type == PPTUPLE_LEFT)))
			{
			    (void) LstInsLast(symlist, (char *)
					      SymCons(SK_CONSTANT,
						      StrCopy(nekot->token)));
			    continue;
			}
		    var = TrVarLookup(nekot->token, tran);
		    if (!var)
			{
			    var = VarCons(StrCopy(nekot->token));
			    VarSIsGat(var);
			    (void) LstInsLast(TraGVariables(tran),
					      (char *) var);
			}
		    else
			VarSIsGat(var);
		    (void) LstInsLast(symlist, (char *)
				      SymCons(SK_VARIABLE, (char *) var));
		}
	    else
		(void) LstInsLast(symlist, (char *) 
				  SymCons(SK_CONSTANT, StrCopy(nekot->token)));
	}
    LstDestroy(list, TRUE);
    (void) LstInsLast(TraGGates(tran), (char *)
		      SymCons(SK_OPERATION, (char *) symlist));
    return(TRUE);
 recover: ;
    if (dToken)
	PpTokDiscard(dToken);
    LstDestroy(list, TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  trans_comp
* Purpose:
* Arguments: tran - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BOOL, trans_comp,
    TRANSITION *, tran)
{
    PP_TOKEN *token;
    PP_TOKEN *dToken;
    LIST *list;

    list = (LIST *) NULL;
    dToken = gTSkEolAndSpace();
    TOKEN_MATCH(dToken, PPOPENING_BRACKET,
		((void) sprintf(errStr,
				"\"{\" after \"comp\" in transition %s",
				TraGName(tran)), errStr));
    PpTokDiscard(dToken);
    list = LstCreate(&ppTokTools);
    (void) LstInsFirst(ppLsLs, (char *) list);
    dToken = AppPrs(PPCLOSING_BRACKET, TRUE);
    if (((LIST *) LstDelFirst(ppLsLs, FALSE)) != list)
	Troba1("Internal error in trans_comp");
    TOKEN_MATCH(dToken, PPCLOSING_BRACKET,
		((void) sprintf(errStr,
				"end \"}\" of comp block of transition %s",
				TraGName(tran)), errStr));
    PpTokDiscard(dToken);
    LSTFOR (list, token, PP_TOKEN)
	{
	    if(!(token->token))
		continue;
	    (void) LstInsLast(TraGComputation(tran), (char *)
			      StrCopy(token->token));
	}
    LstDestroy(list, TRUE);
    return(TRUE);
 recover: ;
    if (dToken)
	PpTokDiscard(dToken);
    if (list)
	LstDestroy(list, TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  trans
* Purpose:
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, trans)
{
    PP_TOKEN *token;
    TRANSITION *tran;
    int gotout, gotin, gotlet;
    VARIABLE *var;
    ARC *arc;
    SYMMULTUP *symmultup;
    SYMBOL *symbol, *cffSym;
    LIST *cffList;
    TUPLE *low, *high;
    unsigned i;
    DOMAIN x;
    char *format;
    BOOL endFound;

    token = (PP_TOKEN *) NULL;
    endFound = FALSE;
    gotlet = gotout = gotin = 0;
    tran = PrsTransLine();
    read_tokens_match_newline(); 
    if(!tran)
	goto recover;
    for(;;)
	{
	    token = gTSkEolAndSpace();
	    if (!token)
		{
		    (void) sprintf(errStr,
				   "Transition %s: unexpected end of file",
				   TraGName(tran));
		    panic(errStr);
		}
	    if (token->type == PPDV)
		{
		    PpTokDiscard(token);
		    token = gTSkSpace();
		    TOKEN_MATCH(token, PPNAME,
				((void) sprintf(errStr,
						"#endtr of transition %s",
						TraGName(tran)), errStr));
		    if (strcmp(token->token, "endtr")) 
			TOKEN_MATCH(token, PPNEVERMATCH,
				    ((void) sprintf(errStr,
						    "#endtr of transition %s",
						    TraGName(tran)), errStr));
		    PpTokDiscard(token);
		    endFound = TRUE;
		    if (!TraGComputation(tran))
			LSTFOR (TraGVariables(tran), var, VARIABLE)
			    if (VarIsOut(var))
				{
				    (void) sprintf(errStr,
						   "Transition %s: output variable %s\nexists, but there is no comp block",
						   TraGName(tran),
						   VarGName(var)); 
				    error(errStr);
				    goto recover;
				}
		    LSTFOR (TraGInputs(tran), arc, ARC)
			(void) LstInsLast(TraGArcs(tran),
					  (char *) arc);
		    LSTFOR (TraGOutputs(tran), arc, ARC)
			(void) LstInsLast(TraGArcs(tran),
					  (char *) arc);
		    LSTFOR (TraGVariables(tran), var, VARIABLE)
			{
			    VarReset(var);
			    VarCHard(var);
			    VarCBad(var);
			    VarCUgly(var);
			}
		    LSTFOR (TraGInputs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if (SytGCffSym(symmultup))
				    continue;
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (SymIsVariable(symbol =
						      SytGSymbol(symmultup,
								 i)))
					VarSet(SymGVariable(symbol));
			    }
		    LSTFOR (TraGInputs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if (SytGCffSym(symmultup))
				    continue;
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (!SymVars(SytGSymbol(symmultup, i)))
					goto recover;
			    }
		    LSTFOR (TraGInputs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if ((!(cffSym = SytGCffSym(symmultup))) ||
				    !SymSatisfied(cffSym))
				    continue;
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (SymIsVariable(symbol =
						      SytGSymbol(symmultup, i))
					&& !VarUnified(var =
						       SymGVariable(symbol)))
					{
					    VarSHard(var);
					    VarSUgly(var);
					}
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (!SymVars(SytGSymbol(symmultup, i)))
					goto recover;
				for (i = 0; i < SytGArity(symmultup);
				     i++)
				    if (SymIsVariable(symbol =
						      SytGSymbol(symmultup,
								 i)))
					VarCHard(SymGVariable(symbol));
			    }
		    LSTFOR (TraGInputs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if (!(cffSym = SytGCffSym(symmultup)))
				    continue;
				for (i = 0; i < SytGArity(symmultup); i++)
				    {
					symbol =
					    SytGSymbol(symmultup, i);
					if (!SymIsVariable(symbol))
					    continue;
					var = SymGVariable(symbol);
					if (VarBad(var) || VarIsCoeff(var) ||
					    !VarUgly(var))
					    continue;
					VarSet(var);
					if (!VarGCoeffs(var))
					    VarSCoeffs(var);
					if (LstEmpty(cffList = VarGCoeffs(var))
					    || (cffSym != (SYMBOL *)
						LstLast(cffList)))
					    (void) LstInsLast(cffList, (char *)
							      cffSym);
				    }
			    }
		    LSTFOR (TraGVariables(tran), var, VARIABLE)
			{
			    if (VarBad(var))
				continue;
			    if (VarIsIn(var) && !VarUnified(var))
				{
				    VarSBad(var);
				    (void) sprintf(errStr,
						   "Input variable %s of %s is not unifiable",
						   VarGName(var),
						   TraGName(tran));
				    error(errStr);
				    goto recover;
				}
			    if (VarIsFold(var) && (!VarIsIn(var))
				&& !VarIsOut(var))
				{
				    VarSBad(var);
				    (void) sprintf(errStr,
						   "Folding variable %s of %s is neither an input",
						   VarGName(var),
						   TraGName(tran));
				    error(errStr);
				    error("nor output variable");
				    goto recover;
				}
			    if (VarIsGat(var) && !VarIsIn(var))
				{
				    VarSBad(var);
				    (void) sprintf(errStr,
						   "Gate variable %s of %s is not an input variable",
						   VarGName(var),
						   TraGName(tran));
				    error(errStr);
				    goto recover;
				}
                            if (VarIsLet(var) && !VarIsIn(var))
				{
				    VarSBad(var);
				    (void) sprintf(errStr,
                                                   "\"let\" variable %s of %s is not an input variable",
						   VarGName(var),
						   TraGName(tran));
				    error(errStr);
				    goto recover;
				}
                            if ((!placeLimits) && VarIsIn(var) &&
                                !VarIsLet(var))
				{
				    VarSBad(var);
				    (void) sprintf(errStr,
                                                   "Input variable %s of %s is not a \"let\" variable",
						   VarGName(var),
						   TraGName(tran));
				    error(errStr);
				    goto recover;
				}
			}
		    LSTFOR (TraGArcs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if (SytGCffSym(symmultup))
				    continue;
				low = NULL;
				high = NULL;
				if (PlaGLow(ArcGPlace(arc)))
				    LSTFOR (PlaGLow(ArcGPlace(arc)),
					    low, TUPLE)
					{
					    if (TupGArity(low) ==
						SytGArity(symmultup))
						break;
					    else if (TupGArity(low) >
						     SytGArity(symmultup))
						{
						    low = NULL;
						    break;
						}
					}
				if (PlaGHigh(ArcGPlace(arc)))
				    LSTFOR (PlaGHigh(ArcGPlace(arc)),
					    high, TUPLE)
					{
					    if (TupGArity(high) ==
						SytGArity(symmultup))
						break;
					    else if (TupGArity(high) >
						     SytGArity(symmultup))
						{
						    high = NULL;
						    break;
						}
					}
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (SymIsVariable(symbol =
						      SytGSymbol(symmultup,
								 i)))
					{
					    var = SymGVariable(symbol);
					    VarSHard(var);
					    if (low &&
						((x = TupGField(low, i))
						 > VarGLow(var)))
						VarSLow(var, x);
					    if (high &&
						((x = TupGField(high, i))
						 < VarGHigh(var)))
						VarSHigh(var, x);
					}
			    }
		    LSTFOR (TraGVariables(tran), var, VARIABLE)
			if (VarIsIn(var) && !VarHard(var))
			    {
				VarSLow(var, (DOMAIN) MAXDOMAIN);
				VarSHigh(var, (DOMAIN) MINDOMAIN);
			    }
		    LSTFOR (TraGArcs(tran), arc, ARC)
			LSTFOR (ArcGSymsum(arc), symmultup, SYMMULTUP)
			    {
				if (!SytGCffSym(symmultup))
				    continue;
				low = NULL;
				high = NULL;
				if (PlaGLow(ArcGPlace(arc)))
				    LSTFOR (PlaGLow(ArcGPlace(arc)),
					    low, TUPLE)
					{
					    if (TupGArity(low) ==
						SytGArity(symmultup))
						break;
					    else if (TupGArity(low) >
						     SytGArity(symmultup))
						{
						    low = NULL;
						    break;
						}
					}
				if (PlaGHigh(ArcGPlace(arc)))
				    LSTFOR (PlaGHigh(ArcGPlace(arc)),
					    high, TUPLE)
					{
					    if (TupGArity(high) ==
						SytGArity(symmultup))
						break;
					    else if (TupGArity(high) >
						     SytGArity(symmultup))
						{
						    high = NULL;
						    break;
						}
					}
				for (i = 0; i < SytGArity(symmultup); i++)
				    if (SymIsVariable(symbol =
						      SytGSymbol(symmultup, i))
					&& VarIsIn(var = SymGVariable(symbol))
					&& !VarHard(var))
					{
					    if (low &&
						((x = TupGField(low, i))
						 < VarGLow(var)))
						VarSLow(var, x);
					    else if (!low)
						VarSLow(var, (DOMAIN)
							MINDOMAIN);
					    if (high &&
						((x = TupGField(high, i))
						 > VarGHigh(var)))
						VarSHigh(var, x);
					    else if (!high)
						VarSHigh(var, (DOMAIN)
							 MAXDOMAIN);
					}
			    }
		    LSTFOR (TraGVariables(tran), var, VARIABLE)
			if ((VarGLow(var) > VarGHigh(var)) && !VarBad(var))
			    {
				VarSBad(var);
				(void) sprintf(errStr,
					       "Variable %s of %s has an empty range",
					       VarGName(var),
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "(low limit %s, high limit %s)",
					       DOMFORMAT, DOMFORMAT);
				format = StrCopy(errStr);
				(void) sprintf(errStr, format,
					       VarGLow(var), VarGHigh(var));
				Deallocate((char *) format);
				error(errStr);
				goto recover;
			    }
		    LSTFOR (TraGVariables(tran), var, VARIABLE)
			if (VarIsIn(var) && (!VarBad(var)) &&
			    (unfMaxDiff < (UDOMAIN)
			     (VarGHigh(var) - VarGLow(var))))
			    {
				(void) sprintf(errStr,
					       "Input variable %s of %s has a too wide range",
					       VarGName(var),
					       TraGName(tran));
				error(errStr);
				(void) sprintf(errStr,
					       "((high limit %s) - (low limit %s) > %s)",
					       DOMFORMAT, DOMFORMAT,
					       UDOMFORMAT);
				format = StrCopy(errStr);
				(void) sprintf(errStr, format,
					       VarGHigh(var), VarGLow(var),
					       unfMaxDiff);
				Deallocate((char *) format);
				error(errStr);
				goto recover;
			    }
		    (void) LstInsLast(TransLst, (char *) tran);
		    return;
		}
	    TOKEN_MATCH(token, PPNAME,
			((void) sprintf(errStr,
                                        "%sin, out, gate, comp, or #endtr in transition %s",
                                        placeLimits ? "" : "let, ",
                                        TraGName(tran)), errStr));
            if (!strcmp(token->token, "let"))
                {
                    if (placeLimits)
                        {
			    (void) sprintf(errStr,
                                           "Transition %s: unexpected \"let\"\nIf you want to use \"let\", run %s with the -R option",
                                           TraGName(tran), myName);
			    error(errStr);
			    goto recover;
                        }
                    if (gotlet++)
			{
			    (void) sprintf(errStr,
                                           "Transition %s: \"let\" allowed only once",
					   TraGName(tran));
			    error(errStr);
			    goto recover;
			}
                    if (!trans_let(tran))
                        goto recover;
                }
            else if (!strcmp(token->token, "in"))
		{
                    if (gotin++)
			{
			    (void) sprintf(errStr,
					   "Transition %s: \"in\" allowed only once",
					   TraGName(tran));
			    error(errStr);
			    goto recover;
			}
		    if (!trans_in_out(tran, TRUE))
			goto recover;
		}
            else if (!strcmp(token->token, "out"))
		{
                    if (gotout++)
			{
			    (void) sprintf(errStr,
					   "Transition %s: \"out\" allowed only once",
					   TraGName(tran));
			    error(errStr);
			    goto recover;
			}
		    if (!trans_in_out(tran, FALSE))
			goto recover;
		}
	    else if (!strcmp(token->token, "gate"))
		{
		    if (!trans_gate(tran))
			goto recover;
		}
	    else if (!strcmp(token->token, "comp"))
		{
		    if (TraGComputation(tran))
			{
			    (void) sprintf(errStr,
					   "Transition %s: comp allowed only once",
					   TraGName(tran));
			    error(errStr);
			    goto recover;
			}
		    TraSComputation(tran);
		    if (!trans_comp(tran))
			goto recover;
		}
	    else
		TOKEN_MATCH(token, PPNEVERMATCH,
			    ((void) sprintf(errStr,
                                            "%sin, out, gate, comp, or #endtr in transition %s",
                                            placeLimits ? "" : "let, ",
					    TraGName(tran)), errStr));
	    PpTokDiscard(token);
	}
 recover: ;
    if (((PP_TOKEN *) NULL) != token)
	PpTokDiscard(token);
    if (((TRANSITION *) NULL) != tran)
	TraDiscard(tran);
    if (endFound)
	return;
    warning("#trans failed, skipping to next #endtr");
    for(token = get_token(); token;)
	{
	    while (token && (token->type != PPDV))
		{
		    PpTokDiscard(token);
		    token = get_token();
		}
	    if (token)
		{
		    PpTokDiscard(token);
		    token = gTSkSpace();
		}
	    if (token && (token->type == PPNAME) &&
		!strcmp(token->token, "endtr"))
		{
		    PpTokDiscard(token);
		    return;
		}
	}
    panic("#trans failed, could not find #endtr when trying to recover");
}
